%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : asn1ct_constructed_ber.erl
%%% Author  : Kenneth Lundin <kenneth@erix.ericsson.se>
%%% Purpose : 
%%% Created :  6 Mar 1997 by Kenneth Lundin <kenneth@erix.ericsson.se>
%%% Functionality added :  4 Nov 1997 by Esko Vierumaki <esko@erix.ericsson.se>

-module(asn1ct_constructed_ber).
-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

-export([gen_encode_sequence/2]).
-export([gen_decode_sequence/3]).
-export([gen_encode_set/2]).
-export([gen_decode_set/3]).
-export([gen_encode_sof/3]).
-export([gen_decode_sof/3]).
-export([gen_encode_choice/2]).
-export([gen_decode_choice/3]).

-include("asn1_records.hrl").

%-compile(export_all).

-import(asn1ct_gen, [emit/1,demit/1]).

% the encoding of class of tag bits 8 and 7
-define(UNIVERSAL,   0).
-define(APPLICATION, 16#40).
-define(CONTEXT,     16#80).
-define(PRIVATE,     16#C0).

% primitive or constructed encoding % bit 6
-define(PRIMITIVE,   0).
-define(CONSTRUCTED, 2#00100000).




%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode/decode SEQUENCE
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_encode_sequence(Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(term),
    asn1ct_name:new(bytes),
 
    {'SEQUENCE',CompList} = D#type.def,

    case Optionals = optionals(CompList) of
	[] ->
	    emit({"   ",{var,asn1ct_name:next(val)}," = asn1rt_ber:list_to_record(",
		  {asis,Typename},", ",{var,asn1ct_name:curr(val)},"),",nl});
	_ ->
	    emit({"   {",{var,asn1ct_name:next(val)},
		  ",Opt} = asn1rt_ber:fixoptionals(",
		  {asis,Optionals},",",
		  {var,asn1ct_name:curr(val)},"),",nl})
    end,

    asn1ct_name:new(val),
    gen_enc_sequence_call(Typename,CompList,1),


    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(D#type.tag),
    case TopEnc_type of
	explicit ->
	    case asn1ct_name:curr(att) of
		att -> 	    
		    emit({nl,"   InnerTag = "}),
		    enc_tag_seq({tag, 'UNIVERSAL', 
				 asn1ct_gen_ber:decode_type('SEQUENCE'), 
				 'IMPLICIT'}),
		    emit({com,nl,"   InnerLen = 0,",nl}),
		    emit({"   ["}),
		    enc_tag_seq(D#type.tag),
		    emit({"    length(InnerTag), InnerTag, InnerLen].",nl,nl});
		_ ->
		    emit({nl,"   InnerTag = "}),
		    enc_tag_seq({tag, 'UNIVERSAL', 
				 asn1ct_gen_ber:decode_type('SEQUENCE'), 
				 'IMPLICIT'}),
		    emit({"   InnerLen = "}),
		    list_len(asn1ct_name:all(att)),
		    emit({com,nl,"   InnerLenEnc = asn1rt_ber:encode_length(InnerLen),"}),
		    emit({nl,"   InnerTagLen = length(InnerTag),"}),
		    emit({nl,"   InnerLenEncLen = length(InnerLenEnc),"}),
		    emit({nl,"   ["}),
		    enc_tag_seq(D#type.tag),
		    emit({"    asn1rt_ber:encode_length(InnerTagLen + InnerLenEncLen + InnerLen), ",nl}),
		    emit({"    InnerTag, InnerLenEnc, ",nl,"    "}),
		    mkvlist(asn1ct_name:all(att)),
		    emit({"].",nl,nl})
	    end;

	%% universal ->
	%% implicit ->
	_ ->
	    case asn1ct_name:curr(att) of
		att -> 	    
		    emit({nl,"   ["}),
		    enc_tag_seq(D#type.tag),
		    emit({"    0].",nl,nl});
		_ ->
		    emit({nl,"   ["}),
		    enc_tag_seq(D#type.tag),
		    emit({"    asn1rt_ber:encode_length("}),
		    list_len(asn1ct_name:all(att)),
		    emit({"), ",nl,"    "}),
		    mkvlist(asn1ct_name:all(att)),
		    emit({"].",nl,nl})
	    end
    end.






    



gen_decode_sequence(Module,Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(term),
    asn1ct_name:new(tag),
    {'SEQUENCE',CompList} = D#type.def,

    emit({"   %%-------------------------------------------------",nl}),
    emit({"   %% decode tag and length ",nl}),
    emit({"   %%-------------------------------------------------",nl}),

    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(D#type.tag),
    case TopEnc_type of
	explicit ->
	    %% Explicit tag
	    asn1ct_name:new(rb),
	    dec_tag_seq(D#type.tag),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length(),

	    %% SET tag
	    asn1ct_name:new(rb),
	    dec_tag_seq({tag, 'UNIVERSAL', asn1ct_gen_ber:decode_type('SEQUENCE'), 'IMPLICIT'}),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length();


	%% universal ->
	%% implicit ->
	_ ->
	    asn1ct_name:new(rb),
	    dec_tag_seq(D#type.tag),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length()

    end,
    
    case CompList of
	[] -> true;
	[{'EXTENSIONMARK',_,_}] -> true; % only_extension_mark
	_ ->
	    emit({"{",{var,asn1ct_name:next(bytes)},
		  ",RemBytes} = asn1rt_ber:split_list(",
		  {var,asn1ct_name:curr(bytes)},
		  ",", {var,asn1ct_name:prev(len)},"),",nl}),
	    asn1ct_name:new(bytes)
    end,
    
    AnyTerms = gen_dec_sequence_call(Module,Typename,CompList,1),

    case AnyTerms of
	no_terms -> 	    
	    emit({nl,nl}),
	    Cnames = mkcnamelist(CompList),
	    demit({"Result = "}), %dbg
	    %% return value as record
	    asn1ct_name:new(rb),
	    emit({"   {{",{asis,Typename},"}, ",{var,asn1ct_name:curr(bytes)},",",nl,"    "}),
	    asn1ct_gen_ber:add_removed_bytes(),
	    emit({"}.",nl});
	only_extension_mark ->
	    emit({nl,nl}),
	    Cnames = mkcnamelist(CompList),
	    demit({"Result = "}), %dbg
	    %% return value as record
	    asn1ct_name:new(rb),
	    emit({"   {{",{asis,Typename},", undefined}, ",{var,asn1ct_name:curr(bytes)},",",nl,"    "}),
	    asn1ct_gen_ber:add_removed_bytes(),
	    emit({"}.",nl});
	_ ->
	    emit({com,nl,nl}),
	    Cnames = mkcnamelist(CompList),
	    demit({"Result = "}), %dbg
	    %% return value as record
	    asn1ct_name:new(rb),
	    emit({"   {{",{asis,Typename},", "}),
	    mkvlist(asn1ct_name:all(term)),
	    emit({"}, asn1rt_ber:restbytes(RemBytes, ",
		  {var,asn1ct_name:next(bytes)},"),",nl,"    "}),
	    asn1ct_gen_ber:add_removed_bytes(),
	    emit({"}.",nl})
    end.


%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode/decode SET
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_encode_set(Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(term),
    asn1ct_name:new(bytes),
    
    {'SET',CompList} = D#type.def,
    
    case Optionals = optionals(CompList) of
	[] ->
	    emit({"   ",{var,asn1ct_name:next(val)}," = asn1rt_ber:list_to_record(",
		  {asis,Typename},", ",{var,asn1ct_name:curr(val)},"),",nl}),
	    asn1ct_name:new(val);
	_ ->
	    emit({"   {",{var,asn1ct_name:next(val)},
		  ",Opt} = asn1rt_ber:fixoptionals(",
		  {asis,Optionals},",",
		  {var,asn1ct_name:curr(val)},"),",nl})
    end,
    gen_enc_sequence_call(Typename,CompList,1),
        
    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(D#type.tag),
    case TopEnc_type of
	explicit ->
	    case asn1ct_name:curr(att) of
		att -> 	    
		    emit({nl,"   InnerTag = "}),
		    enc_tag_set({tag, 'UNIVERSAL', asn1ct_gen_ber:decode_type('SET'), 'IMPLICIT'}),
		    emit({com,nl,"   InnerLen = 0,",nl}),
		    emit({"   ["}),
		    enc_tag_seq(D#type.tag),
		    emit({"    length(InnerTag), InnerTag, InnerLen].",nl,nl});
		_ ->
		    emit({nl,"   InnerTag = "}),
		    enc_tag_set({tag, 'UNIVERSAL', asn1ct_gen_ber:decode_type('SET'), 'IMPLICIT'}),
		    emit({"   InnerLen = "}),
		    list_len(asn1ct_name:all(att)),
		    emit({com,nl,"   InnerLenEnc = asn1rt_ber:encode_length(InnerLen),"}),
		    emit({nl,"   InnerTagLen = length(InnerTag),"}),
		    emit({nl,"   InnerLenEncLen = length(InnerLenEnc),"}),
		    emit({nl,"   ["}),
		    enc_tag_set(D#type.tag),
		    emit({"    asn1rt_ber:encode_length(InnerTagLen + InnerLenEncLen + InnerLen), ",nl}),
		    emit({"    InnerTag, InnerLenEnc, ",nl,"    "}),

		    mkvlist(asn1ct_name:all(att)),
		    emit({"].",nl,nl})
	    end;

	%% universal ->
	%% implicit ->
	_ ->
	    case asn1ct_name:curr(att) of
		att -> 	    
		    emit({nl,"   ["}),
		    enc_tag_set(D#type.tag),
		    emit({"    0].",nl,nl});
		_ ->
		    emit({nl,"   ["}),
		    enc_tag_set(D#type.tag),
		    emit({"    asn1rt_ber:encode_length("}),
		    list_len(asn1ct_name:all(att)),
		    emit({"), ",nl,"    "}),
		    mkvlist(asn1ct_name:all(att)),
		    emit({"].",nl,nl})
	    end
    end.






gen_decode_set(Module,Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(term),
    asn1ct_name:new(tag),
    {'SET',CompList} = D#type.def,

    emit({"   %%-------------------------------------------------",nl}),
    emit({"   %% decode tag and length ",nl}),
    emit({"   %%-------------------------------------------------",nl}),

    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(D#type.tag),
    case TopEnc_type of
	explicit ->
	    %% Explicit tag
	    asn1ct_name:new(rb),
	    dec_tag_set(D#type.tag),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length(),

	    %% SET tag
	    asn1ct_name:new(rb),
	    dec_tag_set({tag, 'UNIVERSAL', asn1ct_gen_ber:decode_type('SET'), 'IMPLICIT'}),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length(),

	    emit({"   {SetTerm, SetBytes} = 'dec_",Typename,"_set_loop'(",
		  {var,asn1ct_name:curr(bytes)},", Telltype, [], Len1),",nl}),
	    
	    asn1ct_name:new(rb),
	    emit({"   'dec_",Typename,"_result'(lists:sort(SetTerm), SetBytes, "}),
	    asn1ct_gen_ber:add_removed_bytes(),
	    emit({"+Len1).",nl,nl,nl});

	%% universal ->
	%% implicit ->
	_ ->
	    asn1ct_name:new(rb),
	    dec_tag_set(D#type.tag),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length(),

	    emit({"   {SetTerm, SetBytes} = 'dec_",Typename,"_set_loop'(",
		  {var,asn1ct_name:curr(bytes)},", Telltype, [], Len),",nl}),
	    
	    asn1ct_name:new(rb),
	    emit({"   'dec_",Typename,"_result'(lists:sort(SetTerm), SetBytes, "}),
	    asn1ct_gen_ber:add_removed_bytes(),
	    emit({"+Len).",nl,nl,nl})
    end,
    


    emit({"%%-------------------------------------------------",nl}),
    emit({"%% Set loop for ",{asis,Typename},nl}),
    emit({"%%-------------------------------------------------",nl}),

    emit({"'dec_",Typename,"_set_loop'(Bytes, Telltype,  S_Term, S_Rb) when S_Rb < 0 ->",nl}),
    emit({"   exit({error,{asn1,{'SET_length_error',{remaining_bytes,Bytes}}}});",nl,nl}),
    emit({"'dec_",Typename,"_set_loop'([], Telltype,  S_Term, S_Rb) when S_Rb > 0 ->",nl}),
    emit({"   exit({error,{asn1,{'SET_length_error',{no_of_missing_bytes, S_Rb}}}});",nl,nl}),
    emit({"'dec_",Typename,"_set_loop'(Bytes, Telltype,  S_Term, 0) ->",nl}),
    emit({"   {S_Term, Bytes};",nl,nl}),
    asn1ct_name:clear(),
    asn1ct_name:new(term),
    emit({"'dec_",Typename,"_set_loop'(",{var,asn1ct_name:curr(bytes)},
	  ", Telltype,  S_Term, S_Rb) ->",nl}),
    
    asn1ct_name:new(bytes),
    gen_dec_set(Module,Typename,CompList,1),

    emit({"      %% tag not found",nl}),
    emit({indent(6),"_ -> {S_Term, Bytes}",nl}),
    emit({indent(3),"end.",nl,nl,nl}),


    emit({"%%-------------------------------------------------",nl}),
    emit({"%% Result ",{asis,Typename},nl}),
    emit({"%%-------------------------------------------------",nl}),

    asn1ct_name:clear(),
    emit({"'dec_",Typename,"_result'(TermList, Bytes, Rb) ->",nl}),
    AnyTerms = gen_dec_set_result(Module,Typename,CompList,1),

    case AnyTerms of
	no_terms -> 	    
	    %% return value as record
	    asn1ct_name:new(rb),
	    emit({"   {{",{asis,Typename},"}, Bytes, Rb}.",nl});
	only_extension_mark ->
	    asn1ct_name:new(rb),
	    emit({"   {{",{asis,Typename},", undefined}, Bytes, Rb}.",nl});
	_ ->
	    emit({nl,"   case ",{var,asn1ct_name:curr(termList)}," of",nl}),
	    emit({"      [] -> {{",{asis,Typename},", "}),
	    mkvlist(asn1ct_name:all(term)),
	    emit({"}, Bytes, Rb};",nl}),    
	    emit({"      ExtraAtt -> exit({error,{asn1,{too_many_attributes, ExtraAtt}}})",nl}),
	    emit({"   end.",nl}),

	    emit({nl,nl,nl})
    end.



    






%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode/decode SEQUENCE OF and SET OF
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_encode_sof(Typename,InnerTypename,D) when record(D,type) ->
    asn1ct_name:start(),
    {SeqOrSetOf, TypeTag, TypeNameSuffix, Cont} = 
	case D#type.def of
	    {'SET OF',_Cont} -> {'SET OF','SET',"_SETOF",_Cont};
	    {'SEQUENCE OF',_Cont} -> {'SEQUENCE OF','SEQUENCE',"_SEQOF",_Cont}
	end,

    emit({"   SOfEnc = 'enc_",Typename,"_components'(Val,[]),",nl}),
    case D#type.tag of
	{tag,Class,TagNr,'EXPLICIT'} ->
	    emit({"   InnerTag = "}),
	    enc_tag({tag, 'UNIVERSAL', 
		     asn1ct_gen_ber:decode_type(TypeTag), 
		     'IMPLICIT'},
		    TypeTag),
	    emit({"   InnerLen = length(lists:flatten(SOfEnc)),",nl}),
	    emit({"   InnerLenEnc = asn1rt_ber:encode_length(InnerLen),",nl}),
	    emit({"   InnerTagLen = length(InnerTag),",nl}),
	    emit({"   InnerLenEncLen = length(InnerLenEnc),",nl}),
	    emit({"   ["}),
	    enc_tag(D#type.tag,TypeTag),
%	    emit({"    InnerTagLen + InnerLenEncLen + InnerLen, InnerTag, InnerLenEnc, SOfEnc].",nl,nl,nl});
		    emit({"    asn1rt_ber:encode_length(InnerTagLen + InnerLenEncLen + InnerLen), ",nl}),
		    emit({"    InnerTag, InnerLenEnc, SOfEnc].",nl,nl,nl});
	%% no tag ->
	%% implicit ->
	_ ->
	    emit({indent(3),"["}),
	    enc_tag(D#type.tag,TypeTag),
	    emit({indent(4),"asn1rt_ber:encode_length(length(lists:flatten(SOfEnc)))",",SOfEnc].",nl,nl,nl})
    end,

    emit({"'enc_",Typename,"_components'([],Acc) -> ",nl,
	  indent(3),"lists:reverse(Acc);",nl,nl}),
    emit({"'enc_",Typename,"_components'([H|T],Acc) ->",nl}),
    emit({indent(3),"'enc_",Typename,"_components'(T,"}), 
    emit({"["}),
    %% the component encoder
    Conttype = asn1ct_gen:get_inner(Cont#type.def),
    case asn1ct_gen:type(Conttype) of
	{primitive,bif} ->
	    asn1ct_gen_ber:gen_encode_prim(ber,Cont,tag,"H");
	{constructed,bif} ->
	    NewTypename = list_to_atom(lists:concat([Typename,TypeNameSuffix])),
	    emit({"'enc_",NewTypename,"'(H)"});
	{Form,user} ->
	    emit({"'enc_",Conttype,"'(H)"});
	#'Externaltypereference'{module=EMod,type=EType} ->
	    emit({"'",EMod,"':'enc_",EType,"'(H)"});
	_ ->
	    emit({"'enc_",Conttype,"'(H)"})
    end,
    emit({" | Acc]).",nl,nl}).


gen_decode_sof(Typename,InnerTypename,D) when record(D,type) ->
    asn1ct_name:start(),
    {SeqOrSetOf, TypeTag, TypeNameSuffix, Cont} = 
	case D#type.def of
	    {'SET OF',_Cont} -> {'SET OF','SET',"_SETOF",_Cont};
	    {'SEQUENCE OF',_Cont} -> {'SEQUENCE OF','SEQUENCE',"_SEQOF",_Cont}
	end,
    
    emit({"   %%-------------------------------------------------",nl}),
    emit({"   %% decode tag and length ",nl}),
    emit({"   %%-------------------------------------------------",nl}),
    
    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(D#type.tag),
    case TopEnc_type of
	explicit ->
	    %% Explicit tag
	    asn1ct_name:new(rb),
	    dec_tag(D#type.tag,TypeTag),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length(),
	    
	    %% SET tag
	    asn1ct_name:new(rb),
	    dec_tag({tag, 'UNIVERSAL', asn1ct_gen_ber:decode_type(TypeTag), 'IMPLICIT'},TypeTag),
	    asn1ct_name:new(bytes),
	    asn1ct_gen_ber:decode_length(),
	    
	    asn1ct_name:new(rb),
	    emit({"   'dec_",Typename,"_components'(Len1+"}),
	    asn1ct_gen_ber:add_removed_bytes(),
	    emit({", Len1, ",{var,asn1ct_name:curr(bytes)},", Telltype,[]).",nl,nl,nl});
	
	%% universal ->
	%% implicit ->
	_ ->
	    emit({"   {Tag, Bytes1, Rb1} = asn1rt_ber:decode_tag(Bytes),",nl}),
	    emit({"   {{Num, Bytes2}, Rb2} = asn1rt_ber:decode_length(Bytes1),",nl}),
	    emit({indent(3),"'dec_",Typename,"_components'(Num+Rb1+Rb2,Num,Bytes2,Telltype,[]).",nl,nl})
    end,
    
    emit({"'dec_",Typename,"_components'(Rb,Num,Bytes,Telltype,Acc) when Num == 0 ->",nl,
	  indent(3),"{lists:reverse(Acc),Bytes,Rb};",nl,nl}),
    emit({"'dec_",Typename,"_components'(Rb,Num,Bytes,Telltype,Acc) when Num < 0 ->",nl,
	  indent(3),"exit({error,{asn1,{length_error,",{asis,SeqOrSetOf},"}}});",nl,nl}),
    emit({"'dec_",Typename,"_components'(Rb,Num,Bytes,Telltype,Acc) ->",nl}),
    emit({indent(3),"{Term, Remain, Rb1} = "}),
    Conttype = asn1ct_gen:get_inner(Cont#type.def),
    case asn1ct_gen:type(Conttype) of
	{primitive,bif} ->
	    asn1ct_gen_ber:gen_dec_prim(ber,Cont,"Bytes",tag,no_length,?PRIMITIVE),
	    emit({com,nl});
	{constructed,bif} ->
	    NewTypename = list_to_atom(lists:concat([Typename,TypeNameSuffix])),
	    emit({"'dec_",NewTypename,"'(Bytes,Telltype),",nl});
	{Form,user} ->
	    emit({"'dec_",Conttype,"'(Bytes,Telltype),",nl});
	#'Externaltypereference'{module=EMod,type=EType} ->
	    emit({"'",EMod,"':'dec_",EType,"'(Bytes,Telltype),",nl});
	_ ->
	    emit({"'dec_",Conttype,"'(Bytes,Telltype),",nl})
    end,
    emit({indent(3),"'dec_",Typename,"_components'(Rb, Num-Rb1, Remain, Telltype, [Term|Acc]).",nl}). 



%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode/decode CHOICE
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_encode_choice(Typename,D) when record(D,type) ->
    ChoiceTag = D#type.tag,
    {'CHOICE',CompList} = D#type.def,
    case (extensible(CompList)) of
	{true,Pos} -> 
	    gen_enc_choice(Typename,ChoiceTag,CompList,ext);
	_ -> 
	    gen_enc_choice(Typename,ChoiceTag,CompList,noext)
    end,
    emit({nl,nl}).

gen_decode_choice(Module,Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(bytes),
    ChoiceTag = D#type.tag,
    {'CHOICE',CompList1} = D#type.def,
    
    CompList = [C || C <- CompList1, record(C,'ComponentType')],

    NumChoices = length(mkcnamelist(CompList)),
    case (extensible(CompList)) of
	{true,Pos} -> 
	    gen_dec_choice(Module,Typename,ChoiceTag,CompList,ext,NumChoices);
	_ -> 
	    gen_dec_choice(Module,Typename,ChoiceTag,CompList,noext,NumChoices)
    end,
    emit({".",nl}).







%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode SEQUENCE
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_enc_sequence_call(TopType,[{'ComponentType',_Pos,Cname,Type,Prop,Tags}|Rest],Num) ->
%    asn1ct_name:new(tag),
%    asn1ct_name:new(enc),
    asn1ct_name:new(encTmp),
    asn1ct_name:new(att),
    case Prop of
	'OPTIONAL' ->
	    gen_enc_component_optional(TopType,Cname,Type,Num);
	{'DEFAULT', {_,_,DefVal}} ->
	    gen_enc_component_default(TopType,Cname,Type,Num,DefVal);
	{'DEFAULT', DefVal} ->
	    gen_enc_component_default(TopType,Cname,Type,Num,DefVal);
	_ ->
	    gen_enc_component_mandatory(TopType,Cname,Type,Num)
    end,
    case Rest of
	[] ->
	    emit({com,nl});
	_ ->
	    emit({com,nl}),
	    gen_enc_sequence_call(TopType,Rest,Num+1)
    end;

gen_enc_sequence_call(TopType,[{'EXTENSIONMARK',_Pos,_}|Rest],Num) ->
    gen_enc_sequence_call(TopType,Rest,Num+1);

gen_enc_sequence_call(_TopType,[],_Num) ->
	true.



%%===============================================================================
%% Encode DEFAULT
%%===============================================================================
gen_enc_component_default(TopType,Cname,Type,Pos,DefVal) ->
    Element = io_lib:format("asn1rt_ber:cindex(~w,Val1,~w)",[Pos+1,Cname]),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos," DEFAULT = ",DefVal,3),
    emit({"   ",{var,asn1ct_name:curr(att)}," = ",nl}),
    emit({"      case ",Element," of",nl}),
    emit({"         asn1_DEFAULT -> [];",nl}),
    emit({"         _ ->",nl}),
    gen_enc_line(TopType,Cname,Type,Element,12,false),
    emit({nl,"      end"}).

%%===============================================================================
%% Encode OPTIONAL
%%===============================================================================
gen_enc_component_optional(TopType,Cname,Type,Pos) ->
    Element = io_lib:format("asn1rt_ber:cindex(~w,Val1,~w)",[Pos+1,Cname]),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos," OPTIONAL","",3),
    emit({"   ",{var,asn1ct_name:curr(att)}," = ",nl}),
    emit({"      case ",Element," of",nl}),
    emit({"         asn1_NOVALUE -> [];",nl}),
    emit({"         _ ->",nl}),
    gen_enc_line(TopType,Cname,Type,Element,12,false),
    emit({nl,"      end"}).

%%===============================================================================
%% Encode MANDATORY
%%===============================================================================
gen_enc_component_mandatory(TopType,Cname,Type,Pos) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos,"","",3),
    Element = io_lib:format("asn1rt_ber:cindex(~w,~s,~w)",
			    [Pos+1,asn1ct_gen_ber:mk_var(asn1ct_name:curr(val)),Cname]),
    gen_enc_line(TopType,Cname,Type,Element,3,true).






%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Decode SEQUENCE
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_dec_sequence_call(Module,TopType,[{'ComponentType',_Pos,Cname,Type,Prop,Tags}|Rest],Num) ->
    case Prop of
	'OPTIONAL' ->
	    gen_dec_component_optional(Module,TopType,Cname,Type,Num,optional);
	{'DEFAULT', {_,_,DefVal}} ->
	    gen_dec_component_optional(Module,TopType,Cname,Type,Num,{default,DefVal});
	{'DEFAULT', DefVal} ->
	    gen_dec_component_optional(Module,TopType,Cname,Type,Num,{default,DefVal});
	_ ->
	    gen_dec_component_mandatory(Module,TopType,Cname,Type,Num)
    end,
    case Rest of
	[] ->
	    true;
	[{'EXTENSIONMARK',_,_}] ->
	    true;
	[{'EXTENSIONMARK',_,_}|_] ->
	    emit({com,nl}),
	    asn1ct_name:new(term),
	    asn1ct_name:new(bytes),
	    gen_dec_sequence_call(Module,TopType,Rest,Num+1);
	_ ->
	    emit({com,nl}),
	    asn1ct_name:new(term),
	    asn1ct_name:new(bytes),
	    gen_dec_sequence_call(Module,TopType,Rest,Num+1)
    end;

gen_dec_sequence_call(Module,TopType,[{'EXTENSIONMARK',_Pos,_}],1) ->
    only_extension_mark;
gen_dec_sequence_call(Module,TopType,[{'EXTENSIONMARK',_Pos,_}],Num) ->
    true;
gen_dec_sequence_call(Module,TopType,[{'EXTENSIONMARK',_Pos,_}|Rest],Num) ->
    gen_dec_sequence_call(Module,TopType,Rest,Num+1);
gen_dec_sequence_call(Module,_TopType,[],1) ->
    no_terms;
gen_dec_sequence_call(Module,_TopType,[],Num) ->
    true.





%%================================
%%SEQUENCE    OPTIONAL and DEFAULT
%%================================
gen_dec_component_optional(Module,TopType,Cname,Type,Pos,OptOrDef) ->
    Element = io_lib:format("element(~w,Opt)",[Pos]),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    case OptOrDef of
	optional ->
	    print_attribute_comment(InnerType,Pos," OPTIONAL","",3);
	{default,DefVal} ->
	    print_attribute_comment(InnerType,Pos," DEFAULT = ",DefVal,3)
    end,

    case Type#type.def of
	{'CHOICE',NewCompList} when OptOrDef == optional -> 
	    emit({"   {",{var,asn1ct_name:curr(term)},", ",{var,asn1ct_name:next(bytes)},", ",
		  {var,asn1ct_name:next(rb)},"} = ",nl}),
	    gen_dec_component_optional_choise(Module,TopType,Cname);
	{'CHOICE',NewCompList} -> 
	    emit({"   {",{var,asn1ct_name:curr(term)},", ",{var,asn1ct_name:next(bytes)},", ",
		  {var,asn1ct_name:next(rb)},"} = ",nl}),
	    gen_dec_component_default_choise(Module,TopType,Cname,OptOrDef);
	{typereference,_,Name} -> 
	    emit({"   {",{var,asn1ct_name:curr(term)},", ",{var,asn1ct_name:next(bytes)},", ",
		  {var,asn1ct_name:next(rb)},"} = ",nl}),

	    gen_dec_component_optional_typeref(Module,Type,OptOrDef),
	    asn1ct_name:new(rb);


	{'Externaltypereference',_,XModule,Name} -> 
	    emit({"   {",{var,asn1ct_name:curr(term)},", ",{var,asn1ct_name:next(bytes)},", ",
		  {var,asn1ct_name:next(rb)},"} = ",nl}),

	    gen_dec_component_optional_typeref(Module,Type,OptOrDef),
	    asn1ct_name:new(rb);
	_ ->
	    asn1ct_name:new(bytes),
	    asn1ct_name:new(bytes),
	    asn1ct_name:new(bytes),
	    emit({"   {",{var,asn1ct_name:curr(term)},", ",{var,asn1ct_name:next(bytes)},", ",
		  {var,asn1ct_name:next(rb)},"} = ",nl}),
	    asn1ct_name:delete(bytes),
	    asn1ct_name:delete(bytes),
	    asn1ct_name:delete(bytes),
	    asn1ct_name:new(rb),
	    emit({indent(6),"case ",{var,asn1ct_name:curr(bytes)}," of",nl}),
	    emit({indent(9), "[] -> ",nl}),
	    case OptOrDef of
		optional ->
		    emit({indent(12),"{asn1_NOVALUE, ",{var,asn1ct_name:curr(bytes)},", 0};",nl});
		{default,DefVal1} ->
		    emit({indent(12),"{",{asis,DefVal1},", ",{var,asn1ct_name:curr(bytes)},", 0};",nl})
	    end,
	    emit({indent(9),"_ -> ",nl}),


	    case asn1ct_gen:type(InnerType) of
		{primitive,bif} ->
		    gen_dec_component_optional_prim(TopType,Cname,Type,Pos,OptOrDef);
		%%{Form,user} ->
		%%{notype,_} ->
		%%{constructed,bif} ->
		_ ->
		    gen_dec_component_optional_constr(TopType,Cname,Type,Pos,OptOrDef)
	    end
    end.


gen_dec_component_optional_choise(Module,TopType,Cname) ->
    NewTypename = lists:concat([TopType,"_",Cname]),
    asn1ct_name:new(rb),
    emit({indent(6),"case catch 'dec_",NewTypename,
	  "'(",{var,asn1ct_name:curr(bytes)},", tellbytes) of",nl}),
    asn1ct_name:new(tag),
    emit({indent(9), "{'EXIT', _} -> {asn1_NOVALUE,",{var,asn1ct_name:curr(bytes)},",0};",nl}),
    emit({indent(9),{var,asn1ct_name:curr(optChoRes)}," -> ",
	  {var,asn1ct_name:curr(optChoRes)}}),
    emit({nl,indent(6),"end"}),
    asn1ct_name:new(optChoRes).



gen_dec_component_default_choise(Module,TopType,Cname,{default,DefValue}) ->
    NewTypename = lists:concat([TopType,"_",Cname]),
    asn1ct_name:new(rb),
    emit({indent(6),"case catch 'dec_",NewTypename,
	  "'(",{var,asn1ct_name:curr(bytes)},", tellbytes) of",nl}),
    asn1ct_name:new(tag),
    emit({indent(9), "{'EXIT', _} -> {",{asis,DefValue},", ",{var,asn1ct_name:curr(bytes)},",0};",nl}),
    emit({indent(9),{var,asn1ct_name:curr(optChoRes)}," -> ",
	  {var,asn1ct_name:curr(optChoRes)}}),
    emit({nl,indent(6),"end"}),
    asn1ct_name:new(optChoRes).



gen_dec_component_optional_typeref(ThisModule,Type,OptOrDef) ->
    {Module,Name} = case Type#type.def of
			 {'Externaltypereference',_,XModule,Name} ->
			       {XModule, Name};
			   {typereference,_,Name} ->
			       {ThisModule, Name}
		    end,
    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),

    case TopEnc_type of
	universal ->
	    Function = list_to_atom("dec_" ++ atom_to_list(Name)),
	    emit({indent(6),"asn1rt_ber:sequence_optional_no_tag("}),
	    emit({{var,asn1ct_name:curr(bytes)}}),
	    emit({", Telltype, ",{asis,Module},", ",{asis,Function},", ",{asis,OptOrDef},")"});
    
	implicit ->
	    Function = list_to_atom("dec_" ++ atom_to_list(Name)),
	    emit({indent(6),"asn1rt_ber:sequence_optional_implicit("}),
	    emit({{var,asn1ct_name:curr(bytes)},", {",Top_class,", ", Top_tagNr}),
	    emit({"}, Telltype, ",{asis,Module},", ",{asis,Function},", ",{asis,OptOrDef},")"});

	explicit ->
	    Function = list_to_atom("dec_" ++ atom_to_list(Name)),
	    emit({indent(6),"asn1rt_ber:sequence_optional_explicit("}),
	    emit({{var,asn1ct_name:curr(bytes)},", {",Top_class,", ", Top_tagNr}),
	    emit({"}, Telltype, ",{asis,Module},", ",{asis,Function},", ",{asis,OptOrDef},")"})
    end.





gen_dec_component_optional_prim(TopType,Cname,Type,Pos,OptOrDef) ->
    {Enc_type, Class, Number} = asn1ct_gen_ber:check_tag(Type#type.tag, Type#type.def),
    emit({indent(12),"{",
	  {var,asn1ct_name:curr(tag)},",",
	  {var,asn1ct_name:next(bytes)},",",
	  {var,asn1ct_name:next(rbTmp)}}),
    emit({"} = asn1rt_ber:decode_tag(",{var,asn1ct_name:curr(bytes)},"),",nl}),
    emit({indent(12),"case ",{var,asn1ct_name:curr(tag)}," of",nl}),
    asn1ct_name:new(bytes),
    asn1ct_name:new(tag),
    asn1ct_name:new(rbTmp),
    emit({indent(15),"{",{asis,Class},",",{var,asn1ct_name:curr(form)},", ",
	  {asis,Number},"} -> ",nl}),
    case Enc_type of
	universal ->
	    emit(indent(15)),
	    asn1ct_gen_ber:decode_length_temp(),
	    emit(indent(18)),
	    emit({"{",{var,asn1ct_name:curr(termTmp)},",",{var,asn1ct_name:next(bytes)},",",
		  {var,asn1ct_name:next(rbTmp)},"} = ",nl}),
	    emit(indent(21)),
	    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
	    emit({",",nl,indent(18),"{",{var,asn1ct_name:curr(termTmp)},", ",
		  {var,asn1ct_name:next(bytes)},", "}),
	    asn1ct_name:delete(rbTmp),
	    asn1ct_name:delete(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)}," + "}),
	    asn1ct_name:new(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)}," + "}),
	    asn1ct_name:new(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)},"};",nl});
    
	implicit ->
	    emit(indent(15)),
	    asn1ct_gen_ber:decode_length_temp(),
	    emit(indent(18)),
	    emit({"{",{var,asn1ct_name:curr(termTmp)},",",{var,asn1ct_name:next(bytes)},",",
		  {var,asn1ct_name:next(rbTmp)},"} = ",nl}),
	    emit(indent(21)),
	    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
	    emit({",",nl,indent(18),"{",{var,asn1ct_name:curr(termTmp)},", ",
		  {var,asn1ct_name:next(bytes)},", "}),
	    asn1ct_name:delete(rbTmp),
	    asn1ct_name:delete(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)}," + "}),
	    asn1ct_name:new(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)}," + "}),
	    asn1ct_name:new(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)},"};",nl});
	explicit ->
	    emit(indent(15)),
	    asn1ct_gen_ber:decode_length_temp(),
	    emit(indent(18)),
	    emit({"{",{var,asn1ct_name:curr(termTmp)},",",{var,asn1ct_name:next(bytes)},",",
		  {var,asn1ct_name:next(rbTmp)},"} = ",nl}),
	    emit(indent(21)),
	    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
	    emit({",",nl,indent(18),"{",{var,asn1ct_name:curr(termTmp)},", ",
		  {var,asn1ct_name:next(bytes)},", "}),
	    asn1ct_name:delete(rbTmp),
	    asn1ct_name:delete(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)}," + "}),
	    asn1ct_name:new(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)}," + "}),
	    asn1ct_name:new(rbTmp),
	    emit({{var,asn1ct_name:next(rbTmp)},"};",nl})
    end,


    emit({indent(15),"_ -> ",nl}),
    asn1ct_name:new(termTmp),
    asn1ct_name:new(rbTmp),
    asn1ct_name:delete(bytes),
    asn1ct_name:delete(bytes),
    case OptOrDef of
	optional ->
	    emit({indent(18),"{asn1_NOVALUE, ",
		  {var,asn1ct_name:curr(bytes)},", 0}",nl});
	{default,DefVal2} ->
	    emit({indent(18),"{",{asis,DefVal2},", "
		  ,{var,asn1ct_name:curr(bytes)},", 0}",nl})
    end,
    emit({indent(12),"end",nl}),
    emit({indent(6),"end"}),
    asn1ct_name:new(form),
    asn1ct_name:new(bytes),
    asn1ct_name:new(bytes),
    asn1ct_name:new(bytes).



gen_dec_component_optional_constr(TopType,Cname,Type,Pos,OptOrDef) ->
    {Enc_type, Class, Number} = asn1ct_gen_ber:check_tag(Type#type.tag, Type#type.def),
    emit({indent(12),"{",
	  {var,asn1ct_name:curr(tag)},",",
	  {var,asn1ct_name:next(bytes)},",",
	  {var,asn1ct_name:next(rbTmp)}}),
    emit({"} = asn1rt_ber:decode_tag(",{var,asn1ct_name:curr(bytes)},"),",nl}),
    emit({indent(12),"case ",{var,asn1ct_name:curr(tag)}," of",nl}),
    asn1ct_name:new(tag),
    asn1ct_name:new(rbTmp),
    emit({indent(15),"{",{asis,Class},",",{var,asn1ct_name:curr(form)},", ",
	  {asis,Number},"} -> ",nl}),
    asn1ct_name:new(form),
    
    emit(indent(18)),
    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
    emit({";",nl}), 
    
    emit({indent(15),"_ -> ",nl}),
    asn1ct_name:new(termTmp),
    asn1ct_name:new(rbTmp),
    case OptOrDef of
	optional ->
	    emit({indent(18),"{asn1_NOVALUE, ",
		  {var,asn1ct_name:curr(bytes)},", 0}",nl});
	{default,DefVal2} ->
	    emit({indent(18),"{",{asis,DefVal2},", "
		  ,{var,asn1ct_name:curr(bytes)},", 0}",nl})
    end,
    emit({indent(12),"end",nl}),
    emit({indent(6),"end"}),
    asn1ct_name:new(bytes),
    asn1ct_name:new(bytes),
    asn1ct_name:new(bytes).




%%=======================
%%SEQUENCE mandatory
%%=======================
gen_dec_component_mandatory(Module,TopType,Cname,Type,Pos) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos,"","",3),

    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),

    Enc_type = 
	case asn1ct_gen:type(InnerType) of
	    #'Externaltypereference'{module=Mod,type=EType} ->
		case TopEnc_type of
		    explicit ->
			emit({"   {",{var,asn1ct_name:next(bytes)},", ",
			      {var,asn1ct_name:next(rb)},"} = asn1rt_ber:check_tag({",
			      {asis,Top_class},",",
			      {asis,?CONSTRUCTED},",",
			      {asis,Top_tagNr},"}, {no_impl_class,no_tag_no}, ",
			      {var,asn1ct_name:curr(bytes)},"), ",nl}),
			asn1ct_name:new(bytes),
			asn1ct_name:new(rb),
    
			emit({"   {{",{var,asn1ct_name:curr(len)},", ",
			      {var,asn1ct_name:next(bytes)},"}, ",
			      {var,asn1ct_name:next(rb)},"} = asn1rt_ber:decode_length(",
			      {var,asn1ct_name:curr(bytes)},"),",nl}),
			asn1ct_name:new(len),
			asn1ct_name:new(rb),
			asn1ct_name:new(bytes);
		    _ ->
		    %% universal ->
		    %% implicit ->
			continue
		end,
		not_valid;

	    {primitive,bif} ->
		Enc_type1 = asn1ct_gen_ber:dec_tag_prim(Type#type.tag, Type#type.def),
		case Enc_type1 of
		    universal ->
			asn1ct_gen_ber:decode_length(),
			emit("   "),
			continue;
		    implicit ->
			asn1ct_gen_ber:decode_length(),
			emit("   "),
			continue;
		    explicit ->
			asn1ct_gen_ber:decode_length(),
			emit("   ")
		end,
		asn1ct_name:new(tag),
		Enc_type1;

	    {Form,user} ->
		case TopEnc_type of
		    explicit ->
			emit({"   {",{var,asn1ct_name:next(bytes)},", ",
			      {var,asn1ct_name:next(rb)},"} = asn1rt_ber:check_tag({",
			      {asis,Top_class},",",
			      {asis,?CONSTRUCTED},",",
			      {asis,Top_tagNr},"}, {no_impl_class,no_tag_no}, ",
			      {var,asn1ct_name:curr(bytes)},"), ",nl}),
			asn1ct_name:new(bytes),
			asn1ct_name:new(rb),
    
			emit({"   {{",{var,asn1ct_name:curr(len)},", ",
			      {var,asn1ct_name:next(bytes)},"}, ",
			      {var,asn1ct_name:next(rb)},"} = asn1rt_ber:decode_length(",
			      {var,asn1ct_name:curr(bytes)},"),",nl}),
			asn1ct_name:new(len),
			asn1ct_name:new(rb),
			asn1ct_name:new(bytes);
		    _ ->
		    %% universal ->
		    %% implicit ->
			continue
		end,
		emit("   "),
		not_valid;

	    {notype,_} ->
		emit("   "),
		not_valid;

	    {constructed,bif} ->
		emit("   "),
		not_valid
	end,

    emit({"{",{var,asn1ct_name:curr(term)},",",
	  {var,asn1ct_name:next(bytes)},",",
	  {var,asn1ct_name:next(rb)},"} = "}),
    asn1ct_name:new(rb),
    gen_dec_line(TopType,Cname,Type,Pos,Enc_type),
    asn1ct_name:new(form).






%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Decode SET
%%===============================================================================
%%===============================================================================
%%===============================================================================

gen_dec_set(Module,TopType,CompList,Pos) ->
    asn1ct_name:new(rbCho),
    TagList = get_all_choice_tags(CompList),
    emit({indent(3),
	  {var,asn1ct_name:curr(tagList)}," = ",{asis,TagList},",",nl}),
    emit({indent(3),
	  "{Tag, Bytes1, ",{var,asn1ct_name:curr(rbCho)},
	  "} = asn1rt_ber:decode_tag(Bytes),",nl,nl}),
    emit({indent(3),
	  "case asn1rt_ber:check_if_valid_tag(",{var,asn1ct_name:curr(tag)},",",
	  {var,asn1ct_name:curr(tagList)},") of",nl}),
    asn1ct_name:new(tagList),
    asn1ct_name:new(rbCho),
    asn1ct_name:new(choTags),
    gen_dec_set_cases(Module,TopType,CompList,TagList,Pos),
    asn1ct_name:new(tag),
    asn1ct_name:new(bytes).




gen_dec_set_cases(Module,TopType,[],List,Pos) ->
    ok;
gen_dec_set_cases(Module,TopType,[H|T],List,Pos) ->
    case H of
	{'EXTENSIONMARK', _, _} ->
	    gen_dec_set_cases(Module,TopType,T,List,Pos);
	_ ->
	    Name = H#'ComponentType'.name,
	    Type = H#'ComponentType'.typespec,
	    TypeDef = Type#type.def,    
	    Tags = H#'ComponentType'.tags,

	    emit({indent(6),"'",Name,"' ->",nl}),
	    case Type#type.def of
		{'CHOICE',NewCompList} -> 
		    gen_dec_set_cases_choice(Module,TopType,H,Pos);
		{typereference,_,TRName} -> 
		    gen_dec_set_cases_typeref(Module,TopType,H,Pos);
		{'Externaltypereference',_,XModule,XName} -> 
		    gen_dec_set_cases_ext_typeref(Module,TopType,H,Pos);
		_ ->
		    gen_dec_set_cases_type(Module,TopType,H,Pos)
	    end,
	    gen_dec_set_cases(Module,TopType,T,List,Pos+1)
    end.



gen_dec_set_cases_choice(Module,TopType,H,Pos) ->
    Cname = H#'ComponentType'.name,
    asn1ct_name:new(rbCho),
    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
    emit({"'dec_",TopType,"_",Cname,"'(Bytes,Telltype),",nl}),
    emit({indent(9),"{{",{asis,Cname},", Dec}, Rest, "}),
    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl}).


gen_dec_set_cases_typeref(Module,TopType,H,Pos) ->
    Cname = H#'ComponentType'.name,
    Type = H#'ComponentType'.typespec,

    {typereference, _, Name} = Type#type.def,
    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),
    asn1ct_name:new(rbCho),

    case TopEnc_type of
	universal ->
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({"'dec_",Name,"'(Bytes,Telltype),",nl}),
	    emit({indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
		  "S_Term++[{",Pos,", Dec}], S_Rb-"}),
	    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});

	implicit ->
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({"'dec_",Name,"'(Bytes,Telltype,{",Top_class,",",Top_tagNr,"}),",nl}),
	    emit({indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
		  "S_Term++[{",Pos,", Dec}], S_Rb-"}),
	    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});

	explicit ->
	    emit({indent(9),"{Bytes2, RbTag} = asn1rt_ber:check_tag({",
		  {asis,Top_class},",",
		  {asis,?CONSTRUCTED},",",
		  {asis,Top_tagNr},"}, {no_impl_class, no_tag_no}, Bytes), ",nl}),
	    emit({indent(9),"{{Len, Bytes3},RbLen} = asn1rt_ber:decode_length(Bytes2),",nl}),

	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({"'dec_",Name,"'(Bytes3,Telltype),",nl}),
	    emit({indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
		  "S_Term++[{",Pos,", Dec}], S_Rb - RbTag - RbLen - "}),
	    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl})

    end.




gen_dec_set_cases_ext_typeref(Module,TopType,H,Pos) ->
    Cname = H#'ComponentType'.name,
    Type = H#'ComponentType'.typespec,
    {'Externaltypereference', _, XModule,Name} = Type#type.def,

    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),
    asn1ct_name:new(rbCho),

    case TopEnc_type of
	universal ->
	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({"'",XModule,"':'dec_",Name,"'(Bytes,Telltype),",nl}),
	    emit({indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
		  "S_Term++[{",Pos,", Dec}], S_Rb-"}),
	    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});

	implicit ->
	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({"'",XModule,"':'dec_",Name,"'(Bytes,Telltype,{",Top_class,",",Top_tagNr,"}),",nl}),
	    emit({indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
		  "S_Term++[{",Pos,", Dec}], S_Rb-"}),
	    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});

	explicit ->
	    emit({indent(9),"{Bytes2, RbTag} = asn1rt_ber:check_tag({",
		  {asis,Top_class},",",
		  {asis,?CONSTRUCTED},",",
		  {asis,Top_tagNr},"}, {no_impl_class, no_tag_no}, Bytes), ",nl}),
	    emit({indent(9),"{{Len, Bytes3},RbLen} = asn1rt_ber:decode_length(Bytes2),",nl}),

	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({"'",XModule,"':'dec_",Name,"'(Bytes3,Telltype),",nl}),
	    emit({indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
		  "S_Term++[{",Pos,", Dec}], S_Rb - RbTag - RbLen - "}),
	    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl})

    end.




gen_dec_set_cases_type(Module,TopType,H,Pos) ->
    Cname = H#'ComponentType'.name,
    Type = H#'ComponentType'.typespec,
    {Enc_type, Class, Number} = asn1ct_gen_ber:check_tag(Type#type.tag, Type#type.def),
    emit({indent(9),
	  "{Class, ",{var,asn1ct_name:curr(form)},", TagNumber} = ",
	  {var,asn1ct_name:curr(tag)},",",nl}),
    case Enc_type of
	universal ->
	    case Type#type.def of
	       X when atom(X) ->
		    emit(indent(6)),
		    decode_length_choice(),
		    asn1ct_name:new(rbCho),
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
		    
		    emit({com,nl,indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
			  "S_Term++[{",Pos,", Dec}], S_Rb-RbCho1-"}),
		    asn1ct_name:delete(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},"-"}),
		    asn1ct_name:new(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl}),
		    asn1ct_name:delete(bytes);
		{X,List} when atom(X) ->
		    asn1ct_name:new(rbCho),
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    asn1ct_name:delete(bytes),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
		    asn1ct_name:new(bytes),

		    emit({com,nl,indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
			  "S_Term++[{",Pos,", Dec}], S_Rb-"}),
		    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});
		X ->
		    exit({error,{asn1,{unrecognized_type,X}}})
	    end;

	implicit ->
	    case Type#type.def of
	       X when atom(X) ->
		    emit(indent(6)),
		    decode_length_choice(),
		    asn1ct_name:new(rbCho),
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type),

		    emit({com,nl,indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
			  "S_Term++[{",Pos,", Dec}], S_Rb - RbCho1 -"}),
		    asn1ct_name:delete(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},"-"}),
		    asn1ct_name:new(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl}),
		    asn1ct_name:delete(bytes);

		{X,List} when atom(X) ->
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    asn1ct_name:delete(bytes),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
		    asn1ct_name:new(bytes),
		    emit({com,nl,indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
			  "S_Term++[{",Pos,", Dec}], S_Rb - "}),
		    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});
		X ->
		    exit({error,{asn1,{unrecognized_type,X}}})
	    end;

	explicit ->
	    case Type#type.def of
	       X when atom(X) ->
		    emit(indent(6)),
		    decode_length_choice(),
		    asn1ct_name:new(rbCho),
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type), 
		    
		    emit({com,nl,indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
			  "S_Term++[{",Pos,", Dec}], S_Rb - RbCho1 - "}),
		    asn1ct_name:delete(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},"-"}),
		    asn1ct_name:new(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl}),
		    asn1ct_name:delete(bytes);

		{X,List} when atom(X) ->
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    asn1ct_name:delete(bytes),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type), 
		    asn1ct_name:new(bytes),
		    
		    emit({com,nl,indent(9),"'dec_",TopType,"_set_loop'(Rest, Telltype, ",
			  "S_Term++[{",Pos,", Dec}], S_Rb - "}),
		    emit({{var,asn1ct_name:curr(rbCho)},");",nl,nl});
		X ->
		    exit({error,{asn1,{unrecognized_type,X}}})
	    end

    end.
    
 




%%===============================================================================
%%  Decode SET result
%%===============================================================================
gen_dec_set_result(Module,TopType,[{'ComponentType',_Pos,Cname,Type,Prop,Tags}|Rest],Num) ->

    case Prop of
	'OPTIONAL' ->
	    gen_dec_set_result_optional(Module,TopType,Cname,Type,Num);
	{'DEFAULT', {_,_,DefVal}} ->
	    gen_dec_set_result_default(Module,TopType,Cname,Type,Num,DefVal);
	{'DEFAULT', DefVal} ->
	    gen_dec_set_result_default(Module,TopType,Cname,Type,Num,DefVal);
	_ ->
	    gen_dec_set_result_mandatory(Module,TopType,Cname,Type,Num)
    end,
    case Rest of
	[] ->
	    true;
	[{'EXTENSIONMARK',_,_}] ->
	    true;
	[{'EXTENSIONMARK',_,_}|More] ->
	    gen_dec_set_result(Module,TopType,Rest,Num+1);
	_ ->

	    gen_dec_set_result(Module,TopType,Rest,Num+1)
    end;

gen_dec_set_result(Module,TopType,[{'EXTENSIONMARK',_Pos,_}],1) ->
    only_extension_mark;
gen_dec_set_result(Module,TopType,[{'EXTENSIONMARK',_Pos,_}],Num) ->
    true;
gen_dec_set_result(Module,TopType,[{'EXTENSIONMARK',_Pos,_}|Rest],Num) ->
    gen_dec_set_result(Module,TopType,Rest,Num);
gen_dec_set_result(_Module,_TopType,[],1) ->
    no_terms;
gen_dec_set_result(_Module,_TopType,[],Num) ->
    true.



gen_dec_set_result_optional(Module,TopType,Cname,Type,Pos) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos," OPTIONAL","",6),

    emit({"   {",{var,asn1ct_name:next(term)},com,{var,asn1ct_name:next(termList)},"} =",nl}),
    emit({"      case ",{var,asn1ct_name:curr(termList)}," of",nl}),
    emit({"          [{",Pos,com,{var,asn1ct_name:curr(termTmp)},"}|",
	                 {var,asn1ct_name:curr(rest)},"] -> "}),
    emit({"{",{var,asn1ct_name:curr(termTmp)},com,
	      {var,asn1ct_name:curr(rest)},"};",nl}),
    emit({"          _ -> {asn1_NOVALUE, ",{var,asn1ct_name:curr(termList)},"}",nl}),
    emit({"      end,",nl}),
    asn1ct_name:new(rest),
    asn1ct_name:new(term),
    asn1ct_name:new(termList),    
    asn1ct_name:new(termTmp).    


gen_dec_set_result_default(Module,TopType,Cname,Type,Pos,DefVal) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos," DEFAULT = ",DefVal,6),

    emit({"   {",{var,asn1ct_name:next(term)},com,{var,asn1ct_name:next(termList)},"} =",nl}),
    emit({"      case ",{var,asn1ct_name:curr(termList)}," of",nl}),
    emit({"          [{",Pos,com,{var,asn1ct_name:curr(termTmp)},"}|",
	                 {var,asn1ct_name:curr(rest)},"] -> "}),
    emit({"{",{var,asn1ct_name:curr(termTmp)},com,
	      {var,asn1ct_name:curr(rest)},"};",nl}),
    emit({"          _ -> {",{asis,DefVal},", ",{var,asn1ct_name:curr(termList)},"}",nl}),
    emit({"      end,",nl}),
    asn1ct_name:new(rest),
    asn1ct_name:new(term),
    asn1ct_name:new(termList),    
    asn1ct_name:new(termTmp).    


gen_dec_set_result_mandatory(Module,TopType,Cname,Type,Pos) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    print_attribute_comment(InnerType,Pos,"","",6),

    emit({"   {",{var,asn1ct_name:next(term)},com,{var,asn1ct_name:next(termList)},"} =",nl}),
    emit({"      case ",{var,asn1ct_name:curr(termList)}," of",nl}),
    emit({"          [{",Pos,com,{var,asn1ct_name:curr(termTmp)},"}|",
	                 {var,asn1ct_name:curr(rest)},"] -> "}),
    emit({"{",{var,asn1ct_name:curr(termTmp)},com,
	      {var,asn1ct_name:curr(rest)},"};",nl}),
    emit({"          _ -> exit({error,{asn1,{mandatory_attribute_no, ",Pos,", missing}}})",nl}),
    emit({"      end,",nl}),
    asn1ct_name:new(rest),
    asn1ct_name:new(term),
    asn1ct_name:new(termList),    
    asn1ct_name:new(termTmp).    





%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode CHOICE
%%===============================================================================
%%===============================================================================
%%===============================================================================
%% for BER we currently don't care if the choice has an EXTENSIONMARKER

%%==========================
%% No tag defined
%%==========================
gen_enc_choice(TopType,undefined,CompList,_Ext) ->
    emit({"   case element(1,Val) of",nl}),
    gen_enc_choice2(TopType,CompList),
    emit({nl,"   end."});

%%==========================
%% EXPLICIT tag
%%==========================
gen_enc_choice(TopType,{tag,Class,TagNr,'EXPLICIT'},CompList,_Ext) ->
    emit({"   Att = case element(1,Val) of",nl}),
    gen_enc_choice2(TopType,CompList),
    emit({nl,"   end,",nl,nl}),
    
    emit({"   ["}),
    emit({"asn1rt_ber:encode_tag_val({",
	  {asis,decode_class(Class)},",",
	  {asis,?CONSTRUCTED},",",
	  {asis,TagNr},"}), ",nl}),

    emit({"    asn1rt_ber:encode_length(length(lists:flatten(Att))), Att].",nl,nl}).




gen_enc_choice2(TopType,[H1,H2|T]) 
when record(H1,'ComponentType'), record(H2,'ComponentType') ->
    Cname = H1#'ComponentType'.name,
    Type = H1#'ComponentType'.typespec,
    emit({"      ",{asis,Cname}," ->",nl}),    
    gen_enc_line(TopType,Cname,Type,"element(2,Val)",9,false),
    emit({";",nl}),
    gen_enc_choice2(TopType,[H2|T]);

gen_enc_choice2(TopType,[H1,H2|T]) when record(H1,'ComponentType') ->
    gen_enc_choice2(TopType,[H1|T]); % skip extensionmark

gen_enc_choice2(TopType,[H1|T]) when record(H1,'ComponentType') ->
    Cname = H1#'ComponentType'.name,
    Type = H1#'ComponentType'.typespec,
    emit({"      ",{asis,H1#'ComponentType'.name}," ->",nl}),
    gen_enc_line(TopType,Cname,Type,"element(2,Val)",9,false),
    emit({";",nl,indent(6), "Else -> ",nl,
	  indent(9),"exit({error,{asn1,{invalid_choice_type,Else}}})"}),
    gen_enc_choice2(TopType,[T]);

gen_enc_choice2(TopType,[H1|T]) ->
    gen_enc_choice2(TopType,T);% skip extensionmark
gen_enc_choice2(TopType,[])  ->
    true.


decode_class('UNIVERSAL') ->
    ?UNIVERSAL;
decode_class('APPLICATION') ->
    ?APPLICATION;
decode_class('CONTEXT') ->
    ?CONTEXT;
decode_class('PRIVATE') ->
    ?PRIVATE.
 
   

%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Decode CHOICE
%%===============================================================================
%%===============================================================================
%%===============================================================================

%%=======================
%% no tag
%%=======================
gen_dec_choice(Module,TopType,undefined,CompList,_,NumChoices) ->
    asn1ct_name:new(rbCho),
    TagList = get_all_choice_tags(CompList),
    emit({indent(3),
	  {var,asn1ct_name:curr(tagList)}," = ",{asis,TagList},",",nl}),
    emit({indent(3),"RbExp = 0,",nl}),
    emit({indent(3),"RbExp1 = 0,",nl}),
    emit({indent(3),
	  "{Tag, Bytes1, ",{var,asn1ct_name:curr(rbCho)},
	  "} = asn1rt_ber:decode_tag(Bytes),",nl,nl}),
    emit({indent(3),
	  "case asn1rt_ber:check_if_valid_tag(",{var,asn1ct_name:curr(tag)},",",
	  {var,asn1ct_name:curr(tagList)},") of",nl}),
    asn1ct_name:new(tagList),
    asn1ct_name:new(rbCho),
    asn1ct_name:new(choTags),
    gen_dec_choice_cases(Module,TopType,CompList,TagList),
    emit({indent(6), {var,asn1ct_name:curr(else)}," -> ",nl,
	  indent(9),"exit({error,{asn1,{invalid_choice_tag,",{var,asn1ct_name:curr(else)},"}}})"}),
    emit({nl,"   end"}),
    asn1ct_name:new(tag),
    asn1ct_name:new(else);

%%=======================
%% EXPLICIT tag
%%=======================
gen_dec_choice(Module,TopType,{tag,Class,TagNo,'EXPLICIT'},CompList,_,NumChoices) ->
    asn1ct_name:delete(bytes),
    asn1ct_name:new(rbCho),
    TagList = get_all_choice_tags(CompList),
    emit({indent(3),
	  {var,asn1ct_name:curr(tagList)}," = ",{asis,TagList},",",nl,nl}),

    emit({"   {",{var,asn1ct_name:next(bytes)},", RbExp} = asn1rt_ber:check_tag({",
	  {asis,decode_class(Class)},",",
	  {asis,?CONSTRUCTED},",",
	  {asis,TagNo},"}, {ImplTagClass, ImplTagNo}, ",{var,asn1ct_name:curr(bytes)},"), ",nl}),
    asn1ct_name:new(bytes),
    
    emit({"   {{",{var,asn1ct_name:curr(len)},", ",{var,asn1ct_name:next(bytes)},
	  "}, RbExp1} = asn1rt_ber:decode_length(",{var,asn1ct_name:curr(bytes)},"),",nl}),
    asn1ct_name:new(len),
    asn1ct_name:new(bytes),

    emit({indent(3),
	  "{Tag, ",{var,asn1ct_name:next(bytes)},", ",{var,asn1ct_name:curr(rbCho)},
	  "} = asn1rt_ber:decode_tag(",{var,asn1ct_name:curr(bytes)},"),",nl,nl}),
    emit({indent(3),
	  "case asn1rt_ber:check_if_valid_tag(",{var,asn1ct_name:curr(tag)},",",
	  {var,asn1ct_name:curr(tagList)},") of",nl}),
    asn1ct_name:new(bytes),
    asn1ct_name:new(tagList),
    asn1ct_name:new(rbCho),
    asn1ct_name:new(choTags),
    gen_dec_choice_cases(Module,TopType,CompList,TagList),
    emit({indent(6), {var,asn1ct_name:curr(else)}," -> ",nl,
	  indent(9),"exit({error,{asn1,{invalid_choice_tag,",{var,asn1ct_name:curr(else)},"}}})"}),
    emit({nl,"   end"}),
    asn1ct_name:new(tag),
    asn1ct_name:new(else).



gen_dec_choice_cases(Module,TopType,[],List) ->
    ok;
gen_dec_choice_cases(Module,TopType,[H|T],List) ->
    case H of
	{'EXTENSIONMARK', _, _} ->
	    gen_dec_choice_cases(Module,TopType,T,List);
	_ ->
	    Name = H#'ComponentType'.name,
	    Type = H#'ComponentType'.typespec,
	    TypeDef = Type#type.def,    
	    Tags = H#'ComponentType'.tags,
	    
	    emit({indent(6),"'",Name,"' ->",nl}),
	    case Type#type.def of
		{'CHOICE',NewCompList} -> 
		    gen_dec_choice_cases_choice(Module,TopType,H);
		{typereference,_,TRName} -> 
		    gen_dec_choice_cases_typeref(Module,TopType,H);
		{'Externaltypereference',_,XModule,XName} -> 
		    gen_dec_choice_cases_typeref(Module,TopType,H);
		_ ->
		    gen_dec_choice_cases_type(Module,TopType,H)
	    end,
	    gen_dec_choice_cases(Module,TopType,T,List)
    end.



gen_dec_choice_cases_choice(Module,TopType,H) ->
    Cname = H#'ComponentType'.name,
    asn1ct_name:new(rbCho),
    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
    emit({"'dec_",TopType,"_",Cname,"'(Bytes,Telltype),",nl}),
    emit({indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + "}),
    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl}).


gen_dec_choice_cases_typeref(ThisModule,TopType,H) ->
    Cname = H#'ComponentType'.name,
    Type = H#'ComponentType'.typespec,

    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),
    {ModuleStr, Name} = case Type#type.def of
			 {typereference, _, Name} ->
			     {"", Name};
			 {'Externaltypereference',_,XModule,Name} ->
			       {lists:concat(["'",XModule,"':"]), Name}
		     end,
    case TopEnc_type of
	universal ->
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({ModuleStr,"'dec_",Name,"'(Bytes,Telltype),",nl}),
	    emit({indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + "}),
	    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl});
	implicit ->
	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({ModuleStr,"'dec_",Name,"'(Bytes,Telltype,{",Top_class,",",Top_tagNr,"}),",nl}),
	    emit({indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + "}),
	    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl});
	explicit ->
	    emit(indent(6)),
	    decode_length_choice(),
	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    emit({ModuleStr,"'dec_",Name,"'(",{var,asn1ct_name:curr(bytes)},",Telltype),",nl}),
	    emit({indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + RbCho1 + "}),
	    asn1ct_name:delete(bytes),
	    asn1ct_name:delete(rbCho),
	    emit({{var,asn1ct_name:curr(rbCho)}," + "}),
	    asn1ct_name:new(rbCho),
	    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl}),
	    asn1ct_name:new(rbCho)
    end.

gen_dec_choice_cases_type(Module,TopType,H) ->
    Cname = H#'ComponentType'.name,
    Type = H#'ComponentType'.typespec,
    {Enc_type, Class, Number} = asn1ct_gen_ber:check_tag(Type#type.tag, Type#type.def),
    emit({indent(9),
	  "{Class, ",{var,asn1ct_name:curr(form)},", TagNumber} = ",
	  {var,asn1ct_name:curr(tag)},",",nl}),
    case Enc_type of
	universal ->
	    case Type#type.def of
	       X when atom(X) ->
		    emit(indent(6)),
		    decode_length_choice(),
		    asn1ct_name:new(rbCho),
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
		    asn1ct_name:delete(bytes),
		    emit({",",nl,indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + RbCho1 + "}),
		    asn1ct_name:delete(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)}," + "}),
		    asn1ct_name:new(rbCho),
		    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl});
		{X,List} when atom(X) ->
		    asn1ct_name:new(rbCho),
		    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
		    asn1ct_name:delete(bytes),
		    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
		    asn1ct_name:new(bytes),
		    emit({",",nl,indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + "}),
		    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl});
		X ->
		    exit({error,{asn1,{unrecognized_type,X}}})
	    end;

	implicit ->
	    emit(indent(6)),
	    decode_length_choice(),
	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    gen_dec_line(TopType,Cname,Type,pos,Enc_type),
	    asn1ct_name:delete(bytes),
	    emit({",",nl,indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + RbCho1 + "}),
	    asn1ct_name:delete(rbCho),
	    emit({{var,asn1ct_name:curr(rbCho)}," + "}),
	    asn1ct_name:new(rbCho),
	    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl});
	explicit ->
	    emit(indent(6)),
	    decode_length_choice(),
	    asn1ct_name:new(rbCho),
	    emit({indent(9),"{Dec, Rest, ",{var,asn1ct_name:curr(rbCho)},"} = "}),
	    gen_dec_line(TopType,Cname,Type,pos,Enc_type), 
	    asn1ct_name:delete(bytes),
	    emit({",",nl,indent(9),"{{",{asis,Cname},", Dec}, Rest, RbExp + RbExp1 + RbCho1 + "}),
	    asn1ct_name:delete(rbCho),
	    emit({{var,asn1ct_name:curr(rbCho)}," + "}),
	    asn1ct_name:new(rbCho),
	    emit({{var,asn1ct_name:curr(rbCho)},"};",nl,nl})
    end.
    


decode_length_choice() ->
    emit({"   {{",{var,asn1ct_name:curr(len)},",",
		   {var,asn1ct_name:next(bytes)},"},",
		    {var,asn1ct_name:next(rbCho)},"} = asn1rt_ber:decode_length(",
		    {var,asn1ct_name:curr(bytes)},"),",nl}),
    asn1ct_name:new(bytes),
    asn1ct_name:new(len),
    asn1ct_name:new(rbCho).
 





get_all_choice_tags(ComponentTypeList) ->
    get_all_choice_tags(ComponentTypeList,[]).

get_all_choice_tags([],TagList) ->
    TagList;
get_all_choice_tags([H|T],TagList)  ->
    case H of
	{'EXTENSIONMARK', _, _} ->
	    get_all_choice_tags(T,TagList);
	_ ->
	    Tags = H#'ComponentType'.tags,
	    Type = H#'ComponentType'.typespec,
	    get_all_choice_tags(T, TagList ++ [{H#'ComponentType'.name, Tags}])
    end.









%%===============================================================================
%%===============================================================================
%%===============================================================================
%% Generate the encode/decode code 
%%===============================================================================
%%===============================================================================
%%===============================================================================
gen_enc_line(TopType,Cname,Type,Element,Indent,Att) when list(Element) ->
    IndDeep = indent(Indent),
    InnerType = asn1ct_gen:get_inner(Type#type.def),

    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),

    case asn1ct_gen:type(InnerType) of
	#'Externaltypereference'{module=EMod,type=EType} ->
	    case TopEnc_type of
		universal ->
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit(IndDeep)
		    end,
		    emit({"'",EMod,"':'enc_",EType,"'(",Element,")"});

		implicit ->
		    emit({IndDeep,{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'",EMod,"':'enc_",EType,"'(",Element,")",com,nl}),
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit({IndDeep})
		    end,
		    emit({"asn1rt_ber:change_tag({",
			  {asis,Top_class},",",
			  {asis,Top_tagNr},"}, ",
			  {var,asn1ct_name:curr(encTmp)},")"}),
		    asn1ct_name:new(encTmp);

		explicit ->
%		    emit({IndDeep,{var,asn1ct_name:curr(tag)}," = "}),
%		    emit({"asn1rt_ber:encode_tag_val({",
%			  {asis,Top_class},",",
%			  {asis,?CONSTRUCTED},",",
%			  {asis,Top_tagNr},"}), ",nl}),
		    emit({IndDeep,{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'",EMod,"':'enc_",EType,"'(",Element,"),",nl}),
%		    emit({com,nl,IndDeep,{var,asn1ct_name:curr(enc)}," = "}),
%		    emit({"[asn1rt_ber:encode_length("}),
%		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
%		    emit({var,asn1ct_name:curr(encTmp)}),
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit({IndDeep})
		    end,
		    emit({"asn1rt_ber:encode_tag_val({",
			  {asis,Top_class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Top_tagNr},"}) ++ "}),
		    emit({"[asn1rt_ber:encode_length("}),
		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
		    emit({var,asn1ct_name:curr(encTmp)})



%		    emit({{var,asn1ct_name:curr(tag)}," ++ ",{var,asn1ct_name:curr(enc)}})
	    end;


	{Form,user} ->
	    case TopEnc_type of
		universal ->
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit(IndDeep)
		    end,
		    emit({"'enc_",InnerType,"'(",Element,")"});

		implicit ->
		    emit({IndDeep,{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'enc_",InnerType,"'(",Element,")",com,nl}),
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit({IndDeep})
		    end,
		    emit({"asn1rt_ber:change_tag({",
			  {asis,Top_class},",",
			  {asis,Top_tagNr},"}, ",
			  {var,asn1ct_name:curr(encTmp)},")"}),
		    asn1ct_name:new(encTmp);

		explicit ->
%		    emit({IndDeep,{var,asn1ct_name:curr(tag)}," = "}),
%		    emit({"asn1rt_ber:encode_tag_val({",
%			  {asis,Top_class},",",
%			  {asis,?CONSTRUCTED},",",
%			  {asis,Top_tagNr},"}), ",nl}),
		    emit({IndDeep,{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'enc_",InnerType,"'(",Element,"),",nl}),
%		    emit({com,nl,IndDeep,{var,asn1ct_name:curr(enc)}," = "}),
%		    emit({"[asn1rt_ber:encode_length("}),
%		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
%		    emit({var,asn1ct_name:curr(encTmp)}),
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit({IndDeep})
		    end,

		    emit({"asn1rt_ber:encode_tag_val({",
			  {asis,Top_class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Top_tagNr},"}) ++ "}),
		    emit({"[asn1rt_ber:encode_length("}),
		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
		    emit({var,asn1ct_name:curr(encTmp)})



%		    emit({{var,asn1ct_name:curr(tag)}," ++ ",{var,asn1ct_name:curr(enc)}})
	    end;


	{notype,_} ->
	    emit({"'enc_",
		  InnerType,"'(",Element,")"});


	{primitive,bif} ->
	    {Enc_type, Class, Number} = asn1ct_gen_ber:check_tag(Type#type.tag, Type#type.def),
	    case Enc_type of
		universal ->
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit(IndDeep)
		    end,
		    asn1ct_gen_ber:gen_encode_prim(ber,Type,tag,Element);

		implicit ->
%		    emit({IndDeep,{var,asn1ct_name:curr(tag)}," = "}),
%		    emit({"asn1rt_ber:encode_tag_val({",
%			  {asis,Class},",",
%			  {asis,?PRIMITIVE},",",
%			  {asis,Number},"}), ",nl}),
%		    emit({IndDeep,{var,asn1ct_name:curr(enc)}," = "}),
%		    asn1ct_gen_ber:gen_encode_prim(ber,Type,notag,Element),
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit({IndDeep})
		    end,
		    emit({"asn1rt_ber:encode_tag_val({",
			  {asis,Class},",",
			  {asis,?PRIMITIVE},",",
			  {asis,Number},"}) ++ "}),
		    asn1ct_gen_ber:gen_encode_prim(ber,Type,notag,Element);

%		    emit({{var,asn1ct_name:curr(tag)}," ++ ",{var,asn1ct_name:curr(enc)}});

		explicit ->
%		    emit({IndDeep,{var,asn1ct_name:curr(tag)}," = "}),
%		    emit({"asn1rt_ber:encode_tag_val({",
%			  {asis,Class},",",
%			  {asis,?CONSTRUCTED},",",
%			  {asis,Number},"}), ",nl}),
		    emit({IndDeep,{var,asn1ct_name:curr(encTmp)}," = "}),
		    asn1ct_gen_ber:gen_encode_prim(ber,Type,tag,Element),
		    emit({com,nl}),
%		    emit({com,nl,IndDeep,{var,asn1ct_name:curr(enc)}," = "}),
%		    emit({"[asn1rt_ber:encode_length("}),
%		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
%		    emit({{var,asn1ct_name:curr(encTmp)},com,nl}),
		    case Att of
			true ->
			    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
			_ ->
			    emit({IndDeep})
		    end,
		    emit({"asn1rt_ber:encode_tag_val({",
			  {asis,Class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Number},"}) ++ "}),
		    emit({"[asn1rt_ber:encode_length("}),
		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
		    emit({{var,asn1ct_name:curr(encTmp)}})


%		    emit({{var,asn1ct_name:curr(tag)}," ++ ",{var,asn1ct_name:curr(enc)}})
	    end;
	    

	{constructed,bif} ->
	    NewTypename = lists:concat([TopType,"_",Cname]),
	    case Att of
		true ->
		    emit({IndDeep,{var,asn1ct_name:curr(att)}," = "});
		_ ->
		    emit(IndDeep)
	    end,
	    emit({"'enc_",NewTypename,"'(",Element,")"})
    end.




gen_dec_line(TopType,Cname,Type,Pos,Enc_type)  ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    BytesVar = asn1ct_gen_ber:mk_var(asn1ct_name:curr(bytes)),
    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),
    case asn1ct_gen:type(InnerType) of
	#'Externaltypereference'{module=Mod,type=EType} ->
	    case TopEnc_type of
		universal ->
		    emit({"'",Mod,"':'dec_",EType,"'(",{var,asn1ct_name:curr(bytes)},
			  ",Telltype)"});
		implicit ->
		    emit({"'",Mod,"':'dec_",EType,"'(",{var,asn1ct_name:curr(bytes)},
			  ",Telltype, {",Top_class,",",Top_tagNr,"})"});
		explicit ->
		    emit({"'",Mod,"':'dec_",EType,"'(",{var,asn1ct_name:curr(bytes)},
			  ",Telltype)"})
	    end;

	{primitive,bif} ->
	    case Enc_type of
		universal ->
		    asn1ct_name:delete(len),
		    Length = asn1ct_gen_ber:mk_var(asn1ct_name:curr(len)),
		    asn1ct_name:new(len),
		    asn1ct_gen_ber:gen_dec_prim(ber,Type,BytesVar,notag,Length,?PRIMITIVE);
		implicit ->
		    asn1ct_name:delete(len),
		    Length = asn1ct_gen_ber:mk_var(asn1ct_name:curr(len)),
		    asn1ct_name:new(len),
		    asn1ct_gen_ber:gen_dec_prim(ber,Type,BytesVar,notag,Length,?PRIMITIVE);
		explicit ->
		    asn1ct_gen_ber:gen_dec_prim(ber,Type,BytesVar,tag,no_length,?CONSTRUCTED)
	    end;

	{Form,user} ->
	    case TopEnc_type of
		universal ->
		    emit({"'dec_",InnerType,"'(",{var,asn1ct_name:curr(bytes)},",Telltype)"});
		implicit ->
		    emit({"'dec_",InnerType,"'(",{var,asn1ct_name:curr(bytes)}, 
			  ",Telltype, {",Top_class,",",Top_tagNr,"})"});
		explicit ->
		    emit({"'dec_",InnerType,"'(",{var,asn1ct_name:curr(bytes)},",Telltype)"})
	    end;

	{notype,_} ->
	    emit({"'dec_",InnerType,"'(",
		  {var,asn1ct_name:curr(bytes)},
		  ",Telltype)"});

	{constructed,bif} ->
	    asn1ct_name:new(form),
	    NewTypename = lists:concat([TopType,"_",Cname]),
	    emit({"'dec_",NewTypename,"'(",{var,asn1ct_name:curr(bytes)},",Telltype)"})	    
    end.




%%===============================================================================
%%===============================================================================
%%===============================================================================
%%  Encode/decode tag and length
%%===============================================================================
%%===============================================================================
%%===============================================================================

enc_tag(TagIn,TheTag) ->
    {Tag, Class, Number} = asn1ct_gen_ber:check_tag(TagIn, TheTag),
    emit({"asn1rt_ber:encode_tag_val({",
	  {asis,Class},",",
	  {asis,?CONSTRUCTED},",",
	  {asis,Number},"}), ",nl}).
    
enc_tag_seq(TagIn) ->
    enc_tag(TagIn,'SEQUENCE').

enc_tag_set(TagIn) ->
    enc_tag(TagIn,'SET').


dec_tag(TagIn,TheTag) ->
    {Tag, Class, Number} = asn1ct_gen_ber:check_tag(TagIn, TheTag),
    emit({"   {",{var,asn1ct_name:next(bytes)},",",{var,asn1ct_name:curr(rb)},
	  "} = asn1rt_ber:check_tag({",
	  {asis,Class},",",
	  {asis,?CONSTRUCTED},",",
	  {asis,Number},"}, {ImplTagClass, ImplTagNo}, ",
	  {var,asn1ct_name:curr(bytes)},"), ",nl}).

dec_tag_seq(TagIn) ->
    dec_tag(TagIn,'SEQUENCE').


dec_tag_set(TagIn) ->
    dec_tag(TagIn,'SET').


%%===============================================================================
%%===============================================================================
%%===============================================================================
%% General and special help functions (not exported)
%%===============================================================================
%%===============================================================================
%%===============================================================================

get_top_enc_type(Tag) ->
    case Tag of
	undefined ->
	    {universal,?UNIVERSAL,undefined};
	{tag, 'UNIVERSAL', TagNr, 'IMPLICIT'} ->
	    {universal,?UNIVERSAL,'IMPLICIT'};
	{tag, 'UNIVERSAL', TagNr, 'EXPLICIT'} ->
	    {universal,?UNIVERSAL,'EXPLICIT'};
	{tag, 'APPLICATION', TagNr, 'IMPLICIT'} ->
	    {implicit,?APPLICATION,TagNr};
	{tag, 'APPLICATION', TagNr, 'EXPLICIT'} ->
	    {explicit,?APPLICATION,TagNr};
	{tag, 'PRIVATE', TagNr, 'IMPLICIT'} ->
	    {implicit,?PRIVATE,TagNr};
	{tag, 'PRIVATE', TagNr, 'EXPLICIT'} ->
	    {explicit,?PRIVATE,TagNr};
	{tag, 'CONTEXT', TagNr, 'IMPLICIT'} ->
	    {implicit,?CONTEXT,TagNr};
	{tag, 'CONTEXT', TagNr, 'EXPLICIT'} ->
	    {explicit,?CONTEXT,TagNr};
	UnknownTag ->
	    exit({error,{asn1,{not_recognized_tag,UnknownTag}}})
    end.



indent(0) -> "";
indent(1) -> " ";
indent(2) -> "  ";
indent(3) -> "   ";
indent(4) -> "    ";
indent(5) -> "     ";
indent(6) -> "      ";
indent(7) -> "       ";
indent(8) -> "        ";
indent(9) -> "         ";
indent(10) -> "          ";
indent(11) -> "           ";
indent(12) -> "            ";
indent(13) -> "             ";
indent(14) -> "              ";
indent(15) -> "               ";
indent(16) -> "                ";
indent(17) -> "                 ";
indent(18) -> "                  ";
indent(19) -> "                   ";
indent(20) -> "                    ";
indent(21) -> "                     ";
indent(Else) -> "                      ".



list_len([H,T1|T]) ->
    emit({"length(lists:flatten(",{var,H},")) + "}),
    list_len([T1|T]);
list_len([H|T]) ->
    emit({"length(lists:flatten(",{var,H},"))"}),
    list_len(T);
list_len([]) ->
    true.


mkvlist([H,T1|T]) ->
    emit({{var,H},", "}),
    mkvlist([T1|T]);
mkvlist([H|T]) ->
    emit({{var,H}}),
    mkvlist(T);
mkvlist([]) ->
    true.

mkcnamelist(L) ->
    mkcnamelist(L,[]).

mkcnamelist([{'ComponentType',_Pos,Name,_Type,_Prop,Tags}|Rest],Acc) ->
    mkcnamelist(Rest,[Name|Acc]);
mkcnamelist([H|T],Acc) ->
    mkcnamelist(T,Acc);
mkcnamelist([],Acc) ->
    lists:reverse(Acc).

extensible(L) ->
    extensible(L,1).

extensible([{'EXTENSIONMARK',_,_}|Rest],Pos) ->
    {true,Pos};
extensible([H|Rest],Pos) ->
    extensible(Rest,Pos+1);
extensible([],_Pos) ->
    false.

gen_dec_extension_value(Bytesvar) ->
    emit({"{Ext,",{var,asn1ct_name:next(bytes)},"} = asn1rt_ber:getext(",{var,asn1ct_name:curr(bytes)},")"}),
    asn1ct_name:new(bytes).


optionals(L) -> optionals(L,[],1).

optionals([{'EXTENSIONMARK',_,_}|Rest],Acc,Pos) ->
    optionals(Rest,Acc,Pos); % optionals in extension are currently not handled
optionals([{'ComponentType',_Pos,Name,_Type,'OPTIONAL',Tags}|Rest],Acc,Pos) ->
		 optionals(Rest,[{Name,Pos}|Acc],Pos+1);

optionals([{'ComponentType',_Pos,Name,_Type,'DEFAULT',Tags}|Rest],Acc,Pos) ->
		 exit({nyi,'DEFAULT'});
optionals([{'ComponentType',_Pos,_Name,_Type,_,Tags}|Rest],Acc,Pos) ->
		 optionals(Rest,Acc,Pos+1);
optionals([],Acc,_) ->
    lists:reverse(Acc).




 
print_attribute_comment(InnerType,Pos,OptDef,Def,Ind) ->
    emit({nl,indent(Ind),"%%-------------------------------------------------"}),
    case InnerType of
	{typereference,_,Name} -> 
	    emit({nl,indent(Ind),"%% attribute number ",Pos," with type ",InnerType});
	{'Externaltypereference',_,XModule,Name} -> 
	    emit({nl,indent(Ind),"%% attribute number ",Pos,"   External ",XModule,":",Name});
	_ ->
	    emit({nl,indent(Ind),"%% attribute number ",Pos," with type ",InnerType})
    end,
    case {OptDef,Def} of
	{"",_} ->
	    continue;
	{" DEFAULT = ",""} ->
	    emit({" ",OptDef," []"});
	{" DEFAULT = ",_} ->
	    emit({" ",OptDef," ",Def});
	{" OPTIONAL",_} ->
	    emit({" ",OptDef})
    end,
    emit({nl,indent(Ind),"%%-------------------------------------------------",nl}).





