%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : asn1ct_constructed_per.erl
%%% Author  : Kenneth Lundin <kenneth@erix.ericsson.se>
%%% Purpose : 
%%% Created :  6 Mar 1997 by Kenneth Lundin <kenneth@erix.ericsson.se>

-module(asn1ct_constructed_per).
-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/london/11 $ ').

-export([gen_encode_constructed_notag/2,gen_encode_choice_notag/2]).
-export([gen_decode_constructed_notag/2,gen_decode_choice_notag/2]).
-export([gen_encode_seqof/3,gen_decode_seqof/3]).
-include("asn1_records.hrl").
%-compile(export_all).

-import(asn1ct_gen, [emit/1,demit/1]).


%% ENCODE GENERATOR FOR SEQUENCE TYPE  ** **********


gen_encode_constructed_notag(Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(term),
    asn1ct_name:new(bytes),
    {_SeqOrSet,CompList} = D#type.def,
    case Optionals = optionals(CompList) of
	[] ->
	    emit({{var,asn1ct_name:next(val)}," = asn1rt_per:list_to_record(",
		  {asis,Typename},", ",{var,asn1ct_name:curr(val)},"),",nl});
	_ ->
	    emit({"{",{var,asn1ct_name:next(val)},
		  ",Opt} = asn1rt_per:fixoptionals(",
		  {asis,Optionals},",",
		  {var,asn1ct_name:curr(val)},"),",nl})
    end,
    asn1ct_name:new(val),
    emit({"[",nl}),
    case (extensible(CompList)) of
	{true,Pos} -> 
	    emit({"asn1rt_per:setext(",{asis,Pos-1},",size(",
		  {var,asn1ct_name:curr(val)},")-1)"}),
	    emit({",",nl});    
	_ -> true
    end,
    case optionals(CompList) of
	[] -> true;
	Optionals -> 
	    emit({"asn1rt_per:setoptionals(Opt),",nl})
    end,
    gen_enc_components_call(Typename,CompList,1),
    emit({"].",nl}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% generate decode function for SEQUENCE and SET
%%
gen_decode_constructed_notag(Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    {_SeqorSet,CompList} = D#type.def,
    case (extensible(CompList)) of
	{true,Pos} -> 
	    gen_dec_extension_value("Bytes"),
	    emit({",",nl});    
	_ -> true
    end,
    case optionals(CompList) of
	[] -> true;
	Optionals -> 
	    Bcurr = asn1ct_name:curr(bytes),
	    Bnext = asn1ct_name:next(bytes),
	    emit({"{Opt,",{var,Bnext},"} = asn1rt_per:getoptionals(",{var,Bcurr},",",{asis,Optionals},",",{asis,length(mkcnamelist(CompList))},")"}),
	    asn1ct_name:new(bytes),
	    emit({",",nl})    
    end,
    gen_dec_components_call(Typename,CompList,1),
    case asn1ct_name:all(term) of
	[] -> true; % no components at all
	_ -> emit({com,nl})
    end,
% we don't return named lists any more   Cnames = mkcnamelist(CompList), 
    demit({"Result = "}), %dbg
    %% return value as record
    emit({"{{",{asis,Typename}}),
    mkvlist(asn1ct_name:all(term)),
    case extensible(CompList) of
	{true,_} ->
	    emit(",undefined");
	_ -> true
    end,
    emit({"},",{var,asn1ct_name:curr(bytes)},"}"}),
    emit({".",nl}).


%% ENCODE GENERATOR FOR THE CHOICE TYPE *******
%% assume Val = {Alternative,AltType}
%% generate
%%[
%% asn1rt_per:set_choice(element(1,Val),Altnum,Altlist,ext),
%%case element(1,Val) of
%%    alt1 ->
%%	encode_alt1(element(2,Val));
%%    alt2 ->
%%	encode_alt2(element(2,Val))
%%end
%%].

gen_encode_choice_notag(Typename,D) when record(D,type) ->
    {'CHOICE',CompList} = D#type.def,
    emit({"[",nl}),
    case (extensible(CompList)) of
	{true,Pos} -> 
	    gen_enc_choice(Typename,CompList,ext);
	_ -> 
	    gen_enc_choice(Typename,CompList,noext)
    end,
    emit({nl,"].",nl}).

gen_decode_choice_notag(Typename,D) when record(D,type) ->
    asn1ct_name:start(),
    asn1ct_name:new(bytes),
    {'CHOICE',CompList} = D#type.def,
    NumChoices = length(mkcnamelist(CompList)),
    case (extensible(CompList)) of
	{true,Pos} -> 
	    gen_dec_choice(Typename,CompList,ext,NumChoices);
	_ -> 
	    gen_dec_choice(Typename,CompList,noext,NumChoices)
    end,
    emit({".",nl}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Encode generator for SEQUENCE OF type


gen_encode_seqof(Typename,SeqOrSetOf,D) when record(D,type) ->
    asn1ct_name:start(),
% Val = [Component]
% asn1rt_per:encode_length(length(Val)),
% lists:
    {_SeqOrSetOf,ComponentType} = D#type.def,
    emit({"[",nl}),
%Note this must be extended to handle range
    emit({nl,"asn1rt_per:encode_length(undefined,length(Val)),",nl}),
    emit({"'enc_",Typename,
	      "_components'(Val,[])"}),
    emit({nl,"].",nl}),
    gen_encode_seqof_components(Typename,SeqOrSetOf,ComponentType).

gen_decode_seqof(Typename,SeqOrSetOf,D) when record(D,type) ->
    asn1ct_name:start(),
% Val = [Component]
% asn1rt_per:encode_length(length(Val)),
% lists:
    {_SeqOrSetOf,ComponentType} = D#type.def,
    emit({nl,"{Num,Bytes1} = asn1rt_per:decode_length(Bytes),",nl}),
    emit({"'dec_",Typename,
	      "_components'(Num,Bytes1,Telltype,[]).",nl}),
    gen_decode_seqof_components(Typename,SeqOrSetOf,ComponentType).

gen_encode_seqof_components(Typename,SeqOrSetOf,Cont) ->
    emit({"'enc_",Typename,"_components'([],Acc) -> lists:reverse(Acc);",nl,nl}),
    emit({"'enc_",Typename,"_components'([H|T],Acc) ->",nl}),
    emit({"'enc_",Typename,"_components'(T,"}), 
    emit({"["}),
    %% the component encoder
    Constructed_Suffix = case SeqOrSetOf of
			     'SET OF' -> "_SETOF";
			     'SEQUENCE OF' -> "_SEQOF"
			 end,
    Conttype = asn1ct_gen:get_inner(Cont#type.def),
    case asn1ct_gen:type(Conttype) of
	{primitive,bif} ->
	    asn1ct_gen_per:gen_encode_prim(per,Cont,false,"H");
	{constructed,bif} ->
	    NewTypename = list_to_atom(lists:concat([Typename,Constructed_Suffix])),
	    emit({"'enc_",NewTypename,"'(H)",nl,nl});
	{Form,user} ->
	    emit({"'enc_",Conttype,"'(H)",nl,nl});
	#'Externaltypereference'{module=EMod,type=EType} ->
	    emit({"'",EMod,"':'enc_",EType,"'(H)",nl,nl});
	_ ->
	    emit({"'enc_",Conttype,"'(H)",nl,nl})
    end,
    emit({" | Acc]).",nl}).

gen_decode_seqof_components(Typename,SeqOrSetOf,Cont) ->
    emit({"'dec_",Typename,"_components'(0,Bytes,Telltype,Acc) ->",nl,"{lists:reverse(Acc),Bytes};",nl}),
    emit({"'dec_",Typename,"_components'(Num,Bytes,Telltype,Acc) ->",nl}),
    emit({"{Term,Remain} = "}),
    Constructed_Suffix = case SeqOrSetOf of
			     'SET OF' -> "_SETOF";
			     'SEQUENCE OF' -> "_SEQOF"
			 end,
    Conttype = asn1ct_gen:get_inner(Cont#type.def),
    case asn1ct_gen:type(Conttype) of
	{primitive,bif} ->
	    asn1ct_gen_per:gen_dec_prim(per,Cont,"Bytes"),
	    emit({com,nl});
	{constructed,bif} ->
	    NewTypename = list_to_atom(lists:concat([Typename,Constructed_Suffix])),
	    emit({"'dec_",NewTypename,"'(Bytes,Telltype),",nl});
	{Form,user} ->
	    emit({"'dec_",Conttype,"'(Bytes,Telltype),",nl});
	#'Externaltypereference'{module=EMod,type=EType} ->
	    emit({"'",EMod,"':'dec_",EType,"'(Bytes,Telltype),",nl});
	_ ->
	    emit({"'dec_",Conttype,"'(Bytes,Telltype),",nl})
    end,
    emit({"'dec_",Typename,"_components'(Num-1,Remain,Telltype,[Term|Acc]).",nl}). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% General and special help functions (not exported)

mkvlist([H|T]) ->
    emit(","),
    mkvlist2([H|T]);
mkvlist([]) ->
    true.
mkvlist2([H,T1|T]) ->
    emit({{var,H},","}),
    mkvlist2([T1|T]);
mkvlist2([H|T]) ->
    emit({{var,H}}),
    mkvlist2(T);
mkvlist2([]) ->
    true.

mkcnamelist(L) ->
    mkcnamelist(L,[]).

mkcnamelist([#'ComponentType'{name=Name}|Rest],Acc) ->
    mkcnamelist(Rest,[Name|Acc]);
mkcnamelist([H|T],Acc) ->
    mkcnamelist(T,Acc);
mkcnamelist([],Acc) ->
    lists:reverse(Acc).

extensible(L) ->
    extensible(L,1).

extensible([{'EXTENSIONMARK',_,_}|Rest],Pos) ->
    {true,Pos};
extensible([H|Rest],Pos) ->
    extensible(Rest,Pos+1);
extensible([],_Pos) ->
    false.


gen_dec_extension_value(Bytesvar) ->
    emit({"{Ext,",{var,asn1ct_name:next(bytes)},"} = asn1rt_per:getext(",{var,asn1ct_name:curr(bytes)},")"}),
    asn1ct_name:new(bytes).


optionals(L) -> optionals(L,[],1).

optionals([{'EXTENSIONMARK',_,_}|Rest],Acc,Pos) ->
    optionals(Rest,Acc,Pos); % optionals in extension are currently not handled
optionals([#'ComponentType'{name=Name,prop='OPTIONAL'}|Rest],Acc,Pos) ->
		 optionals(Rest,[{Name,Pos}|Acc],Pos+1);
optionals([#'ComponentType'{name=Name,prop={'DEFAULT',_}}|Rest],Acc,Pos) ->
		 optionals(Rest,[{Name,Pos}|Acc],Pos+1);
optionals([#'ComponentType'{}|Rest],Acc,Pos) ->
		 optionals(Rest,Acc,Pos+1);
optionals([],Acc,_) ->
    lists:reverse(Acc).



gen_enc_components_call(TopType,[#'ComponentType'{name=Cname,typespec=Type,prop=Prop}|Rest],Num) ->
    case Prop of
	'OPTIONAL' ->
	    gen_enc_component_optional(TopType,Cname,Type,Num);
	{'DEFAULT',DefVal} ->
	    gen_enc_component_default(TopType,Cname,Type,Num);
	_ ->
	    gen_enc_component_mandatory(TopType,Cname,Type,Num)
    end,
    case Rest of
	[] ->
	    emit(nl);
	_ ->
	    emit({com,nl}),
	    gen_enc_components_call(TopType,Rest,Num+1)
    end;
gen_enc_components_call(TopType,[{'EXTENSIONMARK',_Pos,_}|Rest],Num) ->
    emit({"[]",tab,"%EXTENSIONMARK",nl}),
    gen_enc_components_call(TopType,Rest,Num+1);
gen_enc_components_call(_TopType,[],_Num) ->
	true.

gen_enc_component_default(TopType,Cname,Type,Pos) ->
    Element = io_lib:format("asn1rt_per:cindex(~w,Val1,~w)",[Pos+1,Cname]),
    emit({"case ",Element," of",nl}),
    emit({"asn1_DEFAULT -> [];",nl}),
    emit({"_ ->",nl}),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,"%% attribute number ",Pos," with type ",
	      InnerType,nl}),
    gen_enc_line(TopType,Cname,Type,Element),
    emit({nl,"end"}).

gen_enc_component_optional(TopType,Cname,Type,Pos) ->
    Element = io_lib:format("asn1rt_per:cindex(~w,Val1,~w)",[Pos+1,Cname]),
    emit({"case ",Element," of",nl}),
    emit({"asn1_NOVALUE -> [];",nl}),
    emit({"_ ->",nl}),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,"%% attribute number ",Pos," with type ",
	      InnerType,nl}),
    gen_enc_line(TopType,Cname,Type,Element),
    emit({nl,"end"}).

gen_enc_component_mandatory(TopType,Cname,Type,Pos) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,"%% attribute number ",Pos," with type ",
	      InnerType,nl}),
    gen_enc_line(TopType,Cname,Type,Pos).

gen_enc_line(TopType,Cname,Type,Pos) when integer(Pos) ->
    Element = io_lib:format("asn1rt_per:cindex(~w,~s,~w)",[Pos+1,asn1ct_gen_per:mk_var(asn1ct_name:curr(val)),Cname]),
    gen_enc_line(TopType,Cname,Type,Element);
gen_enc_line(TopType,Cname,Type,Element) when list(Element) ->
    Atype = asn1ct_gen:get_inner(Type#type.def),
    case asn1ct_gen:type(Atype) of
	#'Externaltypereference'{module=Mod,type=EType} ->
	    emit({"'",Mod,"':'enc_",
		  EType,"'(",Element,")"});
	{Form,user} ->
		    emit({"'enc_",
			  Atype,"'(",Element,")"});
	{notype,_} ->
		    emit({"'enc_",
			  Atype,"'(",Element,")"});
	{primitive,bif} ->
	    asn1ct_gen_per:gen_encode_prim(per,Type,false,Element);
	{constructed,bif} ->
	    NewTypename = lists:concat([TopType,"_",Cname]),
	    emit({"'enc_",
		      NewTypename,"'(",Element,")"})
    end.


gen_dec_components_call(TopType,[#'ComponentType'{name=Cname,typespec=Type,prop=Prop}|Rest],Num) ->
    asn1ct_name:new(term),
    case Prop of
	'OPTIONAL' ->
	    gen_dec_component_optional(TopType,Cname,Type,Num);
	{'DEFAULT',{_,_,DefVal}} ->
	    gen_dec_component_default(TopType,Cname,Type,Num,DefVal);
	{'DEFAULT',DefVal} ->
	    gen_dec_component_default(TopType,Cname,Type,Num,DefVal);
	_ ->
	    gen_dec_component_mandatory(TopType,Cname,Type,Num)
    end,
    case Rest of
	[] ->
	    true;
	[{'EXTENSIONMARK',_,_}|_] ->
	    true;
	_ ->
	    emit({com,nl}),
	    gen_dec_components_call(TopType,Rest,Num+1)
    end;
gen_dec_components_call(TopType,[{'EXTENSIONMARK',_Pos,_}|Rest],Num) ->
%    emit({"[]",tab,"%EXTENSIONMARK",nl}),
    gen_dec_components_call(TopType,Rest,Num+1);
gen_dec_components_call(_TopType,[],_Num) ->
	true.

gen_dec_component_default(TopType,Cname,Type,Pos,DefVal) ->
    Element = io_lib:format("element(~w,Opt)",[Pos]),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,"%% attribute number ",Pos," with type ",
	      InnerType,nl}),
    emit({"{",{var,asn1ct_name:curr(term)},",",{var,asn1ct_name:next(bytes)},"} = "}),
    emit({"case ",Element," of",nl}),
    emit({"0 -> {",{asis,DefVal},",",{var,asn1ct_name:curr(bytes)},"};",nl}),
    emit({"1 ->",nl}),
    gen_dec_line(TopType,Cname,Type,Pos),
    asn1ct_name:new(bytes),
    emit({nl,"end"}).

gen_dec_component_optional(TopType,Cname,Type,Pos) ->
    Element = io_lib:format("element(~w,Opt)",[Pos]),
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,"%% attribute number ",Pos," with type ",
	      InnerType,nl}),
    emit({"{",{var,asn1ct_name:curr(term)},",",{var,asn1ct_name:next(bytes)},"} = "}),
    emit({"case ",Element," of",nl}),
    emit({"0 -> {asn1_NOVALUE,",{var,asn1ct_name:curr(bytes)},"};",nl}),
    emit({"1 ->",nl}),
    gen_dec_line(TopType,Cname,Type,Pos),
    asn1ct_name:new(bytes),
    emit({nl,"end"}).

gen_dec_component_mandatory(TopType,Cname,Type,Pos) ->
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,"%% attribute number ",Pos," with type ",
	      InnerType,nl}),
    emit({"{",{var,asn1ct_name:curr(term)},",",{var,asn1ct_name:next(bytes)},"} = "}),
    gen_dec_line(TopType,Cname,Type,Pos),
    asn1ct_name:new(bytes).

gen_dec_line(TopType,Cname,Type,Pos)  ->
    BytesVar = asn1ct_gen_per:mk_var(asn1ct_name:curr(bytes)),
    Atype = asn1ct_gen:get_inner(Type#type.def),
    case asn1ct_gen:type(Atype) of
	#'Externaltypereference'{module=Mod,type=EType} ->
	    emit({"'",Mod,"':'dec_",EType,"'(",{var,asn1ct_name:curr(bytes)},
		  ",telltype)"});
	{primitive,bif} ->
	    asn1ct_gen_per:gen_dec_prim(per,Type,BytesVar);
	{Form,user} ->
	    emit({"'dec_",Atype,"'(",
		  {var,asn1ct_name:curr(bytes)},
		  ",telltype)"});
	{notype,_} ->
	    emit({"'dec_",Atype,"'(",
		  {var,asn1ct_name:curr(bytes)},
		  ",telltype)"});
	{constructed,bif} ->
	    NewTypename = lists:concat([TopType,"_",Cname]),
	    emit({"'dec_",
			  NewTypename,"'(",{var,asn1ct_name:curr(bytes)},
			  ",Telltype)"})
    end.

gen_enc_choice(TopType,CompList,Ext) ->
    case Ext of 
	ext ->
	    emit({"asn1rt_per:setchoiceext()"}),
	    emit({",",nl});
	_ -> true
    end,
    gen_enc_choice_tag(CompList,[],Ext),
    emit({com,nl}),
    emit({"case element(1,Val) of",nl}),
    gen_enc_choice2(TopType,CompList),
    emit({nl,"end"}).

gen_enc_choice_tag([H|T],Acc,Ext) when record(H,'ComponentType') ->
    gen_enc_choice_tag(T,[H#'ComponentType'.name|Acc],Ext);
gen_enc_choice_tag([H|T],Acc,Ext) -> % skip EXTENSIONMARK
    gen_enc_choice_tag(T,Acc,Ext);
gen_enc_choice_tag([],Acc,Ext) ->
    Length = length(Acc),
    emit({"asn1rt_per:set_choice(element(1,Val),",{asis,Length},",",
	  {asis,lists:reverse(Acc)},",",{asis,Ext},")"}),
    Length.

gen_enc_choice2(TopType,[H1,H2|T]) 
when record(H1,'ComponentType'), record(H2,'ComponentType') ->
    Cname = H1#'ComponentType'.name,
    Type = H1#'ComponentType'.typespec,
    emit({{asis,Cname}," ->",nl}),
    gen_enc_line(TopType,Cname,Type,"element(2,Val)"),
    emit({";",nl}),
    gen_enc_choice2(TopType,[H2|T]);
gen_enc_choice2(TopType,[H1,H2|T]) when record(H1,'ComponentType') ->
    gen_enc_choice2(TopType,[H1|T]); % skip extensionmark
gen_enc_choice2(TopType,[H1|T]) when record(H1,'ComponentType') ->
    Cname = H1#'ComponentType'.name,
    Type = H1#'ComponentType'.typespec,
    emit({{asis,H1#'ComponentType'.name}," ->",nl}),
    gen_enc_line(TopType,Cname,Type,"element(2,Val)"),
    gen_enc_choice2(TopType,[T]);
gen_enc_choice2(TopType,[H1|T]) ->
    gen_enc_choice2(TopType,T);% skip extensionmark
gen_enc_choice2(TopType,[])  ->
    true.

gen_dec_choice(TopType,CompList,ext,NumChoices) ->
    emit({"{Ext,",{var,asn1ct_name:curr(bytes)},"} = asn1rt_per:getbit(Bytes),",nl}),
    asn1ct_name:new(bytes),
    gen_dec_choice1(TopType,CompList,ext,NumChoices);
gen_dec_choice(TopType,CompList,noext,NumChoices) ->
    gen_dec_choice1(TopType,CompList,ext,NumChoices).

gen_dec_choice1(TopType,CompList,Ext,NumChoices) ->
    emit({"{Choice,",{var,asn1ct_name:curr(bytes)},"} = asn1rt_per:getchoice(",{var,asn1ct_name:prev(bytes)},",",NumChoices,",",Ext,"),",nl}),
    emit({"{Cname,{Val,NewBytes}} = case Choice of",nl}),
    gen_dec_choice2(TopType,CompList),
    emit({nl,"end,",nl}),
    emit({nl,"{{Cname,Val},NewBytes}"}).


gen_dec_choice2(TopType,L) ->
    gen_dec_choice2(TopType,L,0).

gen_dec_choice2(TopType,[H1,H2|T],Pos) 
when record(H1,'ComponentType'), record(H2,'ComponentType') ->
    Cname = H1#'ComponentType'.name,
    Type = H1#'ComponentType'.typespec,
    emit({Pos," -> {",{asis,Cname},",",nl}),
    gen_dec_line(TopType,Cname,Type,Pos),
    emit({"};",nl}),
    gen_dec_choice2(TopType,[H2|T],Pos+1);
gen_dec_choice2(TopType,[H1,H2|T],Pos) when record(H1,'ComponentType') ->
    gen_dec_choice2(TopType,[H1|T],Pos); % skip extensionmark
gen_dec_choice2(TopType,[H1|T],Pos) when record(H1,'ComponentType') ->
    Cname = H1#'ComponentType'.name,
    Type = H1#'ComponentType'.typespec,
    emit({Pos," -> {",{asis,Cname},",",nl}),
    gen_dec_line(TopType,Cname,Type,Pos),
    emit({"}"}),
    gen_dec_choice2(TopType,[T],Pos+1);
gen_dec_choice2(TopType,[H1|T],Pos) ->
    gen_dec_choice2(TopType,T,Pos);% skip extensionmark
gen_dec_choice2(TopType,[],_Pos)  ->
    true.
    
    


 







