%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : asn1.erl
%%% Author  : Kenneth Lundin <kenneth@erix.ericsson.se>
%%% Purpose : Runtime functions for ASN.1 (i.e encode, decode)

-module(asn1rt).
-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

%%-compile(export_all).
-export([encode/2,encode/3,decode/3]).
    
encode(Module,{Type,Term}) ->
    encode(Module,Type,Term).

encode(Module,Type,Term) ->
    case catch apply(Module,encoding_rule,[]) of
	{'EXIT',_} ->
	    {error,{asn1,{undef,Module}}};
	Rule ->
	    encode(Rule,Module,Type,Term)
    end.
	    
encode(ber,Module,Type,Term) ->
    Call = list_to_atom(lists:concat(['enc_',Type])),
    case catch apply(Module,Call,[Term]) of
	{'EXIT',undef} ->
	    {error,{asn1,{undef,Module,Call}}};
	{'EXIT',Reason} ->
	    Reason;
	{Bytes,Len} ->
	    {ok,Bytes};
	X ->
	    {ok,X}
    end;
encode(per,Module,Type,Term) ->
    Call = list_to_atom(lists:concat(['enc_',Type])),
    case catch asn1rt_per:complete(apply(Module,Call,[Term])) of
	{'EXIT',undef} ->
	    {error,{asn1,{undef,Module,Call}}};
	{'EXIT',Reason} ->
	    Reason;
	X ->
	    {ok,X}
    end.

decode(Module,Type,Bytes) ->
    Call = list_to_atom(lists:concat(['dec_',Type])),
    case catch apply(Module,Call,[Bytes,telltype]) of
	{'EXIT',undef} ->
	    {error,{asn1,{undef,Module,Call}}};
	{'EXIT',Reason} ->
	    Reason;
	{X,_Rest} ->
	    {ok,X};
	{X,_Rest,_Len} ->
	    {ok,X}
    end.




