%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(oms).

%% External exports
-export([start/2, stop/1, create/0]).

%%%-----------------------------------------------------------------
%%% This module implements the application OMS.
%%%-----------------------------------------------------------------
-behaviour(application).

start(_, []) ->
    EvaLogDir = "ex_log/",
    {ok, "default"} =
	disk_log:open([{name, "default"}, {file, EvaLogDir ++ "default.LOG"},
		       {type, wrap}, {size, {10000, 4}}]),
    case application:get_env(oms, use_snmp) of
	{ok, false} ->
	    eva_sup:start_link_log({"default", 24*3600});
	_ ->
	    %% When we use snmp, we'll have to specify where the SNMP mgr
	    %% can create eva logs (EvaLogDir), and how large the logs in
	    %% that dir can be (120K).
	    eva_sup:start_link_log_snmp({"default", 24*3600}, EvaLogDir, 120000)
    end.

stop(State) ->
    ok.

create() ->
    application:load(oms),
    file:make_dir("ex_log"),
    case application:get_env(oms, use_snmp) of
	{ok, false} ->
	    eva_sup:create_tables_log([node()]);
	_ ->
	    eva_sup:create_tables_log_snmp([node()])
    end.

