%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(ic).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('peterl@erix.ericsson.se').
%%------------------------------------------------------------
%%
%% 
%%
%%
%%    COPYRIGHT Ericsson Telecom AB 1995
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%%
%%    Module documentation:
%%    ---------------------
%%
%%    The IDL to Erlang code generator.
%%
%%
%%------------------------------------------------------------

-export([sgen/1, gen/1, gen/2, help/0, compile/3]).


%%------------------------------------------------------------
%%
%% Internal stuff
%%
%%------------------------------------------------------------

-export([filter_params/2, handle_preproc/4, do_gen/4]).

-import(icgen, [get_opt/2, 
		to_list/1, get_id2/1, 
		push_file/2, pop_file/2, sys_file/2]).


-import(lists, [foldr/3]).


-include("icforms.hrl").
-include("ic.hrl").

-include_lib("stdlib/include/erl_compile.hrl").

-export([make_erl_options/1]).			% For erlc

-export([main/3, do_scan/1, do_parse/2, do_type/2]).


%%------------------------------------------------------------
%%
%% Entry point
%%
%%------------------------------------------------------------

%% compile(AbsFileName, Outfile, Options)
%%   Compile entry point for erl_compile.

compile(File, _OutFile, Options) ->
    case gen(File, make_erl_options(Options)) of
	ok -> ok;
	Other -> Other
    end.


%% Entry for the -s switch
sgen(ArgList) ->
%%%    io:format("sgen called w ~p~n", [ArgList]),
    apply(?MODULE, gen, ArgList).


gen(File) ->
    gen(File, []).

gen(File, Opts) ->
    G = icgen:new(Opts),
    IdlFile = icgen:add_dot_idl(File),
    case get_opt(G, show_opts) of
	true ->
	    io:format("Opts: ~p~n", [icgen:which_opts(G)]);
	_ -> ok
    end,
    icgen:set_idlfile(G, IdlFile),
    case catch gen2(G, File, Opts) of
	{_, {'EXIT', R}} -> 
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    %%exit(R);
	    io:format("Fatal error : ~p~n",[R]),
	    error;
	{_, {'EXIT', _, R}} -> 
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    %%exit(R);
	    io:format("Fatal error : ~p~n",[R]),
	    error;
	{'EXIT', R} -> 
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    %%exit(R);
	    io:format("Fatal error : ~p~n",[R]),
	    error;
	{'EXIT', _, R} -> 
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    %%exit(R);
	    io:format("Fatal error : ~p~n",[R]),
	    error;
	
	%% In this case, the pragma registration 
        %% found errors so this should return error.
	error ->
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    error;
	_ -> 
	    X = icgen:return(G),
	    %io:format("~p~n",[ets:tab2list(icgen:tktab(G))]),
	    %ic_pragma:print_tab(G),
	    %io:format("Options ~p~n",[icgen:which_opts(G)]),
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    X
    end.


gen2(G, File, Opts) ->
    %%    G = icgen:new(Opts),
    %%    IdlFile = icgen:add_dot_idl(File),
    %%    icgen:set_idlfile(G, IdlFile),
    case get_opt(G, time) of
	true -> 
	    time("TOTAL", ic, main, [G, File, Opts]);
	_ -> 
	    %%    icgen:return(G).
	    case main(G, File, Opts) of
		error ->
		    error;
		_ ->
		    ok
	    end
    end.



main(G, File, Opts) ->
    print_version_str(G),
    ?ifopt(G, time, io:format("Time started  :   ~p/~p/~p ~p:~2.2.0p~n", 
			      [element(1,date()),
			       element(2, date()),
			       element(3, date()),
			       element(1, time()),
			       element(2, time())])),
    case get_opt(G, help) of
	true -> help();
	_ ->
	    S = ?ifopt2(G, time,
			time("scan ", ic, do_scan, [G]),
			do_scan(G)),
	    
	    ?ifopt(G, tokens, io:format("TOKENS: ~p~n", [S])),
	    
	    T = ?ifopt2(G, time, 
			time("parse", ic, do_parse, [G, S]),
			do_parse(G, S)),
		    ?ifopt(G, form, io:format("PARSE FORM: ~p~n", [T])),
	    
	    %% Initiate the pragma table
	    case ?ifopt2(G, time,
			 time("pragmas",ic_pragma,pragma_reg,[G,T]),
			 ic_pragma:pragma_reg(G, T)) of
		%% All pragmas were succesfully applied
			{ok,Clean} ->
		    %%ic_pragma:print_tab(G),
		    
		    case catch ?ifopt2(G, time,
				       time("type ", ic, do_type, [G, Clean]),
					       do_type(G, Clean)) of
			{'EXIT',Reason} ->
			    io:format("Error under type appliance : ~p~n",[Reason]),
			    error;
			T2 ->
			    
				    ?ifopt(G, tform, io:format("TYPE FORM: ~p~n", [T2])),
			    
			    case icgen:get_error_count(G) of
				0 ->
				    %% Check if user has sett backend option
				    case get_opt(G, be) of
					false ->
						    %% Use default backend option
					    DefaultBe = icgen:defaultBe(),
					    icgen:add_opt(G,[{be,DefaultBe}],true),
					    ?ifopt2(G, time, 
						    time("gen  ", ic, do_gen, [DefaultBe, G, File, T2]),
						    do_gen(DefaultBe, G, File, T2));
					Be ->
					    %% Use user defined backend
					    ?ifopt2(G, time, 
						    time("gen  ", ic, do_gen, [Be, G, File, T2]),
						    do_gen(Be, G, File, T2))
				    end;
				_ ->
				    ok	    %% Does not matter
			    end
		    end;
		%% Pragma failure
		error ->
		    error
	    end
end.


do_gen(erl_corba, G, File, T) ->
    ic_erlbe:do_gen(G, File, T);
do_gen(erl_genserv, G, File, T) ->
    ic_erlbe:do_gen(G, File, T);
do_gen(c_genserv, G, File, T) ->
    ic_cbe:do_gen(G, File, T);
do_gen(noc, G, File, T) ->
    ic_noc:do_gen(G, File, T);
do_gen(erl_plain, G, File, T) ->
    ic_plainbe:do_gen(G, File, T);
do_gen(c_server, G, File, T) ->
    ic_cserver:do_gen(G, File, T);
do_gen(c_client, G, File, T) ->
    ic_cbe:do_gen(G, File, T);
do_gen(_,_,_,_) -> %% No language choice
    ok.

do_scan(G) ->
    case get_opt(G, use_leex) of
	1 -> 
	    {ok, Bin} =  file:read_file(icgen:idlfile(G)),
	    icscanner:string(binary_to_list(Bin));
	2 -> 
	    {ok, Bin} =  file:read_file(icgen:idlfile(G)),
	    R = icscanner:tokens([], binary_to_list(Bin)),
	    element(5, element(2,R));
	_->
	    icscan:scan(G, icgen:idlfile(G))
    end.
    

do_parse(G, Tokens) ->
    case icparse:parse(Tokens) of
	{ok, L} -> L;
	X when element(1, X) == error -> 
	    Err = element(2, X),
	    icgen:fatal_error(G, {parse_error, element(1, Err), 
				  element(3, Err)});
	X -> exit(X)
    end.


do_type(G, Form) ->
    ictype:type_check(G, Form).
	
time(STR,M,F,A) ->
    case timer:tc(M, F, A) of
	{_, {'EXIT', R}} -> exit(R);
	{_, {'EXIT', _, R}} -> exit(R);
	{_, _X} when element(1, _X)==error -> throw(_X);
	{_T, _R} -> 
	    io:format("Time for ~s:  ~5.2f~n", [STR, _T/1000000]),
	    _R
    end.



%% Filters parameters so that only those with certain attributes are
%% seen. The filter parameter is a list of attributes that will be
%% seen, ex. [in] or [inout, out]
filter_params(Filter, Params) ->
    lists:filter(fun(P) ->
		    lists:member(get_param_attr(P#param.inout), Filter) end,
		 Params).


%% Access primitive to get the attribute name (and discard the line
%% number).
get_param_attr({A, N}) -> A.


%%
%% Fixing the preproc directives
%%
handle_preproc(G, N, line_nr, X) ->
    Id = get_id2(X),
    Flags = X#preproc.aux,
    case Flags of
	[] -> push_file(G, Id);
	_ ->
	    foldr(fun({_, _, "1"}, Gprim) -> push_file(Gprim, Id);
		     ({_, _, "2"}, Gprim) -> pop_file(Gprim, Id);
		     ({_, _, "3"}, Gprim) -> sys_file(Gprim, Id) end,
		  G, Flags)
    end;
handle_preproc(G, N, Other, X) ->
    G.



%%------------------------------------------------------------
%%
%% The help department
%%
%% 
%%
%%------------------------------------------------------------

help() ->
    io:format("No help available at the moment~n", []),
    ok.

print_version_str(G) ->
    case {get_opt(G, silent), get_opt(G, silent2)} of
	{true, _} -> ok;
	{_, true} -> ok;
	_ -> 
	    io:format("Erlang IDL compiler version ~s~n", [?COMPILERVSN])
    end.



%%
%% Converts generic compiler options to specific options.
%% 
%% Used by erlc
%%

make_erl_options(Opts) ->

    %% This way of extracting will work even if the record passed
    %% has more fields than known during compilation.

    Includes0 = Opts#options.includes,
    Defines = Opts#options.defines,
    Outdir = Opts#options.outdir,
    Warning = Opts#options.warning,
    Verbose = Opts#options.verbose,
    Specific = Opts#options.specific,
    Optimize = Opts#options.optimize,
    OutputType = Opts#options.output_type,
    Cwd = Opts#options.cwd,

    Includes1 = 
	case Opts#options.ilroot of
	    undefined ->
		Includes0;
	    Ilroot ->
		[Ilroot|Includes0]
	end,
    PreProc = 
	lists:flatten(
	  lists:map(fun(D) -> io_lib:format("-I~s ", [to_list(D)]) end, 
		    Includes1)++
	  lists:map(
	    fun ({Name, Value}) ->
		    io_lib:format("-D~s=~s ", [to_list(Name), to_list(Value)]);
		(Name) ->
		    io_lib:format("-D~s ", [to_list(Name)])
	    end,
	    Defines)),
    Options =
	case Verbose of
	    true ->  [];
	    false -> []
	end ++
	case Warning of
	    0 -> [nowarn];
	    _ -> ['Wall']
	end ++
	case Optimize of
	    0 -> [];
	    _ -> []
	end,
    
    Options++[{outdir, Outdir}, {preproc_flags, PreProc}]++Specific.


