%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(ict).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('babbis@erix.ericsson.se').

%%------------------------------------------------------------
%%
%% 
%%
%%
%%    COPYRIGHT Ericsson Telecom AB 1995
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%%
%%    Module documentation:
%%    ---------------------
%%
%%    Type Kind generation.
%%
%%
%%	FIX LIST:
%%
%%
%%------------------------------------------------------------

-include("icforms.hrl").
-include("ic.hrl").

-export([dep/1,dep/2]).
-export([main/3, do_scan/1, do_parse/2, mkDep/2]).


-import(icgen, [get_opt/2]).


%%--------------------------------------------------------------------
%%
%% IFR stuff
%%
%%
%%--------------------------------------------------------------------

-define(IFRID(G), mk_name(G, "IFR")).
-define(VARID(G), mk_name(G, "VAR")).
-define(IFRMOD, orber_ifr).







dep(File) ->
    dep(File, []).

dep(File, Opts) ->
    G = icgen:new(Opts),
    IdlFile = icgen:add_dot_idl(File),
    case get_opt(G, show_opts) of
	true ->
	    io:format("Opts: ~p~n", [icgen:which_opts(G)]);
	_ -> ok
    end,
    icgen:set_idlfile(G, IdlFile),
    case catch dep2(G, File, Opts) of
	{_, {'EXIT', R}} -> 
	    exit(R);
	{_, {'EXIT', _, R}} -> 
	    exit(R);
	{'EXIT', R} -> 
	    exit(R);
	{'EXIT', _, R} -> 
	    exit(R);
	X ->
	    icgen:free_table_space(G), %% Free space for all ETS tables
	    X
    end.



dep2(G, File, Opts) ->
    case get_opt(G, time) of
	true -> 
	    time("TOTAL", ?MODULE, main, [G, File, Opts]);
	_ -> 
	    case main(G, File, Opts) of
		error ->
		    error;
		Tree ->
		    Tree
	    end
    end.


main(G, File, Opts) ->
    print_version_str(G),
    ?ifopt(G, time, io:format("Time started  :   ~p/~p/~p ~p:~2.2.0p~n", 
			      [element(1,date()),
			       element(2, date()),
			       element(3, date()),
			       element(1, time()),
			       element(2, time())])),
    case get_opt(G, help) of
	true -> help();
	_ ->
	    S = ?ifopt2(G, time,
			time("scan ", ic, do_scan, [G]),
			do_scan(G)),
	    ?ifopt(G, tokens, io:format("TOKENS: ~p~n", [S])),
	    
	    T = ?ifopt2(G, time, 
			time("parse", ic, do_parse, [G, S]),
			do_parse(G, S)),
	    ?ifopt(G, form, io:format("PARSE FORM: ~p~n", [T])),
	    
	    case icgen:get_error_count(G) of
		0 ->
		    L = get_opt(G, be),
		    ?ifopt2(G, time, 
			    time("dep  ", tools, mkDep, [G,T]),
			    mkDep(G,T));
		_ ->
		    error			       
	    end
    
    end.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mkDep(G,X)  when list(X) -> mkDepList(G,X);
mkDep(G,X) -> [].


mkDepList(G,[]) -> [];
mkDepList(G,List) ->
    CFN = icgen:idlfile(G), 
    [CFN | mkDepList(G,CFN,[],List)].





%%mkDepList(G, TFN, Stack, []) -> io:format("~nStack = ~p~n",[Stack]),[];
%%mkDepList(G, TFN, Stack, [X | Xs]) when record(X,preproc) ->
%%    case element(3,X#preproc.id) of
%%	TFN ->
%%	    [ branch | mkDepList(G,TFN,[TFN|Stack],Xs) ];
%%	CFN ->
%%	    [ CFN | mkDepList(G,TFN,[CFN|Stack],Xs) ]
%%    end;
%%mkDepList(G, TFN, Stack, [X | Xs]) ->
%%    mkDepList(G, TFN, Stack, Xs).


mkDepList(G, TFN, Stack, []) -> io:format("~nStack = ~p~n",[Stack]),[];
mkDepList(G, TFN, Stack, [X | Xs]) when record(X,preproc) ->
    case element(3,X#preproc.id) of
	TFN ->
	    [ branch | mkDepList(G,TFN,[TFN|Stack],Xs) ];
	CFN ->
	    [ CFN | mkDepList(G,TFN,[CFN|Stack],Xs) ]
    end;
mkDepList(G, TFN, Stack, [X | Xs]) ->
    mkDepList(G, TFN, Stack, Xs).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

do_scan(G) ->
    case get_opt(G, use_leex) of
	1 -> 
	    {ok, Bin} =  file:read_file(icgen:idlfile(G)),
	    icscanner:string(binary_to_list(Bin));
	2 -> 
	    {ok, Bin} =  file:read_file(icgen:idlfile(G)),
	    R = icscanner:tokens([], binary_to_list(Bin)),
	    element(5, element(2,R));
	_->
	    icscan:scan(G, icgen:idlfile(G))
    end.
    

do_parse(G, Tokens) ->
    case icparse:parse(Tokens) of
	{ok, L} -> L;
	X when element(1, X) == error -> 
	    Err = element(2, X),
	    icgen:fatal_error(G, {parse_error, element(1, Err), 
				  element(3, Err)});
	X -> exit(X)
    end.



time(STR,M,F,A) ->
    case timer:tc(M, F, A) of
	{_, {'EXIT', R}} -> exit(R);
	{_, {'EXIT', _, R}} -> exit(R);
	{_, _X} when element(1, _X)==error -> throw(_X);
	{_T, _R} -> 
	    io:format("Time for ~s:  ~5.2f~n", [STR, _T/1000000]),
	    _R
    end.

%%------------------------------------------------------------
%%
%% The help department
%%
%% 
%%
%%------------------------------------------------------------

help() ->
    io:format("No help available at the moment~n", []),
    ok.

print_version_str(G) ->
    case {get_opt(G, silent), get_opt(G, silent2)} of
	{true, _} -> ok;
	{_, true} -> ok;
	_ -> 
	    io:format("Erlang IDL tools version ~s~n", [?COMPILERVSN])
    end.














