%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(duplex_main).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('tobbe@erix.ericsson.se').
%%-------------------------------------------------------
%% Created:  1 Aug 1995 by tobbe@erix.ericsson.se
%% Function: An autonomus Server both accepting connection 
%%           attempts from C, and trying to connect 
%%           to C. It is kind of fault tolerant...
%%-------------------------------------------------------
-rcs('$Id: duplex_main.erl,v 1.3 1996/04/09 13:11:28 tobbe Exp tobbe $.').

-export([foo/1,bar/1,stop/1,start/3,init/4]).
-export([get_etime/1,foo/2,bar/3]).

-include("duplex.hrl").

-define( LONG_TIMEOUT  , 5000 ).

foo(P) when pid(P) -> call(P,foo).

bar(P) when pid(P) -> call(P,bar).

stop(P) when pid(P) -> call(P,stop).

call(P,What) ->
    P ! {self(),What},
    receive {P,R} -> R end.
    
start(Address,ListenPort,OtherPort) when integer(ListenPort),list(Address),integer(OtherPort) ->
    start(list_to_atom(Address),ListenPort,OtherPort);
start(Address,ListenPort,OtherPort) when integer(ListenPort),atom(Address),integer(OtherPort) ->
    spawn_link(duplex_main,init,[self(),ListenPort,Address,OtherPort]).

init(From,ListenPort,Address,OtherPort) ->
    process_flag(trap_exit,true),
    Server = duplex:server(ListenPort),
    Client = duplex:client(Address,OtherPort),
    timer:start(),
    loop(From,Server,Client,ListenPort,Address,OtherPort).

loop(From,Server,Client,ListenPort,Address,OtherPort) ->
    receive
	{'EXIT',Client,Reason} ->
	    io:format("No connection to C server !\n"),
	    timer:send_after(?LONG_TIMEOUT,{timeout,client}),
	    loop(From,Server,down,ListenPort,Address,OtherPort);
	{'EXIT',Server,Reason} ->
	    io:format("Erlang server died, restarted... !\n"),
	    timer:send_after(?LONG_TIMEOUT,{timeout,server}),
	    loop(From,down,Client,ListenPort,Address,OtherPort);
	{From,foo} when pid(Client) ->
	    From ! {self(),ok}, 
	    {ok,Foo} = ig:adr_of(Client,foo),
	    duplex:call_foo(Client,Foo),
	    loop(From,Server,Client,ListenPort,Address,OtherPort);
	{From,foo} ->
	    From ! {self(),no_connection_yet}, 
	    loop(From,Server,Client,ListenPort,Address,OtherPort);
	{From,bar} when pid(Client) ->
	    From ! {self(),ok}, 
	    {ok,Bar} = ig:adr_of(Client,bar),
	    duplex:call_bar(Client,Bar),
	    loop(From,Server,Client,ListenPort,Address,OtherPort);
	{From,bar} ->
	    From ! {self(),no_connection_yet}, 
	    loop(From,Server,Client,ListenPort,Address,OtherPort);
	{timeout,client} ->
	    NewClient = duplex:client(Address,OtherPort),
	    loop(From,Server,NewClient,ListenPort,Address,OtherPort);
	{timeout,server} ->
	    NewServer = duplex:server(ListenPort),
	    loop(From,NewServer,Client,ListenPort,Address,OtherPort);
	{From,stop} ->
	    stop_servers([Server,Client]),
	    From ! {self(),stopped},
	    exit(kill);
	{'EXIT',From,Reason} ->
	    exit(kill)
    end.

stop_servers([Pid|T]) when pid(Pid) ->
    duplex:stop(Pid),
    stop_servers(T);
stop_servers([_|T]) ->
    stop_servers(T);
stop_servers([]) -> ok.

%%-------------------------------------------------------
%% Here are the callback routines
%%-------------------------------------------------------

%% Returns an 'etime' struct !!
get_etime(App) ->
    {H,M,S} = time(),
    #etime{hour=H,minute=M,second=S}.

foo(App, Attribute) ->
    io:format("foo ~w: Attribute = ~s~n",[App,Attribute]),
    Result = 42,
    io:format("foo ~w: returning = ~w~n",[App,Result]),
    Result.

bar(App, Attribute, Key) ->
    io:format("bar ~w: Attribute = ~s , Key = ~w~n",[App,Attribute,Key]).

	    

