%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1994,1995, Ellemtel Telecommunications Systems Laboratories
%% Copyright (C) 1996 Ericsson Telecom
%%-----------------------------------------------------------------------
%% Created: 94-10-27 by Tobbe (tobbe@erix.ericsson.se)
%%
%% Function: The C code generator part of the Interface Generator
%%-----------------------------------------------------------------------
-module(iggenC).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-vsn('$Revision: /main/release/free/1').
-author('tobbe@erix.ericsson.se').
-rcs('$Id: iggenC.erl,v 1.14 1996/04/24 10:59:08 tobbe Exp tobbe $.').

-export([start/4,format_error/1, is_float/1]).

-import(ig,[get_options/1]).
-import(iggenerator,[open_file/1,emit/2,resolve_type/2,local_error/1,header_str/1]).

-include("ig.hrl").
-include("igcsasl.hrl").

%%-------------------------------------------------------
%% format_error(ErrorDescriptor)
%%
%% This function will format an error message string,
%% depending on the ErrorDescriptor provided.
%%-------------------------------------------------------
format_error({not_simple_type,SimpleType}) ->
    ?lc(["argument type, not valid : ",SimpleType]);
format_error({unexpected_type,Type}) ->
    ?lc(["unexpected type encountered : ",Type]);
format_error({aggregate_type,Type}) ->
    ?lc(["aggregate type not valid in struct : ",Type]);
format_error({internal,AtLine}) ->
    ?lc(["internal error in file <",?FILE,"> at line : ",AtLine]).

start(S,Fname,AbsForm,Hooks) -> 
    ?TELL("~nC symbol allocation",[]),
    igsym:in(S,h_file,Fname),
    DotC = dotH_to_dotC(Fname),
    Fd = open_file(DotC),
    igsym:in(S,c_file,{DotC,Fd}),
    %%UndefHooks = get_undef_hooks(AbsForm,Hooks),
    %%io:format("~nUndef hooks: ~p~n", [UndefHooks]),
    %% Generate Pass 1 to find out needed symbols
    Pass1 = dummy,
    generator(S,Pass1,AbsForm,Hooks),
    gen_callbacks(S,Pass1,AbsForm),
    gen_ig_setup_func(S,Pass1),
    %% Output header stuff 
    emit(Fd,["/* ",header_str(Fname)," */\n"]),
    emit(Fd,[include(S,Fname),?nl,extern(S,AbsForm),?nl,
	     typedefs(S),?nl,global(S),?nl,local(S,AbsForm)]),
    declare_used_vars(S,Fd),
    %% Generate Pass 2 of the code to file
    ?TELL("~nGenerating C code",[]),
    emit(Fd,[dispatch(),?nl]),
    generator(S,Fd,AbsForm,Hooks),
    ?TELL("~nGenerating callbacks",[]),
    gen_callbacks(S,Fd,AbsForm),
    gen_ig_setup_func(S,Fd),
    ?TELL("~nC code generated to file: ~s~n",[DotC]),
    file:close(Fd),
    ok.

%%-------------------------------------------------------
%% Generate code that declares all used variables
%%-------------------------------------------------------
declare_used_vars(S,Fd) -> 
    emit(Fd,["/*--- Internal Variable Declarations ---*/\n"]),
    Max = igsym:max_sym(S),
    ?andthen(Max>=0, emit(Fd,["static symbol funs[",Max+1,"];\n"])),
    case igsym:get_used_vars(S) of
	{error,_} -> ok;
	Vars -> declare(Fd,Vars)
    end.

declare(Fd,[{Var,Type}|Vs]) ->
    emit(Fd,["static ",Type," ",Var,";\n"]),
    declare(Fd,Vs);
declare(_,[]) -> ok.

%%-------------------------------------------------------
%% Generate the setup function.
%%-------------------------------------------------------
gen_ig_setup_func(S,Fd) ->
    gen_igsetup_fun(S,Fd),
    gen_igsetup_cback(S,Fd).

gen_igsetup_fun(S,Fd) ->
    emit(Fd,["\nvoid igsetup_fun(int in, int out)\n{\n"]),
    Max = igsym:max_sym(S)+1,
    emit(Fd,["\t",?ETERM_PTR,"arr[5];\n",
	     "\tint i,j,max=",Max,
	     ";\n\tchar bp[32];\n",
	     "\tstatic int igsetup=0;\n\n"]),
    emit(Fd,["\t",?IG_FUN_IN," = in;\n"]),
    emit(Fd,["\t",?IG_FUN_OUT," = out;\n"]),
    %% We only have to initiate the array once !!
    emit(Fd,["\n\tif (!igsetup){\n"]),
    emit(Fd,["\t\tigsetup = 1;\n"]),
    init_fun_array(S,Fd),
    emit(Fd,["\t}\n"]),
    ?andthen(Max>=0,init_cback(S,Fd)),
    ?andthen(Max>=0,init_sizeof(S,Fd)),
    emit(Fd,["} /* igsetup_fun */ \n\n"]).

gen_igsetup_cback(S,Fd) ->
    emit(Fd,["\nvoid igsetup_cback(int in, int out)\n{\n"]),
    emit(Fd,["\t",?IG_CBACK_IN," = in;\n"]),
    emit(Fd,["\t",?IG_CBACK_OUT," = out;\n"]),
    emit(Fd,["} /* igsetup_cback */ \n\n"]).


%%-------------------------------------------------------
%% Initiate the array of function pointers. 
%% Note: A quick and dirty hack lurks here as we store the
%% size of variables as function pointers in order to make
%% it easy for us to send the size to Erlang.
%%-------------------------------------------------------
init_fun_array(S,Fd) ->
    init_fun_array(S,Fd,igsym:first_sym(S)).

init_fun_array(S,Fd,{ok,Sym}) when record(Sym,symbol) ->
    No = Sym#symbol.no,
    Tag = Sym#symbol.tag,
    if
	Tag == ?IG_VAR_TAG ->
	    emit(Fd,["\n\t\tfuns[",No,"].f   = (FuncPtr) &",Sym#symbol.name,";\n"]);
	Tag == ?IG_SIZE_TAG ->
	    %% Note the hack with strip/3 , see gen_var_sizeof/2
	    %% Removes the SIZE_PREFIX of the variable name.
	    emit(Fd,["\n\t\tfuns[",No,"].f   = (FuncPtr) sizeof(",
		     string:strip(Sym#symbol.name,left,$_),"); /* Q.A.D */\n"]);
	true ->
	    emit(Fd,["\n\t\tfuns[",No,"].f   = (FuncPtr) ",Sym#symbol.name,";\n"])
    end,
    emit(Fd,[  "\t\tfuns[",No,"].no  = ",No,";\n"]),
    emit(Fd,[  "\t\tfuns[",No,"].tag = ",tag2str(Tag),";\n"]),
    init_fun_array(S,Fd,igsym:next_sym(S,No));
init_fun_array(_,_,_) -> ok.

init_cback(S,Fd) ->
    Tmp = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["\n\t/* Send the address of all kind of callbacks to Erlang */\n"]),
    emit(Fd,["\tfor (j=0 ; j<max ; j++) {\n"]),
    emit(Fd,["\t\tif ((funs[j].tag != IG_fun_tag) && (funs[j].tag != IG_size_tag)) {\n"]),
    emit(Fd,["\t\t\tarr[0] = erl_mk_int( funs[j].no );\n"]),
    emit(Fd,["\t\t\tarr[1] = erl_mk_uint((unsigned int) funs[j].f );\n"]),
    emit(Fd,["\t\t\t",Tmp," = erl_mk_tuple(arr,2);\n"]),
    emit(Fd,["\t\t\ti = erl_encode( ",Tmp," , (unsigned char *) bp );\n"]),
    emit(Fd,["\t\t\terl_free_compound( ",Tmp," );\n"]),
    igsym:free_var(S,?ETERM_PTR,Tmp),
    emit(Fd,["\t\t\tig_write(",?IG_FUN_OUT,", bp, i);\n\t\t}\n\t}\n"]).

init_sizeof(S,Fd) ->
    Tmp = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["\n\t/* Send the size of all variables to Erlang */\n"]),
    emit(Fd,["\tfor (j=0 ; j<max ; j++) {\n"]),
    emit(Fd,["\t\tif (funs[j].tag == IG_size_tag) {\n"]),
    emit(Fd,["\t\t\tarr[0] = erl_mk_int( funs[j].no );\n"]),
    emit(Fd,["\t\t\tarr[1] = erl_mk_int( (int) funs[j].f ); /* Q.A.D */\n"]),
    emit(Fd,["\t\t\t",Tmp," = erl_mk_tuple(arr,2);\n"]),
    emit(Fd,["\t\t\ti = erl_encode( ",Tmp," , (unsigned char *) bp );\n"]),
    emit(Fd,["\t\t\terl_free_compound( ",Tmp," );\n"]),
    igsym:free_var(S,?ETERM_PTR,Tmp),
    emit(Fd,["\t\t\tig_write(",?IG_FUN_OUT,", bp, i);\n\t\t}\n\t}\n"]).



%% - CSASL -
%% Check if user has defined hook, if so call it otherwise 
%% generate default behaviour.
%%
%% Return the hooks that are NOT defined by the user (that is, 
%% those who will need special treatment
%%
get_undef_hooks(AbsForm,[Hook|T]) ->
    case lists:keysearch(?get_name(Hook),3,AbsForm) of
	{value, Val} ->
	    get_undef_hooks(AbsForm,T);
	Other ->
	    [Hook|get_undef_hooks(AbsForm,T)]
    end;
get_undef_hooks(Funs,[]) -> [].

%% - CSASL -
%% Wrapper for taking care of the undefined hooks and the usual 
%% ig funs
generator(S,Fd,Funs,Hooks) ->
    %%io:format("GENERATOR:~n~p~n", [Funs]),
    generator_hooks(S,Fd,Hooks,get_undef_hooks(Funs,Hooks)),
    generator_1(S,Fd,Funs,Hooks).

%% - CSASL -
%%
%%	Generate empty code for those hooks that the user did not
%%	declare (which is those with a reference count of 1) and
%%	normal code for the others
%%
generator_hooks(S,Fd,[Hook|Hooks],UndefHooks) ->
    HookName = ?get_name(Hook),
    %% Add all hooks in the correct order
    igsym:mk_fun_sym(S,?lc(["_ig_",HookName])),

    %% Check if this hook is undefined and then gen 
    %% std emptyfunc
    case lists:keymember(HookName, 1, UndefHooks) of
	true ->
	    generator_default(S,Fd,HookName);
	Other ->
	    true
    end,
    generator_hooks(S,Fd,Hooks,UndefHooks);
generator_hooks(S,Fd,[],UndefHooks) -> ok.

%% - CSASL -
generator_default(S,Fd,Fun) ->
    emit(Fd,["\n/* Standard code for ",Fun," */"]),
    TheName = ?lc(["_ig_",Fun]),
    emit(Fd,["\nstatic void ",TheName,
	     "(",?ETERM_PTR,"args, char *bufp)\n{\n"]),
    gen_retval(Fd,Fun,?CSASLHOOKS),
    emit(Fd,["} /* ",TheName," */ \n"]),
    ?TICK,
    ok.

%% - CSASL -
%% gen_retval generates an empty return value for those hooks that 
%% the user did NOT define
%%
gen_retval(Fd,Fun,Hooks) when Fun==erl_c_terminate ->
    %%emit(Fd,["\tfprintf(stderr,\"Hook: ",Fun,"\\n\");\n"]),
    emit(Fd,["\texit(0);\n"]);
gen_retval(Fd,Fun,Hooks) ->
    case ?get_retval(Fun,Hooks) of
	void ->
	    emit(Fd,["\tint count;\n"]),
	    emit(Fd,["\tETERM * IGetermp0;\n"]),
	    emit(Fd,["/*--- build_term ---*/\n"]),
	    emit(Fd,["\tIGetermp0 = erl_mk_string((char *)\"\");\n"]),
	    emit(Fd,["\tif ((count = erl_encode( IGetermp0 , (unsigned char *) encode_buffer ))>IG_EBUFSIZE)\n"]),
	    emit(Fd,["\t\tig_errno = IG_ENCODE_ERROR;\n"]),
	    emit(Fd,["\terl_free_compound( IGetermp0 );\n"]),
	    emit(Fd,["\tif (ig_write( ig_fun_out , (char *) encode_buffer , count )!=ERL_RPORT_OK)\n"]),
	    emit(Fd,["\t\tig_errno = IG_WRITE_ERROR;\n"]);

	string ->
	    emit(Fd,["\tint count;\n"]),
	    emit(Fd,["\tETERM * IGetermp0;\n"]),
	    emit(Fd,["/*--- build_term ---*/\n"]),
	    emit(Fd,["\tIGetermp0 = erl_mk_string((char *)\"C status\");\n"]),
	    emit(Fd,["\tif ((count = erl_encode( IGetermp0 , (unsigned char *) encode_buffer ))>IG_EBUFSIZE)\n"]),
	    emit(Fd,["\t\tig_errno = IG_ENCODE_ERROR;\n"]),
	    emit(Fd,["\terl_free_compound( IGetermp0 );\n"]),
	    emit(Fd,["\tif (ig_write( ig_fun_out , (char *) encode_buffer , count )!=ERL_RPORT_OK)\n"]),
	    emit(Fd,["\t\tig_errno = IG_WRITE_ERROR;\n"]);
	binary ->
	    emit(Fd,["\tint count;\n"]),
	    emit(Fd,["\tIG_binary tmp;\n"]),
	    emit(Fd,["\tETERM * IGetermp0;\n"]),
	    %%emit(Fd,["\tfprintf(stderr,\"Hook: ",Fun,"\\n\");\n"]),
	    emit(Fd,["\ttmp.binp = NULL;\n"]),
	    emit(Fd,["\ttmp.size = 0;\n"]),
	    emit(Fd,["\t/*--- build_term ---*/\n"]),
	    emit(Fd,["\tIGetermp0 = erl_mk_binary((char *) tmp.binp , tmp.size );\n"]),
	    emit(Fd,["\tif ((count = erl_encode( IGetermp0 , (unsigned char *) encode_buffer ))>IG_EBUFSIZE)\n"]),
	    emit(Fd,["\t\tig_errno = IG_ENCODE_ERROR;\n"]),
	    emit(Fd,["\terl_free_compound( IGetermp0 );\n"]),
	    emit(Fd,["\tif (ig_write( ig_fun_out , (char *) encode_buffer , count )!=ERL_RPORT_OK)\n"]),
	    emit(Fd,["\t\tig_errno = IG_WRITE_ERROR;\n"])
    end.


%%---------------------------------------------------
%% Generate code for each function
%%---------------------------------------------------
generator_1(S,Fd,[{ig_fun,Line,Fun,Rtype,InArgs}|T],Hooks) ->
%%    case lists:keysearch(Fun,1,Hooks) of
%%	{value,_} ->
%%	    generator_1(S,Fd,T,Hooks);
%%	Other ->
	    generator_0(S,Fd,'fun',Line,Fun,Rtype,InArgs),
	    generator_1(S,Fd,T,Hooks);
%%    end;
generator_1(S,Fd,[{ig_void,Line,Fun,Rtype,InArgs}|T],Hooks) ->
    generator_0(S,Fd,ig_void,Line,Fun,Rtype,InArgs),
    generator_1(S,Fd,T,Hooks);
generator_1(S,Fd,[{ig_var,Line,Rtype,Name}|T],Hooks) ->
    generator_0(S,Fd,Line,Rtype,Name),
    generator_1(S,Fd,T,Hooks);
generator_1(S,Fd,[_|T],Hooks) -> generator_1(S,Fd,T,Hooks);
generator_1(_,_,[],Hooks) -> ok.

generator_0(S,Fd,Ftype,Line,Fun,Rtype,InArgs) ->
    case catch generate(S,Fd,Ftype,Fun,Rtype,InArgs) of
	{local_error,What} -> 
	    fatal_error(S,Line,What);
	ok ->
	    ok
    end.

generator_0(S,Fd,Line,Rtype,Name) ->
    case catch gen_var_func(S,Fd,Rtype,Name) of
	{local_error,What} -> 
	    fatal_error(S,Line,What);
	ok ->
	    ok
    end.

gen_var_func(S,Fd,Rtype,Name) ->
    gen_var_get(S,Fd,Rtype,Name),
    gen_var_set(S,Fd,Rtype,Name),
    gen_var_sizeof(S,Fd,Rtype,Name).

gen_var_get(S,Fd,Rtype,Name) ->
    TheName = ?lc(["_igget_",Name]),
    igsym:mk_fun_sym(S,TheName),
    igsym:mk_sym(S,Name,?IG_VAR_TAG),
    emit(Fd,["\nstatic void ",TheName,
	     "(",?ETERM_PTR,"args, char *bufp)\n{\n"]),
    gen_return(S,Fd,Name,Rtype),
    emit(Fd,["} /* ",TheName," */ \n"]),
    ?TICK,
    ok.

gen_var_set(S,Fd,Rtype,Name) ->
    TheName = ?lc(["_igset_",Name]),
    igsym:mk_fun_sym(S,TheName),
    igsym:mk_sym(S,Name,?IG_VAR_TAG),
    emit(Fd,["\nstatic void ",TheName,
	     "("?ETERM_PTR,"args, char *bufp)\n{\n"]),
    [{Src,Type}] = unpack_args(S,Fd,"args",[Rtype]),
    emit(Fd,["\t",Name," = ",Src,";\n"]),
    igsym:free_var(S,Type,Src),
    emit(Fd,["} /* ",TheName," */ \n"]),
    ?TICK,
    ok.

%%-------------------------------------------------------
%% Note: This is a Q.A.D hack ! We add a couple of '_'
%% infront of the variable name which have to be removed
%% when we generate the init routines. This is done to
%% avoid name clash with the same variable name but with
%% a IG_VAR_TAG instead.
%%-------------------------------------------------------
gen_var_sizeof(S,Fd,Rtype,Name) ->
    TheName = ?lc([?SIZE_PREFIX,Name]),
    igsym:mk_sym(S,TheName,?IG_SIZE_TAG),
    add_file_to_size_sym(S,Name,TheName),
    ok.

%% Add the file where the created "size-variable" belong.
add_file_to_size_sym(S,Name,TheName) when list(Name) ->
    add_file_to_size_sym(S,list_to_atom(Name),TheName);
add_file_to_size_sym(S,Name,TheName) when list(TheName) ->
    add_file_to_size_sym(S,Name,list_to_atom(TheName));
add_file_to_size_sym(S,Name,TheName) when atom(Name),atom(TheName) ->
    case catch igsym:out(S,{sym_file,Name}) of
	Fname when list(Fname) ->
	    igsym:in(S,{sym_file,TheName},Fname);
	{error,out} ->
	    io:format("~p: Name = ~p~n",[?FILE,Name])
    end.

generate(S,Fd,Ftype,Fun,Rtype,InArgs) ->
    str_alloc_reset(),				% OTP-1822 OTP-1944
    TheName = ?lc(["_ig_",Fun]),
    igsym:mk_fun_sym(S,TheName),
    emit(Fd,["\nstatic void ",TheName,
	     "(",?ETERM_PTR,"args, char *bufp)\n{\n"]),
    Params = unpack_args(S,Fd,"args",InArgs),
    gen_call(S,Fd,Ftype,Fun,Rtype,Params),
    free_strings(Fd),				% OTP-1822 OTP-1944

    emit(Fd,["} /* ",TheName," */ \n"]),
    ?TICK,
    ok.


%%
%% OTP-1822 OTP-1944 OTP-2006
%%
%% Strings must be freed and they must be freed even if they are
%% declared deep down in structs. Every string that is allocated is
%% stored in the process dict. and freed just before the generated
%% function is finished.
%%
%% OTP-2006 required malloced IG_binaries to be freed.
%%
%%

%% OTP-1822 OTP-1944
%% OTP2006 added suport for IG_binaries (through new case in fun)
free_strings(Fd) ->
    lists:foreach(
      fun({string, Var}) ->
	      emit(Fd, ["\tif ( strcmp(",Var,", \"\") != 0 ) erl_free( ",
			Var," );\n"]);
	 ({binary, Var}) ->
	      emit(Fd, ["\tfree( ",Var," );\n"])
      end,
      str_alloc_get()),
    str_alloc_reset(),
    ok.



%% Dirty primitives for remembering all allocated strings. OTP-1822
%% and OTP-1944, 
%%
%% OTP2006 added suport for IG_binaries (through the Type parameter)
%%
str_alloc(Type, V) ->
    Item = {Type, V},
    case get(string_alloc_list) of
	L when list(L) ->
	    put(string_alloc_list, [Item | L]);
	_ ->
	    put(string_alloc_list, [Item])
    end.

str_alloc_reset() ->
    put(string_alloc_list, []).

str_alloc_get() ->
    get(string_alloc_list).



%% Throw an error in a "standard" format
fatal_error(S,Line,What) ->
    {Fname,Fd} = igsym:out(S,c_file),
    catch file:close(Fd),
    catch file:delete(Fname),
    throw({error,{Fname,Line,iggenC,What}}).

%% Generate code for unpacking the arguments.
unpack_args(S,Fd,Decoded,[D]) when D#descriptor.cat==void ->
    [void];
unpack_args(S,Fd,Decoded,Args) ->
    unpack_args(S,Fd,Decoded,Args,1).

%%-------------------------------------------------------
%% Ok, here we unpacks each of the arguments. During
%% the work we allocates a suitable variable to hold
%% the argument. This variable is to be freed elsewere.
%% We return a tuple/2 for each argument containing:
%%  - the variable allocated that contains the argument
%%  - the type of the variable
%%-------------------------------------------------------
unpack_args(S,Fd,Decoded,[Arg|Args],I) ->
    Res = unpack_the_arg(S,Fd,Decoded,Arg,I),
    [Res|unpack_args(S,Fd,Decoded,Args,I+1)];
unpack_args(_,_,_,[],_) -> [].

unpack_the_arg(S,Fd,Decoded,Arg,I) when record(Arg,descriptor) ->
    %% Determine the type to be allocated.
    Atype = allocated_type_form(Arg),
    %% Allocate the C variable that shall contain the argument.
    Cvar = igsym:alloc_var(S,Atype),
    %% Transform the type argument into a general form
    Gtype = untangel_type(S,Arg),
    %% Phui..at last..unpack the damn thing..
    unpack_arg(S,Fd,Decoded,Cvar,Gtype,I),
    {Cvar,Atype}.

allocated_type_form(D) when record(D,descriptor) -> D#descriptor.ansi_string.

%% Depending on the type description, untangel the type
%% to be generated for.
untangel_type(S,D) when D#descriptor.cat==type -> 
    resolve_type(S,D#descriptor.tname);
untangel_type(_,D) when record(D,descriptor)   -> 
    {D#descriptor.cat,D#descriptor.ansi_string}.

%%-------------------------------------------------------
%% Get the argument in specified slot, and unpack it !!
%%-------------------------------------------------------
unpack_arg(S,Fd,Decoded,Cvar,Gtype,I) ->
    %% Get a variable to hold the slot
    Slot = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["/*--- unpack_arg ",I," ---*/\n"]),
    emit(Fd,["\t",Slot," = erl_element(",I,",",Decoded,");\n"]),
    %% Now unpack the data and store it in the specified C variable
    Res = unpack_data(S,Fd,Slot,Cvar,Gtype),
    igsym:free_var(S,?ETERM_PTR,Slot),
    Res.

%% NOTE: The type information is assumed to be in "general" form
%% see untangel_type/2 (should need a hyperlink here... :-)
unpack_data(S,Fd,Slot,Cvar,{ptr,Type}) ->
    emit(Fd,["\t",Cvar," = (",Type,") ",Slot,"->uval.uival.u;\n"]);
%    unpack_ptr(S,Fd,Slot,Cvar,Type);
unpack_data(S,Fd,Slot,Cvar,{type_struct,Es,_}) when list(Es) ->
    %% Skip the record tag in the first slot
    unpack_struct(S,Fd,Slot,Cvar,Es,2);
unpack_data(S,Fd,Slot,Cvar,{'IG_string',_}) ->
%%    io:format("~nAllocating string: ~p~n", [Cvar]),
    str_alloc(string, Cvar),
    emit(Fd,["/*--- unpack_string ---*/\n"]),
    emit(Fd,["\t",Cvar," = erl_iolist_to_string(",Slot,");\n"]),
    emit(Fd, ["\tif ( ",Cvar," == NULL ) ",Cvar," = \"\";\n"]), %Janne fix (PL)
    ok;
unpack_data(S,Fd,Slot,Cvar,{binaryPtr,_}) ->
    emit(Fd,["/*--- unpack_binary ---*/\n"]),
    str_alloc(binary, Cvar),
    emit(Fd,["\t",Cvar," = (IG_binaryPtr) malloc( sizeof(IG_binary) );\n"]),
    emit(Fd,["\t",Cvar,"->size = ",Slot,"->uval.bval.size;\n"]),
    emit(Fd,["\t",Cvar,"->binp = ",Slot,"->uval.bval.b;\n"]),
    ok;
unpack_data(S,Fd,Slot,Cvar,{std,Type}) ->	% Clause changed: OTP-1450
    emit(Fd,["/*--- unpack_std_type ---*/\n"]),
    case {is_float(Type), is_unsigned(Type)} of
	{true, _} ->
	    emit(Fd,["\t",Cvar," = (",Type,") ",Slot,"->uval.fval.f;\n"]);
	{false, true} ->
	    emit(Fd,["\t",Cvar," = (",Type,") ",Slot,"->uval.uival.u;\n"]);
	{false, false}-> 
	    emit(Fd,["\t",Cvar," = (",Type,") ",Slot,"->uval.ival.i;\n"])
    end,
    ok;
unpack_data(_,_,_,_,Type) ->
    local_error({unexpected_type,Type}).

unpack_struct(S,Fd,Slot,Cvar,[{struct_elem,Type,Name}|Es],I) ->
    emit(Fd,["/*--- unpack_struct ",I," ---*/\n"]),
    %% Get a variable to hold the nested slot
    Tmp = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["\t",Tmp," = erl_element(",I,",",Slot,");\n"]),
    NewCvar = ?lc([Cvar,".",Name]),
    unpack_data(S,Fd,Tmp,NewCvar,Type),
    igsym:free_var(S,?ETERM_PTR,Tmp),
    unpack_struct(S,Fd,Slot,Cvar,Es,I+1);
unpack_struct(S,Fd,Slot,Cvar,[{struct_elem,Type,Name,Len}|Es],I) ->
    emit(Fd,["/*--- unpack_struct ",I," ---*/\n"]),
    unpack_struct_array(S,Fd,Slot,Cvar,Name,I),
    unpack_struct(S,Fd,Slot,Cvar,Es,I+1);
unpack_struct(S,Fd,Slot,Cvar,[{nested_struct,Elements,Name,_}|Es],I) ->
    emit(Fd,["/*--- unpack_struct ",I," ---*/\n"]),
    Tmp = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["\t",Tmp," = erl_element(",I,",",Slot,");\n"]),
    NewCvar = ?lc([Cvar,".",Name]),
    unpack_struct(S,Fd,Tmp,NewCvar,Elements,2),
    igsym:free_var(S,?ETERM_PTR,Tmp),
    unpack_struct(S,Fd,Slot,Cvar,Es,I+1);
unpack_struct(_,Fd,_,_,[],_) -> 
    emit(Fd,["/*--- unpack_struct finished ---*/\n"]),
    ok.

unpack_struct_array(S,Fd,Slot,Cvar,Name,I) ->
    emit(Fd,["/*--- unpack_struct_array ---*/\n"]),
    Tmp = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["\t",Tmp," = erl_element(",I,",",Slot,");\n"]),
    emit(Fd,["\tmemcpy((void *) ",Cvar,".",Name,",",
	       "(const void *) ",Tmp,"->uval.bval.b,",
	       "(size_t) ",Tmp,"->uval.bval.size);\n"]),
    igsym:free_var(S,?ETERM_PTR,Tmp).

%% Generate the actual function call, 
gen_call(S,Fd,ig_void,Fun,_,Params) -> 
    CallArgs = gen_call_args(Params),
    emit(Fd,["\t",Fun,"(",CallArgs,");\n"]),
    free_call_args(S,Fd,Params),
    ok;
%%gen_call(S,Fd,_,Fun,R,[void]) when R#descriptor.cat==void -> 
%%    io:format("~ngen_call 1: for: ~p~n",[Fun]),
%%    emit(Fd,["\t",Fun,"();\n"]),
%%    ok;
%%gen_call(S,Fd,_,Fun,R,Params) when R#descriptor.cat==void -> 
%%    io:format("~ngen_call 2: for: ~p~n",[Fun]),
%%    CallArgs = gen_call_args(Params),
%%    emit(Fd,["\t",Fun,"(",CallArgs,");\n"]),
%%    free_call_args(S,Fd,Params),
%%    ok;
gen_call(S,Fd,'fun',Fun,Rtype,Params)  when Rtype#descriptor.cat==void -> 
    CallArgs = gen_call_args(Params),
    emit(Fd,["\t",Fun,"(",CallArgs,");\n"]),
    free_call_args(S,Fd,Params),
    %% Note Cvar is nil because its not used in gen_return of a void func
    gen_return(S,Fd,nil,Rtype),
    ok;
%%gen_call(S,Fd,_,Fun,R,[void]) when R#descriptor.cat==void -> 
%%    io:format("~ngen_call 1: for: ~p~n",[Fun]),
%%    emit(Fd,["\t",Fun,"();\n"]),
%%    ok;
%%gen_call(S,Fd,_,Fun,R,Params) when R#descriptor.cat==void -> 
%%    io:format("~ngen_call 2: for: ~p~n",[Fun]),
%%    CallArgs = gen_call_args(Params),
%%    emit(Fd,["\t",Fun,"(",CallArgs,");\n"]),
%%    free_call_args(S,Fd,Params),
%%    ok;
gen_call(S,Fd,'fun',Fun,Rtype,Params)  when Rtype#descriptor.cat==void -> 
    CallArgs = gen_call_args(Params),
    emit(Fd,["\t",Fun,"(",CallArgs,");\n"]),
    free_call_args(S,Fd,Params),
    %% Note Cvar is nil because its not used in gen_return of a void func
    gen_return(S,Fd,nil,Rtype),
    ok;
gen_call(S,Fd,'fun',Fun,Rtype,Params) -> 
    Atype = allocated_type_form(Rtype),
    Cvar  = igsym:alloc_var(S,Atype),
    CallArgs = gen_call_args(Params),
    emit(Fd,["\t",Cvar," = ",cast_or_not(Rtype)," ",
	     Fun,"(",CallArgs,");\n"]),
    free_call_args(S,Fd,Params),
    gen_return(S,Fd,Cvar,Rtype),
    igsym:free_var(S,Atype,Cvar),
    ok.

%% ANSI C forbids casting nonscalar to the same type
cast_or_not(D) when D#descriptor.cat==type -> "";
cast_or_not(D) -> ?lc(["(",D#descriptor.ansi_string,")"]).

%% ..and write back the result..
gen_return(S,Fd,Src,Rtype) ->
    Dest = igsym:alloc_var(S,?ETERM_PTR),
    build_term(S,Fd,[{Rtype,Src,Dest}]),
    Tmp = igsym:alloc_var(S,"int"),
    emit(Fd,["\tif ((",Tmp," = erl_encode( ",Dest," , (unsigned char *) encode_buffer ))>IG_EBUFSIZE)\n",
	     "\t\t",?IG_ERRNO," = IG_ENCODE_ERROR;\n"]),
    emit(Fd,["\terl_free_compound( ",Dest," );\n"]),
    igsym:free_var(S,?ETERM_PTR,Dest),
    emit(Fd,["\tif (ig_write( ",?IG_FUN_OUT," , (char *) encode_buffer , ",Tmp," )!=ERL_RPORT_OK)\n",
	     "\t\t",?IG_ERRNO," = IG_WRITE_ERROR;\n"]),
    igsym:free_var(S,"int",Tmp).

%% Generate a string of parameters
gen_call_args([void]) -> [];
gen_call_args([{Sym,Type}|Tail]) ->
    [Sym|gen_call_args_0(Tail)].

gen_call_args_0([]) -> [];
gen_call_args_0([{Sym,Type}|Tail]) ->
    [?lc([" , ",Sym])|gen_call_args_0(Tail)].

%% Return the call-arg symbols
%% 
free_call_args(_,_,[void]) -> [];
free_call_args(S,Fd,[{Sym,Type}|Tail]) ->
    igsym:free_var(S,Type,Sym),
    free_call_args(S,Fd,Tail);
free_call_args(_,_,[]) -> [].

%%-------------------------------------------------------
%% Generate C callback functions (if any)
%%-------------------------------------------------------
gen_callbacks(S,_,[]) -> ok;
gen_callbacks(S,Fd,[{ig_call,Line,Fun,Rtype,Params}|T]) ->
    gen_callback(S,Fd,Fun,Rtype,Params),
    ?TICK,
    gen_callbacks(S,Fd,T);
gen_callbacks(S,Fd,[{ig_cast,Line,Fun,Rtype,Params}|T]) ->
    gen_castback(S,Fd,Fun,Rtype,Params),
    ?TICK,
    gen_callbacks(S,Fd,T);
gen_callbacks(S,Fd,[{ig_extern,Line,Fun,Rtype,Params}|T]) ->
    ?TICK,
    igsym:mk_sym(S,Fun,?IG_EXTERN_TAG),
    gen_callbacks(S,Fd,T);
gen_callbacks(S,Fd,[_|T]) ->
    gen_callbacks(S,Fd,T).
    
gen_callback(S,Fd,Fun,Rtype,Params) when record(Rtype,descriptor) ->
    emit(Fd,["\n/*--- callback routine: ",Fun," ---*/\n"]),
    ?DBG("~n/*--- callback routine: ~w , params: ~p~n",[Fun,Params]),
    igsym:mk_sym(S,Fun,?IG_CALL_TAG),
    {Args,StrArgs} = gen_callback_args(Params),
    Ptype = Rtype#descriptor.ansi_string,
    emit(Fd,[Ptype," ",Fun,"(",StrArgs,")\n{\n"]),
    encode_the_args(S,Fd,Fun,Args),
    emit(Fd,["/*--- unpack_rvalue ---*/\n"]),
    emit(Fd,["\tif ((len = ig_read(",?IG_CBACK_IN,")) > 0) {\n"]),
    unpack_rvalue(S,Fd,Rtype),
    emit(Fd,["\t}\n"]),
    %% This is rather delicate ! What should we do when for example
    %% the connection breaks down when we are waiting for the result ?
    %% We cant Exit since that should crash the whole application !
    %% So we will instead set 'ig_errno' , which makes it possible
    %% for the surrounding application to test it. However, we must 
    %% return something so we just returns garbage !
    Atype = allocated_type_form(Rtype),
    Cvar = igsym:alloc_var(S,Atype),
    emit(Fd,["\telse {\n",
	     "\t\t",?IG_ERRNO," = IG_READ_ERROR;\n",
	     "\t\treturn ",Cvar,";\n\t}\n"]),
    igsym:free_var(S,Atype,Cvar),
    emit(Fd,["} /* ",Fun," */ \n"]).

%% Unpack the returned value
unpack_rvalue(S,Fd,Rtype) when record(Rtype,descriptor) ->
    Decoded = igsym:alloc_var(S,?ETERM_PTR),
    emit(Fd,["\t\t",Decoded," = erl_decode((unsigned char *) igbufp );\n"]),
    Atype = allocated_type_form(Rtype),
    Cvar = igsym:alloc_var(S,Atype),
    Gtype = untangel_type(S,Rtype),
    unpack_data(S,Fd,Decoded,Cvar,Gtype),
    case Gtype of
	{binaryPtr, _} -> ok; % Do not free binaryptr
	_ ->
	    emit(Fd,["\terl_free_compound(",Decoded,");\n"])
    end,
    emit(Fd,["\treturn ",Cvar,";\n"]),
    igsym:free_var(S,?ETERM_PTR,Decoded),
    igsym:free_var(S,Atype,Cvar),
    ok.

gen_castback(S,Fd,Fun,Rtype,Params) ->
    emit(Fd,["\n/*--- cast routine: ",Fun," ---*/\n"]),
    ?DBG("~n/*--- cast routine: ~w , params: ~w~n",[Fun,Params]),
    igsym:mk_sym(S,Fun,?IG_CAST_TAG),
    {Args,StrArgs} = gen_callback_args(Params),
    Ptype = Rtype#descriptor.ansi_string,
    emit(Fd,[Ptype," ",Fun,"(",StrArgs,")\n{\n"]),
    encode_the_args(S,Fd,Fun,Args),
    emit(Fd,["} /* ",Fun," */ \n"]).

gen_callback_args([Arg]) when Arg#descriptor.cat==void -> {[],""};
gen_callback_args(Params) ->
    gen_callback_args(Params,0,[],[]).

%% Generate a variable for each parameter, and
%% the ANSI parameter string to be generated.
gen_callback_args([Arg|T],N,Args,StrArgs) when record(Arg,descriptor) ->
    Var = ?lc(["V",N]),
    Astr = Arg#descriptor.ansi_string,
    StrVar = ?lc([Astr," ",Var]),
    gen_callback_args(T,N+1,[{Arg,Var}|Args],[StrVar|StrArgs]);
gen_callback_args([],_,Args,StrArgs) ->
    {lists:reverse(Args),
     iggenerator:mk_comma_str(lists:reverse(StrArgs))}.

%% Build an Erlang term structure containing the arguments,
%% encode it, and send it to Erlang
encode_the_args(S,Fd,Fun,Args) ->
    emit(Fd,["/*--- encode_the_args ---*/\n"]),
    DecArgs = declare_vars(Fd,Args),
    %% Reserve space for the length field
    Size = length(DecArgs)+1,  
    emit(Fd,["\t",?ETERM_PTR,"arr[",Size,"];\n"]),
    emit(Fd,["\t"?ETERM_PTR,"result;\n"]),
    emit(Fd,["\tint len;\n\n"]),
    emit(Fd,["\t",?IG_ERRNO," = IG_OK;\n"]),
    build_term(S,Fd,DecArgs),
    {ok,Sym} = igsym:get_sym(S,Fun),
    emit(Fd,["/*--- fill_array ---*/\n"]),
    emit(Fd,["\tarr[0] = erl_mk_int(",Sym#symbol.no,");\n"]),
    fill_array(S,Fd,DecArgs,1),
    emit(Fd,["\tresult = erl_mk_list(arr,",Size,");\n"]),
    emit(Fd,["\tif ((len = erl_encode(result,(unsigned char *) encode_buffer))>IG_EBUFSIZE)\n",
	     "\t\t",?IG_ERRNO," = IG_ENCODE_ERROR;\n"]),
    emit(Fd,["\terl_free_compound(result);\n"]),
    emit(Fd,["\tif (ig_write(",?IG_CBACK_OUT,",(char *) encode_buffer,len)!=ERL_RPORT_OK)\n",
	     "\t\t",?IG_ERRNO," = IG_WRITE_ERROR;\n"]).

%% Declare the variables that shall contain the corresponding
%% formal parameters in Erlang term format.
declare_vars(Fd,[{Arg,Src}|T]) ->
    Dest = ?lc(["D",Src]),
    emit(Fd,["\t",?ETERM_PTR,Dest,";\n"]),
    [{Arg,Src,Dest}|declare_vars(Fd,T)];
declare_vars(_,[]) -> [].

fill_array(S,Fd,[{_,_,Dest}|T],I) ->
    emit(Fd,["\tarr[",I,"] = ",Dest,";\n"]),
    fill_array(S,Fd,T,I+1);
fill_array(_,_,[],_) -> ok.

tag2str(?IG_FUN_TAG)    -> "IG_fun_tag";
tag2str(?IG_CALL_TAG)   -> "IG_call_tag";
tag2str(?IG_CAST_TAG)   -> "IG_cast_tag";
tag2str(?IG_EXTERN_TAG) -> "IG_extern_tag";
tag2str(?IG_VOID_TAG)   -> "IG_void_tag";
tag2str(?IG_VAR_TAG)    -> "IG_var_tag";
tag2str(?IG_SIZE_TAG)   -> "IG_size_tag".

build_term(S,Fd,[{Arg,Src,Dest}|T]) when record(Arg,descriptor) ->
    emit(Fd,["/*--- build_term ---*/\n"]),
    Gtype = untangel_type(S,Arg),
    build_term(S,Fd,[{Gtype,Src,Dest}|T]);
build_term(S,Fd,[{{std,Tstring},Src,Dest}|T]) -> % Clause changed: OTP-1450
    case {is_float(Tstring), is_unsigned(Tstring)} of
	{true, _} ->
	    emit(Fd,["\t",Dest," = erl_mk_float((double) ",Src,");\n"]);
	{false, true} ->
	    emit(Fd,["\t",Dest," = erl_mk_uint((",Tstring,") ",Src,");\n"]);
	{false, false}->
	    emit(Fd,["\t",Dest," = erl_mk_int((",Tstring,") ",Src,");\n"])
    end,
    build_term(S,Fd,T);
build_term(S,Fd,[{{'IG_string',_},Src,Dest}|T]) ->
    emit(Fd,["\t",Dest," = erl_mk_string((char *)",Src,");\n"]),
    build_term(S,Fd,T);
build_term(S,Fd,[{{binaryPtr,_},Src,Dest}|T]) ->
    emit(Fd,["\t",Dest," = erl_mk_binary((char *) ",Src,"->binp , ",Src,"->size );\n"]),
    build_term(S,Fd,T);
build_term(S,Fd,[{{ptr,_},Src,Dest}|T]) ->
    emit(Fd,["\t",Dest," = erl_mk_uint((unsigned int) ",Src,");\n"]),
    build_term(S,Fd,T);
build_term(S,Fd,[{{type_struct,Es,Name},Src,Dest}|T]) ->
    %% We only enter here the very first time !
    build_type_struct(S,Fd,Src,Dest,Es,Name),
    build_term(S,Fd,T);
build_term(S,Fd,[{{void,_},Src,Dest}|T]) ->
    emit(Fd,["\t",Dest," = erl_mk_string((char *)\"\");\n"]),
    build_term(S,Fd,T);
build_term(S,Fd,[{Gtype,Src,Dest}|T]) ->
    io:format("~nGtype: ~p, Src: ~p~n",[Gtype,Src]),
    local_error({internal,?LINE});
build_term(_,_,[]) -> ok.


%% Declare an array to be used when building the tuple,
%% and create the tuple to be encoded.
build_type_struct(S,Fd,Src,Dest,Es,Name) ->
    emit(Fd,["/*--- build_type_struct ---*/\n"]),
    Arr = igsym:alloc_var(S,?ETERM_PTR,local),
    Elen = length(Es)+1,
    emit(Fd,["\t{",?ETERM_PTR,Arr,"[",Elen,"];\n"]),
    %% Add the record tag
    emit(Fd,["\t",Arr,"[0] = erl_mk_atom(\"",Name,"\");\n"]),
    build_struct(S,Fd,Src,Arr,Es,1),
    emit(Fd,["\t",Dest," = erl_mk_tuple( ",Arr," , ",Elen," );\n"]),
    igsym:free_var(S,?ETERM_PTR,Arr),
    emit(Fd,["\t}\n"]),
    emit(Fd,["/*--- End of: build_type_struct  ---*/\n"]).

build_struct(S,Fd,Src,Dest,[{struct_elem,Type,Name}|Es],I) ->
    NewSrc = ?lc([Src,".",Name]),
    NewDest = ?lc([Dest,"[",I,"]"]),
    build_term(S,Fd,[{Type,NewSrc,NewDest}]),
    build_struct(S,Fd,Src,Dest,Es,I+1);
build_struct(S,Fd,Src,Dest,[{struct_elem,{std,Type},Name,Len}|Es],I) ->
    emit(Fd,["\t",Dest,"[",I,"] = erl_mk_binary((char *) ",
	     Src,".",Name," , sizeof(",Type,")*",Len," );\n"]),
    build_struct(S,Fd,Src,Dest,Es,I+1);
build_struct(S,Fd,Src,Dest,[{struct_elem,{ptr,Type},Name,Len}|Es],I) ->
    emit(Fd,["\t",Dest,"[",I,"] = mk_binary((char *) ",
	     Src,".",Name," , sizeof(",Type,")*",Len," );\n"]),
    build_struct(S,Fd,Src,Dest,Es,I+1);
build_struct(S,Fd,Src,Dest,[{nested_struct,Elems,Name,Tname}|Es],I) ->
    NewSrc = ?lc([Src,".",Name]),
    NewDest = ?lc([Dest,"[",I,"]"]),
    build_type_struct(S,Fd,NewSrc,NewDest,Elems,Tname),
    build_struct(S,Fd,Src,Dest,Es,I+1);
build_struct(_,_,_,_,[],_) -> ok.

include(S,Fname) ->
    ?lc(["#include <string.h>\n",
	 "#include \"erl_interface.h\"\n",
	 "#include \"ig.h\"\n",
	 "#include \"",Fname,"\"\n",
	 include_files(S),
	 "\n/*--- Options used when generating: ",
	 iggenerator:term2str(get_options(S)),
	 " ---*/\n"]).

include_files(S) ->
    case igsym:out(S,{option,files}) of
	false -> "";
	Files when list(Files) -> include_files(S,Files,[])
    end.

include_files(S,[F|Fs],Ack) ->
    Str = ?lc(["#include \"",F,"\"\n"]),
    include_files(S,Fs,lists:append(Str,Ack));
include_files(_,[],Ack) -> Ack.

extern(S,AbsForm) ->
    ?lc(["/*--- Extern declarations ---*/\n",
	 "extern int ig_read(int);\n",
	 "extern int ig_write(int,char*,int);\n",
	 "extern char *igbufp;\n",
	 "extern int igbuflen;\n"|
	 extern_0(S,AbsForm)]).

extern_0(S,[{ig_var,_,D,Name}|T]) when D#descriptor.storage_class==extern -> 
    Dstr = D#descriptor.ansi_string,
    [?lc(["extern ",Dstr," ",Name,";\n"])|extern_0(S,T)];
%%--------------------------------------------------------
%% 21 Aug 1995 by tobbe
%% Removed this because of the ellipsis problem. Let the
%% prototypes be visible through ig.h instead.
%%--------------------------------------------------------
%%extern_0(S,[{ig_fun,_,Fun,Rtype,InArgs}|T]) -> 
%%    {Args,StrArgs} = gen_callback_args(InArgs),
%%    Rstr = Rtype#descriptor.ansi_string,
%%    [?lc(["extern ",Rstr," ",Fun,"(",StrArgs,");\n"])|extern_0(S,T)];
%%extern_0(S,[{ig_extern,_,Fun,Rtype,InArgs}|T]) -> 
%%    {Args,StrArgs} = gen_callback_args(InArgs),
%%    Rstr = Rtype#descriptor.ansi_string,
%%    [?lc(["extern ",Rstr," ",Fun,"(",StrArgs,");\n"])|extern_0(S,T)];
extern_0(S,[_|T]) -> 
    extern_0(S,T);
extern_0(S,[]) -> ["\n"].

local(S,AbsForm) ->
    ?lc(["/*--- Local variable declarations ---*/\n",
	 "#define IG_EBUFSIZE (1024*64)\n",
	 "static unsigned char encode_buffer[IG_EBUFSIZE];\n"|
	 local_0(S,AbsForm)]).

local_0(S,[{ig_var,_,Rtype,Name}|T]) when Rtype#descriptor.storage_class=/=extern -> 
    Rstr = Rtype#descriptor.ansi_string,
    Sclass = Rtype#descriptor.storage_class,
    [?lc([skip_auto(Sclass)," ",Rstr," ",Name,";\n"])|local_0(S,T)];
local_0(S,[_|T]) -> 
    local_0(S,T);
local_0(S,[]) -> ["\n"].

%% Q.A.D hack !!
skip_auto(auto) -> "";
skip_auto(X) -> X.

typedefs(S) ->
    ?lc(["/*--- Typedef's ---*/\n",
	"typedef void (*FuncPtr)(",?ETERM_PTR,", char*);\n"
	"typedef struct {\n"
	"\tFuncPtr f;\n\tint no;\n\tint tag;\n}symbol;\n"]).

global(S) ->
    ?lc(["/*--- Global Declarations ---*/\n",
	 "int ",?IG_FUN_IN,"=",ig:stdin(S,'fun'),
	 ",",?IG_FUN_OUT,"=",ig:stdout(S,'fun'),";\n",
	 "int ",?IG_CBACK_IN,"=",ig:stdin(S,cback),
	 ",",?IG_CBACK_OUT,"=",ig:stdout(S,cback),";\n",
	 "int ",?IG_ERRNO,";\n"]).

dispatch() ->
    ?lc(["\nvoid ig_dispatch(void)\n{\n",
	 "\t",?ETERM," *root,*tag,*args;\n\n",
	 "\tif ((root = erl_decode((unsigned char *) igbufp)) == NULL) {\n",
	 "\t\tig_errno = IG_DECODE_ERROR;\n",
	 "\t\treturn;\n\t}\n",
	 "\ttag = erl_element(1,root);\n",
	 "\targs = erl_element(2,root);\n",
	 "\tfuns[tag->uval.ival.i].f( args , igbufp );\n",
	 "\terl_free_compound( root );\n",
	 "} /* ig_dispatch */\n"]).

dotH_to_dotC([$.,$h|T]) -> [$_,$s,$t,$u,$b,$.,$c|T];
dotH_to_dotC([H|T]) -> [H|dotH_to_dotC(T)].

ucase(Atom) when atom(Atom)    -> ucase(atom_to_list(Atom));
ucase([H|T]) when H>=65,H=<90  -> [H|ucase(T)];
ucase([H|T]) when H>=97,H=<122 -> [H-32|ucase(T)];
ucase([H|T])                   -> [H|ucase(T)];
ucase([])                      -> [].

is_unsigned(String) when list(String) ->
    case string:index(String,"unsigned") of
	N when N>0 -> true;
	_ -> false
    end.

%% Exported! (PL)
is_float("float ") -> true;
is_float("double ") -> true;
is_float(_) -> false.

