%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1994,1995, Ellemtel Telecommunications Systems Laboratories
%%-----------------------------------------------------------------------
%% Created: 94-10-27 by Tobbe (tobbe@erix.ericsson.se)
%% 
%% Function: The Erlang code generator part of the Interface Generator
%% 
%%-----------------------------------------------------------------------
%% Modified: 25 Jan 1999 by tobbe@serc.rmit.edu.au
%% Function: Added possibility to run the Erlang node in the background.
%%-----------------------------------------------------------------------
-module(iggenErl).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('tobbe@erix.ericsson.se').
-rcs('$Id: iggenErl.erl,v 1.6 1996/04/09 12:57:45 tobbe Exp $.').

-export([start/4,length_plus_one/1,format_error/1]).

-import(ig,[get_option/2,stdio/1]).
-import(iggenerator,[open_file/1,emit/2,resolve_type/2,
		     local_error/1,mk_comma_str/1,header_str/1]).

-include("ig.hrl").
-include("igcsasl.hrl").

format_error({unexpected_type,Type}) ->
    ?lc(["unexpected type encountered : ",Type]).

start(S,Fname,AbsForm,Hooks) -> 
    ?TELL("Generating Erlang code",[]),
    DotErl = dotH_to_dotErl(Fname),
    Fd = open_file(DotErl),
    igsym:in(S,erl_file,{DotErl,Fd}),
    emit(Fd,["%% ",header_str(Fname),"\n\n"]),
    Funs = get_funs(S,AbsForm,Hooks),
    Mod = filename:basename(Fname, ".h"),
    emit(Fd,[eheader(S,Mod,Funs),"\n"]),
    ?DBG("calling start: ~w~n",[Funs]),
    generator(S,Fd,Mod,Funs,Hooks),
    emit(Fd,[gen_callback(S),"\n"]),
    ?TELL("Erlang code generated to file: ~s~n",[DotErl]),
    gen_hrl(S,Mod,Fname,AbsForm),
    file:close(Fd),
    ok.


%% Extract the stuff we are going to generate, and
%% get the key symbol for each item.
%%
%% - CSASL -
%% Do not extract those functions that are hooks.
get_funs(S,[{ig_fun,L,N,R,P}|T],Hooks) -> 
    case ?is_hook(N,Hooks) of
	true ->
	    get_funs(S,T,Hooks);
	Other ->
	    TheName = ?lc(["_ig_",N]),
	    {ok,Sym} = igsym:get_sym(S,TheName),
	    [{ig_fun,L,N,R,P,Sym}|get_funs(S,T,Hooks)]
    end;
get_funs(S,[{ig_void,L,N,R,P}|T],Hooks) -> 
    case ?is_hook(N,Hooks) of
	true ->
	    get_funs(S,T,Hooks);
	Other ->
	    TheName = ?lc(["_ig_",N]),
	    {ok,Sym} = igsym:get_sym(S,TheName),
	    [{ig_void,L,N,R,P,Sym}|get_funs(S,T,Hooks)]
    end;
get_funs(S,[{ig_var,L,R,N}|T],Hooks) -> 
    TheFun1 = ?lc(["_igget_",N]),
    {ok,Sym1} = igsym:get_sym(S,TheFun1),
    TheFun2 = ?lc(["_igset_",N]),
    {ok,Sym2} = igsym:get_sym(S,TheFun2),
    [{igget_var,L,N,R,[],Sym1}, 
     {igset_var,L,N,R,[R],Sym2}|get_funs(S,T,Hooks)];
get_funs(S,[_|T],Hooks) -> get_funs(S,T,Hooks);
get_funs(_,[],Hooks) -> [].

eheader(S,Mod,Funcs) ->
    ExpList = mk_export_list(Funcs),
    ?lc(["-module(",Mod,").\n",
	 "-export([",
	 mk_comma_str(["start/0","start/1","stop/0","stop/1","client/2",
		       "server/1","sup_start/1","sup_start/2",
		       "\n\t'IG_callbacks'/0"]),
	 "]).\n",
	 "-export([",mk_comma_str(ExpList),"]).\n\n",
	 "-define( LONG_MIN , -2147483647 ).\n",
	 "-define( LONG_MAX ,  2147483647 ).\n",
	 "-define( ULONG_MAX,  4294967295 ).\n",
	 "\nstart() ->\n",
	 "\tigserver:start({{spawn,'",Mod,"'},[",stdio(S),"]}, ",
	 "'IG_callbacks'()).\n\n",

	 "start(Exec) when list(Exec) ->\n",
	 "\tstart(list_to_atom(Exec));\n",
	 "start(Exec) when atom(Exec) ->\n",
	 "\tigserver:start({{spawn,Exec},[",stdio(S),"]},",
	 "'IG_callbacks'());\n\n"
	 "start({fd,In,Out}) when integer(In),integer(Out) ->\n",
	 "\tigserver:start({{fd,In,Out},[",stdio(S),"]},",
	 "'IG_callbacks'()).\n\n"

	 "client(Address,PortNo) when list(Address) ->\n",
	 "\tclient(list_to_atom(Address),PortNo);\n"
	 "client(Address,PortNo) when atom(Address),integer(PortNo) ->\n",
	 "\tigserver:start({client,Address,PortNo},'IG_callbacks'()).\n\n",

	 "server(PortNo) when integer(PortNo) ->\n",
	 "\tigserver:start({server,PortNo},'IG_callbacks'()).\n\n",

	 "sup_start(LibName) ->\n",
	 "\tsup_start(LibName,",Mod,").\n",
	 "sup_start(LibName,ProgName) ->\n",
	 "\tExec = {sup_port,LibName,ProgName,[",stdio(S),"]},\n",
	 "\tigserver:sup_start({local,",Mod,"},Exec,'IG_callbacks'()).\n\n",
	 "stop() -> ",Mod," ! {self(),stop},ok.\n",
	 "stop(S) -> S ! {self(),stop},ok.\n"]).
	
mk_export_list(Funcs) ->
    Fnames = lists:map({erlang,atom_to_list},[],
		       rmap({erlang,element},[3],Funcs)),
    Card = lists:map({iggenErl,length_plus_one},[],
		     rmap({erlang,element},[5],Funcs)),
    export_merge(Fnames,Card).

%% The size(3) below is a Q.A.D hack !!
rmap(F, As, [Hd|Tail]) when size(Hd)>=3 ->
    [apply(F, lists:append(As,[Hd]))|rmap(F, As, Tail)];
rmap(F, As, [_|Tail]) -> rmap(F,As,Tail);
rmap(F, As, []) -> [].

length_plus_one([D]) when D#descriptor.cat==void -> 1;
length_plus_one(L) when list(L) -> length(L) + 1.

%%export_merge([[H|T]|B],[X|Y]) when H>=$A,H=<$Z ->
%%    [?lc(["'",[H|T],"'/",X])|export_merge(B,Y)];
export_merge([A|B],[X|Y]) ->
    [?lc([quote(A),"/",X])|export_merge(B,Y)];
%%    [?lc([quote(A),"/",X-1]),?lc([quote(A),"/",X])|export_merge(B,Y)];
export_merge([],[]) -> [].
	
%% Generate a function that returns information
%% about the callback functions.
gen_callback(S) ->
    CbackSyms = get_cback_syms(S),
    Str = case get_option(S,cback_mod) of
	      '_' ->
		  iggenerator:term2str(build_cback_list(S,CbackSyms));
	      CbackMod ->
		  iggenerator:term2str([{CbackMod,sym_to_atom(CbackSyms)}])
	  end,
    ?lc(["\n'IG_callbacks'() ->\n\t",Str,".\n"]).

%% Each callback should be implemented in its corresponding
%% <DotHfile>_cb.erl file, so we now have to build a structure
%% representing this.
build_cback_list(S,Syms) ->
    build_cback_list(S,Syms,[]).

build_cback_list(S,[Sym|Syms],Cbl) ->
    case catch igsym:out(S,{sym_file,list_to_atom(Sym#symbol.name)}) of
	Fname when list(Fname) ->
	    Base = filename:basename(filename:rootname(Fname)),
	    CbackMod = list_to_atom(Base ++ "_cb"),
	    build_cback_list(S,Syms,add_to_cbl(Cbl,CbackMod,Sym))
    end;
build_cback_list(S,[],Cbl) -> Cbl.

%% Make the symbol names atomic
sym_to_atom([Sym|Syms]) when list(Sym#symbol.name) ->
    [Sym#symbol{name=list_to_atom(Sym#symbol.name)}|sym_to_atom(Syms)];
sym_to_atom([Sym|Syms]) ->
    [Sym|sym_to_atom(Syms)];
sym_to_atom([]) -> [].

%% Store the Module and CbackFunc as atoms to speed 
%% up the dispatching at runtime.
add_to_cbl([{Fname,Syms}|T],Fname,Sym) -> 
    [{Fname,[Sym#symbol{name=list_to_atom(Sym#symbol.name)}|Syms]}|T];
add_to_cbl([H|T],Fname,Sym) -> 
    [H|add_to_cbl(T,Fname,Sym)];
add_to_cbl([],Fname,Sym) -> 
    [{Fname,[Sym#symbol{name=list_to_atom(Sym#symbol.name)}]}].

get_cback_syms(S) ->
    get_cback_syms(S,igsym:first_sym(S)).

get_cback_syms(S,{ok,Sym}) when Sym#symbol.tag=/=?IG_FUN_TAG ->
    [Sym|get_cback_syms(S,igsym:next_sym(S,Sym#symbol.no))];
get_cback_syms(S,{ok,Sym}) when record(Sym,symbol) ->
    get_cback_syms(S,igsym:next_sym(S,Sym#symbol.no));
get_cback_syms(S,_) -> [].


%% -CSASL-
%%
%%	The Hook parameter is to suppress generation of the hooks in
%%	Erlang.
%%
generator(S,Fd,Mod,[{Ftype,Line,Fun,Rtype,InArgs,Sym}|T],Hooks) ->
    case catch generate(S,Fd,Mod,Ftype,Fun,Rtype,InArgs,Sym) of
	{local_error,What} -> 
	    fatal_error(S,Line,What);
	ok ->
	    generator(S,Fd,Mod,T,Hooks)
    end;
generator(S,Fd,Mod,[],Hooks) -> 
    ?TELL("\n",[]),
    [].


generate(S,Fd,Mod,Ftype,Fun,Rtype,InArgs,Sym) ->
    evar_clear(S),
    {Args,Guards} = gen_args(S,InArgs),

%%%    %% First add a wrapper for auto registered name
%%%    emit(Fd,["\n",quote(Fun),"(",mk_comma_str([Args]),") ->\n"]),
%%%    emit(Fd,["\t",quote(Fun),"(",mk_comma_str([Mod|Args]),").\n"]),

    %% Then add a wrapper for registered names
    Reg = evar_get(S),
    RegGuard = ?lc(["atom(",Reg,")"]),

    %% Generate the clause head...
    emit(Fd,[quote(Fun),"(",mk_comma_str([Reg|Args]),
	     ") when ",
	     mk_comma_str([RegGuard])," ->\n"]),
    emit(Fd,["\t",quote(Fun),"(",mk_comma_str([?lc(["whereis(",Reg,")"])|Args]),
	     ");\n"]),
    %% Then generate the Pid case
    Port = Reg,
    PortGuard = ?lc(["pid(",Port,")"]),
    %% Generate the clause head...
    emit(Fd,[quote(Fun),"(",mk_comma_str([Port|Args]),
	     ") when ",
	     mk_comma_str([PortGuard|Guards])," ->\n"]),
    %% ...and the packing of the tagged data...
    Bin = evar_get(S),
    emit(Fd,["\t",Bin," = term_to_binary({",
	     Sym#symbol.no,",",
	     "{",mk_comma_str(Args),"}}),\n"]),
    %% ...the sending to the port, and the returned 
    %% result (if any)
    gen_send(S,Fd,Ftype,Rtype,Port,Bin),
    ?TICK,
    ok.

gen_send(S,Fd,ig_void,_,Port,Bin) ->
    gen_async(Fd,Port,Bin);
gen_send(S,Fd,igset_var,_,Port,Bin) ->
    gen_async(Fd,Port,Bin);
%%gen_send(S,Fd,_,D,Port,Bin) when D#descriptor.cat==void ->
%%    gen_async(Fd,Port,Bin);
gen_send(S,Fd,_,_,Port,Bin) ->
    gen_sync(S,Fd,Port,Bin).

gen_async(Fd,Port,Bin) ->
    emit(Fd,["\t",Port," ! ","{self(),{async,",Bin,"}},\n"]),
    emit(Fd,["\tok.\n"]).

gen_sync(S,Fd,Port,Bin) ->
    emit(Fd,["\t",Port," ! ","{self(),{sync,",Bin,"}},\n"]),
    Res = evar_get(S),
    emit(Fd,["\treceive {",Port,",{data,",Res,"}} ",
	     "-> {ok,",Res,"} end.\n"]).


quote(Atom) when atom(Atom)   -> quote(atom_to_list(Atom));
quote([H|T]) when H>=$A,H=<$Z -> ?lc(["'",[H|T],"'"]);
quote(L) -> L.


gen_args(S,InArgs) ->
    ?DBG("gen_args: ~w~n",[InArgs]),
    gen_args(S,InArgs,[],[]).

gen_args(S,[D],A,G) when D#descriptor.cat==void -> 
    gen_args(S,[],A,G);
gen_args(S,[H|T],A,G) when record(H,descriptor) -> 
    Gtype = untangel_type(S,H),
    {Arg,Guard} = gen_arg(S,Gtype),
    gen_args(S,T,[Arg|A],[Guard|G]);
gen_args(_,[],A,G) -> 
    {lists:reverse(A),lists:reverse(G)}.

untangel_type(S,D) when D#descriptor.cat==type -> 
    resolve_type(S,D#descriptor.tname);
untangel_type(S,D) when record(D,descriptor) -> 
    {D#descriptor.cat,D#descriptor.ansi_string}.

gen_arg(S,{ptr,_}) -> 
    Evar = evar_get(S),
    Guard = ?lc(["integer(",Evar,"),",Evar,">=0,",Evar,"=<?ULONG_MAX"]),
    {Evar,Guard};
%    gen_pointer(S);
gen_arg(S,{type_struct,Es,Name}) -> gen_type_struct(S,Es,Name);
gen_arg(S,{'IG_string',_}) -> 
    Evar = evar_get(S),
    Guard = ?lc(["list(",Evar,")"]),
    {Evar,Guard};
gen_arg(S,{binaryPtr,_}) -> 
    Evar = evar_get(S),
    Guard = ?lc(["binary(",Evar,")"]),
    {Evar,Guard};
gen_arg(S,{std,StdType}) ->			% Clause changed: OTP-1450
    Evar = evar_get(S),
    Guard = case iggenC:is_float(StdType) of
		true ->
		    ?lc(["float(",Evar,")"]);
		false ->
		    gen_simple_guard(Evar,StdType)
	    end,
    %%    Guard = gen_simple_guard(Evar,StdType),
    {Evar,Guard};
gen_arg(S,Type) ->
    local_error({unexpected_type,Type}).

gen_simple_guard(Evar,Type) ->
    case is_unsigned(Type) of
	true -> ?lc(["integer(",Evar,"),",Evar,">=0,",Evar,"=<?ULONG_MAX"]);
	_    -> ?lc(["integer(",Evar,"),",Evar,">=?LONG_MIN,",Evar,"=<?LONG_MAX"])
    end.

is_unsigned(StdType) ->
    case string:index(StdType,"unsigned") of
	0 -> false;
	_ -> true
    end.

%gen_pointer(S) ->
%    B1 = evar_get(S),B2 = evar_get(S),    
%    B3 = evar_get(S),B4 = evar_get(S),
%    Ptr = ?lc(["{",B1,",",B2,",",B3,",",B4,"}"]),
%    Guards = ?lc(["integer(",B1,"),",B1,">=0,",
%		  "integer(",B2,"),",B2,">=0,",
%		  "integer(",B3,"),",B3,">=0,",
%		  "integer(",B4,"),",B4,">=0"]),
%    {Ptr,Guards}.

gen_type_struct(S,Es,Name) ->
    {Args,Guards} = gen_type_struct(S,Es,[],[]),
    %% Ugly, should include proper .hrl file instead !!
    {?lc(["{",atom_as_list(Name),",", 
	  lists:flatten(mk_comma_str(Args)),"}"]),
     lists:flatten(mk_comma_str(Guards))}.
    
gen_type_struct(S,[{struct_elem,Type,_}|T],A,G) ->
    {A1,G1} = gen_arg(S,Type),
    gen_type_struct(S,T,[A1|A],[G1|G]);
gen_type_struct(S,[{struct_elem,Type,_,Len}|T],A,G) ->
    Evar  = evar_get(S),
    %% Perhaps we also should check: 
    %%   size(Evar)=<(sizeof(Type)*Len) 
    Guard = ?lc(["binary(",Evar,")"]),
    gen_type_struct(S,T,[Evar|A],[Guard|G]);
gen_type_struct(S,[{nested_struct,Es,Name,Tname}|T],A,G) ->
    {A1,G1} = gen_type_struct(S,Es,Tname),
    gen_type_struct(S,T,[A1|A],[G1|G]);
gen_type_struct(S,[],A,G) ->
    {lists:reverse(A),lists:reverse(G)}.


%% Here comes some functions for handling 
%% the allocation of Erlang variables
evar_clear(S) ->
    igsym:in(S,evar,0).

evar_get(S) ->
    case igsym:out(S,evar) of
	N when integer(N) -> 
	    igsym:in(S,evar,N+1),
	    ?lc(["V",N])
    end.

dotH_to_dotErl([$.,$h|T]) -> [$.,$e,$r,$l|T];
dotH_to_dotErl([H|T]) -> [H|dotH_to_dotErl(T)].

dotH_to_dotHrl([$.,$h|T]) -> [$.,$h,$r,$l|T];
dotH_to_dotHrl([H|T]) -> [H|dotH_to_dotHrl(T)].

dotHrl_to_def([$.,$h,$r,$l|T]) -> [$_,$h,$r,$l|T];
dotHrl_to_def([H|T]) -> [H|dotHrl_to_def(T)].

%% Throw an error in a "standard" format
fatal_error(S,Line,What) ->
    {Fname,Fd} = igsym:out(S,erl_file),
    catch file:close(Fd),
    catch file:delete(Fname),
    throw({error,{Fname,Line,iggenErl,What}}).

%% Generate a .hrl-file 
gen_hrl(S,Mod,Fname,AbsForm) ->
    case gen_hrl_p(AbsForm) of
	true ->
	    DotHrl = dotH_to_dotHrl(Fname),
	    Fd = open_file(DotHrl),
	    igsym:in(S,hrl_file,{DotHrl,Fd}),
	    DefHrl = dotHrl_to_def(filename:basename(DotHrl)),
	    emit(Fd,["-ifndef(_",DefHrl,").\n"]),
	    emit(Fd,["-define(_",DefHrl,",true).\n"]),
	    emit(Fd,["%% ",header_str(Fname),"\n"]),
	    emit(Fd,["\n%% Record definitions\n\n"]),
	    gen_records(S,Fd,AbsForm),
	    emit(Fd,["\n%% Defines\n\n"]),
%%	    RevAbsForm = lists:reverse(AbsForm),
	    RevAbsForm = AbsForm,
	    gen_defines(S,Fd,RevAbsForm,RevAbsForm),
	    emit(Fd,["\n-endif.\n\n"]),
	    ?TELL("Erlang definitions generated to file: ~s~n",[DotHrl]),
	    file:close(Fd);
	_ ->
	    ok
    end.

gen_records(S,Fd,[{type,Line,Type,Def}|T]) when list(Def) ->
    Stype = atom_as_list(Type),
    emit(Fd,["-record(",Stype,",{"]),
    Ns = get_record_names(Def),
    emit(Fd,[mk_comma_str(Ns),"}).\n"]),
    gen_records(S,Fd,T);
gen_records(S,Fd,[_|T]) ->
    gen_records(S,Fd,T);
gen_records(S,Fd,[]) -> ok.

get_record_names([M|T]) when atom(M#member.name) -> 
    [M#member.name|get_record_names(T)];
get_record_names([M|T]) when tuple(M#member.name) -> 
    [element(1,M#member.name)|get_record_names(T)];
get_record_names([]) -> [].

gen_defines(S,Fd,[{hdef,Line,Name,What}|T],Hdefs) ->
    emit(Fd,["-define( ",Name," , ",format_hdef(What,Hdefs)," ).\n"]),
    gen_defines(S,Fd,T,Hdefs);
gen_defines(S,Fd,[{ifdef,Line,Name}|T],Hdefs) ->
    emit(Fd,["-ifdef( ",Name," ).\n"]),
    gen_defines(S,Fd,T,Hdefs);
gen_defines(S,Fd,[{ifndef,Line,Name}|T],Hdefs) ->
    emit(Fd,["-ifndef( ",Name," ).\n"]),
    gen_defines(S,Fd,T,Hdefs);
gen_defines(S,Fd,[{else,Line}|T],Hdefs) ->
    emit(Fd,["-else.\n"]),
    gen_defines(S,Fd,T,Hdefs);
gen_defines(S,Fd,[{endif,Line}|T],Hdefs) ->
    emit(Fd,["-endif.\n"]),
    gen_defines(S,Fd,T,Hdefs);
gen_defines(S,Fd,[_|T],Hdefs) -> gen_defines(S,Fd,T,Hdefs);
gen_defines(S,Fd,[],_) -> ok.

format_hdef({identifier,_,Name},Hdefs) ->
    case isa_hdef(Name,Hdefs) of
	{true,_} -> ?lc(["?",Name]);
	_ -> atom_to_list(Name)
    end;
format_hdef({octalconstant,_,L},_) -> ?lc(["8#",L]);
format_hdef({hexconstant,_,L},_) -> ?lc(["16#",L]);
format_hdef({stringconstant,_,L},_) -> ?lc(["\"",L,"\""]);
format_hdef({_,_,What},_) -> What.

isa_hdef(Hdef,[{hdef,_,Hdef,What}|_]) -> {true,What};
isa_hdef(Hdef,[_|T]) -> isa_hdef(Hdef,T);
isa_hdef(_,[]) -> false.

gen_hrl_p([{type,Line,Type,Def}|T]) when list(Def) -> true;
gen_hrl_p([{hdef,Line,Name,What}|T]) -> true;
gen_hrl_p([_|T]) -> gen_hrl_p(T);
gen_hrl_p([]) -> false.

atom_as_list(Atom) when atom(Atom) ->
    atom_as_list(atom_to_list(Atom));
atom_as_list(List) when list(List) ->
    case string:words(List) of
	1 when hd(List)==$_ -> lists:append([$'|List],[$']);
	1 when hd(List)>=65,hd(List)=<90 -> lists:append([$'|List],[$']);
	N when N>1 -> lists:append([$'|List],[$']);
	_ -> List
    end.

