%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1994,1995, Ellemtel Telecommunications Systems Laboratories
%%-----------------------------------------------------------------------
%% Created: 94-10-27 by Tobbe (tobbe@erix.ericsson.se)
%% 
%% Function: The generator part of the Interface Generator
%% 
%%-----------------------------------------------------------------------
-module(iggenerator).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('tobbe@erix.ericsson.se').
-rcs('$Id: iggenerator.erl,v 1.4 1996/04/24 10:59:31 tobbe Exp $.').

-export([gen/3,open_file/1,emit/2,term2str/1,mk_comma_str/1,local_error/1]).
-export([resolve_type/2,header_str/1]).

-include("ig.hrl").
-include("igcsasl.hrl").

report_error(Fname,Line,Emsg) ->
    io:format("~n~s~n",[?lc([Fname,":",Line," ",Emsg])]).

gen(S,Fname,AbsForm) -> 
    case catch gen0(S,Fname,AbsForm) of
	{error,{_,Line,Mod,What}}  -> 
	    Emsg = apply(Mod,format_error,[What]),
	    report_error(Fname,Line,Emsg),
	    {error,generator};
	{'EXIT',Reason} -> 
	    io:format("~nGenerator crashed, flame the responsible!~n"),
	    {error,Reason};
	Gen ->
	    ?TELL("Generator ready~n", []),
	    Gen
    end.


gen0(S,Fname,AbsForm) -> 
    ?DBG("Enter iggenerator: ~w~n",[AbsForm]),
    igsym:store_types(S,AbsForm),
    iggenC:start(S,Fname,AbsForm,?CSASLHOOKS),
    iggenErl:start(S,Fname,AbsForm,?CSASLHOOKS).


%% Return a string of elements where the elements
%% are separated by a comma.
mk_comma_str([H|[]]) -> H;
mk_comma_str([H|T]) -> 
    ?lc([H,",",mk_comma_str(T)]);
mk_comma_str([]) -> [].


open_file(Fname) ->
    case file:open(Fname,write) of
	{error,Reason} -> throw({error,Reason});
	{ok,Fd} -> Fd
    end.

to_binary([Bin|T]) when binary(Bin) ->
    [Bin|to_binary(T)];
to_binary([List|T]) when list(List) ->
    [list_to_binary(List)|to_binary(T)];
to_binary([]) -> [].


emit(dummy,Stmt) -> 
%%    io:format("~s",[?lc(Stmt)]),
    ok;
emit(Fd,Stmt) ->
    io:put_chars(Fd,?lc(Stmt)).
%%    io:format("~s",[?lc(Stmt)]).


term2str(Term) -> 
    lists:flatten(io_lib:format("~p",[Term])).


%%-------------------------------------------------------
%% Resolve a type definition, i.e untangel it to only
%% consist of predefined datatypes. A typedef-struct is 
%% tagged with 'type_struct' at the top-level and may 
%% contain 'struct_elem'- and 'nested_struct' tagged data.
%%-------------------------------------------------------
resolve_type(S,Tname) when atom(Tname) ->
    case igsym:get_type_def(S,Tname) of
	{type,_,Tname,D} when record(D,descriptor) ->
	    case D#descriptor.cat of
		type -> resolve_type(S,D#descriptor.tname);
		Cat  -> {Cat,D#descriptor.ansi_string}
	    end;
	{type,_,Tname,D} when list(D) ->
	    {type_struct,resolve_type_struct(S,D),Tname}
    end.

%% Resolve each element in a typedef struct.
resolve_type_struct(S,[M|Tail]) when record(M,member) ->
    Member = resolve_type_struct_member(S,M#member.desc,M#member.name),
    [Member|resolve_type_struct(S,Tail)];
resolve_type_struct(S,[]) -> [].

resolve_type_struct_member(S,D,Name) when D#descriptor.cat==type ->
    case igsym:get_type_def(S,D#descriptor.tname) of
	{type,_,Tname,D2} when record(D2,descriptor) ->
	    case D2#descriptor.cat of
		type -> 
		    case resolve_type(S,D2#descriptor.tname) of
			{type_struct,Es,Tname} ->
			    {nested_struct,Es,Name,Tname};
			ResType ->
			    {struct_elem,ResType,Name}
		    end;
		Cat  -> 
		    {struct_elem,{Cat,D2#descriptor.ansi_string},Name}
	    end;
	{type,_,Tname,M} when list(M) ->
	    {nested_struct,resolve_type_struct(S,M),Name,Tname}
    end;
resolve_type_struct_member(S,D,Name) when D#descriptor.cat=/=type,atom(Name) ->
    X = {D#descriptor.cat,D#descriptor.ansi_string},
    {struct_elem,X,Name};
resolve_type_struct_member(S,D,{Name,Len}) when D#descriptor.cat=/=type ->
    X = {D#descriptor.cat,D#descriptor.ansi_string},
    {struct_elem,X,Name,Len};
resolve_type_struct_member(S,_,_) -> 
    local_error({internal,?LINE}).


%% To be used when to escape deep down...
local_error(What) -> throw({local_error,What}).


header_str(From) ->
    ?lc(["Created: ",date_str(), ", by IG from the input file: ",From]).

date_str() ->
    {Y,M,D} = date(),
    ?lc([D," ",month(M)," ",Y]).

month(1)  -> "Jan";
month(2)  -> "Feb";
month(3)  -> "Mar";
month(4)  -> "Apr";
month(5)  -> "May";
month(6)  -> "Jun";
month(7)  -> "Jul";
month(8)  -> "Aug";
month(9)  -> "Sep";
month(10) -> "Oct";
month(11) -> "Nov";
month(12) -> "Dec".
