%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1994,1995, Ellemtel Telecommunications Systems Laboratories
%%-----------------------------------------------------------------------
%% Created: 94-10-27 by Tobbe (tobbe@erix.ericsson.se)
%% 
%% Function: The IG server
%%
%%-----------------------------------------------------------------------
%% Modified: 25 Jan 1999 by tobbe@serc.rmit.edu.au
%% Function: Added possibility to run the Erlang node in the background.
%%-----------------------------------------------------------------------
-module(igserver).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('tobbe@erix.ericsson.se').

-rcs('$Id: igserver.erl,v 1.7 1996/04/24 10:59:22 tobbe Exp $.'). 

%% Fnutt's fixing thing


%% Public interface
-export([start/2, sup_start/3, adr_of/2, sizeof/2]).

-include("ig.hrl").
-include("igcsasl.hrl").



%% Spawn exports
-export([init_it/4, listener/2, linit/3, sloop/6, listener/2, killer/6]).

%% CSASL exports 
-export([system_continue/3, system_terminate/4, system_code_change/4,
	 print_event/3]).



%% Really bad macros to handle reporting of all received messages
%% while still matching out stuff from within messages. Note also that
%% the variable _NewDebug_ is used in a bad, bad way at some of the
%% calls of this macro.
-define(rec_case(Debug,Case,Res),
	Case when Debug==[] -> 
	       _NewDebug_ = [],
	       Res;
	Case ->
	       _NewDebug_ = sys:handle_debug(Debug,{igserver,print_event},
					     self(),{in,Case}),
	       Res).

-define(wait_for(Debug,Case,Result),
	receive
	    ?rec_case(Debug,Case,Result)
	end).
-define(wait_for2(Debug,Case1,Result1,Case2,Result2),
	receive
	    ?rec_case(Debug,Case1,Result1);
	    ?rec_case(Debug,Case2,Result2)
	end).

-define(wait_for3(Debug,Case1,Result1,Case2,Result2,Case3,Result3),
	receive
	    ?rec_case(Debug,Case1,Result1);
	    ?rec_case(Debug,Case2,Result2);
	    ?rec_case(Debug,Case3,Result3)
	end).





%% Note that these messages received here are received outside the
%% igserver (these functions are a public interface)
adr_of(S,Name) when list(Name) ->
    adr_of(S,list_to_atom(Name));
adr_of(S,Name) when pid(S),atom(Name) ->
    S ! {self(),{cb_adr,Name}},
    receive {S,Res} -> Res end.

sizeof(S,Name) when pid(S) ->
    S ! {self(),{sizeof,list_to_atom(?lc([?SIZE_PREFIX,Name]))}},
    receive {S,Res} -> Res end.


%% Public start functions
start(Exec, Callbacks) ->
    element(2, proc_lib:start_link(igserver, init_it, 
				   [self(), Exec, Callbacks, []])).
sup_start(Name, Exec, Callbacks) ->
    proc_lib:start_link(igserver, init_it, [self(), Exec, Callbacks, []]).


%%%% More cases could be added later
%%igregister({local, Name}, Pid) ->
%%    register(Name, igget_pid(Pid)).

%%igget_pid(P) when pid(P) -> P;
%%igget_pid({ok, P}) when pid(P) -> P.


init_it(From, Exec, Callbacks, Options) ->
    process_flag(trap_exit,true),
    Debug = [],					% Should use sys:debug_opts
    case catch init(From, Exec, Callbacks, Debug) of
	{server, Pid} ->
	    proc_lib:init_ack(From, {ok, self()}),
	    killer(From, Pid, Debug, Exec, Callbacks, []);
	{ok, Conn, S} ->
	    proc_lib:init_ack(From, {ok, self()}),
	    sloop(From, Conn, S, Debug, Exec, Callbacks);
	Else ->
	    Error=Else,
%%	    Error = {bad_return_value, Else},
	    proc_lib:init_ack(From, {error, Error}),
	    Error
    end.


%% Init shall set up a connection (complete with call back tables if
%% needed) and return a handle for the connection and the cb table
init(From,Exec,Callbacks,Debug) ->
    init(From,Exec,Callbacks,{[]},Debug).
init(From,Exec,Callbacks,InitState,Debug) ->
    case setup_connection(Exec) of
	{ok,Conn} ->
	    S = igsym:new(),
	    Len = store_cbacks(S,Callbacks),
	    sinit(From, Conn, S, Len, Debug), % init callbacks
	    call_hook(Conn, init, InitState),
	    {ok, Conn, S};
	Other ->
	    Other
    end.


setup_connection({sup_port, LibName, ProgName, Args}) ->
    PortProg = ?lc([code:priv_dir(LibName),"/bin/",ProgName]),
    Port = open_port({spawn,PortProg},[binary,{packet,2}|Args]),
    Conn = {port,Port},
    {ok,Conn};
setup_connection({{spawn,Exec},Stdio}) ->
    Port = open_port({spawn,Exec},[binary,{packet,2}|Stdio]),
    Conn = {port,Port},
    {ok,Conn};
setup_connection({client,Adr,PortNo}) ->
    socket:start(),
    Socket = socket:client('STREAM','AF_INET',{Adr, PortNo},{binary_packet,2}),
    Conn   = {client,Socket},
    {ok,Conn};
setup_connection({server,PortNo}) ->
    socket:start(),
    Pid = spawn_link(igserver,listener,[self(),PortNo]),
    {server,Pid};
setup_connection({{fd,In,Out},Stdio}) ->
    Port = open_port({fd,In,Out},[binary,{packet,2}|Stdio]),
    Conn = {port,Port},
    {ok,Conn};
setup_connection(_) ->
    false.




store_cbacks(S,[{CbackMod,Cbacks}|T]) ->
    L = store_callbacks(S,CbackMod,Cbacks),
    L + store_cbacks(S,T);
store_cbacks(_,[]) -> 0.

store_callbacks(S,CbackMod,Cbacks) ->
    store_callbacks(S,CbackMod,Cbacks,0).
store_callbacks(S,CbackMod,[H|T],Len) when record(H,symbol) ->
    igsym:in(S,{cback_mod,H#symbol.no},CbackMod),
    igsym:in(S,H#symbol.no,H),
    store_callbacks(S,CbackMod,T,Len+1);
store_callbacks(_,_,[],Len) -> Len.
    

sinit(From,{server,Socket},S,_,Debug) ->
    %% A server won't get any callback addresses
    ok;
sinit(From,Conn,S,0,Debug) ->
    ok;
sinit(From,Conn,S,N,Debug) when N>0 ->
    %% Collect all the callback addresses
    case get_cback_msg(Conn,Debug) of
	Res when binary(Res) ->
	    case binary_to_term(Res) of
		{No,Data}  ->
		    case catch igsym:out(S,No) of
			X when X#symbol.tag==?IG_SIZE_TAG ->
			    igsym:in(S,X#symbol.name,Data),
			    sinit(From,Conn,S,N-1,Debug);
			X when record(X,symbol) ->
			    igsym:in(S,X#symbol.name,Data),
			    sinit(From,Conn,S,N-1,Debug);
			_ ->
			    io:format("Ignoring no: ~p~n",[No]),
			    sinit(From,Conn,S,N,Debug)
		    end;
		XX ->
		    io:format("~p:~p Mega error in sinit, got: ~p~n",
			      [?FILE,?LINE,XX]),
		    exit(kill)
	    end;
	{'EXIT',Reason} ->
	    exit(Reason)
    end.


%% -CSASL -
%%
%%	recursive call to sloop moved from outside case into each
%%	branch to let system messages change the server state
%%
sloop(From,Conn,S,Debug,Exec,Callbacks) ->
    Msg = receive
	      Input ->
		  Input
	  end,
    case Msg of
	{system, Who, Req} ->
	    handle_sys_msg(Who,Req,From,Conn,S,Debug,Exec,Callbacks);
	{'EXIT', From, Reason} ->
	    terminate(From,Conn,S);
	_Msg when Debug == [] ->
	    handle_msg(Msg,From,Conn,S,Debug,Exec,Callbacks);
	_Msg ->
	    Debug1 = sys:handle_debug(Debug, {igserver, print_event}, 
				      self(), {in, Msg}),
	    handle_msg(Msg,From,Conn,S,Debug1,Exec,Callbacks)
    end.


handle_msg(Msg,From,Conn,S,Debug,Exec,Callbacks) ->
    case Msg of
	{_,{async,Bin}}   -> 
	    Debug1 = reply(Conn,Bin,Debug),
	    %%send_fun_msg(Conn,Bin),
	    igserver:sloop(From,Conn,S,Debug1,Exec,Callbacks);
	{Who,{sync,Bin}} ->
	    Debug1 = reply(Conn,Bin,Debug),
	    %%send_fun_msg(Conn,Bin),
%%	    {data,Res,Debug2} = get_fun_msg(Conn,Debug1),
	    case get_fun_msg(Conn, Debug1) of
		{data, Res, Debug2} ->
		    Debug3 = reply(Who,{self(),
					{data,binary_to_term(Res)}},Debug2),
		    igserver:sloop(From,Conn,S,Debug3,Exec,Callbacks);
		{closed,Reason} -> 
		    terminate(From,no_conn,S),
		    Reason;
		{'EXIT',Reason} -> 
		    terminate(From,no_conn,S),
		    exit(Reason)
	    end;
	{_,stop} ->
	    terminate(From,Conn,S),
	    stopped;
	{Who,{cb_adr,Name}} -> 
	    Debug1 = case catch igsym:out(S,Name) of
			 {error,Reason} -> %%Who ! {self(),{error,Reason}};
			     reply(Who,{self(),{error,Reason}},Debug);
			 Ptr -> %%Who ! {self(),{ok,Ptr}}
			     reply(Who,{self(),{ok,Ptr}},Debug)
		     end,
	    igserver:sloop(From,Conn,S,Debug1,Exec,Callbacks);
	{Who,{sizeof,Name}} ->
	    Debug1 = case catch igsym:out(S,Name) of
			 {error,Reason} -> %%Who ! {self(),{error,Reason}};
			     reply(Who,{self(),{error,Reason}},Debug);
			 Size -> %%Who ! {self(),{ok,Size}}
			     reply(Who,{self(),{ok,Size}},Debug)
		     end,
	    igserver:sloop(From,Conn,S,Debug1,Exec,Callbacks);
	Other ->
	    %% Connection specific message
	    case get_conn_msg(Conn,Msg) of
		{data,Res} ->
		    Debug1 = dispatch_callback(From,Conn,S,
					       binary_to_term(Res),Debug),
		    igserver:sloop(From,Conn,S,Debug1,Exec,Callbacks);
		{closed,Reason} -> 
		    terminate(From,no_conn,S),
		    Reason;
		{'EXIT',Reason} -> 
		    terminate(From,no_conn,S),
		    exit(Reason);
		XX ->
		    io:format("ig server got unexpected msg: ~p~n",[XX]),
		    igserver:sloop(From,Conn,S,Debug,Exec,Callbacks)
	    end
    end.

get_conn_msg({server,Socket},Msg) ->
    case Msg of
	{Socket,{fromsocket,Res}}	-> {data,Res};
	{Socket,{socket_closed,Reason}}	-> {closed,Reason};
	{'EXIT',Socket,Reason}		-> {'EXIT',Reason};
	XX -> {msg,XX}
    end;
get_conn_msg({client,Socket},Msg) ->
    case Msg of
	{Socket,{fromsocket,Res}}	-> {data,Res};
	{Socket,{socket_closed,Reason}}	-> {closed,Reason};
	{'EXIT',Socket,Reason}		-> {'EXIT',Reason};
	XX				-> {msg,XX}
    end;
get_conn_msg({port,Port},Msg) ->
    case Msg of 
	{Port,{data,Res}}	-> {data,Res};
	{'EXIT',Port,Reason}	-> {closed,Reason};
	XX			-> {msg,XX}
    end.
	

%% Shall return a new debug structure
reply(To,Msg,Debug) ->
    %%io:format("~nreply(_, _, ~p)~n",[Debug]),
    if  Debug == [] -> real_reply(To,Msg), [];
	true ->
	    real_reply(To,Msg),
	    sys:handle_debug(Debug,{igserver,print_event},self(),
			     {out,Msg,To})
    end.

real_reply({port,Port},Msg) ->
    Port ! {self(),{command,Msg}};
real_reply({client,Socket},Msg) ->
    Socket ! {self(),{deliver,Msg}};
real_reply({server,Socket},Msg) ->
    Socket ! {self(),{deliver,Msg}};
real_reply(Pid,Msg) ->
    %% We take a chance here, assuming Pid is either a pid or a
    %% registered name
    Pid!Msg.


%% Must return {data, Result, NewDebug}
get_fun_msg({port,Port},Debug) ->
    %% Note extremely ugly use of macro variable _NewDebug_
    ?wait_for2(Debug,{Port,{data,Res}}, {data,Res,_NewDebug_},
	       {'EXIT',Port,Reason}, {closed,Reason});

get_fun_msg({server,Socket},Debug) ->
    %% Note extremely ugly use of macro variable _NewDebug_
    ?wait_for3(Debug,{Socket,{fromsocket,Res}},{data,Res,_NewDebug_},
	       {Socket,{socket_closed,Reason}}, {closed,Reason},
	       {'EXIT',Socket,Reason}, {'EXIT',Reason});

get_fun_msg({client,Socket},Debug) ->
    %% Note extremely ugly use of macro variable _NewDebug_
    ?wait_for3(Debug,{Socket,{fromsocket,Res}},{data,Res,_NewDebug_},
	       {Socket,{socket_closed,Reason}}, {closed,Reason},
	       {'EXIT',Socket,Reason}, {'EXIT',Reason}).

get_cback_msg({port,Port},Debug) ->
    ?wait_for2(Debug,{Port,{data,Res}},Res,
	       {'EXIT',Port,Reason},{'EXIT',Reason});
get_cback_msg({client,Socket},Debug) ->
    ?wait_for3(Debug,{Socket,{fromsocket,Res}},Res,
	       {Socket,{socket_closed,Reason}},{'EXIT',Reason},
	       {'EXIT',Socket,Reason},{'EXIT',Reason});
get_cback_msg({server,Socket},Debug) ->
    ?wait_for3(Debug,{Socket,{fromsocket,Res}},Res,
	       {Socket,{socket_closed,Reason}},{'EXIT',Reason},
	       {'EXIT',Socket,Reason},{'EXIT',Reason}).



%% Must return a new debug structure (such as reply/3 returns)
dispatch_callback(From,Conn,S,[No|Args],Debug) ->
    CbackMod = igsym:out(S,{cback_mod,No}),
    case catch igsym:out(S,No) of
 	X when X#symbol.tag==?IG_CAST_TAG ->
	    spawn(CbackMod,X#symbol.name,[From|Args]),
	    Debug;
 	X when X#symbol.tag==?IG_CALL_TAG ->
	    %% If we crashes, then fine by us...
	    reply(Conn, term_to_binary(apply(CbackMod,
					     X#symbol.name,
					     [From|Args])),
		  Debug);
	_ ->
	    io:format("<ERROR> in ig server, no callback found: ~p~n",
		      [{CbackMod,[No|Args]}]),
	    Debug
    end.


%%-------------------------------------------------------
%% The 'Killer' acts as a supervisor over the 'Listener'
%% and all the servers that may be running. If the initiating
%% application dies, the killer ensures that nothing will be
%% left hanging.
%%-------------------------------------------------------

killer(From,Listener,Debug,Exec,Cb,Servers) ->
    receive
	%% A new connection. Start a server for it !!
	{Server,iam_new} ->
	    %%link(Listener),
	    link(Server),
	    Server ! {killer,self(),Exec,Cb,Debug},
	    %%Server ! {self(),SS,Len},
	    igserver:killer(From,Listener,Debug,Exec,Cb,[Server|Servers]);
	{From,stop} ->
	    lists:foreach(fun(X) -> exit(X, normal) end, Servers),
	    stopped;
	{'EXIT',From,Reason} ->
	    exit(Reason);
	{'EXIT',Listener,Reason} ->
	    exit(listener_died);
	{'EXIT',Server,Reason} ->
	    NewServers = lists:filter(fun(X) when X==Server -> false;
				      (X) -> true end,
				      Servers),
	    igserver:killer(From,Listener,Debug,Exec,Cb,NewServers);
	Msg ->
	    lists:foreach(fun(X) -> catch (X!Msg) end, Servers),
	    igserver:killer(From,Listener,Debug,Exec,Cb,Servers)
    end.

%%-------------------------------------------------------
%% We may receive several connections so we always have
%% one process waiting for a connect attempt.
%%-------------------------------------------------------

listener(Killer,PortNo) ->
    process_flag(trap_exit,true),
    ListenSocket = socket:listen('STREAM','AF_INET',PortNo,{binary_packet,2}),
    listen(Killer,ListenSocket).

listen(Killer,ListenSocket) ->
    Pid = spawn_link(igserver,linit,[self(),Killer,ListenSocket]),
    listen_wait(Killer,ListenSocket,Pid).

listen_wait(Killer,ListenSocket,Pid) ->
    receive
	{Pid,ok} -> listen(Killer,ListenSocket);
	{'EXIT',Killer,Reason} -> exit(Reason);
	_ -> listen_wait(Killer,ListenSocket,Pid)
    end.

linit(Listener,Killer,ListenSocket) ->
    Sock = socket:accept(ListenSocket),
    Listener ! {self(),ok},
    Killer ! {self(),iam_new},
    receive
	{killer,Killer,Exec,CallBacks,Debug}-> 
	    SS = igsym:new(),
	    Len = store_cbacks(SS,CallBacks),
	    igserver:sloop(Killer,{server,Sock},SS,Debug,Exec,CallBacks)
    end.



%%-------------------------------------------------------
%% - CSASL -
%%
%%	Handle system messages
%%
%%-------------------------------------------------------
handle_sys_msg(Who,Req,From,Conn,S,Debug,Exec,Callbacks)  ->
    Extra = case Req of
		suspend ->
		    call_hook(Conn,suspend,{}),
		    [];
		get_status ->
		    {ok,Status} = call_hook(Conn,get_status,{}),
		    {c_status,Status};
		{debug, {trace,Flag}} ->
		    call_hook(Conn,trace,{flag_to_int(Flag)}),
		    [];
		{debug, {log,Flag}} ->
		    call_hook(Conn,log,{flag_to_int(Flag)}),
		    [];
		{debug, {log_to_file,FName}} ->
		    call_hook(Conn,log_to_file,{FName}),
		    [];
		{debug, no_debug} ->
		    call_hook(Conn,no_debug,{}),
		    [];
		Other ->
		    []
	    end,
    handle_sys_msg2(Who,Req,From,Conn,S,Debug,Exec,Callbacks,Extra).

    
handle_sys_msg2(Who,Req,From,Conn,S,Debug,Exec,Callbacks,Extra) ->
    sys:handle_system_msg(Req, Who, self(), ?MODULE, Debug, 
			  [Req,From,Conn,S,Exec,Callbacks,Extra]).

flag_to_int(true) -> 1;
flag_to_int(_) -> 0.

%%-------------------------------------------------------
%% - CSASL -
%%
%%	Callbacks from sys.erl
%%
%%-------------------------------------------------------
system_continue(_Parent,Debug,Misc) ->
    %%io:format("~nMisc was: ~p~n",[Misc]),
    [Why,From,Conn,S,Exec,Callbacks,Extra] = Misc,
    if  Why==suspend -> 
	    call_hook(Conn,resume,{}),
	    igserver:sloop(From,Conn,S,Debug,Exec,Callbacks);
	Why==stop -> 
	    terminate(From,Conn,S),
	    stopped;
	true -> 
	    igserver:sloop(From,Conn,S,Debug,Exec,Callbacks)
    end.

system_terminate(Reason,Parent,Debug,Misc) ->
    [Why,From,Conn,S,Exec,Callbacks,Extra] = Misc,
    terminate(From,Conn,S),
    {ok, Misc}.


%% - CSASL -
%% system_code_change
%%
%%	The code have been changed on the Erlang side and now we must
%%	change the C program which we do by killing it and trying to
%%	restart it. We may have a new start specification supplied in
%%	the How parameter. Note that we're in a catch so we only try
%%	to be successfull (not covering all possibilities)
%%
system_code_change(Misc,_Mod,_Vsn,How) ->
    [Why,From,Conn,S,Exec,Callbacks,Extra] = Misc,
    do_code_change(Why,From,Conn,S,Exec,Callbacks,Extra,How).


%% - CSASL -
%%
%%	Actual code change. Note that servers generally don't change
%%	code, but rather dies when resumed.
%%
do_code_change(Why, From, {server,Socket}, S, Exec, Callbacks, Extra, How) ->
    {ok,[stop,From,{server,Socket},S,Exec,Callbacks,Extra]};
do_code_change(Why, From, Conn, S, Exec, Callbacks, Extra, How) ->
    {ok, Saved} = call_hook(Conn, get_state, {}),
    terminate(From, Conn, S),
    case How of
	{newexec, NewExec} ->
	    {ok, NewConn, NewS} = init(From, NewExec, Callbacks, {Saved},[]),
	    {ok,[Why, From, NewConn, NewS, NewExec, Callbacks, Extra]};
	{startspec, {M, F, A}} ->
	    catch apply(M, F, A),
	    {ok, NewConn, NewS} = init(From, Exec, Callbacks, {Saved}, []),
	    {ok,[Why, From, NewConn, NewS, Exec, Callbacks, Extra]};
	Other ->
	    {ok,NewConn,NewS} = init(From,Exec,Callbacks,{Saved},[]),
	    {ok,[Why,From,NewConn,NewS,Exec,Callbacks,Extra]}
    end.




%%
%% call_hook
%%
%%	The name passed is used to find the exact hook name and its
%%	index in the list of funs (on the C side). Note that hooks
%%	never should result in any debug information traces and
%%	therefore [] is sent to reply/3
%%
call_hook(Conn,Name,Args) ->
    %%io:format("~p: Call hook: ~p~n",[self(),Name]),
    case ?get_hook(Name,?CSASLHOOKS) of
	{Hook,Ix} ->
	    V = term_to_binary({Ix,Args}),
	    reply(Conn,V,[]),
	    %%send_fun_msg(Conn,V),
	    case get_fun_msg(Conn,[]) of
		{data,Result,_} ->
		    {ok,binary_to_term(Result)};
		{closed,Reason} -> 
		    exit(Reason);
		{'EXIT',Reason} -> 
		    exit(Reason)
	    end;
	Other ->
	    Other
    end.
    
cast_hook(Conn,Name,Args) ->
    %%io:format("~p: Cast hook: ~p~n",[self(),Name]),
    case ?get_hook(Name,?CSASLHOOKS) of
	{Hook,Ix} ->
	    V = term_to_binary({Ix,Args}),
	    reply(Conn,V,[]);
%%	    send_fun_msg(Conn,V);
	Other ->
	    Other
    end.

terminate(From,Conn,S) ->
    igsym:delete(S),
    close_conn(Conn).
%%    ok.

		
%% Do not wait forever on sockets
close_conn(no_conn) -> ok;
close_conn({client,Socket}) ->
    cast_hook({client,Socket},terminate,{}),
    receive
	{Socket,{socket_closed,normal}} -> normal
    after 100 -> normal
    end;
close_conn({server,Socket}) ->
    cast_hook({server,Socket},terminate,{}),
    receive
	{Socket,{socket_closed,normal}} -> normal
    after 100 -> normal
    end;
close_conn({port,Port}) ->
    cast_hook({port,Port},terminate,{}),
    ?wait_for([],{'EXIT',Port,Reason},Reason).	%use [] for Debug



%%-----------------------------------------------------------------
%% Format debug messages.  Print them as the call-back module sees
%% them, not as the real erlang messages.  Use trace for that.
%%-----------------------------------------------------------------
print_event(Dev, {in, Msg}, Name) ->
    case Msg of
	{From,{async,Bin}}   -> 
	    io:format(Dev, "*DBG* ~p got cast ~p from ~w~n",
		      [Name,binary_to_term(Bin),From]);
	{Who,{sync,Bin}} ->
	    io:format(Dev, "*DBG* ~p got call ~p from ~w~n",
		      [Name,binary_to_term(Bin),Who]);
	{Who,stop} ->
	    io:format(Dev, "*DBG* ~p got stop from ~w~n",
		      [Name,Who]);
	{Who,{cb_adr,Fun}} -> 
	    io:format(Dev, "*DBG* ~p got callback adr request ~p from ~w~n",
		      [Name,Fun,Who]);
	{Who,{sizeof,Fun}} ->
	    io:format(Dev, "*DBG* ~p got sizeof request ~p from ~w~n",
		      [Name,Fun,Who]);
	{Socket,{fromsocket,Res}} ->
	    io:format(Dev, "*DBG* ~p got socket data ~p from ~w~n",
		      [Name,binary_to_term(Res),Socket]);
	{Socket,{socket_closed,Reason}} ->
	    io:format(Dev, "*DBG* ~p got socket closed from ~w~n",
		      [Name,Socket]);
	{'EXIT',Socket,Reason} ->
	    io:format(Dev, "*DBG* ~p got socket EXIT from ~w~n",
		      [Name,Socket]);
	{Port,{data,Res}} when port(Port) ->
	    io:format(Dev, "*DBG* ~p got port data ~p from ~w~n",
		      [Name,binary_to_term(Res),Port]);
	{Who,{data,Res}} ->
	    io:format(Dev, "*DBG* ~p got data ~p from ~w~n",
		      [Name,binary_to_term(Res),Who]);
	_ ->
	    io:format(Dev, "*DBG* ~p got ~p~n", [Name, Msg])
    end;
print_event(Dev, {out, Msg, To}, Name) ->
    io:format(Dev, "*DBG* ~p sent ~p to ~w~n", [Name, Msg, To]);
print_event(Dev, Event, Name) ->
    io:format(Dev, "*DBG* ~p dbg  ~p~n", [Name, Event]).
