%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(inets_sup).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('jocke@erix.ericsson.se').
-export([crock/0,start/2,stop/1,init/1]).

%% crock (Used for debugging!)

crock() ->
  application:start(sasl),
  application:start(inets).

%% start

start(Type,State) ->
  supervisor:start_link({local,inets_sup},inets_sup,[]).

%% stop

stop(State) ->
  ok.

%% init

init([]) ->
  case get_services() of
    {error,Reason} ->
      {error,Reason};
    Services ->
      SupFlags={one_for_one,10,3600},
      {ok,{SupFlags,child_spec(Services,0)}}
  end.

get_services() ->
  case catch application:get_env(inets,services) of
    {ok,Services} ->
      Services;
    _ ->
      []
  end.

child_spec([],_) ->
  [];
child_spec([{httpd,ConfigFile}|Rest],N) ->
    [{{httpd,N},{httpd,start_link,[ConfigFile]},transient,brutal_kill,worker,
      [ftp,
       httpd,
       httpd_conf,
       httpd_example,
       httpd_listener,
       httpd_parse,
       httpd_request,
       httpd_response,
       httpd_socket,
       httpd_util,
       inets_sup,
       mod_actions,
       mod_alias,
       mod_auth,
       mod_cgi,
       mod_dir,
       mod_disk_log,
       mod_esi,
       mod_get,
       mod_head,
       mod_include,
       mod_log,
       mod_auth_mnesia,
       mod_auth_plain,
       mod_auth_dets,
       mod_security]}|child_spec(Rest,N+1)].

