%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(mod_actions).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('jocke@erix.ericsson.se').
-export([do/1,load/2]).

-include("httpd.hrl").

%% do

do(Info) ->
  case httpd_util:key1search(Info#mod.data,status) of
    %% A status code has been generated!
    {StatusCode,PhraseArgs,Reason} ->
      {proceed,Info#mod.data};
    %% No status code has been generated!
    undefined ->
      case httpd_util:key1search(Info#mod.data,response) of
	%% No response has been generated!
	undefined ->
	  Path=mod_alias:path(Info#mod.data,Info#mod.config_db,
			      Info#mod.request_uri),
	  Suffix=httpd_util:suffix(Path),
	  MimeType=httpd_util:lookup_mime(Info#mod.config_db,Suffix,
					  "text/plain"),
	  Actions=httpd_util:multi_lookup(Info#mod.config_db,action),
	  case action(Info#mod.request_uri,MimeType,Actions) of
	    {yes,RequestURI} ->
	      {proceed,[{new_request_uri,RequestURI}|Info#mod.data]};
	    no ->
	      Scripts=httpd_util:multi_lookup(Info#mod.config_db,script),
	      case script(Info#mod.request_uri,Info#mod.method,Scripts) of
		{yes,RequestURI} ->
		  {proceed,[{new_request_uri,RequestURI}|Info#mod.data]};
		no ->
		  {proceed,Info#mod.data}
	      end
	  end;
	%% A response has been generated or sent!
	Response ->
	  {proceed,Info#mod.data}
      end
  end.

action(RequestURI,MimeType,[]) ->
  no;
action(RequestURI,MimeType,[{MimeType,CGIScript}|Rest]) ->
  {yes,CGIScript++RequestURI};
action(RequestURI,MimeType,[_|Rest]) ->
  action(RequestURI,MimeType,Rest).

script(RequestURI,Method,[]) ->
  no;
script(RequestURI,Method,[{Method,CGIScript}|Rest]) ->
  {yes,CGIScript++RequestURI};
script(RequestURI,Method,[_|Rest]) ->
  script(RequestURI,Method,Rest).

%%
%% Configuration
%%

%% load

load([$A,$c,$t,$i,$o,$n,$ |Action],[]) ->
  case regexp:split(Action," ") of
    {ok,[MimeType,CGIScript]} ->
      {ok,[],{action,{MimeType,CGIScript}}};
    {ok,_} ->
      {error,?NICE(httpd_conf:clean(Action)++" is an invalid Action")}
  end;
load([$S,$c,$r,$i,$p,$t,$ |Script],[]) ->
  case regexp:split(Script," ") of
    {ok,[Method,CGIScript]} ->
      {ok,[],{script,{Method,CGIScript}}};
    {ok,_} ->
      {error,?NICE(httpd_conf:clean(Script)++" is an invalid Script")}
  end.
