%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : setup.erl
%%% Author  : Dan Gudmundsson <dgud@erix.ericsson.se>
%%% Purpose : Initilize the mnesia__corba_session application in erlang
%%% Created : 28 Apr 1998 by Dan Gudmundsson <dgud@erix.ericsson.se>
%%%----------------------------------------------------------------------

%% Start the server with:
%%  erl -sname server -orber \"domain server\"
%%  and run setup:server()

%% Start the client with:
%%  erl -sname client -orber iiop_port 5001 -orber bootstrap_port 5002 -orber domain \"client\"
%%  and run setup:client()

-module(setup).
-author('dgud@erix.ericsson.se').

-export([orber_install/1, start_server_apps/1, start_client_apps/1, 
	 server/0, client/0]).

server() ->
    Ns = [node()],
    orber_install(Ns),
    start_server_apps(Ns),
    
    %% Both server and client needs to register the types
    oe_person:oe_register().

client() ->
    Ns = [node()],
    orber_install(Ns),
    start_client_apps(Ns),

    %% Both server and client needs to register the types
    oe_person:oe_register().

%% Orber needs to installed
orber_install(Ns) ->
    ok = mnesia:delete_schema(Ns),
    ok = mnesia:create_schema(Ns),
    [rpc:call(Node, mnesia, start, []) || Node <- Ns],
    orber:install(Ns).

%% Start the needed applications mnesia, orber and mnesia_corba_session.
start_server_apps(Ns) ->    
    [rpc:call(Node, mnesia, start, []) || Node <- Ns],
    orber:start(),
    %% Use the application parameters, in a real appl.
    mnesia_session_lib:start([{enable_corba, true}]).

%% Start the needed applications mnesia, orber and mnesia_corba_session.
start_client_apps(Ns) ->
    [rpc:call(Node, mnesia, start, []) || Node <- Ns],
    orber:start().


