%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : mnesia_connector_impl.erl
%%% Author  : Dan Gudmundsson <dgud@legolas>
%%% Purpose : Callback module which handles a connector
%%% Created : 3 Apr 1998 by Dan Gudmundsson <dgud@legolas>
%%%----------------------------------------------------------------------

-module(mnesia_connector_impl).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('dgud@erix.ericsson.se').

%%-compile(export_all).
-export([init/1, terminate/2, 
	 connect/1, disconnect/2]).

-record(state, {debug}).
-define(SUPERVISOR, mnesia_session_sup).

-define(DBG(F, A, D), mnesia_session_lib:debug(F, A, D)).

init([]) -> 
    {ok, #state{debug = mnesia_session_lib:get_env(debug)}}.

terminate(Reason, State) ->
    ok.

connect(State) ->
    Res = supervisor:start_child(?SUPERVISOR, []),
    ?DBG("connect() -> ~p~n", [Res], State#state.debug),
    case Res of
	{ok, SessionPid} ->
	    {reply, SessionPid, State};
	{error, Reason} ->
	    {reply, {error, Reason}, State}
    end.

disconnect(State, SessionPid) ->
    ?DBG("disconnect(~p)~n", [SessionPid], State#state.debug),
    exit(SessionPid, shutdown),
    {reply, ok, State}.


