%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : mnesia_corba_connector_impl.erl
%%% Author  : Dan Gudmundsson <dgud@legolas>
%%% Purpose : Callback module which handles a corba connector
%%% Created : 3 Apr 1998 by Dan Gudmundsson <dgud@legolas>
%%%----------------------------------------------------------------------

-module(mnesia_corba_connector_impl).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('dgud@erix.ericsson.se').

%%-compile(export_all).
-export([init/1, terminate/2, 
	 connect/1, disconnect/2]).

-record(state, {debug}).
-define(SUPERVISOR, mnesia_corba_session_sup).

-define(DBG(F, A, D), mnesia_session_lib:debug(F, A, D)).
-define(VERBOSE(F, A, D), mnesia_session_lib:verbose(F, A, D)).

init([]) ->
    State = #state{debug = mnesia_session_lib:get_env(debug)},
    {ok,State}.

terminate(Reason, State) -> 
    ok.

connect(State) ->
    Res = supervisor:start_child(?SUPERVISOR, []),
    ?DBG("connect() -> ~p~n", [Res], State#state.debug),
    case Res of
	{ok, Pid} ->
	    receive
		{mnesia_corba_session_key, Pid, Key} ->
		    {reply, Key, State}
	    after 0 ->
		    Reason = {corba_session_key_lost, Pid},
		    ?VERBOSE("<ERROR> ~p~n", [Reason], State#state.debug),
		    {reply, {error, Reason}, State}
	    end;
	{error, Reason} ->
	    {reply, {error, Reason}, State}
    end.

disconnect(State, SessionObjectkey) ->
    ?DBG("disconnect(~p)~n", [SessionObjectkey], State#state.debug),
    Res = corba:dispose(SessionObjectkey),
    {reply, Res, State}.


