%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1997
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%%--------------------------------------------------------------------
%% File: corba.erl
%% Author: Lars Thorsen, Peter Lundell
%% 
%% Description:
%%    This file contains the CORBA::ORB interface plus some 
%%    Orber specific functions.
%%
%% Creation date: 970115
%%-----------------------------------------------------------------
-module(corba).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').
-author('peterl@erix.ericssons.se').

-include_lib("orber/include/corba.hrl").

%%-----------------------------------------------------------------
%% Standard interface CORBA
%%-----------------------------------------------------------------
%%-export(['ORB_init'/2]).
%%-----------------------------------------------------------------
%% Standard interface CORBA::ORB
%%-----------------------------------------------------------------
-export([%create_list/2,
	 %create_operation_list/2,
	 %% get_default_context/1,
	 %% 'ORB_init'/2,
	 %% 'BOA_init/2,
	 resolve_initial_references/1,
	 list_initial_services/0,
	 resolve_initial_references_remote/2,
	 list_initial_services_remote/1,
	 object_to_string/1,
	 string_to_object/1]).

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([create/2,
	 create/3,
	 create/4,
	 create_link/2,
	 create_link/3,
	 create_link/4,
	 dispose/1,
	 create_subobject_key/2,
	 get_subobject_key/1,
	 get_pid/1,
	 raise/1]).


%%-----------------------------------------------------------------
%% Internal (inside orber implementation) exports
%%-----------------------------------------------------------------
-export([call/4, call/5,
	 cast/4,
	 request_from_iiop/5,
	 mk_objkey/3,
	 objkey_to_string/1,
	 string_to_objkey/1,
	 check_exception_type/1,
	 %%call_internal/2,
	 get_system_exception_typedef/1]).



%%------------------------------------------------------------
%%
%% Implementation of CORBA CORBA::ORB interfaces
%%
%%------------------------------------------------------------

create_list(Count) ->
    corba_nvlist:create_list(Count).

create_operation_list(OpDef) ->
    corba_nvlist:create_operation_list(OpDef).

%%'ORB_init'(Args, Name) ->
%%    ok.

%%-----------------------------------------------------------------
%% Initial reference handling
%%-----------------------------------------------------------------
resolve_initial_references(ObjectId) ->   
    orber_initial_references:get(string_to_objkey("INIT"), ObjectId).

list_initial_services() ->  
    orber_initial_references:list(string_to_objkey("INIT")).

resolve_initial_references_remote(ObjectId, []) ->
    corba:raise(#'BAD_PARAM'{completion_status=?COMPLETED_NO});
resolve_initial_references_remote(ObjectId, [RemoteModifier| Rest]) ->
    case lists:prefix("iiop://", RemoteModifier) of
       true ->
	    [_, Host, Port] = string:tokens(RemoteModifier, ":/"),
	    orber_iiop:request({Host, list_to_integer(Port), "INIT"}, 'get',
			       [ObjectId],
			       {{'tk_objref', 12, "object"},
				[{'tk_string', 0}],
				[]}, 'true');
       false ->
	    resolve_initial_references_remote(ObjectId, Rest)
    end.

list_initial_services_remote([]) ->
    corba:raise(#'BAD_PARAM'{completion_status=?COMPLETED_NO});
list_initial_services_remote([RemoteModifier| Rest]) ->
    case lists:prefix("iiop://", RemoteModifier) of
       true ->
	    [_, Host, Port] = string:tokens(RemoteModifier, ":/"),
	    orber_iiop:request({Host, list_to_integer(Port), "INIT"}, 'list', [],
			       {{'tk_sequence', {'tk_string',0},0},
				[], []}, 'true');
	false -> 
	    list_initial_services_remote(Rest)
    end.



%%-----------------------------------------------------------------
%% Objectreference convertions
%%-----------------------------------------------------------------
object_to_string(Object) ->
    iop_ior:string_code(Object).

string_to_object(IORString) ->
    {ObjRef, R, L} = iop_ior:string_decode(IORString),
    ObjRef.


%%------------------------------------------------------------
%%
%% Implementation of NON-standard functions
%%
%%------------------------------------------------------------
create(Module, TypeID) ->
    create(Module, TypeID, []).

create(Module, TypeID, Env) ->
    common_create(Module, TypeID, Env, [], 'start').

create(Module, TypeID, Env, {T, RegName}) when atom(RegName) ->
    common_create(Module, TypeID, Env, {T, RegName}, 'start').


create_link(Module, TypeID) ->
    create_link(Module, TypeID, []).

create_link(Module, TypeID, Env) ->
    common_create(Module, TypeID, Env, [], 'start_link').

create_link(Module, TypeID, Env, {T, RegName}) when atom(RegName) ->
    common_create(Module, TypeID, Env, {T, RegName}, 'start_link').


common_create(Module, TypeID, Env, RegName, StartMethod) ->
    case apply(Module, StartMethod, [Env]) of
	{ok, Pid} ->
	    case catch mk_objkey(TypeID, Pid, RegName) of
		{'EXCEPTION', E} ->
		    %% This branch is only used if we couldn't register 
		    %% our new objectkey due to an internal error in orber.
		    gen_server:call(Pid, stop),
		    corba:raise(E);
		{'EXIT', R} ->
		    %% This branch takes care of exit values
		    %% which aren't expected (due to bug).
		    gen_server:call(Pid, stop),
		    exit(R);
		Objkey ->
		    %% The normal case.
		    Objkey
	    end;
	X ->
	    X
    end.


dispose(Obj) ->
    corba_boa:dispose(Obj).

create_subobject_key(Objkey, B) when binary(B) ->
    iop_ior:set_privfield(Objkey, B);
create_subobject_key(Objkey, T) ->
    create_subobject_key(Objkey, term_to_binary(T)).

get_subobject_key(Objkey) ->
    iop_ior:get_privfield(Objkey).

get_pid(Objkey) ->
    {Location, Key} = iop_ior:get_key(Objkey),
     if
	Location == 'internal' ->
	     orber_objectkeys:get_pid(Key);
	Location == 'internal_registered' -> 
	     case whereis(Key) of
		 undefined ->
		     corba:raise(#'OBJECT_NOT_EXIST'{completion_status=?COMPLETED_NO});
		 Pid ->
		     Pid
	     end;
	Location == 'external' -> 
	     corba:raise(#'INV_OBJREF'{completion_status=?COMPLETED_NO})
    end.

raise(E) ->
    throw({'EXCEPTION', E}).

%%-----------------------------------------------------------------
%% Corba:call - the function for requests
%%-----------------------------------------------------------------
call(Obj, Func, Args, Types) ->
    call(Obj, Func, Args, Types, infinity).

call(Obj, Func, Args, Types, Timeout) ->
    {Location, Key} = iop_ior:get_key(Obj),
    if
	Location == 'internal' ->
	    Pid = orber_objectkeys:get_pid(Key),
	    call_internal(Pid, {Obj, Func, Args, Types}, Timeout);
	Location == 'internal_registered' -> 
		    call_internal(Key, {Obj, Func, Args, Types}, Timeout);
	Location == 'external' -> 
	    orber_iiop:request(Key, Func, Args, Types, 'true')
    end.

call_internal(Pid, {Obj, Func, Args, Types}, Timeout) when pid(Pid) ->
    case gen_server:call(Pid, {Obj, Func, Args}, Timeout) of
	{'EXCEPTION', E} ->
	    corba:raise(E);
	Res ->
	    Res
    end;
call_internal(Registered, {Obj, Func, Args, Types}, Timeout)  when atom(Registered)->
    case whereis(Registered) of
	undefined ->
	    {Host, Port} = iop_ior:get_port(Obj),
	    ObjKey = iop_ior:get_objkey(Obj),
	    orber_iiop:request({Host, Port, objkey_to_string(ObjKey)},
			       Func, Args, Types, 'true');
	P ->
	    case gen_server:call(P, {Obj, Func, Args}, Timeout) of
		{'EXCEPTION', E} ->
		    corba:raise(E);
		Res ->
		    Res
	    end
    end.
	
%%-----------------------------------------------------------------
%% Corba:cast - the function for ONEWAY requests
%%-----------------------------------------------------------------
cast(Obj, Func, Args, Types) ->
    {Location, Key} = iop_ior:get_key(Obj),
    if
	Location == 'internal' ->
	    Pid = orber_objectkeys:get_pid(Key),
	    cast_internal(Pid, {Obj, Func, Args, Types});
	Location == 'internal_registered' -> 	
	    cast_internal(Key, {Obj, Func, Args, Types});
	Location == 'external' -> 
	    orber_iiop:request(Key, Func, Args, Types, 'false')
    end.

cast_internal(Pid, {Obj, Func, Args, Types}) when pid(Pid) ->
    gen_server:cast(Pid, {Obj, Func, Args});
cast_internal(Registered, {Obj, Func, Args, Types}) ->
    case whereis(Registered) of
	undefined ->
	    {Host, Port} = iop_ior:get_port(Obj),
	    ObjKey = iop_ior:get_objkey(Obj),
	    orber_iiop:request({Host, Port, objkey_to_string(ObjKey)}, 
			       Func, Args, Types, 'false');
	P ->
	    gen_server:cast(P, {Obj, Func, Args})
    end.
	


%%-----------------------------------------------------------------
%% Incomming request from iiop
%%-----------------------------------------------------------------
request_from_iiop({Id, Type, Key, UserDef}, Func, Args, Types,
		  ResponseExpected) ->
    case ResponseExpected of
	'true' ->
	    gen_server:call(convert_key_to_pid(Key),
			    {{Id, Type, Key, UserDef}, Func, Args},
			    infinity);
	'false' ->
	    gen_server:cast(convert_key_to_pid(Key),
			    {{Id, Type, Key, UserDef}, Func, Args})
    end.

%%------------------------------------------------------------
%% Internal stuff
%%------------------------------------------------------------

convert_key_to_pid(Key) when binary(Key) ->
    orber_objectkeys:get_pid(Key);
convert_key_to_pid(Name) when atom(Name) ->
    Name.

mk_objkey(Id, Pid, []) when pid(Pid) ->
    Key = make_objkey(),
    case orber_objectkeys:register(Key, Pid) of
	ok ->
	    {list_to_binary(Id), 'key', Key, term_to_binary(undefined)};
	_ ->
	    corba:raise(#'INTERNAL'{completion_status=?COMPLETED_NO})
    end;
mk_objkey(Id, Pid, {'global', RegName}) when pid(Pid), atom(RegName) ->
    Key = term_to_binary(RegName),
    case orber_objectkeys:register(Key, Pid) of
	ok ->
	    {list_to_binary(Id), 'key', Key, term_to_binary(undefined)};
	_ ->
	    corba:raise(#'INTERNAL'{completion_status=?COMPLETED_NO})
    end;
mk_objkey(Id, Pid, {'local', RegName}) when pid(Pid), atom(RegName) ->
    register(RegName, Pid),
    Key = make_objkey(),
    case orber_objectkeys:register(Key, Pid) of
	ok ->
	    {list_to_binary(Id), 'registered', RegName,
	     term_to_binary(undefined)};
	_ ->
	    corba:raise(#'INTERNAL'{completion_status=?COMPLETED_NO})
    end.

make_objkey() ->
    term_to_binary({now(), node()}).

objkey_to_string({Id, 'key', Key, UserDef}) ->
    orber:domain() ++ ":" ++ binary_to_list(Id) ++ ":key:" ++
	binary_to_list(Key) ++ ":" ++ binary_to_list(UserDef);
objkey_to_string({Id, 'registered', 'orber_init', UserDef}) ->
    "INIT";
objkey_to_string({Id, 'registered', Atom, UserDef}) ->
    orber:domain() ++ ":" ++ binary_to_list(Id) ++ ":registered:" ++
	atom_to_list(Atom) ++ ":" ++ binary_to_list(UserDef); 
objkey_to_string(External_object_key) ->
    External_object_key.

string_to_objkey("INIT") ->
    {list_to_binary("IDL:CORBA/InitialReferences:1.0"), 'registered', 
     'orber_init', term_to_binary(undefined)};
string_to_objkey(String) -> 
    case prefix(orber:domain(), String) of
	false ->
	    String;
	Rest ->
	    string_to_objkey_2(Rest)

    end.

prefix([], L2) ->
    L2;
prefix([E |L1], [E | L2]) ->
    prefix(L1, L2);
prefix(_, _) ->
    false.

string_to_objkey_2([$:| Rest]) ->
    {I, R1} = read_to_colon(Rest),
    {P, R2} = read_to_colon(R1),
    {V, R3} = read_to_colon(R2),
    {T, R4} = read_to_colon(R3),
    {O, U} = read_to_colon(R4),
    Type = list_to_atom(T),
    case Type of
	'key' ->
	    Object = list_to_binary(O);
	'registered' -> 
	    Object = list_to_atom(O)
    end,
    {list_to_binary(I ++ ":" ++ P ++ ":" ++ V), Type, Object, list_to_binary(U)}.

read_to_colon([$: |R]) ->
    {[], R};
read_to_colon([C |R]) ->
    {Acc, R2} = read_to_colon(R),
    {[C|Acc], R2}.
    
%%-----------------------------------------------------------------
%% Check if CORBA system exception or user defined
%%-----------------------------------------------------------------
check_exception_type('UNKNOWN') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('BAD_PARAM') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('NO_MEMORY') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('IMP_LIMIT') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('COMM_FAILURE') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('INV_OBJREF') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('NO_PERMISSION') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('INTERNAL') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('MARSHAL') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('INITIALIZE') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('NO_IMPLEMENT') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('BAD_TYPECODE') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('BAD_OPERATION') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('NO_RESOURCES') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('NO_RESPONSE') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('PERSIST_STORE') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('BAD_INV_ORDER') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('TRANSIENT') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('FREE_MEM') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('INV_IDENT') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('INV_FLAG') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('INTF_REPOS') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('BAD_CONTEXT') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('OBJ_ADAPTER') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('DATA_CONVERSION') ->
    ?SYSTEM_EXCEPTION;
check_exception_type('OBJECT_NOT_EXIST') ->
    ?SYSTEM_EXCEPTION;
check_exception_type(_) ->
    ?USER_EXCEPTION.


%%-----------------------------------------------------------------
%% Generate typedef of system exception
%%-----------------------------------------------------------------
get_system_exception_typedef(ExcName) ->
    Name = atom_to_list(ExcName),
    {'tk_except', "IDL:CORBA/" ++ Name ++ ":1.0", Name,
     [
      {"minor",'tk_ulong'},
      {"completed",{'tk_enum', "", "completion_status",
	   ["COMPLETED_YES", "COMPLETED_NO",
	    "COMPLETED_MAYBE"]}}
     ]}.
