%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: corba_object.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains the CORBA::Object interface
%%
%% Creation date: 970709
%%
%%-----------------------------------------------------------------
-module(corba_object).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-include_lib("orber/include/corba.hrl").
-include_lib("orber/src/orber_iiop.hrl").

%%-----------------------------------------------------------------
%% Standard interface CORBA::Object
%%-----------------------------------------------------------------
-export([%get_implementation/1, 
	 get_interface/1,
	 is_nil/1, 
%	 duplicate/1, 
%	 release/1,
	 is_a/2,
	 non_existent/1,
	 is_equivalent/2,
	 hash/2,
	 create_request/6]).

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([]).

%%------------------------------------------------------------
%% Implementation of standard interface
%%------------------------------------------------------------
%%get_implementation(Obj) ->
%%    exit(#'NO_IMPLEMENT'{completion_status=?COMPLETED_NO}).

get_interface(Obj) ->
    TypeId = iop_ior:get_typeID(Obj),
    Rep = orber_ifr:find_repository(),
    Int = orber_ifr_repository:lookup_id(Rep, TypeId),   
    orber_ifr_interfacedef:describe_interface(Int).

is_nil(Object) when record(Object, 'IOP_IOR') ->
    iop_ior:check_nil(Object);
is_nil({I,T,K,P}) ->
    iop_ior:check_nil({I,T,K,P});
is_nil(_) ->
    corba:raise(#'INV_OBJREF'{completion_status=?COMPLETED_NO}).

%%duplicate(Obj) ->
%%    Obj.

%%release(Obj) ->
%%    ok.

is_a(Obj, Logical_type_id) ->
    case iop_ior:get_typeID(Obj) of
	Logical_type_id ->
	    true;
	_ ->
	    false
    end.

non_existent(Obj) ->
    {Location, Key} = iop_ior:get_key(Obj),
    if
	Location == 'internal' ->
	    case catch orber_objectkeys:get_pid(Key) of
		{'EXCEPTION', E} when record(E,'OBJECT_NOT_EXIST') ->
		    true;
		{'EXCEPTION', X} ->
		    corba:raise(X);
		{'EXIT', R} ->
		    exit(R);
		_ ->
		    false
	    end;
	Location == 'internal_registered' ->
	    case whereis(Key) of
		undefined ->
		    true;
		P ->
		    false
	    end;
	true -> 	
	    false
    end.

is_equivalent(Obj, Other_object) ->
    case Obj of
	Other_object ->
	    true;
	_ ->
	    false
    end.

hash(Obj, Maximum) ->
    erlang:hash(iop_ior:get_key(Obj), Maximum).


create_request(Obj, Ctx, Op, ArgList, NamedValueResult, ReqFlags) ->
    {ok, NamedValueResult, []}.
