%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: iop_ior.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains the IOP::IOR handling
%%
%% Creation date: 970327
%%
%%-----------------------------------------------------------------
-module(iop_ior).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-include_lib("orber/include/corba.hrl").
-include_lib("orber/src/orber_iiop.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([code/3, decode/3, string_code/1, string_decode/1, create/4,
	 get_key/1, get_port/1, get_typeID/1, get_objkey/1, check_nil/1,
	 get_privfield/1, set_privfield/2]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([]).

%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: create/4
%%-----------------------------------------------------------------
create(TypeID, Host, IIOP_port, Objkey) ->
    V=#'IIOP_Version'{major=?IIOP_MAJOR,
		      minor=?IIOP_MINOR},
    PB=#'IIOP_ProfileBody'{iiop_version=V,
			   host=Host, 
			   port=IIOP_port,
			   object_key=Objkey},
    #'IOP_IOR'{type_id=TypeID, profiles=[#'IOP_TaggedProfile'{tag=?TAG_INTERNET_IOP,
							      profile_data=PB}]}.

%%-----------------------------------------------------------------
%% Func: get_key/1
%%-----------------------------------------------------------------
get_key(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{object_key={Id, Type, Key, UserDef}}}]}) when binary(Key) ->
    {'internal', Key};
get_key(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{object_key={Id, Type, RegName, UserDef}}}]})when atom(RegName) ->
    {'internal_registered', RegName};
get_key(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{host=Host,port=IIOP_port,object_key=ObjectKey}}]}) ->
    {'external', {Host, IIOP_port, ObjectKey}};
get_key({_Id, _Type, Key, _UserDef}) when binary(Key) ->
    {'internal', Key};
get_key({_Id, _Type, RegName, _UserDef}) when atom(RegName) ->
    {'internal_registered', RegName}.

%%%-----------------------------------------------------------------
%% Func: get_Port/1
%%-----------------------------------------------------------------
get_port(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{host=Host,port=IIOP_port}}]}) ->
    {Host, IIOP_port};
get_port({_Id, _Type, _Key, _UserDef}) ->
    {orber:host(), orber:iiop_port()}.

%%-----------------------------------------------------------------
%% Func: get_typeID/1
%%-----------------------------------------------------------------
get_typeID(#'IOP_IOR'{type_id=TypeID}) ->
    TypeID;
get_typeID({Id, _Type, _Key, _UserDef}) ->
    binary_to_list(Id).
%%-----------------------------------------------------------------
%% Func: get_objkey/1
%%-----------------------------------------------------------------
get_objkey(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{object_key=ObjectKey}}]}) ->
    ObjectKey;
get_objkey({Id, Type, Key, UserDef}) ->
    {Id, Type, Key, UserDef}.

%%-----------------------------------------------------------------
%% Func: get_privfield/1
%%-----------------------------------------------------------------
get_privfield(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{object_key={Id, Type, Key, UserDef}}}]}) ->
    UserDef;
get_privfield({Id, Type, Key, UserDef}) ->
    UserDef.
%%-----------------------------------------------------------------
%% Func: set_privfield/2
%%-----------------------------------------------------------------
set_privfield(#'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{object_key={Id, Type, Key, _}}}]}, UserData) ->
	    #'IOP_IOR'{profiles=[#'IOP_TaggedProfile'{profile_data=#'IIOP_ProfileBody'{object_key={Id, Type, Key, UserData}}}]};
set_privfield({Id, Type, Key, _}, UserData) ->
	    {Id, Type, Key, UserData}.

%%-----------------------------------------------------------------
%% Func: check_nil/1
%%-----------------------------------------------------------------
check_nil(#'IOP_IOR'{type_id="", profiles=[]}) ->
    true;
check_nil({Id, _, _, _}) ->  
    case binary_to_list(Id) of
	"" ->
	    true; 
	_ ->
	    false
    end;
check_nil(_) ->
    false.
%%-----------------------------------------------------------------
%% Func: string_code/1
%%-----------------------------------------------------------------
string_code(IOR) ->
    {IorByteSeq0, Length0} = cdr_encode:enc_type('tk_octet', 0, [], 0),
    {IorByteSeq, _} = code(IOR, IorByteSeq0, Length0),
    IorByteSeq1 = lists:reverse(IorByteSeq),
    IorByteSeq2 = lists:flatten(IorByteSeq1),
    IorHexSeq = bytestring_to_hexstring(IorByteSeq2),
    [$I,$O,$R,$: | IorHexSeq].
    
%%-----------------------------------------------------------------
%% Func: code/3
%%-----------------------------------------------------------------
code(#'IOP_IOR'{type_id=TypeId, profiles=Profiles}, Bytes, Len) ->
    ProfileSeq = code_profile_datas(Profiles),
    %% Byte order
    cdr_encode:enc_type(?IOR_TYPEDEF,
			#'IOP_IOR'{type_id=TypeId, profiles=ProfileSeq},
			Bytes, Len);
code({Id, Type, Key, UserDef}, Bytes, Len) ->
    IOR = create(binary_to_list(Id), orber:host(), orber:iiop_port(),
		 {Id, Type, Key, UserDef}),
    code(IOR, Bytes, Len).

code_profile_datas([]) ->
    [];
code_profile_datas([#'IOP_TaggedProfile'{tag=0, profile_data=P} | Profiles]) ->
    P1 = convert_objkey_to_byteseq(P),
    NewBytes = code_profile_data(P1),
    [#'IOP_TaggedProfile'{tag=0, profile_data=NewBytes} | code_profile_datas(Profiles)];
code_profile_datas([#'IOP_TaggedProfile'{tag=1, profile_data=P} | Profiles]) ->
        exit({'code_profile_datas', 'multiple components not supported'}).
code_profile_data(ProfileData) ->
    %% Byte order
    {NewBytes, Length} = cdr_encode:enc_type('tk_octet', 0, [], 0),
    {NewBytes1, Length1} = cdr_encode:enc_type(?PROFILEDATA_TYPEDEF,
						ProfileData,
						NewBytes, Length),
    NewBytes2 = lists:reverse(NewBytes1),
    lists:flatten(NewBytes2).


convert_objkey_to_byteseq(#'IIOP_ProfileBody'{iiop_version=#'IIOP_Version'{major=Major,
							      minor=Minor}, 
					      host=Host, port=Port,
					      object_key=ObjKey}) ->
    #'IIOP_ProfileBody'{iiop_version=#'IIOP_Version'{major=Major,
					minor=Minor}, 
			host=Host, port=Port,
			object_key=corba:objkey_to_string(ObjKey)}.

%%-----------------------------------------------------------------
%% Func: string_decode/1
%%-----------------------------------------------------------------
string_decode([$I,$O,$R,$: | IorHexSeq]) ->
    IorByteSeq = hexstring_to_bytestring(IorHexSeq),
    {ByteOrder, IorRest} = cdr_decode:dec_byte_order(IorByteSeq),
    decode(IorRest, 1, ByteOrder).

%%-----------------------------------------------------------------
%% Func: decode/3
%%-----------------------------------------------------------------
decode(IorByteSeq, Len, ByteOrder) ->
    {#'IOP_IOR'{type_id=TypeId, profiles=Profiles}, Rest, Length} =
	cdr_decode:dec_type(?IOR_TYPEDEF, IorByteSeq, Len, ByteOrder),
    L = decode_profiles(Profiles),
    {#'IOP_IOR'{type_id=TypeId, profiles=L}, Rest, Length}.

decode_profiles([]) ->
    [];
decode_profiles([P | Profiles]) ->
    Struct = decode_profile(P),
    L = decode_profiles(Profiles),
    [Struct | L].

decode_profile(#'IOP_TaggedProfile'{tag=0, profile_data=ProfileData}) ->
    {ByteOrder, Rest} = cdr_decode:dec_byte_order(ProfileData),
    Len = 1,
    {Struct, [], _} = cdr_decode:dec_type(?PROFILEDATA_TYPEDEF, Rest, Len,
					  ByteOrder),
    Struct1 = convert_byteseq_to_objkey(Struct), 
    #'IOP_TaggedProfile'{tag=0, profile_data=Struct1};
decode_profile(#'IOP_TaggedProfile'{tag=1, profile_data=ProfileData}) ->
    exit({'decode_profile_data', 'multiple components not supported'}).
    

convert_byteseq_to_objkey(#'IIOP_ProfileBody'{iiop_version=#'IIOP_Version'{major=Major,
							      minor=Minor}, 
					      host=Host, port=Port,
					      object_key=ObjKey}) ->
    #'IIOP_ProfileBody'{iiop_version=#'IIOP_Version'{major=Major, minor=Minor}, 
			host=Host, port=Port, object_key=corba:string_to_objkey(ObjKey)}.
%%-----------------------------------------------------------------
%% Func: parseIORfromServiceLog/1
%% Was only used when testing against JacORB !!!!!!!!!!
%%-----------------------------------------------------------------
parseIORfromServiceLog([HexSeq|_]) ->
    StartPos = string:str(HexSeq,"IOR:"),
    NewHexSeq = string:substr(HexSeq, StartPos),
    EndPos = string:str(NewHexSeq,"\n"),
    IorString = string:substr(NewHexSeq, 1, EndPos - 1),
    string_decode(IorString).

%%-----------------------------------------------------------------
%% Func: hexstring_to_bytestring/1
%%-----------------------------------------------------------------
hexstring_to_bytestring(HexString) ->
    ByteString = hexstring_to_bytestring(HexString, []),
    lists:reverse(ByteString).

hexstring_to_bytestring([], Acc) ->
    Acc;
hexstring_to_bytestring([H1, H2 |Rest], Acc) ->
    I1 = hex_to_int(H1),
    I2 = hex_to_int(H2),
    I = I1 * 16 + I2,
    Acc2 = cdrlib:enc_octet(I, Acc),
    hexstring_to_bytestring(Rest, Acc2).


hex_to_int(H) when H >= $a ->
    10 + H - $a;
hex_to_int(H) when H >= $A ->
    10 + H -$A;
hex_to_int(H) ->
    H - $0.
%%-----------------------------------------------------------------
%% Func: bytestring_to_hexstring/1
%% Args: A byte string
%% Returns: A list of hexadecimal digits (onebyte will be represented as 
%%          two hexadecimal digits).
%%-----------------------------------------------------------------
bytestring_to_hexstring(ByteString) ->
    HexString = bytestring_to_hexstring(ByteString, []),
    lists:reverse(HexString).

bytestring_to_hexstring([], Acc) ->
    Acc;
bytestring_to_hexstring([B |Rest], Acc) ->
    [C1, C2] = int_to_hex(B),
    bytestring_to_hexstring(Rest,[C2, C1| Acc]).

int_to_hex(B) when B < 256, B >= 0 ->
    N1 = B div 16,
    N2 = B rem 16,
    [code_character(N1),
     code_character(N2)].

code_character(N) when N < 10 ->
    $0 + N;
code_character(N) ->
    $a + (N - 10).

