%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : orber_ifr_typecode.erl
%%% Author  : Per Danielsson <pd@gwaihir>
%%% Purpose : Code for Typecode
%%% Created : 14 May 1997 by Per Danielsson <pd@gwaihir>
%%%----------------------------------------------------------------------

%%% NOTE: 
%%% Only make_typcode is for real here. All of the TypeCode interfaces
%%% specified in the IDL specification needs to be implemented.
%%%

-module(orber_ifr_typecode).
-author('pd@gwaihir').

%%-compile(export_all).
-export([
	 %%make_typecode/9,
	 equal/2,
	 kind/1,
	 id/1,
	 name/1,
	 member_count/1,
	 member_name/2,
	 member_type/2,
	 member_label/2,
	 discriminator_type/1,
	 default_index/1,
	 '_length'/1,
	 content_type/1,
	 param_count/1,
	 parameter/2
	]).

-import(orber_ifr_utils,[get_field/2]).

-include("orber_ifr.hrl").
-include("ifr_objects.hrl").
%-include_lib("orber/include/ifr_types.hrl").
-include_lib("orber/include/corba.hrl").

%%%======================================================================
%%% TypeCode

%%% "Internal" functions

%%% ir_TypeCode does not have a field ir_Internal_ID. TypeCodes are
%%% never explicitly written to the database as separate DB-records.
%%% TypeCodes are stored as full records whenever they are used in an
%%% IFR-object.

%%% p. 12-11 in CORBA 2.0, also check Peters HTML-doc for the IDL-compiler.
%%% eg. for struct: {tk_struct, Id, Name, [{foo,bar}, {barf,fronk} ...]

%%% *** This function should be removed. It is obsolete, and was never
%%% necessary in the first place...

make_typecode(ObjType,ObjID,Id,Name,Discriminator_type,Original_type,Members,
	      Bound,Kind) ->
    case Kind of
	dk_Struct ->
	    {tk_struct, Id, Name, Members};
%%	     lists:map(fun({Type,ID}) ->
%%			       {?__get_field(Type,ID,name),
%%				?__get_field(Type,ID,type)}
%%		       end,
%%		       Members)};
	dk_Union ->
	    XXX = element(2, Discriminator_type),
	    {tk_unions, Id, Name, 
	     get_field({ir_IDLType,XXX},type),
	     -1, Members};
%%	     lists:map(fun({Type,ID}) ->
%%			       {?__get_field(Type,ID,label),
%%				?__get_field(Type,ID,name),
%%				?__get_field(Type,ID,type)}
%%		       end,
%%		       Members)};
	dk_Enum ->
	    {tk_enum, Id, Name, Members};
	dk_Alias ->
	    {Orig_type,Orig_ID} = Original_type,
	    {tk_alias, Id, Name, get_field({Orig_type,Orig_ID},type)};
	dk_String ->
	    {tk_string, Bound};
	dk_Sequence ->
	    {tk_sequence, Original_type, Bound};
	dk_Array ->
	    {tk_array, Original_type,Bound};
	dk_Interface ->
	    {tk_objref, Id, Name};
	Other ->
	    ?ifr_exception("Cannot make typecode based on definitionkind:",
			   Other)
    end.

%%%----------------------------------------------------------------------
%%% Inherited interfaces

%% none %%

%%%----------------------------------------------------------------------
%%% Non-inherited interfaces

equal({ObjType, ObjID}, {Tc_ObjType, Tc_ObjID})
?tcheck(ir_TypeCode, ObjType) ->
    get_field({ObjType,ObjID},kind) == get_field({Tc_ObjType,Tc_ObjID},kind).

kind({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

id({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

name({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

member_count({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

member_name({ObjType, ObjID}, Index) ->
    {ok, {ObjType, ObjID}}.

member_type({ObjType, ObjID}, Index) ->
    {ok, {ObjType, ObjID}}.

member_label({ObjType, ObjID}, Index) ->
    {ok, {ObjType, ObjID}}.

discriminator_type({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

default_index({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

'_length'({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

content_type({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

param_count({ObjType, ObjID}) ->
    {ok, {ObjType, ObjID}}.

parameter({ObjType, ObjID}, Index) ->
    {ok, {ObjType, ObjID}}.
