%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber_iiop_bridge.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains the IIOP "proxy" for incomming connections
%%
%% Creation date: 970115
%%
%%-----------------------------------------------------------------
-module(orber_iiop_bridge).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-behaviour(gen_server).

-include_lib("orber/src/orber_iiop.hrl").
-include_lib("orber/include/corba.hrl").
-include_lib("orber/src/orber_debug.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([start/0, start/1]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([init/1, handle_info/2, terminate/2]).


%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: start/0
%%-----------------------------------------------------------------
start() ->
    ignore.
%%-----------------------------------------------------------------
%% Func: start/1
%%-----------------------------------------------------------------
start(Opts) ->
    ?PRINTDEBUG2("start bridge: ~p ", [self()]),
    gen_server:start_link(orber_iiop_bridge, Opts, []).

%%-----------------------------------------------------------------
%% Func: send_reply/1
%%-----------------------------------------------------------------
send_reply(Pid,Rep) ->
    gen_server:call(Pid, Rep).

%%-----------------------------------------------------------------
%% Server functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: init/1
%%-----------------------------------------------------------------
init({connection, Socket}) ->
    ?PRINTDEBUG2("orber_iiop_bridge ini: ~p ", [self()]),
    {ok, [Socket]}.

%%-----------------------------------------------------------------
%% Func: terminate/2
%%-----------------------------------------------------------------
terminate(From, Reason) ->
    ok.

%%-----------------------------------------------------------------
%% Func: handle_info/2
%%-----------------------------------------------------------------
handle_info({tcp_closed, Socket}, [Socket]) ->
    {stop, normal, []};
handle_info({tcp, Socket, Bytes}, [Socket]) ->
    case catch cdr_decode:dec_message(null, Bytes) of
	{'EXCEPTION', DecodeException} ->
	    Reply = cdr_encode:enc_message_error(),
	    ?IIOP_SOCKET_MOD:write(Socket, Reply);
	{'EXIT', _} ->
	    Reply = cdr_encode:enc_message_error(),
	    ?IIOP_SOCKET_MOD:write(Socket, Reply);
	F ->
	    handle_message(F, Socket)	    
    end,
    {noreply, [Socket]};
handle_info({Socket, {socket_closed, normal}}, [Socket]) ->
    {stop, normal, []};
handle_info({Socket, {socket_closed, Error}}, [Socket]) ->
    {stop, "Connection broken", []};
handle_info({Socket, {fromsocket, Bytes}}, [Socket]) ->
    case catch cdr_decode:dec_message(null, Bytes) of
	{'EXCEPTION', DecodeException} ->
	    Reply = cdr_encode:enc_message_error(),
	    ?IIOP_SOCKET_MOD:write(Socket, Reply);
	{'EXIT', _} ->
	    Reply = cdr_encode:enc_message_error(),
	    ?IIOP_SOCKET_MOD:write(Socket, Reply);
	F ->
	    handle_message(F, Socket)	    
    end,
    {noreply, [Socket]};
handle_info(X,[Socket]) ->
    {noreply, [Socket]}.


%%-----------------------------------------------------------------
%% Func: handle_message/2
%%-----------------------------------------------------------------

handle_message({Hdr, Par, TypeCodes}, Socket) when record(Hdr, request_header) ->
    Result = corba:request_from_iiop(Hdr#request_header.object_key,
			       list_to_atom(Hdr#request_header.operation),
			       Par, [], Hdr#request_header.response_expected),
    case Hdr#request_header.response_expected of
	'true' ->
	    case result_to_list(Result, TypeCodes) of
		{'EXCEPTION', Exception} ->
		    {TypeOfException, ExceptionTypeCode} =
			orber_typedefs:get_exception_def(Exception),
		    Reply = cdr_encode:enc_reply(Hdr#request_header.request_id,
						 TypeOfException,
						 {ExceptionTypeCode, [], []}, 
						 Exception, []);
		[Res |OutPar] ->
		    Reply = cdr_encode:enc_reply(
			      Hdr#request_header.request_id,
			      'no_exception',
			      TypeCodes,
			      Res, OutPar);
		_ ->
		    E = #'INTERNAL'{completion_status=?COMPLETED_MAYBE},
		    {TypeOfException, ExceptionTypeCode} =
			orber_typedefs:get_exception_def(E),
		    Reply = cdr_encode:enc_reply(Hdr#request_header.request_id,
						 TypeOfException,
						 {ExceptionTypeCode, [], []}, 
						 E, [])
	    end, 
	    ?IIOP_SOCKET_MOD:write(Socket, Reply);
	'false' ->
	    %% ONEWAY requests
	    ok
    end;

handle_message('close_connection', _) ->
    ok;

handle_message('message_error', _) ->
    ok;

handle_message(Hdr, Socket) when record(Hdr,locate_request_header) ->
    Location = orber_objectkeys:check(Hdr#locate_request_header.object_key),
    Reply = cdr_encode:enc_locate_reply(Hdr#locate_request_header.request_id,
					Location),
    ?IIOP_SOCKET_MOD:write(Socket, Reply);
handle_message(Hdr,_) when record(Hdr,cancel_request_header) ->
    ok;
handle_message(_, Socket) ->
    %% This cluase takes care of all erranous messages.
    Reply = cdr_encode:enc_message_error(),
    ?IIOP_SOCKET_MOD:write(Socket, Reply).


result_to_list({'EXCEPTION', E}, _) ->
    {'EXCEPTION', E};
result_to_list(Result, {TkRes, _, TkOut}) ->
   case length(TkOut) of
       0 ->
	   [Result];
       N ->
	   tuple_to_list(Result)
   end.
