%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber_iiop_net.erl
%% Author: Lars Thorsen, Peter Lundell
%% 
%% Description:
%%    This file contains the IIOP communication server
%%
%% Creation date: 970115
%%
%%-----------------------------------------------------------------
-module(orber_iiop_net).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-behaviour(gen_server).

-include_lib("orber/src/orber_iiop.hrl").
-include_lib("orber/src/orber_debug.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([start/1]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([init/1, terminate/2, handle_call/3, net_accept/2]).
-export([handle_cast/2, handle_info/2]).

%%-----------------------------------------------------------------
%% Server state record
%%-----------------------------------------------------------------
-record(state, {db=[],ports=[]}).

%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: start/1
%%-----------------------------------------------------------------
start(Opts) ->
    gen_server:start_link({local, orber_iiop_net}, orber_iiop_net, Opts, []).

%%-----------------------------------------------------------------
%% Server functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: init/1
%%-----------------------------------------------------------------
init(Options) ->
    ?PRINTDEBUG("init netserver"),
    State = parse_options(Options, #state{}),
    {ok, State}.

%%-----------------------------------------------------------------
%% Func: terminate/1
%%-----------------------------------------------------------------
terminate(From, Reason) ->
    ok.

%%-----------------------------------------------------------------
%% Func: handle_call/3
%%-----------------------------------------------------------------
handle_call({connect, Socket}, From, State) ->
    Pid = orber_iiop_netsup:start_connection(Socket),
    {reply, Pid, State}.

%%-----------------------------------------------------------------
%% Internal Functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: parse_options/1
%%-----------------------------------------------------------------
parse_options([{port, Port} | Rest], State) ->
    Listen = ?IIOP_SOCKET_MOD:listen(Port, []),
    ?PRINTDEBUG2("listen at port ~p", [Port]),
    Pid = proc_lib:spawn_link(orber_iiop_net, net_accept, [Listen, self()]),
    parse_options(Rest, State#state{ports=[Port | State#state.ports]});
parse_options([], State) ->
    State.

%%-----------------------------------------------------------------
%% Func: net_accept/2
%%-----------------------------------------------------------------
net_accept(ListenFd, Parent) ->
    ?PRINTDEBUG("orber_iiop_net waiting in socket:accept"),
    S = ?IIOP_SOCKET_MOD:accept(ListenFd),
    case gen_server:call(orber_iiop_net, {connect, S}, infinity) of
	{ok, Pid} ->
	    ?PRINTDEBUG2("changed controlling process to ~p",[Pid]),
	    ?IIOP_SOCKET_MOD:controlling_process(S, Pid);
	_ ->
	    ?IIOP_SOCKET_MOD:close(S)	
    end,
    net_accept(ListenFd, Parent).



%%------------------------------------------------------------
%% Standard gen_server cast handle
%%------------------------------------------------------------
handle_cast(_, State) ->
    {noreply,  State}.


%%------------------------------------------------------------
%% Standard gen_server handles
%%------------------------------------------------------------
handle_info(_, State) ->
    {noreply,  State}.
