%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber_iiop_netsup.erl
%% Author: Lars Thorsen, Peter Lundell
%% 
%% Description:
%%    This file contains the IIOP communication supervisor which
%%    holds all active "proxies"
%%
%% Creation date: 970115
%%
%%-----------------------------------------------------------------
-module(orber_iiop_netsup).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-behaviour(supervisor).

-include_lib("orber/src/orber_debug.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([start/2,start_connection/1]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([init/1, terminate/2]).

%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: start/2
%%-----------------------------------------------------------------
start(sup, Opts) ->
    ?PRINTDEBUG2("start sup in ~p", [self()]),
    supervisor:start_link({local, orber_iiop_netsup}, orber_iiop_netsup,
			  {sup, Opts});
start(A1, A2) -> 
    ?PRINTDEBUG2("start/2 called with parameters ~p ~p", [A1, A2]),
    ok.


%%-----------------------------------------------------------------
%% Server functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: init/1
%%-----------------------------------------------------------------
init({sup, Opts}) ->
    ?PRINTDEBUG2("orber_iiop_netsup init supervisor ~p", [self()]),
    SupFlags = {simple_one_for_one, 500, 100},
    ChildSpec = [
		 {name, {orber_iiop_bridge, start, []}, temporary, 
		  10000, worker, [orber_iiop_bridge]}
		],
    {ok, {SupFlags, ChildSpec}};
init(Opts) ->
    ?PRINTDEBUG2("plain init ~p", [self()]),
    {ok, []}.


%%-----------------------------------------------------------------
%% Func: terminate/1
%%-----------------------------------------------------------------
terminate(From, Reason) ->
    ok.

%%-----------------------------------------------------------------
%% Func: start_connection/1
%%-----------------------------------------------------------------
start_connection(Socket) ->
    ?PRINTDEBUG2("supervisor ~p starting bridge for socket ~p",
		 [self(), Socket]),
    supervisor:start_child(orber_iiop_netsup, [{connection,Socket}]).

