%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%%----------------------------------------------------------------------
%% File    : orber_imr.erl
%% Author  : Per Danielsson <pd@gwaihir>
%% Purpose : Implementation Repository for CORBA
%% Created : 22 May 1997 by Per Danielsson <pd@gwaihir>
%%----------------------------------------------------------------------
-module(orber_imr).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('pd@erix.ericsson.se').

%%-compile(export_all).
-export([init/2,create_impl/1,find_impl/1]).

-include_lib("orber/include/corba.hrl").
-include_lib("orber/include/imr.hrl").
-include("orber_ifr.hrl").

init(Nodes,Timeout) ->
    %% Do we have the Impementation Repository table? If not, create it.
    case lists:member(imprec,mnesia:system_info(tables)) of
	true ->
	    ok;
	_ ->
	    ?debug_print("creating ", imprec),
	    R = mnesia:create_table(imprec,
				    [{disc_copies,Nodes},
				     {attributes,
				      record_info(fields,imprec)}]),
	    case R of
		{atomic,ok} ->
		    ok;
		_ ->
		    ?ifr_exception("Cannot create IMR tables",R)
	    end
    end,
    Wait = mnesia:wait_for_tables([imprec], Timeout),
    case Wait of
	ok ->
	    ok;
	_ ->
	    ?ifr_exception("Database table waiting failed.",Wait)
    end.

create_impl(Impl_rec) ->
    _F = ?write_function(Impl_rec),
    orber_ifr:ifr_transaction_read_write(_F).

find_impl(What) ->
    _F = ?read_function(What),
    orber_ifr:ifr_transaction_read_write(_F).
