%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber_initial_references.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains the CORBA::InitialReferences interface
%%
%% Creation date: 970827
%%
%%-----------------------------------------------------------------
-module(orber_initial_references).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-behaviour(gen_server).

-include_lib("orber/include/corba.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([start/1, shutdown/1, init/1,
	 terminate/2, handle_call/3,
	 get/2, list/1]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([handle_cast/2, handle_info/2]).

%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
start(Env) ->
    gen_server:start_link({local, 'orber_init'}, ?MODULE, Env, []).

shutdown(EO_this) ->
    gen_server:call(EO_this, stop).

%%-----------------------------------------------------------------
%% InitialReferences Interface 
%%-----------------------------------------------------------------
get(EO_this, Id) ->
    corba:call(EO_this, 'get', [Id], {{'tk_objref', 12, "object"},
				      [{"id", {'tk_string', 0}}],
				      []}).

list(EO_this) ->
    corba:call(EO_this, 'list', [], {{'tk_sequence',{"id", {'tk_string', 0}, 0}},
				     [], []}).


%%-----------------------------------------------------------------
%% Internal interface functions
%%-----------------------------------------------------------------
init([]) ->
    Id = ets:new('InitialReferences', []), 
    NSObjKey = {list_to_binary("IDL:CosNaming/NamingContext:1.0"),
		'registered', 'orber_nameservice', term_to_binary(undefined)},
    NSior = iop_ior:create("IDL:CosNaming/NamingContext:1.0", orber:host(),
		       orber:iiop_port(), NSObjKey),
    ets:insert(Id, {"NameService", NSior}),
    {ok, Id}.

terminate(Reason, TableName) ->
    ets:delete(TableName),
    ok.


%%-----------------------------------------------------------------
%% Handle incomming calls 
handle_call({EO_this, 'get', [Id]}, From, TableName) ->
    case ets:lookup(TableName, Id) of
	[{_, ObjRef} ] ->
	    {'reply', ObjRef, TableName};
	_ ->
	    corba:raise(#'InvalidName'{})
    end;
handle_call({EO_this, 'list', []}, From, TableName) ->
    ObjIdList = get_all_objectids(ets:tab2list(TableName), []),
    {'reply', ObjIdList, TableName};

handle_call({EO_THIS, oe_get_interface, []},
            EO_From, EO_State) ->
    {'reply', [{"get", {{'tk_objref', 12, "object"},
			     [{'tk_string', 0}],
			     []}},
		    {"list", {{'tk_sequence',{'tk_string', 0}, 0},
			      [],
			      []}}
		   ], EO_State};
handle_call('stop', From, State) ->
    {'stop', normal, 'ok', State};
handle_call(Req, From,State) ->
    {'reply', {'ok', 'nil', 'nil'}, State}.


%%------------------------------------------------------------
%% Standard gen_server cast handle
%%
handle_cast(_, State) ->
    {noreply,  State}.


%%------------------------------------------------------------
%% Standard gen_server handles
%%
handle_info(_, State) ->
    {noreply,  State}.

%%------------------------------------------------------------
%% get_all_objectids/2
%%------------------------------------------------------------
get_all_objectids([], Acc) ->
    Acc;
get_all_objectids([{Key, _} | Rest], Acc) ->
    get_all_objectids(Rest, [ Key | Acc]).

