%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber_iiop_new_socket.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains a standard interface to socket to handle the change
%%    of the socket module in OTP. This module handles the calls to the old socket 
%%    interface (orber_iiop_new_socket contains the new interface).
%%
%% Creation date: 970115
%%
%%-----------------------------------------------------------------
-module(orber_old_socket).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-include_lib("orber/include/corba.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([start/0, connect/2, listen/2, accept/1, read/1, write/2,
	 controlling_process/2, close/1]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([]).

%%-----------------------------------------------------------------
%% External functions
%%-----------------------------------------------------------------
start() ->	
    application:start(sockets).

%%-----------------------------------------------------------------
%% Connect to IIOP Port at Host in CDR mode, in order to 
%% establish a connection.
%%
connect(Host, Port) ->
    socket:client('STREAM', 'AF_INET', {Host, Port}, {packet, 0}).

%%-----------------------------------------------------------------
%% Create a listen socket at Port in CDR mode for 
%% data connection.
%%
listen(Port, _) ->
    socket:listen('STREAM', 'AF_INET', Port, {packet, 0}).

%%-----------------------------------------------------------------
%% Wait in accept on the socket
%% 
accept(ListenSocket) ->
    socket:accept(ListenSocket).

%%-----------------------------------------------------------------
%% Close the socket
%% 
close(Socket) ->
    socket:close(Socket).


%%-----------------------------------------------------------------
%% Read from socket
%% 
read(Socket) ->
    receive
        {Socket, {fromsocket, Bytes}} ->
	    Bytes;
        {Socket, {socket_closed, normal}} ->
            corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_NO});
        {Socket, {socket_closed, Error}} ->
            corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_NO})
    end.

%%-----------------------------------------------------------------
%% Write to socket
%% 
write(Socket, Bytes) ->
    Socket ! {self(), {'deliver', Bytes}}.

%%-----------------------------------------------------------------
%% Change the controlling process for the socket
%% 
controlling_process(Socket, Pid) ->
    socket:controlling_process(Socket, Pid).

