%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% 
%% The release installation files are:
%%
%%	ReleaseName.rel - the file contains a specification of the
%%	version of the release and which applications (and versions
%%	of) it contains. Each release in the Up and Down lists must
%%	have this file.
%%
%%	relup - the file contains low-level release instructions for
%%	Up (or Down) grading the release. This file is the output of
%%	mk_relup.
%%
%%
%% Implementation notes
%% --------------------
%%
%%	A release is made up of a number of applications of specified
%%	version. An application is made up of a number of modules of
%%	specified version. The specification of application versions
%%	is done in the ReleaseName.rel file and the specification of
%%	the module versions is done in the .app file of the
%%	application (which should be in the ebin directory of the
%%	application). The instructions how to Up (or Down) an
%%	application is put in the .appup file of the application
%%	(which also should be in ebin).
%%
%%	mk_relup will read the ReleaseName.rel file and find all .app
%%	and .appup files for the target release. It will then scan
%%	through the Up and Down lists and for each item find its .rel
%%	file, compare its application versions with the target and if
%%	the versions differ it will record the release upgrade
%%	instructions for each application.
%%
%%
%% Used terms (parameters and terms mixed)
%% ----------
%%
%%	The product which is released has a name such as "OTP APN 123
%%	456" or "VOLVO 850 GLS" or something similar which officially
%%	identifies the product.
%%
%%	Each (official or internal) release of the product has a
%%	version, such as "P1G".
%%
%%	Each release of the product also has a ReleaseName. It is an
%%	unofficial name given by the release creator so that he can
%%	easily administer his releases. (examples are: "test",
%%	"october"). The name is used for identifying the release
%%	specification (in the ReleaseName.rel file) and there may be
%%	several releases (with different names) with the same version.
%%
%%	UpNameList and DownNameList are lists of release names (as
%%	above). These lists contains those releases that can be Up (or
%%	Down) graded to.
%% 
%%	Path is those directories in which to search for the release
%%	files (the ReleaseName.rel, the .app and .appup files), it
%%	defaults to the standard Erlang code search path and if the
%%	path option is used then that path is prepended to the
%%	standard path.
%% 
%% 
%% 
%%-----------------------------------------------------------------


-module(systools_relup).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('peterl@erix.ericsson.se').

-export([mk_relup/3, mk_relup/4, format_error/1, format_warning/1]).

-include("systools.hrl").




%%-----------------------------------------------------------------
%% Options is a list of {path, Path} | silent | noexec where path sets
%% search path, silent supresses error message printing on console,
%% noexec supresses writing the output "relup" file
%%
%% Note side effect in mk_relup2 which writes to disk.
%%-----------------------------------------------------------------
mk_relup(RelName, UpList, DnList) ->
    mk_relup(RelName, UpList, DnList, []).
mk_relup(RelName, UpList, DnList, Opts) ->
    R = (catch mk_relup2(RelName, UpList, DnList, get_path(Opts), Opts)),
    case {get_opt(silent, Opts), get_opt(noexec, Opts)} of
	{false, false} ->
	    case R of
		{ok, Res, Mod, Ws} -> warn2(Ws),
				      ok;
		Other -> error2(Other),
			 error
	    end;
	_ -> R
    end.



mk_relup2(RelName, UpList, DnList, Path, Opts) ->
    case systools_make:get_release(to_list(RelName), Path, false) of
	{ok, Rel, AppList2, Warns} ->
	    AppList = lists:map(fun(X) -> element(2, X) end, AppList2),
	    {Up, Warns2} = foreach_up(up, Rel, AppList, UpList, Path, 
				      Opts, Warns, []),
	    %% Down
	    {Dn, Warns99} = foreach_up(dn, Rel, AppList, DnList, Path, 
				       Opts, Warns2, []),
	    
	    Relup = {Rel#release.vsn, Up, Dn},
	    write_relup_file(Relup, Opts),
	    {ok, Relup, ?MODULE, Warns99};
	Other -> throw(Other)
    end.


%% foreach_up checks each source release
foreach_up(Mode, Rel, AppList, [N|Ns], Path, Opts, Ws, Out) ->
    case systools_make:read_release(get_name(N), Path) of
	{ok, Src} ->
	    {RU1, Ws2} = check_erts_vsn(Rel, Src, Ws),

	    {RUabstr2, Ws3} = 
		foreach_app(Mode, AppList, Src, Path, Opts, Ws2, RU1),

	    {RUabstr, Ws99} = 
		check_del_apps(AppList, Src#release.applications, 
			       RUabstr2, Ws3),

	    {Alist, NewWs} =
		if  Mode == dn ->
			case systools_make:get_release(to_list(N),Path, false)
			    of
			    {ok, _, AppList2, Warns} ->
				{lists:map(fun(X) -> element(2, X) end, 
					   AppList2), Warns++Ws99};
			    Other -> throw(Other)
			end;
		    true -> {AppList, Ws99}
		end,

	    case systools_rc:translate_scripts(Mode, RUabstr, Alist) of
		{ok, RU} ->
		    foreach_up(Mode, Rel, AppList, Ns, Path, Opts, NewWs, 
			       [{Src#release.vsn, get_descr(N), RU} | Out]);
		XXX -> throw(XXX)
	    end;
	Other -> throw({error, ?MODULE, {no_rel_file, N, Other}})
    end;
foreach_up(Mode, Rel, AppList, [], Path, Opts, Ws, Out) -> {Out, Ws}.



%% Shall return {NewRU scriope, New warnings
check_erts_vsn(#release{erts_vsn = Vsn1, name = N1}, 
	       #release{erts_vsn = Vsn2, name = N2}, Ws) when Vsn1 /= Vsn2 ->
    {[[restart_new_emulator]], [{erts_vsn_changed, {N1, N2}} | Ws]};
check_erts_vsn(_, _, Ws) -> {[], Ws}.
		    
check_del_apps(AppList, SrcList, RU, Ws) ->
    {lists:foldl(fun({N,V,_T}, Acc) ->
			 case lists:keysearch(N, #application.name, AppList) of
			     {value, _} ->
				 Acc;
			     _ ->
				 [[{remove_application, N}] | Acc]
			 end end,
		 RU, SrcList), Ws}.


%% foreach_app checks every app in target against the source apps,
%% shall find the relup script if versions differ.
foreach_app(Mode, [A|As], Src, Path, Opts, Ws, Out) ->
    case find_app(A#application.name, Src#release.applications) of
	{N2, V2, _T2} ->
	    if  A#application.vsn == V2 ->
		    %% Easy case, same version
		    foreach_app(Mode, As, Src, Path, Opts, Ws, Out);
		true ->
		    %% Check if we have an upgrade script for this vsn
		    {Ws2, Out2} = find_relup(Mode, A, V2, Path, Opts, Ws, Out),
		    foreach_app(Mode, As, Src, Path, Opts, Ws2, Out2)
	    end;
	R ->
	    %% Oops, this is a new application.
	    foreach_app(Mode, As, Src, Path, Opts, Ws, 
			[[{add_application, A#application.name}] | Out])
    end;

foreach_app(Mode, [], Src, Path, Opts, Ws, Out) -> {Out, Ws}.


find_relup(Mode, A, Vsn, Path, Opts, Ws, Out) ->
    %% Find the bloody file
    F = A#application.dir ++ "/" ++ to_list(A#application.name) ++ ".appup",
    {S, V} = case systools_lib:read_term(F) of
		 {ok, {Vsn2, Up, Dn}} when Mode == up -> {Up, Vsn2};
		 {ok, {Vsn2, Up, Dn}} when Mode == dn -> {Dn, Vsn2};
		 X -> throw({error, ?MODULE, {file_problem, {F, X}}})
	     end,
    Ws2 = if  A#application.vsn == V -> Ws;
	      true -> [{bad_vsn,{ V, A#application.vsn}} | Ws]
	  end,
    case lists:keysearch(Vsn, 1, S) of
	{value, {_, Script}} ->
	    {Ws2, [Script | Out]};
	_ ->
	    throw({error, ?MODULE, {no_relup, F, A, Vsn}})
    end.

find_app(Name, List) ->
    case lists:keysearch(Name, 1, List) of
	{value, V} -> V;
	_ -> false
    end.

%% Primitives for the "lists of release names" that we upgrade from
%% and to.
get_name({N, D}) -> to_list(N);
get_name(N) -> to_list(N).

get_descr({N, D}) -> D;
get_descr(N) -> [].




to_list(X) when atom(X) -> atom_to_list(X);
to_list(X) when list(X) -> X.



%%-----------------------------------------------------------------
%% Shall write a relup file and return the RelupScript
%%-----------------------------------------------------------------
write_relup_file(RelupList, Opts) ->
    case get_opt(noexec, Opts) of
	true -> ok;
	_ ->
	    case file:open("relup", write) of
		{ok, Fd} ->
		    io:format(Fd, "~p.~n", [RelupList]),
		    file:close(Fd);
		{error, Reason} ->
		    throw({error, ?MODULE, {file_problem, {"relup", Reason}}})
	    end
    end,
    {ok, RelupList}.    

get_path(Opts) ->
    case get_opt(path, Opts) of
	false ->
	    code:get_path();
	P1 ->
	    P2 = lists:map(fun(Dir) when atom(Dir) -> atom_to_list(Dir);
			      (Dir)                -> Dir
			   end, P1),
	    P3 = systools_lib:get_path(P2),
	    make_set(P3 ++ code:get_path())
    end.

get_opt(Opt, Opts) ->
    case lists:keysearch(Opt, 1, Opts) of
	{value, {_, Val}} -> Val;
	_ ->
	    case lists:member(Opt, Opts) of
		true -> true;
		_ -> default(Opt)
	    end
    end.

make_set([]) -> [];
make_set([H|T]) ->
    [H | [ Y || Y<- make_set(T),
		Y =/= H]].

default(path)   -> false;
default(noexec) -> false;
default(silent) -> false.


error2({'EXIT', Err}) -> error2(Err);
error2({error, Mod, Error}) ->
    S = apply(Mod, format_error, [Error]),
    io:format(S, []);
error2(Other) ->
    io:format("Error: ~p~n", [Other]).

format_error({file_problem, {File, What}}) ->
    io_lib:format("Could not ~p file ~p~n", [get_reason(What), File]);
format_error({no_rel_file, N, What}) ->
    io_lib:format("No release file for ~p, reason: ~p~n", [get_name(N),
							   get_reason(What)]);
format_error({no_relup, File, App, Vsn}) ->
    io_lib:format("No release upgrade script entry for ~p-~s to ~p-~s in file ~p~n",
		  [App#application.name, App#application.vsn, 
		   App#application.name, Vsn, File]);

format_error(Error) ->
    io:format("~p~n", [Error]).


warn2(L) when list(L) ->
    lists:foreach(fun(W) -> warn2(W) end, L);
warn2(W) ->
    S = format_warning(W),
    io:format("~s", [S]).

format_warning({erts_vsn_changed, {Rel1, Rel2}}) ->
    io_lib:format("*WARNING* The ERTS version changed between ~p and ~p~n",
		  [Rel1, Rel2]);
format_warning(What) ->
    io_lib:format("*WARNING* ~p~n",[What]).


get_reason({error, {open, _, _}}) -> open;
get_reason({error, {read, _, _}}) -> read;
get_reason({error, {parse, _, _}}) -> parse;
get_reason({error, {open, _}}) -> open;
get_reason({error, {read, _}}) -> read;
get_reason({error, {parse, _}}) -> parse;
get_reason({open, _}) -> open;
get_reason({read, _}) -> read;
get_reason({parse, _}) -> parse;
get_reason(open) -> open;
get_reason(read) -> read;
get_reason(parse) -> parse.



