%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_app).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

-behaviour(application).

%%%-----------------------------------------------------------------
%%%  This module implements the SNMP application.
%%%-----------------------------------------------------------------
-export([start/2, stop/1]).

start(Type, []) ->
    DbDir =
	case application:get_env(snmp, snmp_db_dir) of
	    {ok, Dir2} when list(Dir2) -> Dir2;
	    {ok, Bad} -> exit({bad_config, {snmp, {snmp_db_dir, Bad}}});
	    _ -> exit({undefined_config, {snmp, snmp_db_dir}})
    end,
    Prio =
	case application:get_env(snmp, snmp_priority) of
	    {ok, Pr} when atom(Pr) -> Pr;
	    _ -> normal
    end,
    MultiT =
	case application:get_env(snmp, snmp_multi_threaded) of
	    {ok, true} -> true;
	    {ok, false} -> false;
	    {ok, Bad2} -> exit({bad_config,{snmp,{snmp_multi_threaded,Bad2}}});
	    _ -> false
    end,
    Vsns = snmp_misc:get_vsns(),
    case application:get_env(snmp, snmp_agent_type) of
	{ok, sub} ->
	    Opts = [{priority, Prio},
		    {snmp_vsn, Vsns},
		    {multi_threaded, MultiT}],
	    case snmp_supervisor:start_sub(DbDir, Opts) of
		{ok, Pid} ->
		    {ok, Pid, []};
		Error ->
		    Error
	    end;
	_ ->
	    ConfDir =
		case application:get_env(snmp, snmp_config_dir) of
		    {ok, Dir1} when list(Dir1) -> Dir1;
		    {ok, Bud} ->exit({bad_config,{snmp,{snmp_config_dir,Bud}}});
		    _ -> exit({undefined_config, {snmp, snmp_config_dir}})
		end,
	    Mibs =
		case application:get_env(snmp, snmp_master_agent_mibs) of
		    {ok, Ms} when list(Ms) -> 
			lists:map(fun(Mib) -> ConfDir ++ Mib end, Ms);
		    _ -> []
		end,
	    Opts = [{priority, Prio},
		    {snmp_vsn, Vsns},
		    {multi_threaded, MultiT},
		    {mibs, Mibs},
		    {name, {local, snmp_master_agent}}],
	    case snmp_supervisor:start_master(DbDir, ConfDir, Opts) of
		{ok, Pid} when Type == normal ->
		    {ok, Pid};
		{ok, Pid} ->
		    {takeover, Node} = Type,
		    case rpc:call(Node, snmp, info, [snmp_master_agent]) of
			{badrpc, R} ->
			    error_logger:info_msg("snmp: could not takeover "
						  "loaded mibs: ~p~n", [R]);
			List ->
			    {value, {_, LMibs}} =
				lists:keysearch(loaded_mibs, 1, List),
			    lists:foreach(fun takeover_mib/1, LMibs)
		    end,
		    {ok, Pid};
		Else ->
		    Else
	    end
    end.

stop(_) ->
    ok.

takeover_mib({MibName, _Symbolic, FileName}) ->
    case MibName of
	'STANDARD-MIB' -> ok;
	'SNMPv2-MIB' -> ok;
	_ -> 
	    case snmp:load_mibs(snmp_master_agent, [FileName]) of
		ok -> ok;
		{error, R} ->
		    error_logger:info_msg("snmp: could not reload mib ~p: ~p~n",
					  [FileName, R])
	    end
    end.

