%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%----------------------------------------------------------------------
%% This is parts stolen from the asn1 compiler by klacke.
%% Functions come from asn1_bits, asn1_decode, asn1_misc.
%% 
%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%%----------------------------------------------------------------------

-module(snmp_asn1).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-export([dec_len/1]).
-author('eklas@erlang.ericsson.se'). % is there a thief attribute :-)
-revision('$Revision: 1.2 $').
-date('$Date: 1997/06/23 11:01:27 $').

%%----------------------------------------------------------------------
%% This function was all I needed. It's to nasty to rewrite.
%%----------------------------------------------------------------------
dec_len([0,0|Tail]) ->
    L = eat_indefinite(Tail),
    {L,Tail};

dec_len([Hd|Tl]) ->
    case isset(Hd,8) of
	0 ->
	    {Hd,Tl};
	1->
	    No = clear(Hd,8),
	    {DigList,Rest} = head(No,Tl),
	    Size = dec_integer_len(DigList),
	    {Size,Rest}
    end.

eat_indefinite([0,0|_]) ->
    2;

eat_indefinite(Bytes) ->
    {Num,Tail} = eat_part(Bytes),
    Num + eat_indefinite(Tail).

eat_part(Bytes) ->
    {_,This,Rest} = dec_head(Bytes),
    {length(This),Rest}.

dec_integer_len([D]) ->
    D;
dec_integer_len([A,B]) ->
    (A bsl 8) bor B;
dec_integer_len([A,B,C]) ->
    (A bsl 16) bor (B bsl 8) bor C.

dec_head(Bytes) -> 
    case has_long_tag(Bytes) of
	true ->
	    [Hi,Lo|Tail] = Bytes,
	    dec_head2([Hi,Lo],Tail);
	false ->
	    [Tag|Tail] = Bytes,
	    dec_head2([Tag],Tail)
    end.

dec_head2(Tag,[Hd|Tl]) ->
    case isset(Hd,8) of
	0 ->
	    Size = Hd,
	    {V,R} = head(Size,Tl),
	    {Tag,lists:append(Tag,[Hd|V]),R};
	1 ->
	    No = clear(Hd,8),
	    {DigList,R} = head(No,Tl),
	    Size = dec_integer_len(DigList),
	    {This,Next} = head(Size,R),
	    {Tag,lists:concat([Tag,[Hd|DigList],This]),Next}
    end.

has_long_tag([A|Tail]) ->
    if
	A bor 2#11100000 == 255 -> true;
	true -> false
    end.

head(L,List) when length(List) == L -> {List,[]};

head(L,List) ->
    case catch head(L,List,[]) of
	{'EXIT',_} ->
	    exit({asn1_error,{'Can not get entire length of bytes ',L}});
	X ->
	    X
    end.


head(0,L,Res) ->
    {lists:reverse(Res),L};

head(Int,[H|Tail],Res) ->
    head(Int-1,Tail,[H|Res]).

clear(Byte,8) -> 
    Byte band 127;
clear(Byte,Pos) when Pos < 9 ->
    Mask = bnot set(0,Pos),
    Mask band Byte.

isset(Byte,8) ->
    if
	Byte > 127 -> 1;
	true -> 0
    end;

isset(Byte,Pos) when Pos < 9 ->
    if
	Byte band ( 1 bsl (Pos -1)) > 0 ->
	    1;
	true ->
	    0
    end.

set(Byte,8) -> 
    Byte bor 2#10000000;
set(Byte,Pos)  when Pos < 9 ->
    Mask = 1  bsl (Pos-1),
    Byte bor Mask.

