%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_config).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-include("snmp_types.hrl").
-export([config/0, write_files/7, write_files/8]).

%%----------------------------------------------------------------------
%% Handy SNMP configuration
%%----------------------------------------------------------------------

config() ->
    case catch config2() of
	ok -> ok;
	{error, Reason} -> {error, Reason};
	Q -> {error,failed}
    end.

config2() ->
    case catch config3() of
	ok -> ok;
	{error, Reason} -> {error, Reason};
	Q -> {error, dialog_failed}
    end.

config3() ->
    io:format("~nSimple SNMP configuration tool (v~s)~n", [?version]),
    p("----------------------------------------------"),
    p("Note: Non-trivial configurations still has to be done manually."),
    p("IP addresses may be entered as dront.ericsson.se (UNIX only) "
      "or 123.12.13.23~n"),
    UsersName='$USER'(),
    SysName=ask(["1. System name (sysName standard variable) [",UsersName,"]"],
		 UsersName),
    AgentUDP=ask("2. The UDP port the agent listens to. (standard 161) [4000]",
		 "4000"),
    Host = '$HOST'(),
    AgentIP=make_ip(ask(["3. IP address for the agent (only used as id\n"
			 "   when sending traps) [", Host, "]"], Host)),
    ManagerIP=make_ip(ask(["4. IP address for the manager (only this manager "
			   "will have access\n   to the agent, traps are sent "
			   "to this one) [", Host, "]"], Host)),
    TrapUdp=ask("5. To what UDP port at the manager should traps\n"
		"   be sent (standard 162)? [5000]", "5000"),
    Ver = ask("6. What SNMP version should be used (1,2,1&2)? [2]", "2"),
    NotifType=
	case Ver of
	    "2" ->
		ask("6.b. Should a notification be sent as trap or\n"
		    "     inform? [trap]", "trap");
	    _ ->
		"trap"
	end,
    {ok, DefDir} = file:get_cwd(),
    Dir=ask("7. Where is the configuration directory (absolute)? [" ++ DefDir ++ "]",
	    DefDir),
    verify_dir(Dir),
    NewVer = verify_ver(Ver),
    NewNotifType = verify_notif_type(NotifType),
    "y" = ask("8. Current configuration files will now be overwritten. "
	      "Ok [y]/n?", "y"),
    Dir2 = snmp_misc:ensure_trailing_dir_delimiter(Dir),
    case catch write_files(Dir2,NewVer,ManagerIP,TrapUdp,
			   AgentIP,AgentUDP,SysName,NewNotifType) of
	ok ->
	   p("------------------------"),
	   p("Info: 1. Community \"public\" has read-only access to all mibs."),
	   p("      2. Community \"all-rights\" has read-write access to all "
	     "mibs."),
	    p("      3. Standard traps (community \"standard trap\") is sent "
	      "to the manager."),
	    p("The following files were written: address.conf, agent.conf, "
	      "community.conf,\n   standard.conf, target_addr.conf, "
	      "target_params.conf\n   notify.conf view.conf, sys.config"),
	    p("------------------------"),
	    ok;
	Q -> {error, 'write files failed'}
    end.

verify_ver("1") -> [1];
verify_ver("2") -> [2];
verify_ver([$1,_,$2]) -> [1,2];
verify_ver([$2,_,$1]) -> [1,2];
verify_ver(_) -> throw({error, 'version syntax error'}).
    
verify_dir(Dir) ->
    case filename:pathtype(Dir) of
	absolute -> ok;
	Q -> throw({error,bad_dir})
    end.
	    
verify_notif_type("trap") -> trap;
verify_notif_type("inform") -> inform;
verify_notif_type(_) -> throw({error, 'notifcation type syntax error'}).
    

ip(Host) ->
    case catch snmp_misc:ip(Host) of
	{ok, IPtuple} -> tuple_to_list(IPtuple);
	{error, Reason} -> throw({error, Reason});
	Q -> throw({error, {'ip conversion failed', Host}})
    end.

make_ip(Str) ->
    case catch snmp_misc:ip(Str) of
	{ok, IPtuple} -> tuple_to_list(IPtuple);
	Q -> ip(Str)
    end.
		 
    
p(Str) ->
    io:format(Str), io:format("~n").

ask(Str, Defval) ->
    io:format(Str),
    Answ = io:get_line(''),
    case remove_newline(Answ) of
	"" -> Defval;
	Q -> Q
    end.

'$HOST'() ->
    case os:type() of
	{unix, _} ->
	    remove_newline(unix:cmd("echo $HOST"));
	_ -> "127.0.0.1"
    end.

'$USER'() ->
    case os:type() of
	{unix, _} ->
	    lists:append(remove_newline(unix:cmd("echo $USER")), "'s agent");
	{_,_} -> "my agent"
    end.

remove_newline(Str) -> 
    lists:delete($\n, Str).

%%======================================================================
%% File generation
%%======================================================================

%%----------------------------------------------------------------------
%% Dir: string()  (ex: "../conf/")
%% ManagerIP, AgentIP: [int(),int(),int(),int()]
%% TrapUdp, AgentUDP: string()    (ex: "4000")
%% SysName: string()
%%----------------------------------------------------------------------
write_files(Dir,Ver,ManagerIP,TrapUdp,AgentIP,AgentUDP,SysName) ->
    write_files(Dir,Ver,ManagerIP,TrapUdp,AgentIP,AgentUDP,SysName,"trap").
write_files(Dir,Ver,ManagerIP,TrapUdp,AgentIP,AgentUDP,SysName,NotifType) ->
    write_address_conf(Dir, ManagerIP, TrapUdp),
    write_agent_conf(Dir, AgentIP, AgentUDP),
    write_community_conf(Dir, ManagerIP),
    write_standard_conf(Dir, SysName),
    write_target_addr_conf(Dir, ManagerIP, TrapUdp, Ver),
    write_target_params_conf(Dir, Ver),
    write_notify_conf(Dir, NotifType),
    write_view_conf(Dir),
    write_sys_config(Dir,Ver),
    ok.

write_address_conf(Dir, ManagerIP, TrapUdp) ->
    Comment = 
"%% This files defines the address info for the managers\n"
"%% known to the agent.\n"
"%% The data is inserted into the intAddressTable defined\n"
"%% in OTP-SNMPEA-MIB.\n"
"%% Each row is a 3-tuple:\n"
"%% {ManagerIpAddress, UDPPort, MaxPacketSize}.\n"
"%% MaxPacketSize is usually 484 for UDP based communication.\n"
"%% For example\n"
"%% {[127,42,17,6], 5000, 484}.\n"
"%% {[127,42,17,8], 162, 484}.\n"
"%%\n",
    {ok, Fid} = file:open(filename:join(Dir,"address.conf"),write),
    ok = io:format(Fid, "~s~s{~w,~s,~w}.~n",
		   [header(), Comment, ManagerIP, TrapUdp, 484]),
    file:close(Fid).

write_agent_conf(Dir, AgentIP, AgentUDP) -> 
    Comment = 
"%% This files defines the Agent address info\n"
"%% The data is inserted into the intAgent* variables defined\n"
"%% in OTP-SNMPEA-MIB.\n"
"%% Each row is a 2-tuple:\n"
"%% {IntAgentVariable, Value}.\n"
"%% For example\n"
"%% {intAgentUDPPort, 4000}.\n"
"%% The ip address for the agent is sent as id in traps.\n"
"%% {intAgentIpAddress, [127,42,17,5]}.\n"
"%% {intAgentMaxPacketSize, 484}.\n",
    {ok, Fid} = file:open(filename:join(Dir,"agent.conf"),write),
    ok = io:format(Fid, "~s~s~n{intAgentUDPPort, ~s}.~n~w."
		   "~n{intAgentMaxPacketSize, 484}.~n",
		   [header(), Comment, AgentUDP, {intAgentIpAddress, AgentIP}]),
    file:close(Fid).

write_community_conf(Dir, ManagerIP) ->
    Comment = 
"%% This files defines the community info\n"
"%% The data is inserted into the intCommunityTable defined\n"
"%% in OTP-SNMPEA-MIB.\n"
"%% Each row is a 4-tuple:\n"
"%% {ManagerIpAddress, CommunityString, MibView, Access}.\n"
"%% For example\n"
"%% {[127,42,17,6], \"public\", 1, readWrite}.\n"
"%% {[127,42,17,6], \"standard trap\", 1, read}.\n"
"%% {[0,0,0,0], \"public\", 2, read}.\n"
"%%\n",
    {ok, Fid} = file:open(filename:join(Dir,"community.conf"),write),
    ok = io:format(Fid, "~s~s{~w, \"public\", 1, read}.~n"
		   "{~w, \"all-rights\", 1, readWrite}.~n"
		   "{~w, \"standard trap\", 1, read}.~n",
		   [header(), Comment, ManagerIP, ManagerIP, ManagerIP]),
    file:close(Fid).


write_standard_conf(Dir, SysName) ->
        Comment = 
"%% This files defines the STANDARD-MIB info.\n"
"%% Each row is a 2-tuple:\n"
"%% {StandardVariable, Value}.\n"
"%% For example\n"
"%% {sysDescr, \"Erlang SNMPv1 agent\"}.\n"
"%% {sysObjectID, [1,2,3]}.\n"
"%% {sysContact, \"(mbj,eklas)@erlang.ericsson.se\"}.\n"
"%% {sysName, \"test\"}.\n"
"%% {sysLocation, \"erlang\"}.\n"
"%% {sysServices, 72}.\n"
"%% {snmpEnableAuthenTraps, enabled}.\n"
"%%\n"
"{sysDescr, \"Erlang SNMPv1 agent\"}.\n"
"{sysObjectID, [1,2,3]}.\n"
"{sysContact, \"(mbj,eklas)@erlang.ericsson.se\"}.\n"
"{sysLocation, \"erlang\"}.\n"
"{sysServices, 72}.\n"
"{snmpEnableAuthenTraps, enabled}.\n",
    {ok, Fid} = file:open(filename:join(Dir,"standard.conf"),write),
    ok = io:format(Fid, "~s~s{sysName, \"~s\"}.~n",
		   [header(), Comment, SysName]),
    file:close(Fid).
	
write_target_addr_conf(Dir, ManagerIp, UDP, Vers) -> 
        Comment = 
"%% This files defines the target address parameters.\n"
"%% The data is inserted into the snmpTargetAddrTable defined\n"
"%% in SNMP-TARGET-MIB.\n"
"%% Each row is a 7-tuple:\n"
"%% {Name, Ip, Udp, Timeout, RetryCount, TagList, ParamsName}.\n"
"%% For example\n"
"%% {\"1.2.3.4 v1\", [1,2,3,4], 162, 1500, 3, \"std_inform\", \"otp_v2\"}.\n"
"%%\n",
    {ok, Fid} = file:open(filename:join(Dir,"target_addr.conf"),write),
    io:format(Fid, "~s~s", [header(), Comment]),
    lists:foreach(fun(Ver) ->
			  ok = io:format(Fid, "{\"~s\", ~w, ~s, 1500, 3, "
					 "\"std_trap\", \"~s\"}.~n",
					 [mk_ip(ManagerIp, Ver),
					  ManagerIp, UDP, mk_param(Ver)])
		  end,
		  Vers),
    file:close(Fid).

mk_param(1) -> "otp_v1";
mk_param(2) -> "otp_v2".
     
mk_ip([A,B,C,D], Ver) ->
    io_lib:format("~w.~w.~w.~w v~w", [A,B,C,D,Ver]).

write_target_params_conf(Dir, Vers) -> 
        Comment = 
"%% This files defines the target parameters.\n"
"%% The data is inserted into the snmpTargetParamsTable defined\n"
"%% in SNMP-TARGET-MIB.\n"
"%% Each row is a 5-tuple:\n"
"%% {Name, MPModel, SecurityModel, SecurityName, SecurityLevel}.\n"
"%% For example\n"
"%% {\"otp_v1\", v1, otp_sec, \"\", noAuthNoPriv}.\n"
"%%\n",
    {ok, Fid} = file:open(filename:join(Dir,"target_params.conf"),write),
    io:format(Fid, "~s~s", [header(), Comment]),
    lists:foreach(fun(Ver) ->
			  SM = if Ver == 1 -> v1;
				  Ver == 2 -> v2c
			       end,
			  ok = io:format(Fid, "{\"otp_v~w\", ~w, otp_sec, "
					 "\"\", noAuthNoPriv}.~n",
					 [Ver, SM])
		  end,
		  Vers),
    file:close(Fid).

write_notify_conf(Dir, NotifyType) -> 
        Comment = 
"%% This files defines the notification parameters.\n"
"%% The data is inserted into the snmpNotifyTable defined\n"
"%% in SNMP-NOTIFICATION-MIB.\n"
"%% The Name is used as CommunityString for v1 and v2c.\n"
"%% Each row is a 3-tuple:\n"
"%% {Name, Tag, Type}.\n"
"%% For example\n"
"%% {\"standard trap\", \"std_trap\", trap}.\n"
"%% {\"standard inform\", \"std_inform\", inform}.\n"
"%%\n",
    {ok, Fid} = file:open(filename:join(Dir,"notify.conf"),write),
    ok = io:format(Fid, "~s~s{\"standard trap\", \"std_trap\", ~s}.~n",
		   [header(), Comment, NotifyType]),
    file:close(Fid).

ver_to_trap_str([1]) -> "v1";
ver_to_trap_str([2]) -> "v2";
% default is to use the latest snmp version
ver_to_trap_str([1,2]) -> "v2".


write_view_conf(Dir) -> 
        Comment = 
"%% This files defines the Mib Views\n"
"%% The data is inserted into the intViewTable defined\n"
"%% in OTP-SNMPEA-MIB.\n"
"%% Each row is a 4-tuple:\n"
"%% {ViewIndex, ViewSubtree, ViewStatus, ViewMask}.\n"
"%% For example\n"
"%% {1, [1,3,6,1,2,1,1], included, null}.\n"
"%% {1, [1,3,6,1,3], included, null}. % for EX1-MIB\n"
"%% {2, [1,3], included, null}.\n"
"%%\n",
    {ok, Fid} = file:open(filename:join(Dir,"view.conf"),write),
    ok = io:format(Fid, "~s~s{1, [1,3,6], included, null}.~n", [header(),
								Comment]),
    file:close(Fid).

write_sys_config(Dir,Version) ->
    {ok, Fid} = file:open(filename:join(Dir,"sys.config"),write),
    ok = io:format(Fid, "~s[{snmp, [{snmp_config_dir, \"~s\"},~n"
		   "         {snmp_db_dir, \"~s\"},~n         ~s]}].~n",
		   [header(),Dir,Dir,ver_to_sys_str(Version)]),
    file:close(Fid).


ver_to_sys_str([1]) -> "{v1, true}, {v2, false}";
ver_to_sys_str([2]) -> "{v1, false}, {v2, true}";
ver_to_sys_str([1,2]) -> "{v1, true}, {v2, true}".

header() ->
    {Y,Mo,D} = date(),
    {H,Mi,S} = time(),
    io_lib:format("%% This file was generated by "
		  "snmp_config (v~s)   ~w-~w-~w ~w:~w:~w~n",
		  [?version,Y,Mo,D,H,Mi,S]).
