%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_mgr_misc).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('eklas@erlang.ericsson.se').
-revision('$Revision: 1.9 $').
-date('$Date: 1997/09/09 13:59:35 $').
%% c(snmp_mgr_misc).

%% API
-export([start_link_packet/6, stop/1, send_pdu/2, send_msg/4, error/2]).

%% internal exports
-export([init_packet/7]).

-include("snmp_types.hrl").
-record(mini_mib_elem,{aliasname, oid, type}).

%%----------------------------------------------------------------------
%% The InHandler process will receive messages on the form {snmp_pdu, Pdu}.
%%----------------------------------------------------------------------
start_link_packet(InHandler, AgentIp, UdpPort, TrapUdp, Community, Version) 
  when integer(UdpPort) ->
    proc_lib:start_link(snmp_mgr_misc, init_packet,
		[self(),InHandler,AgentIp,UdpPort,TrapUdp, Community, Version]).

stop(Pid) ->
    Pid ! {stop, self()},
    receive
	{Pid, stopped} -> ok
    end.
	

send_pdu(Pdu, PacketPid) when record(Pdu, pdu) ->
    PacketPid ! {send_pdu, Pdu}.

send_msg(Msg, PacketPid, Ip, Udp) when record(Msg, message) ->
    PacketPid ! {send_msg, Msg, Ip, Udp}.

%%--------------------------------------------------
%% The SNMP encode/decode process
%%--------------------------------------------------
init_packet(Parent, SnmpMgr, AgentIp, UdpPort, TrapUdp, Community, Version) ->
    {ok, UdpId} = gen_udp:open(TrapUdp, [{reuseaddr, true}]),
    proc_lib:init_ack(Parent, self()),
    packet_loop(SnmpMgr, UdpId, AgentIp, UdpPort, Community, Version).

packet_loop(SnmpMgr, UdpId, AgentIp, UdpPort, Community, Version) ->
    receive
	{send_pdu, Pdu} ->
	    Msg = #message{version = Version, community = Community,
			   data = Pdu},
	    case catch snmp_pdus:enc_message(Msg) of
		{'EXIT', Reason} ->
		    error("Encoding error. Pdu: ~w. Reason: ~w",[Pdu, Reason]);
		B when list(B) -> 
		    gen_udp:send(UdpId, AgentIp, UdpPort, B)
	    end,
	    packet_loop(SnmpMgr, UdpId, AgentIp, UdpPort, Community, Version);
	{send_msg, Msg, Ip, Udp} ->
	    case catch snmp_pdus:enc_message(Msg) of
		{'EXIT', Reason} ->
		    error("Encoding error. Msg: ~w. Reason: ~w",[Msg, Reason]);
		B when list(B) -> 
		    gen_udp:send(UdpId, Ip, Udp, B)
	    end,
	    packet_loop(SnmpMgr, UdpId, AgentIp, UdpPort, Community, Version);
	{udp, UdpId, Ip, UdpPort, Bytes} ->
	    case catch snmp_pdus:dec_message(Bytes) of
		Message when record(Message, message) ->
		    case SnmpMgr of
			{pdu, Pid} ->
			    Pid ! {snmp_pdu, Message#message.data};
			{msg, Pid} ->
			    Pid ! {snmp_msg, Message, Ip, UdpPort}
		    end;
		Reason ->
		    error("Decoding error. Bytes: ~w ~n Reason: ~w "
			  "(UDPport: ~w, Ip: ~w)",
			  [Bytes, Reason, UdpPort, Ip])
	    end,
	    packet_loop(SnmpMgr, UdpId, AgentIp, UdpPort, Community, Version);
	{stop, Pid} ->
	    gen_udp:close(UdpId),
	    Pid ! {self(), stopped},
	    exit(normal);
	Other ->
	    exit({snmp_packet_got_other, Other})
    end.

error(Format, Data) ->
    io:format("* Error: "),
    ok = io:format(Format, Data),
    io:format("~n").

