%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%*********************************************************************
%%%
%%%   Filename:         %M%
%%%
%%%   Author:           EPK/TE  Fredrik Gustafson 
%%% 
%%%   SCCS Delta ID:    %I%
%%%
%%%   SCCS Delta Date:  %E%
%%%
%%%   SCCS ID:          @(#)  %M%  %I%  %E%
%%%
%%%   Description:      
%%%
%%%*********************************************************************
%%%
%%%   COPYRIGHT ERICSSON TELECOM AB 1997
%%%
%%%   THE COPYRIGHT TO THE COMPUTER PROGRAM HEREIN IS THE
%%%   PROPERTY OF ERICSSON TELECOM AB, SWEDEN. THE PROGRAM
%%%   MAY BE USED AND/OR COPIED ONLY WITH THE WRITTEN PERMISSION
%%%   FROM ERICSSON TELECOM AB, OR IN ACCORDANCE WITH THE TERMS
%%%   AND CONDITIONS STIPULATED IN THE AGREEMENT/CONTRACT UNDER
%%%   WHICH THE PROGRAM HAS BEEN SUPPLIED.
%%%
%%%*********************************************************************
%%%
%%%   Revision information:
%%%
%%%   Date      Name                       Description
%%%   ------    -----------------------    -----------
%%%
%%%
%%%*********************************************************************



-module(coast_comm_funcs).
-author('fredrik@erlang.ericsson.se').
-vsn('$Revision: /main/release/free/1').
-copyright('Copyright (C) 1991-97 Ericsson Telecom AB').



-export([remove_from_database/1,
	 get_source_files/1
	]).


-include("coast_server.hrl").






%%%*********************************************************************
%%% EXTERNAL FUNCTIONS
%%%*********************************************************************




remove_from_database(Module) ->
    ets:delete(?TABLE_NAME, {module_initialised, Module}),
    ets:delete(?TABLE_NAME, {first_code_line, Module}),
    %ets:delete(?TABLE_NAME, {end_of_function, Module}),
    OldEntries = ets:match(?TABLE_NAME, 
			   {{Module, '$1', '$2', '$3', '$4', '$5', '$6'}, 
			    '_', '_', '_'}),
    remove_old_entries(Module, OldEntries).




get_source_files(Modules) ->
    get_source_files(Modules, []).



	

%%%*********************************************************************
%%% INTERNAL FUNCTIONS
%%%*********************************************************************




get_source_files([Module | T], Acc) ->
    SrcFile = case catch Module:module_info(attributes) of
		  {'EXIT', Reason} ->
		      {error, {no_such_module, Module}};
		  ModuleInfo ->
		      case catch lists:keysearch(coast_compiled, 1, ModuleInfo) of
			  false ->
			      {error, {not_coast_compiled, Module}};
			  {value, {coast_compiled, [{Module, FileName}]}} ->
			      FileName
		      end
	      end,
    get_source_files(T, [SrcFile | Acc]);
get_source_files([], Acc) ->
    lists:reverse(Acc).



remove_old_entries(Module, [[F, A, I, C, CL, L] | T]) ->
    ets:delete(?TABLE_NAME, {Module, F, A, I, C, CL, L}),
    remove_old_entries(Module, T);
remove_old_entries(Module, []) ->
    ok.





