-module(asn1ct_parser).
-define(THIS_MODULE, asn1ct_parser).
-export([parse/1, parse_and_scan/1, format_error/1]).
-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/1 $').
-include("asn1_records.hrl").
%-record(module,{name,tagdefault,exports,imports, typval}).
%-record(typval,{name,type}).
%-record(typedef,{tag = notag,typespec}).

to_set(V) when list(V) ->
	ordsets:list_to_set(V);
to_set(V) ->
	ordsets:list_to_set([V]).

merge_constraints([Ch|Ct],Initc) ->
	E = Ch#constraint.e,
	case Ch#constraint.c of
		{'SizeConstraint',V} ->
			merge_constraints(Ct,Initc#'Constraint'{'SizeConstraint'=V,e=E});
		{'SingleValue',V} when integer(V) ->
			merge_constraints(Ct,Initc#'Constraint'{'SingleValue'=V,
				'ValueRange'={V,V},e=E});
		{'SingleValue',V} when list(V) ->
			merge_constraints(Ct,Initc#'Constraint'{'SingleValue'=V,
				'ValueRange'={lists:min(V),lists:max(V)},e=E});
		{'ValueRange',V}  ->
			merge_constraints(Ct,Initc#'Constraint'{'ValueRange'=V,e=E});
		{'PermittedAlphabet',{'SingleValue',V}} when list(V) ->
			V2 = {'SingleValue',ordsets:list_to_set(V)},
			merge_constraints(Ct,Initc#'Constraint'{'PermittedAlphabet'=V2,e=E});
		{'PermittedAlphabet',{'SingleValue',V}} ->
			V2 = {'SingleValue',[V]},
			merge_constraints(Ct,Initc#'Constraint'{'PermittedAlphabet'=V2,e=E});
		Other ->
			merge_constraints(Ct,Initc#'Constraint'{'Other'=Other,e=E})
	end;

merge_constraints([],Initc) ->
	Initc.




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 3, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, 'DEFINITIONS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 5, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(4, __Cat, [3 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(4, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ModuleIdentifier', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(5, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('DefinitiveObjIdComponentList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [7 | __Ss], [__T | __Stack]);
yeccpars2(7, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(8, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinitiveObjIdComponent', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinitiveObjIdComponent', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [10 | __Ss], [__T | __Stack]);
yeccpars2(10, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(11, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('DefinitiveObjIdComponent', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('DefinitiveIdentifier', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('DefinitiveObjIdComponentList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, 'AUTOMATIC', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, 'IMPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, 'EXPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'EXPLICIT'), 'EXPLICIT'
  end,
 yeccpars2(19, __Cat, [15 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, 'TAGS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 341, [16 | __Ss], [__T | __Stack]);
yeccpars2(16, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(17, 'TAGS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 340, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(18, 'TAGS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 339, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, 'EXTENSIBILITY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  false,
 yeccpars2(21, __Cat, [19 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(20, 'IMPLIED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 338, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [21 | __Ss], [__T | __Stack]);
yeccpars2(21, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(22, 'BEGIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [22 | __Ss], [__T | __Stack]);
yeccpars2(22, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(23, 'EXPORTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'END', __Ss, __Stack, __T, __Ts, __Tzr) ->
 __Val =  {'ModuleBody',nil,nil,[]},
 yeccpars2(26, 'END', [23 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  {exports,all},
 yeccpars2(25, __Cat, [23 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(24, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 323, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 335, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, 'IMPORTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  {imports,[]},
 yeccpars2(29, __Cat, [25 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(26, 'END', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [26 | __Ss], [__T | __Stack]);
yeccpars2(26, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(27, __Cat, __Ss,  [__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'ModuleBody',Ex,Im,Types} = __7, {{typereference,Pos,Name},Defid} = __1, #module{pos = Pos, name = Name, defid = Defid, tagdefault = __3, extensiondefault = __4, exports = Ex, imports = Im, typeorval = Types}
  end,
 __Nss = lists:nthtail(7, __Ss),
 yeccpars2(yeccgoto('ModuleDefinition', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(28, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 323, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(29, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(30, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('AssignmentList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ModuleBody',__1,__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ModuleBody', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Assignment', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Assignment', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(35, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(36, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(37, 'DEFINED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 310, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, 'STRING', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 306, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = __1},
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, 'STRING', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(43, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [43 | __Ss], [__T | __Stack]);
yeccpars2(43, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ReferencedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, 'PDV', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(49, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [49 | __Ss], [__T | __Stack]);
yeccpars2(49, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'GeneralizedTime',
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 285, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'INTEGER',
 yeccpars2(yeccgoto('IntegerType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = 'NULL'},
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, 'IDENTIFIER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [57 | __Ss], [__T | __Stack]);
yeccpars2(57, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(58, 'STRING', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 283, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'ObjectDescriptor',
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = __1},
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 272, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 271, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 273, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(63, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 244, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 243, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 245, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ReferencedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'EXPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 237, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'IMPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [71 | __Ss], [__T | __Stack]);
yeccpars2(71, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #typedef{pos = element(2,__1), name = element(3,__1), typespec = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TypeAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ConstrainedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'UTCTime',
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, 'PRIVATE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 225, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'APPLICATION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 223, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'UNIVERSAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  'CONTEXT',
 yeccpars2(224, __Cat, [74 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(75, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('CharacterStringType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(79, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externaltypereference'{pos = element(2,__1), module = element(3,__1), type = element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Externaltypereference', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(81, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SelectionType',element(3,__1),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SelectionType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(83, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2],#'Constraint'{})},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [84 | __Ss], [__T | __Stack]);
yeccpars2(84, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2,__3],#'Constraint'{})},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2,__3,__4],#'Constraint'{})},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(86, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2,__3,__4,__5],#'Constraint'{})},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(88, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 220, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(89, 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(90, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(91, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(92, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(93, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(94, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  undefined,
 yeccpars2(208, __Cat, [94 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(95, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ConstraintSpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('Elems', hd(__Ss)), 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(96, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('IntersectionElements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(97, 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'FALSE',
 yeccpars2(yeccgoto('BooleanValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(99, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(100, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'INTERSECTION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(101, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Intersections', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(103, 'INTERSECTION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('IElems', hd(__Ss)), 'INTERSECTION', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(103, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('IElems', hd(__Ss)), '^', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(103, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Unions', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('LowerEndpoint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, '..', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [105 | __Ss], [__T | __Stack]);
yeccpars2(105, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(106, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'MIN',
 yeccpars2(yeccgoto('LowerEndValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(108, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RestrictedCharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(110, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('CharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [112 | __Ss], [__T | __Stack]);
yeccpars2(112, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(113, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Elements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(115, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'TRUE',
 yeccpars2(yeccgoto('BooleanValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(116, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RestrictedCharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'UNION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 185, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('ElementSetSpec', hd(__Ss)), '!', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(118, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('ElementSetSpec', hd(__Ss)), ')', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(118, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UElems', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, '..', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('LowerEndValue', hd(__Ss)), '..', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(119, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('LowerEndValue', hd(__Ss)), '<', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(119, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SingleValue',__1},
 yeccpars2(yeccgoto('SubTypeElements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'COMPONENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('RestrictedCharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(123, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('SignedNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(127, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(128, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [128 | __Ss], [__T | __Stack]);
yeccpars2(128, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(129, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ComponentValueList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(130, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ValueList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(131, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [131 | __Ss], [__T | __Stack]);
yeccpars2(131, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(132, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(133, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [133 | __Ss], [__T | __Stack]);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('SignedNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SAndSOfValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [136 | __Ss], [__T | __Stack]);
yeccpars2(136, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [136 | __Ss], [__T | __Stack]);
yeccpars2(136, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(137, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(138, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'Tuple',__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('Tuple', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [139 | __Ss], [__T | __Stack]);
yeccpars2(139, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(140, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, __Cat, __Ss,  [__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'Quadruple',__2,__4,__6,__8},
 __Nss = lists:nthtail(8, __Ss),
 yeccpars2(yeccgoto('Quadruple', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(143, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(145, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedValue',element(2,__1),element(3,__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ChoiceValue',element(3,__1),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ChoiceValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(149, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 150, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(150, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(151, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(152, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedNumber',element(3,__1),{'ExternalValue',element(3,__3),element(3,__5)}},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('NamedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedNumber',element(3,__1),element(3,__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('NamedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedNumber',element(3,__1),__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('NamedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SAndSOfValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(156, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(157, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1,space|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ValueList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ValueList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(159, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(160, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentValueList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(162, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SAndSOfValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(163, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [163 | __Ss], [__T | __Stack]);
yeccpars2(163, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(164, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externalvaluereference'{pos = element(2,__1), module = element(3,__1), value = element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Value', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(166, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 181, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(169, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('TypeConstraints', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(171, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'ABSENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'PRESENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(172, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('PresenceConstraint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, 'ABSENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, 'PRESENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedConstraint',element(3,__1),__2,undefined},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(174, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('PresenceConstraint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('PresenceConstraint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedConstraint',element(3,__1),undefined,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedConstraint',element(3,__1),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('NamedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH COMPONENTS',{'FullSpecification',__4}},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(180, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TypeConstraints', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(181, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [181 | __Ss], [__T | __Stack]);
yeccpars2(181, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(182, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 183, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(183, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH COMPONENTS',{'PartialSpecification',__3}},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH COMPONENT',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UnionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(187, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UnionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(188, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case {__1,__3} of {{'SingleValue',V1},{'SingleValue',V2}} -> {'SingleValue',ordsets:union(to_set(V1),to_set(V2))} end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Unions', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SizeConstraint',__2#constraint.c},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, 'MAX', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 191, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(191, 'MAX', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(192, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'MAX',
 yeccpars2(yeccgoto('UpperEndValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(193, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UpperEndpoint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ValueRange',{__1,__3}},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UpperEndValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 - 1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('UpperEndpoint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 + 1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('LowerEndpoint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ContainedSubType',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('IntersectionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(200, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('IntersectionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(202, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case {__1,__3} of {{'SingleValue',V1},{'SingleValue',V2}} -> {'SingleValue',ordsets:intersection(to_set(V1),to_set(V2))} end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Intersections', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'PermittedAlphabet',__2#constraint.c},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(205, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('IntersectionElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(206, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'EXCEPT',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Exclusions', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(208, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(209, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #constraint{c = __2, e = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Constraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ExceptionSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(211, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(212, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 216, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(216, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externalvaluereference'{pos = element(2,__1), module = element(3,__1), value = element(3,__1)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(218, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(219, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ALL',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ElementSetSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  - element(3,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SignedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(221, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(222, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Elements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Class', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(224, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(225, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Class', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(226, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Class', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 236, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 235, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('ClassNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(229, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 234, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('ClassNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(231, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 232, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(232, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 233, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(233, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = #'Externalvaluereference'{pos = element(2,__3), module = element(3,__3), value = element(3,__5)}},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = element(3,__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = element(3,__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(237, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(238, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(239, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2#type{tag = __1#tag{type = get(tagdefault)}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TaggedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(240, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3#type{tag = __1#tag{type = 'IMPLICIT'}},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TaggedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3#type{tag = __1#tag{type = 'EXPLICIT'}},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TaggedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 268, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(243, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(245, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 253, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(246, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  undefined,
 yeccpars2(263, __Cat, [246 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(247, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 261, [247 | __Ss], [__T | __Stack]);
yeccpars2(247, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(248, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 259, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ComponentTypeList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 258, [249 | __Ss], [__T | __Stack]);
yeccpars2(249, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(250, 'DEFAULT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 255, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'NamedType',Pos,{Name,Type}} = __1, #'ComponentType'{pos = Pos, name = Name, typespec = Type, prop = mandatory}
  end,
 yeccpars2(yeccgoto('ComponentType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(251, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('NamedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(252, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(253, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SET',[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SetType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   V1 = __1, {'NamedType',V1#identifier.pos,{V1#identifier.val,__2}}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(255, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(256, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'NamedType',Pos,{Name,Type}} = __1, #'ComponentType'{pos = Pos, name = Name, typespec = Type, prop = 'OPTIONAL'}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(257, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'NamedType',Pos,{Name,Type}} = __1, #'ComponentType'{pos = Pos, name = Name, typespec = Type, prop = {'DEFAULT',__3}}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SET',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('SetType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(259, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(260, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentTypeList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(261, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(262, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'COMPONENTS OF',element(2,__1),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'EXTENSIONMARK',element(2,__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 265, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(265, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(266, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [266 | __Ss], [__T | __Stack]);
yeccpars2(266, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SET OF',__5}, constraint = merge_constraints([#constraint{c = {'SizeConstraint',__3}}],#'Constraint'{})},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SET OF',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SetOfType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(268, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(269, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SET OF',__4}, constraint = merge_constraints([__2],#'Constraint'{})},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(270, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(271, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(272, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(273, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 275, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(274, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 276, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(275, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SEQUENCE',[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SequenceType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(276, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SEQUENCE',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('SequenceType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(277, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 278, [277 | __Ss], [__T | __Stack]);
yeccpars2(277, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(278, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(279, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [279 | __Ss], [__T | __Stack]);
yeccpars2(279, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SEQUENCE OF',__5}, constraint = merge_constraints([#constraint{c = {'SizeConstraint',__3}}],#'Constraint'{})},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(280, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SEQUENCE OF',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SequenceOfType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(281, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(282, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [282 | __Ss], [__T | __Stack]);
yeccpars2(282, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SEQUENCE OF',__4}, constraint = merge_constraints([__2],#'Constraint'{})},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(283, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'OCTET STRING',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(284, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'OBJECT IDENTIFIER',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(285, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 288, [285 | __Ss], [__T | __Stack]);
yeccpars2(285, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(286, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 290, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('NamedNumberList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(287, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(288, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(289, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'INTEGER',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('IntegerType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(290, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 288, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(291, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('NamedNumberList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(292, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 297, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(293, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'EXTENSIONMARK',
 yeccpars2(yeccgoto('EnumerationItem', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(294, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 300, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(295, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('Enumeration', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(296, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('EnumerationItem', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(297, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('EnumerationItem', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 297, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(299, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Enumeration', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(300, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ENUMERATED',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('EnumeratedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'EMBEDDED PDV',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(302, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(303, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(304, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'CHOICE',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ChoiceType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(305, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'CHARACTER STRING',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('CharacterStringType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(306, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [306 | __Ss], [__T | __Stack]);
yeccpars2(306, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'BIT STRING',[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BitStringType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(307, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 288, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(308, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 309, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(309, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'BIT STRING',__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('BitStringType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, 'BY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 311, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(311, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 312, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(312, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ANY DEFINED BY',element(3,__4)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(313, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [313 | __Ss], [__T | __Stack]);
yeccpars2(313, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [313 | __Ss], [__T | __Stack]);
yeccpars2(313, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(314, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(315, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #valuedef{pos = element(2,__1), name = element(3,__1), type = __2, value = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ValueAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(316, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('AssignmentList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {imports,[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Imports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(318, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Symbol', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(319, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 333, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('SymbolList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(320, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(321, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 323, [321 | __Ss], [__T | __Stack]);
yeccpars2(321, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [321 | __Ss], [__T | __Stack]);
yeccpars2(321, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('SymbolsFromModuleList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(322, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [322 | __Ss], [__T | __Stack]);
yeccpars2(322, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(323, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Reference', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(324, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Reference', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(325, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {imports,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Imports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(326, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModuleList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(327, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 328, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(328, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 329, [328 | __Ss], [__T | __Stack]);
yeccpars2(328, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 330, [328 | __Ss], [__T | __Stack]);
yeccpars2(328, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SymbolsFromModule'{symbols = __1, module = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModule', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(329, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SymbolsFromModule'{symbols = __1, module = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModule', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(330, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [330 | __Ss], [__T | __Stack]);
yeccpars2(330, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(331, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 332, [331 | __Ss], [__T | __Stack]);
yeccpars2(331, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(332, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SymbolsFromModule'{symbols = __1, module = __3},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModule', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(333, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 323, [333 | __Ss], [__T | __Stack]);
yeccpars2(333, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [333 | __Ss], [__T | __Stack]);
yeccpars2(333, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(334, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SymbolList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(335, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {exports,[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Exports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(336, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 337, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(337, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {exports,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Exports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(338, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'IMPLIED',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ExtensionDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(339, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'IMPLICIT'), 'IMPLICIT'
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TagDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(340, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'EXPLICIT'), 'EXPLICIT'
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TagDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(341, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'AUTOMATIC'), 'AUTOMATIC'
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TagDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto('ModuleDefinition', 0) ->
 1;
yeccgoto('ModuleIdentifier', 0) ->
 2;
yeccgoto('DefinitiveIdentifier', 3) ->
 4;
yeccgoto('DefinitiveObjIdComponentList', 5) ->
 7;
yeccgoto('DefinitiveObjIdComponentList', 6) ->
 14;
yeccgoto('DefinitiveObjIdComponent', 5) ->
 6;
yeccgoto('DefinitiveObjIdComponent', 6) ->
 6;
yeccgoto('TagDefault', 15) ->
 19;
yeccgoto('ExtensionDefault', 19) ->
 21;
yeccgoto('ModuleBody', 23) ->
 26;
yeccgoto('Exports', 23) ->
 25;
yeccgoto('Imports', 25) ->
 29;
yeccgoto('SymbolsFromModuleList', 28) ->
 322;
yeccgoto('SymbolsFromModuleList', 321) ->
 326;
yeccgoto('SymbolsFromModule', 28) ->
 321;
yeccgoto('SymbolsFromModule', 321) ->
 321;
yeccgoto('SymbolList', 28) ->
 320;
yeccgoto('SymbolList', 321) ->
 320;
yeccgoto('SymbolList', 333) ->
 334;
yeccgoto('SymbolList', 24) ->
 336;
yeccgoto('Symbol', 28) ->
 319;
yeccgoto('Symbol', 321) ->
 319;
yeccgoto('Symbol', 333) ->
 319;
yeccgoto('Symbol', 24) ->
 319;
yeccgoto('Reference', 28) ->
 318;
yeccgoto('Reference', 321) ->
 318;
yeccgoto('Reference', 333) ->
 318;
yeccgoto('Reference', 24) ->
 318;
yeccgoto('AssignmentList', 29) ->
 31;
yeccgoto('AssignmentList', 30) ->
 316;
yeccgoto('Assignment', 29) ->
 30;
yeccgoto('Assignment', 30) ->
 30;
yeccgoto('Externaltypereference', 36) ->
 52;
yeccgoto('Externaltypereference', 80) ->
 52;
yeccgoto('Externaltypereference', 101) ->
 52;
yeccgoto('Externaltypereference', 207) ->
 52;
yeccgoto('Externaltypereference', 69) ->
 52;
yeccgoto('Externaltypereference', 238) ->
 52;
yeccgoto('Externaltypereference', 237) ->
 52;
yeccgoto('Externaltypereference', 252) ->
 52;
yeccgoto('Externaltypereference', 261) ->
 52;
yeccgoto('Externaltypereference', 265) ->
 52;
yeccgoto('Externaltypereference', 243) ->
 52;
yeccgoto('Externaltypereference', 268) ->
 52;
yeccgoto('Externaltypereference', 278) ->
 52;
yeccgoto('Externaltypereference', 271) ->
 52;
yeccgoto('Externaltypereference', 281) ->
 52;
yeccgoto('Externaltypereference', 34) ->
 52;
yeccgoto('DefinedType', 36) ->
 47;
yeccgoto('DefinedType', 80) ->
 47;
yeccgoto('DefinedType', 101) ->
 47;
yeccgoto('DefinedType', 207) ->
 47;
yeccgoto('DefinedType', 69) ->
 47;
yeccgoto('DefinedType', 238) ->
 47;
yeccgoto('DefinedType', 237) ->
 47;
yeccgoto('DefinedType', 252) ->
 47;
yeccgoto('DefinedType', 261) ->
 47;
yeccgoto('DefinedType', 265) ->
 47;
yeccgoto('DefinedType', 243) ->
 47;
yeccgoto('DefinedType', 268) ->
 47;
yeccgoto('DefinedType', 278) ->
 47;
yeccgoto('DefinedType', 271) ->
 47;
yeccgoto('DefinedType', 281) ->
 47;
yeccgoto('DefinedType', 34) ->
 47;
yeccgoto('TypeAssignment', 29) ->
 32;
yeccgoto('TypeAssignment', 30) ->
 32;
yeccgoto('ValueAssignment', 29) ->
 33;
yeccgoto('ValueAssignment', 30) ->
 33;
yeccgoto('Type', 36) ->
 71;
yeccgoto('Type', 80) ->
 81;
yeccgoto('Type', 101) ->
 198;
yeccgoto('Type', 207) ->
 212;
yeccgoto('Type', 69) ->
 239;
yeccgoto('Type', 238) ->
 240;
yeccgoto('Type', 237) ->
 241;
yeccgoto('Type', 252) ->
 254;
yeccgoto('Type', 261) ->
 262;
yeccgoto('Type', 265) ->
 266;
yeccgoto('Type', 243) ->
 267;
yeccgoto('Type', 268) ->
 269;
yeccgoto('Type', 278) ->
 279;
yeccgoto('Type', 271) ->
 280;
yeccgoto('Type', 281) ->
 282;
yeccgoto('Type', 34) ->
 313;
yeccgoto('BuiltinType', 36) ->
 41;
yeccgoto('BuiltinType', 80) ->
 41;
yeccgoto('BuiltinType', 101) ->
 41;
yeccgoto('BuiltinType', 207) ->
 41;
yeccgoto('BuiltinType', 69) ->
 41;
yeccgoto('BuiltinType', 238) ->
 41;
yeccgoto('BuiltinType', 237) ->
 41;
yeccgoto('BuiltinType', 252) ->
 41;
yeccgoto('BuiltinType', 261) ->
 41;
yeccgoto('BuiltinType', 265) ->
 41;
yeccgoto('BuiltinType', 243) ->
 41;
yeccgoto('BuiltinType', 268) ->
 41;
yeccgoto('BuiltinType', 278) ->
 41;
yeccgoto('BuiltinType', 271) ->
 41;
yeccgoto('BuiltinType', 281) ->
 41;
yeccgoto('BuiltinType', 34) ->
 41;
yeccgoto('NamedType', 245) ->
 250;
yeccgoto('NamedType', 259) ->
 250;
yeccgoto('NamedType', 273) ->
 250;
yeccgoto('NamedType', 302) ->
 250;
yeccgoto('ReferencedType', 36) ->
 61;
yeccgoto('ReferencedType', 80) ->
 61;
yeccgoto('ReferencedType', 101) ->
 61;
yeccgoto('ReferencedType', 207) ->
 61;
yeccgoto('ReferencedType', 69) ->
 61;
yeccgoto('ReferencedType', 238) ->
 61;
yeccgoto('ReferencedType', 237) ->
 61;
yeccgoto('ReferencedType', 252) ->
 61;
yeccgoto('ReferencedType', 261) ->
 61;
yeccgoto('ReferencedType', 265) ->
 61;
yeccgoto('ReferencedType', 243) ->
 61;
yeccgoto('ReferencedType', 268) ->
 61;
yeccgoto('ReferencedType', 278) ->
 61;
yeccgoto('ReferencedType', 271) ->
 61;
yeccgoto('ReferencedType', 281) ->
 61;
yeccgoto('ReferencedType', 34) ->
 61;
yeccgoto('Value', 82) ->
 119;
yeccgoto('Value', 127) ->
 130;
yeccgoto('Value', 132) ->
 145;
yeccgoto('Value', 144) ->
 146;
yeccgoto('Value', 130) ->
 130;
yeccgoto('Value', 156) ->
 130;
yeccgoto('Value', 161) ->
 145;
yeccgoto('Value', 186) ->
 119;
yeccgoto('Value', 190) ->
 195;
yeccgoto('Value', 191) ->
 195;
yeccgoto('Value', 200) ->
 119;
yeccgoto('Value', 204) ->
 119;
yeccgoto('Value', 217) ->
 218;
yeccgoto('Value', 87) ->
 119;
yeccgoto('Value', 255) ->
 257;
yeccgoto('Value', 314) ->
 315;
yeccgoto('Value', 330) ->
 130;
yeccgoto('BuiltinValue', 82) ->
 91;
yeccgoto('BuiltinValue', 127) ->
 91;
yeccgoto('BuiltinValue', 132) ->
 91;
yeccgoto('BuiltinValue', 144) ->
 91;
yeccgoto('BuiltinValue', 130) ->
 91;
yeccgoto('BuiltinValue', 156) ->
 91;
yeccgoto('BuiltinValue', 161) ->
 91;
yeccgoto('BuiltinValue', 186) ->
 91;
yeccgoto('BuiltinValue', 190) ->
 91;
yeccgoto('BuiltinValue', 191) ->
 91;
yeccgoto('BuiltinValue', 200) ->
 91;
yeccgoto('BuiltinValue', 204) ->
 91;
yeccgoto('BuiltinValue', 217) ->
 91;
yeccgoto('BuiltinValue', 87) ->
 91;
yeccgoto('BuiltinValue', 255) ->
 91;
yeccgoto('BuiltinValue', 314) ->
 91;
yeccgoto('BuiltinValue', 330) ->
 91;
yeccgoto('NamedValue', 127) ->
 129;
yeccgoto('NamedValue', 159) ->
 129;
yeccgoto('BooleanValue', 82) ->
 90;
yeccgoto('BooleanValue', 127) ->
 90;
yeccgoto('BooleanValue', 132) ->
 90;
yeccgoto('BooleanValue', 144) ->
 90;
yeccgoto('BooleanValue', 130) ->
 90;
yeccgoto('BooleanValue', 156) ->
 90;
yeccgoto('BooleanValue', 161) ->
 90;
yeccgoto('BooleanValue', 186) ->
 90;
yeccgoto('BooleanValue', 190) ->
 90;
yeccgoto('BooleanValue', 191) ->
 90;
yeccgoto('BooleanValue', 200) ->
 90;
yeccgoto('BooleanValue', 204) ->
 90;
yeccgoto('BooleanValue', 217) ->
 90;
yeccgoto('BooleanValue', 87) ->
 90;
yeccgoto('BooleanValue', 255) ->
 90;
yeccgoto('BooleanValue', 314) ->
 90;
yeccgoto('BooleanValue', 330) ->
 90;
yeccgoto('IntegerType', 36) ->
 55;
yeccgoto('IntegerType', 80) ->
 55;
yeccgoto('IntegerType', 101) ->
 55;
yeccgoto('IntegerType', 207) ->
 55;
yeccgoto('IntegerType', 69) ->
 55;
yeccgoto('IntegerType', 238) ->
 55;
yeccgoto('IntegerType', 237) ->
 55;
yeccgoto('IntegerType', 252) ->
 55;
yeccgoto('IntegerType', 261) ->
 55;
yeccgoto('IntegerType', 265) ->
 55;
yeccgoto('IntegerType', 243) ->
 55;
yeccgoto('IntegerType', 268) ->
 55;
yeccgoto('IntegerType', 278) ->
 55;
yeccgoto('IntegerType', 271) ->
 55;
yeccgoto('IntegerType', 281) ->
 55;
yeccgoto('IntegerType', 34) ->
 55;
yeccgoto('NamedNumberList', 285) ->
 287;
yeccgoto('NamedNumberList', 290) ->
 291;
yeccgoto('NamedNumberList', 307) ->
 308;
yeccgoto('NamedNumber', 82) ->
 108;
yeccgoto('NamedNumber', 127) ->
 108;
yeccgoto('NamedNumber', 132) ->
 108;
yeccgoto('NamedNumber', 144) ->
 108;
yeccgoto('NamedNumber', 130) ->
 108;
yeccgoto('NamedNumber', 156) ->
 108;
yeccgoto('NamedNumber', 161) ->
 108;
yeccgoto('NamedNumber', 186) ->
 108;
yeccgoto('NamedNumber', 190) ->
 108;
yeccgoto('NamedNumber', 191) ->
 108;
yeccgoto('NamedNumber', 200) ->
 108;
yeccgoto('NamedNumber', 204) ->
 108;
yeccgoto('NamedNumber', 217) ->
 108;
yeccgoto('NamedNumber', 87) ->
 108;
yeccgoto('NamedNumber', 255) ->
 108;
yeccgoto('NamedNumber', 285) ->
 286;
yeccgoto('NamedNumber', 290) ->
 286;
yeccgoto('NamedNumber', 292) ->
 296;
yeccgoto('NamedNumber', 298) ->
 296;
yeccgoto('NamedNumber', 307) ->
 286;
yeccgoto('NamedNumber', 314) ->
 108;
yeccgoto('NamedNumber', 330) ->
 108;
yeccgoto('SignedNumber', 82) ->
 113;
yeccgoto('SignedNumber', 127) ->
 113;
yeccgoto('SignedNumber', 132) ->
 113;
yeccgoto('SignedNumber', 144) ->
 113;
yeccgoto('SignedNumber', 143) ->
 147;
yeccgoto('SignedNumber', 130) ->
 113;
yeccgoto('SignedNumber', 156) ->
 113;
yeccgoto('SignedNumber', 161) ->
 113;
yeccgoto('SignedNumber', 186) ->
 113;
yeccgoto('SignedNumber', 190) ->
 113;
yeccgoto('SignedNumber', 191) ->
 113;
yeccgoto('SignedNumber', 200) ->
 113;
yeccgoto('SignedNumber', 204) ->
 113;
yeccgoto('SignedNumber', 207) ->
 211;
yeccgoto('SignedNumber', 217) ->
 113;
yeccgoto('SignedNumber', 87) ->
 113;
yeccgoto('SignedNumber', 255) ->
 113;
yeccgoto('SignedNumber', 314) ->
 113;
yeccgoto('SignedNumber', 330) ->
 113;
yeccgoto('EnumeratedType', 36) ->
 51;
yeccgoto('EnumeratedType', 80) ->
 51;
yeccgoto('EnumeratedType', 101) ->
 51;
yeccgoto('EnumeratedType', 207) ->
 51;
yeccgoto('EnumeratedType', 69) ->
 51;
yeccgoto('EnumeratedType', 238) ->
 51;
yeccgoto('EnumeratedType', 237) ->
 51;
yeccgoto('EnumeratedType', 252) ->
 51;
yeccgoto('EnumeratedType', 261) ->
 51;
yeccgoto('EnumeratedType', 265) ->
 51;
yeccgoto('EnumeratedType', 243) ->
 51;
yeccgoto('EnumeratedType', 268) ->
 51;
yeccgoto('EnumeratedType', 278) ->
 51;
yeccgoto('EnumeratedType', 271) ->
 51;
yeccgoto('EnumeratedType', 281) ->
 51;
yeccgoto('EnumeratedType', 34) ->
 51;
yeccgoto('Enumeration', 292) ->
 294;
yeccgoto('Enumeration', 298) ->
 299;
yeccgoto('EnumerationItem', 292) ->
 295;
yeccgoto('EnumerationItem', 298) ->
 295;
yeccgoto('BitStringType', 36) ->
 40;
yeccgoto('BitStringType', 80) ->
 40;
yeccgoto('BitStringType', 101) ->
 40;
yeccgoto('BitStringType', 207) ->
 40;
yeccgoto('BitStringType', 69) ->
 40;
yeccgoto('BitStringType', 238) ->
 40;
yeccgoto('BitStringType', 237) ->
 40;
yeccgoto('BitStringType', 252) ->
 40;
yeccgoto('BitStringType', 261) ->
 40;
yeccgoto('BitStringType', 265) ->
 40;
yeccgoto('BitStringType', 243) ->
 40;
yeccgoto('BitStringType', 268) ->
 40;
yeccgoto('BitStringType', 278) ->
 40;
yeccgoto('BitStringType', 271) ->
 40;
yeccgoto('BitStringType', 281) ->
 40;
yeccgoto('BitStringType', 34) ->
 40;
yeccgoto('SequenceType', 36) ->
 66;
yeccgoto('SequenceType', 80) ->
 66;
yeccgoto('SequenceType', 101) ->
 66;
yeccgoto('SequenceType', 207) ->
 66;
yeccgoto('SequenceType', 69) ->
 66;
yeccgoto('SequenceType', 238) ->
 66;
yeccgoto('SequenceType', 237) ->
 66;
yeccgoto('SequenceType', 252) ->
 66;
yeccgoto('SequenceType', 261) ->
 66;
yeccgoto('SequenceType', 265) ->
 66;
yeccgoto('SequenceType', 243) ->
 66;
yeccgoto('SequenceType', 268) ->
 66;
yeccgoto('SequenceType', 278) ->
 66;
yeccgoto('SequenceType', 271) ->
 66;
yeccgoto('SequenceType', 281) ->
 66;
yeccgoto('SequenceType', 34) ->
 66;
yeccgoto('ComponentTypeList', 245) ->
 249;
yeccgoto('ComponentTypeList', 259) ->
 260;
yeccgoto('ComponentTypeList', 273) ->
 274;
yeccgoto('ComponentTypeList', 302) ->
 303;
yeccgoto('ComponentType', 245) ->
 248;
yeccgoto('ComponentType', 259) ->
 248;
yeccgoto('ComponentType', 273) ->
 248;
yeccgoto('ComponentType', 302) ->
 248;
yeccgoto('ComponentValueList', 127) ->
 128;
yeccgoto('ComponentValueList', 159) ->
 160;
yeccgoto('SequenceOfType', 36) ->
 65;
yeccgoto('SequenceOfType', 80) ->
 65;
yeccgoto('SequenceOfType', 101) ->
 65;
yeccgoto('SequenceOfType', 207) ->
 65;
yeccgoto('SequenceOfType', 69) ->
 65;
yeccgoto('SequenceOfType', 238) ->
 65;
yeccgoto('SequenceOfType', 237) ->
 65;
yeccgoto('SequenceOfType', 252) ->
 65;
yeccgoto('SequenceOfType', 261) ->
 65;
yeccgoto('SequenceOfType', 265) ->
 65;
yeccgoto('SequenceOfType', 243) ->
 65;
yeccgoto('SequenceOfType', 268) ->
 65;
yeccgoto('SequenceOfType', 278) ->
 65;
yeccgoto('SequenceOfType', 271) ->
 65;
yeccgoto('SequenceOfType', 281) ->
 65;
yeccgoto('SequenceOfType', 34) ->
 65;
yeccgoto('SAndSOfValue', 82) ->
 111;
yeccgoto('SAndSOfValue', 127) ->
 111;
yeccgoto('SAndSOfValue', 132) ->
 111;
yeccgoto('SAndSOfValue', 144) ->
 111;
yeccgoto('SAndSOfValue', 130) ->
 111;
yeccgoto('SAndSOfValue', 156) ->
 111;
yeccgoto('SAndSOfValue', 161) ->
 111;
yeccgoto('SAndSOfValue', 186) ->
 111;
yeccgoto('SAndSOfValue', 190) ->
 111;
yeccgoto('SAndSOfValue', 191) ->
 111;
yeccgoto('SAndSOfValue', 200) ->
 111;
yeccgoto('SAndSOfValue', 204) ->
 111;
yeccgoto('SAndSOfValue', 217) ->
 111;
yeccgoto('SAndSOfValue', 87) ->
 111;
yeccgoto('SAndSOfValue', 255) ->
 111;
yeccgoto('SAndSOfValue', 314) ->
 111;
yeccgoto('SAndSOfValue', 330) ->
 111;
yeccgoto('ValueList', 127) ->
 131;
yeccgoto('ValueList', 130) ->
 157;
yeccgoto('ValueList', 156) ->
 158;
yeccgoto('ValueList', 330) ->
 331;
yeccgoto('SetType', 36) ->
 68;
yeccgoto('SetType', 80) ->
 68;
yeccgoto('SetType', 101) ->
 68;
yeccgoto('SetType', 207) ->
 68;
yeccgoto('SetType', 69) ->
 68;
yeccgoto('SetType', 238) ->
 68;
yeccgoto('SetType', 237) ->
 68;
yeccgoto('SetType', 252) ->
 68;
yeccgoto('SetType', 261) ->
 68;
yeccgoto('SetType', 265) ->
 68;
yeccgoto('SetType', 243) ->
 68;
yeccgoto('SetType', 268) ->
 68;
yeccgoto('SetType', 278) ->
 68;
yeccgoto('SetType', 271) ->
 68;
yeccgoto('SetType', 281) ->
 68;
yeccgoto('SetType', 34) ->
 68;
yeccgoto('SetOfType', 36) ->
 67;
yeccgoto('SetOfType', 80) ->
 67;
yeccgoto('SetOfType', 101) ->
 67;
yeccgoto('SetOfType', 207) ->
 67;
yeccgoto('SetOfType', 69) ->
 67;
yeccgoto('SetOfType', 238) ->
 67;
yeccgoto('SetOfType', 237) ->
 67;
yeccgoto('SetOfType', 252) ->
 67;
yeccgoto('SetOfType', 261) ->
 67;
yeccgoto('SetOfType', 265) ->
 67;
yeccgoto('SetOfType', 243) ->
 67;
yeccgoto('SetOfType', 268) ->
 67;
yeccgoto('SetOfType', 278) ->
 67;
yeccgoto('SetOfType', 271) ->
 67;
yeccgoto('SetOfType', 281) ->
 67;
yeccgoto('SetOfType', 34) ->
 67;
yeccgoto('ChoiceType', 36) ->
 45;
yeccgoto('ChoiceType', 80) ->
 45;
yeccgoto('ChoiceType', 101) ->
 45;
yeccgoto('ChoiceType', 207) ->
 45;
yeccgoto('ChoiceType', 69) ->
 45;
yeccgoto('ChoiceType', 238) ->
 45;
yeccgoto('ChoiceType', 237) ->
 45;
yeccgoto('ChoiceType', 252) ->
 45;
yeccgoto('ChoiceType', 261) ->
 45;
yeccgoto('ChoiceType', 265) ->
 45;
yeccgoto('ChoiceType', 243) ->
 45;
yeccgoto('ChoiceType', 268) ->
 45;
yeccgoto('ChoiceType', 278) ->
 45;
yeccgoto('ChoiceType', 271) ->
 45;
yeccgoto('ChoiceType', 281) ->
 45;
yeccgoto('ChoiceType', 34) ->
 45;
yeccgoto('ChoiceValue', 82) ->
 93;
yeccgoto('ChoiceValue', 127) ->
 93;
yeccgoto('ChoiceValue', 132) ->
 93;
yeccgoto('ChoiceValue', 144) ->
 93;
yeccgoto('ChoiceValue', 130) ->
 93;
yeccgoto('ChoiceValue', 156) ->
 93;
yeccgoto('ChoiceValue', 161) ->
 93;
yeccgoto('ChoiceValue', 186) ->
 93;
yeccgoto('ChoiceValue', 190) ->
 93;
yeccgoto('ChoiceValue', 191) ->
 93;
yeccgoto('ChoiceValue', 200) ->
 93;
yeccgoto('ChoiceValue', 204) ->
 93;
yeccgoto('ChoiceValue', 217) ->
 93;
yeccgoto('ChoiceValue', 87) ->
 93;
yeccgoto('ChoiceValue', 255) ->
 93;
yeccgoto('ChoiceValue', 314) ->
 93;
yeccgoto('ChoiceValue', 330) ->
 93;
yeccgoto('SelectionType', 36) ->
 64;
yeccgoto('SelectionType', 80) ->
 64;
yeccgoto('SelectionType', 101) ->
 64;
yeccgoto('SelectionType', 207) ->
 64;
yeccgoto('SelectionType', 69) ->
 64;
yeccgoto('SelectionType', 238) ->
 64;
yeccgoto('SelectionType', 237) ->
 64;
yeccgoto('SelectionType', 245) ->
 251;
yeccgoto('SelectionType', 252) ->
 64;
yeccgoto('SelectionType', 259) ->
 251;
yeccgoto('SelectionType', 261) ->
 64;
yeccgoto('SelectionType', 265) ->
 64;
yeccgoto('SelectionType', 243) ->
 64;
yeccgoto('SelectionType', 268) ->
 64;
yeccgoto('SelectionType', 273) ->
 251;
yeccgoto('SelectionType', 278) ->
 64;
yeccgoto('SelectionType', 271) ->
 64;
yeccgoto('SelectionType', 281) ->
 64;
yeccgoto('SelectionType', 302) ->
 251;
yeccgoto('SelectionType', 34) ->
 64;
yeccgoto('TaggedType', 36) ->
 70;
yeccgoto('TaggedType', 80) ->
 70;
yeccgoto('TaggedType', 101) ->
 70;
yeccgoto('TaggedType', 207) ->
 70;
yeccgoto('TaggedType', 69) ->
 70;
yeccgoto('TaggedType', 238) ->
 70;
yeccgoto('TaggedType', 237) ->
 70;
yeccgoto('TaggedType', 252) ->
 70;
yeccgoto('TaggedType', 261) ->
 70;
yeccgoto('TaggedType', 265) ->
 70;
yeccgoto('TaggedType', 243) ->
 70;
yeccgoto('TaggedType', 268) ->
 70;
yeccgoto('TaggedType', 278) ->
 70;
yeccgoto('TaggedType', 271) ->
 70;
yeccgoto('TaggedType', 281) ->
 70;
yeccgoto('TaggedType', 34) ->
 70;
yeccgoto('Tag', 36) ->
 69;
yeccgoto('Tag', 80) ->
 69;
yeccgoto('Tag', 101) ->
 69;
yeccgoto('Tag', 207) ->
 69;
yeccgoto('Tag', 69) ->
 69;
yeccgoto('Tag', 238) ->
 69;
yeccgoto('Tag', 237) ->
 69;
yeccgoto('Tag', 252) ->
 69;
yeccgoto('Tag', 261) ->
 69;
yeccgoto('Tag', 265) ->
 69;
yeccgoto('Tag', 243) ->
 69;
yeccgoto('Tag', 268) ->
 69;
yeccgoto('Tag', 278) ->
 69;
yeccgoto('Tag', 271) ->
 69;
yeccgoto('Tag', 281) ->
 69;
yeccgoto('Tag', 34) ->
 69;
yeccgoto('ClassNumber', 224) ->
 227;
yeccgoto('Class', 74) ->
 224;
yeccgoto('CharacterStringType', 36) ->
 44;
yeccgoto('CharacterStringType', 80) ->
 44;
yeccgoto('CharacterStringType', 101) ->
 44;
yeccgoto('CharacterStringType', 207) ->
 44;
yeccgoto('CharacterStringType', 69) ->
 44;
yeccgoto('CharacterStringType', 238) ->
 44;
yeccgoto('CharacterStringType', 237) ->
 44;
yeccgoto('CharacterStringType', 252) ->
 44;
yeccgoto('CharacterStringType', 261) ->
 44;
yeccgoto('CharacterStringType', 265) ->
 44;
yeccgoto('CharacterStringType', 243) ->
 44;
yeccgoto('CharacterStringType', 268) ->
 44;
yeccgoto('CharacterStringType', 278) ->
 44;
yeccgoto('CharacterStringType', 271) ->
 44;
yeccgoto('CharacterStringType', 281) ->
 44;
yeccgoto('CharacterStringType', 34) ->
 44;
yeccgoto('RestrictedCharacterStringValue', 82) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 127) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 132) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 144) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 130) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 156) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 161) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 186) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 190) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 191) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 200) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 204) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 217) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 87) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 255) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 314) ->
 110;
yeccgoto('RestrictedCharacterStringValue', 330) ->
 110;
yeccgoto('Quadruple', 82) ->
 109;
yeccgoto('Quadruple', 127) ->
 109;
yeccgoto('Quadruple', 132) ->
 109;
yeccgoto('Quadruple', 144) ->
 109;
yeccgoto('Quadruple', 130) ->
 109;
yeccgoto('Quadruple', 156) ->
 109;
yeccgoto('Quadruple', 161) ->
 109;
yeccgoto('Quadruple', 186) ->
 109;
yeccgoto('Quadruple', 190) ->
 109;
yeccgoto('Quadruple', 191) ->
 109;
yeccgoto('Quadruple', 200) ->
 109;
yeccgoto('Quadruple', 204) ->
 109;
yeccgoto('Quadruple', 217) ->
 109;
yeccgoto('Quadruple', 87) ->
 109;
yeccgoto('Quadruple', 255) ->
 109;
yeccgoto('Quadruple', 314) ->
 109;
yeccgoto('Quadruple', 330) ->
 109;
yeccgoto('Tuple', 82) ->
 116;
yeccgoto('Tuple', 127) ->
 116;
yeccgoto('Tuple', 132) ->
 116;
yeccgoto('Tuple', 144) ->
 116;
yeccgoto('Tuple', 130) ->
 116;
yeccgoto('Tuple', 156) ->
 116;
yeccgoto('Tuple', 161) ->
 116;
yeccgoto('Tuple', 186) ->
 116;
yeccgoto('Tuple', 190) ->
 116;
yeccgoto('Tuple', 191) ->
 116;
yeccgoto('Tuple', 200) ->
 116;
yeccgoto('Tuple', 204) ->
 116;
yeccgoto('Tuple', 217) ->
 116;
yeccgoto('Tuple', 87) ->
 116;
yeccgoto('Tuple', 255) ->
 116;
yeccgoto('Tuple', 314) ->
 116;
yeccgoto('Tuple', 330) ->
 116;
yeccgoto('CharacterStringValue', 82) ->
 92;
yeccgoto('CharacterStringValue', 127) ->
 92;
yeccgoto('CharacterStringValue', 132) ->
 92;
yeccgoto('CharacterStringValue', 144) ->
 92;
yeccgoto('CharacterStringValue', 130) ->
 92;
yeccgoto('CharacterStringValue', 156) ->
 92;
yeccgoto('CharacterStringValue', 161) ->
 92;
yeccgoto('CharacterStringValue', 186) ->
 92;
yeccgoto('CharacterStringValue', 190) ->
 92;
yeccgoto('CharacterStringValue', 191) ->
 92;
yeccgoto('CharacterStringValue', 200) ->
 92;
yeccgoto('CharacterStringValue', 204) ->
 92;
yeccgoto('CharacterStringValue', 217) ->
 92;
yeccgoto('CharacterStringValue', 87) ->
 92;
yeccgoto('CharacterStringValue', 255) ->
 92;
yeccgoto('CharacterStringValue', 314) ->
 92;
yeccgoto('CharacterStringValue', 330) ->
 92;
yeccgoto('ConstrainedType', 36) ->
 46;
yeccgoto('ConstrainedType', 80) ->
 46;
yeccgoto('ConstrainedType', 101) ->
 46;
yeccgoto('ConstrainedType', 207) ->
 46;
yeccgoto('ConstrainedType', 69) ->
 46;
yeccgoto('ConstrainedType', 238) ->
 46;
yeccgoto('ConstrainedType', 237) ->
 46;
yeccgoto('ConstrainedType', 252) ->
 46;
yeccgoto('ConstrainedType', 261) ->
 46;
yeccgoto('ConstrainedType', 265) ->
 46;
yeccgoto('ConstrainedType', 243) ->
 46;
yeccgoto('ConstrainedType', 268) ->
 46;
yeccgoto('ConstrainedType', 278) ->
 46;
yeccgoto('ConstrainedType', 271) ->
 46;
yeccgoto('ConstrainedType', 281) ->
 46;
yeccgoto('ConstrainedType', 34) ->
 46;
yeccgoto('Constraint', 81) ->
 83;
yeccgoto('Constraint', 83) ->
 84;
yeccgoto('Constraint', 84) ->
 85;
yeccgoto('Constraint', 85) ->
 86;
yeccgoto('Constraint', 171) ->
 173;
yeccgoto('Constraint', 165) ->
 184;
yeccgoto('Constraint', 112) ->
 189;
yeccgoto('Constraint', 198) ->
 83;
yeccgoto('Constraint', 99) ->
 203;
yeccgoto('Constraint', 212) ->
 83;
yeccgoto('Constraint', 71) ->
 83;
yeccgoto('Constraint', 239) ->
 83;
yeccgoto('Constraint', 240) ->
 83;
yeccgoto('Constraint', 241) ->
 83;
yeccgoto('Constraint', 63) ->
 242;
yeccgoto('Constraint', 254) ->
 83;
yeccgoto('Constraint', 262) ->
 83;
yeccgoto('Constraint', 244) ->
 264;
yeccgoto('Constraint', 266) ->
 83;
yeccgoto('Constraint', 267) ->
 83;
yeccgoto('Constraint', 269) ->
 83;
yeccgoto('Constraint', 62) ->
 270;
yeccgoto('Constraint', 272) ->
 277;
yeccgoto('Constraint', 279) ->
 83;
yeccgoto('Constraint', 280) ->
 83;
yeccgoto('Constraint', 282) ->
 83;
yeccgoto('Constraint', 313) ->
 83;
yeccgoto('ConstraintSpec', 82) ->
 94;
yeccgoto('TypeWithConstraint', 36) ->
 72;
yeccgoto('TypeWithConstraint', 80) ->
 72;
yeccgoto('TypeWithConstraint', 101) ->
 72;
yeccgoto('TypeWithConstraint', 207) ->
 72;
yeccgoto('TypeWithConstraint', 69) ->
 72;
yeccgoto('TypeWithConstraint', 238) ->
 72;
yeccgoto('TypeWithConstraint', 237) ->
 72;
yeccgoto('TypeWithConstraint', 252) ->
 72;
yeccgoto('TypeWithConstraint', 261) ->
 72;
yeccgoto('TypeWithConstraint', 265) ->
 72;
yeccgoto('TypeWithConstraint', 243) ->
 72;
yeccgoto('TypeWithConstraint', 268) ->
 72;
yeccgoto('TypeWithConstraint', 278) ->
 72;
yeccgoto('TypeWithConstraint', 271) ->
 72;
yeccgoto('TypeWithConstraint', 281) ->
 72;
yeccgoto('TypeWithConstraint', 34) ->
 72;
yeccgoto('ElementSetSpec', 82) ->
 95;
yeccgoto('ElementSetSpec', 87) ->
 221;
yeccgoto('ExceptionSpec', 94) ->
 208;
yeccgoto('ExceptionSpec', 246) ->
 263;
yeccgoto('ExceptionIdentification', 207) ->
 210;
yeccgoto('Unions', 82) ->
 118;
yeccgoto('Unions', 87) ->
 118;
yeccgoto('UnionMark', 117) ->
 186;
yeccgoto('UElems', 82) ->
 117;
yeccgoto('UElems', 87) ->
 117;
yeccgoto('Intersections', 82) ->
 103;
yeccgoto('Intersections', 87) ->
 103;
yeccgoto('IntersectionElements', 82) ->
 102;
yeccgoto('IntersectionElements', 186) ->
 188;
yeccgoto('IntersectionElements', 200) ->
 202;
yeccgoto('IntersectionElements', 87) ->
 102;
yeccgoto('IntersectionMark', 100) ->
 200;
yeccgoto('IElems', 82) ->
 100;
yeccgoto('IElems', 87) ->
 100;
yeccgoto('Elements', 82) ->
 96;
yeccgoto('Elements', 186) ->
 96;
yeccgoto('Elements', 200) ->
 96;
yeccgoto('Elements', 204) ->
 206;
yeccgoto('Elements', 87) ->
 96;
yeccgoto('Elems', 82) ->
 97;
yeccgoto('Elems', 186) ->
 97;
yeccgoto('Elems', 200) ->
 97;
yeccgoto('Elems', 87) ->
 97;
yeccgoto('SubTypeElements', 82) ->
 114;
yeccgoto('SubTypeElements', 186) ->
 114;
yeccgoto('SubTypeElements', 200) ->
 114;
yeccgoto('SubTypeElements', 204) ->
 114;
yeccgoto('SubTypeElements', 87) ->
 114;
yeccgoto('Exclusions', 97) ->
 205;
yeccgoto('Exclusions', 89) ->
 219;
yeccgoto('LowerEndpoint', 82) ->
 105;
yeccgoto('LowerEndpoint', 186) ->
 105;
yeccgoto('LowerEndpoint', 200) ->
 105;
yeccgoto('LowerEndpoint', 204) ->
 105;
yeccgoto('LowerEndpoint', 87) ->
 105;
yeccgoto('UpperEndpoint', 190) ->
 194;
yeccgoto('LowerEndValue', 82) ->
 104;
yeccgoto('LowerEndValue', 186) ->
 104;
yeccgoto('LowerEndValue', 200) ->
 104;
yeccgoto('LowerEndValue', 204) ->
 104;
yeccgoto('LowerEndValue', 87) ->
 104;
yeccgoto('UpperEndValue', 190) ->
 193;
yeccgoto('UpperEndValue', 191) ->
 196;
yeccgoto('TypeConstraints', 167) ->
 170;
yeccgoto('TypeConstraints', 179) ->
 180;
yeccgoto('TypeConstraints', 181) ->
 182;
yeccgoto('NamedConstraint', 167) ->
 169;
yeccgoto('NamedConstraint', 179) ->
 169;
yeccgoto('NamedConstraint', 181) ->
 169;
yeccgoto('PresenceConstraint', 171) ->
 176;
yeccgoto('PresenceConstraint', 173) ->
 177;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


