-module(icparse).
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%

-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('peterl@erix.ericsson.se').


-export([parse/1, parse_and_scan/1, format_error/1]).

-import(lists, [reverse/1]).

-ifdef(JAM).
-compile([{parse_transform,jam_yecc_pj},pj]).
-endif.


-include("icforms.hrl").



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?MODULE, "end_of_file"}};
        {error, Descriptor, Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?MODULE, Reason}};
        {ok, Tokens, Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string({_, _, Other}) when list(Other), number(hd(Other)) ->
    Other;
yecctoken2string({_, _, Other}) ->
    io_lib:format('~p', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
%%-vsn('$Revision: /main/london/1').
%%-author('peterl@erix.ericsson.se').

%%------------------------------------------------------------
%%
%% 
%%
%%
%%    COPYRIGHT Ericsson Telecom AB 1995
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%%
%%    Module documentation:
%%    ---------------------
%%
%%    The IDL parser.
%%
%%
%%------------------------------------------------------------













yeccpars2(0, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 316, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 315, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('OorM_<definition>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(4, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(5, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<interface>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 313, [7 | __Ss], [__T | __Stack]);
yeccpars2(7, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(8, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<interface>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(10, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 241, [10 | __Ss], [__T | __Stack]);
yeccpars2(10, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(11, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [13 | __Ss], [__T | __Stack]);
yeccpars2(13, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(14, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<definition>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<definition>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  reverse(__1),
 yeccpars2(yeccgoto('<specification>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(18, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 218, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(20, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [21 | __Ss], [__T | __Stack]);
yeccpars2(21, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(22, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [22 | __Ss], [__T | __Stack]);
yeccpars2(22, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(23, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 186, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(26, switch, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [26 | __Ss], [__T | __Stack]);
yeccpars2(26, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(27, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(28, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(29, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  icgen:scoped_id_new(__1),
 yeccpars2(yeccgoto('<scoped_name>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<integer_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {unsigned,__1},
 yeccpars2(yeccgoto('<integer_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unsigned_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unsigned_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<boolean_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<char_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_long_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_short_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(48, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<unsigned_long_int>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<unsigned_short_int>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [50 | __Ss], [__T | __Stack]);
yeccpars2(50, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(51, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(55, __Cat, [51 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('OorM_<case>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [53 | __Ss], [__T | __Stack]);
yeccpars2(53, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(54, '}', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  reverse(lists:flatten(__1)),
 yeccpars2(yeccgoto('<switch_body>', hd(__Ss)), '}', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(55, __Cat, [54 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [55 | __Ss], [__T | __Stack]);
yeccpars2(55, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(59, __Cat, [55 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [56 | __Ss], [__T | __Stack]);
yeccpars2(56, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(57, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Uggly_pragmas', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(114, __Cat, [58 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, default, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [59 | __Ss], [__T | __Stack]);
yeccpars2(59, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [59 | __Ss], [__T | __Stack]);
yeccpars2(59, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [59 | __Ss], [__T | __Stack]);
yeccpars2(59, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(60, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(113, __Cat, [60 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(62, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(63, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(64, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_operator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_operator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<shift_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, '&', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<xor_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [71 | __Ss], [__T | __Stack]);
yeccpars2(71, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<primary_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(75, '%', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<add_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [76 | __Ss], [__T | __Stack]);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_exp>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<primary_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<and_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(81, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<mult_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(83, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<or_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<boolean_literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<boolean_literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(86, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_operator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(88, '&', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'xor',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<xor_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(89, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(90, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'and',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<and_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(91, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(92, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(93, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {rshift,__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<shift_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(94, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(95, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(96, '%', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'-',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<add_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(97, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(99, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(100, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'/',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<mult_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'*',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<mult_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(102, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'%',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<mult_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(103, '%', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'+',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<add_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {lshift,__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<shift_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<unary_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(107, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  icgen:scoped_id_add(__1,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<scoped_name>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(108, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(109, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'or',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<or_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(110, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(112, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<primary_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 ++ ([__2] ++ __3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('OorM_<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, default, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(115, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(116, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(118, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(165, __Cat, [118 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(121, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(162, __Cat, [121 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<constr_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(123, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<template_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<template_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(130, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<constr_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(131, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<constr_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<any_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(136, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<floating_pt_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(137, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<floating_pt_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<octet_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [139 | __Ss], [__T | __Stack]);
yeccpars2(139, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(140, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #string{},
 yeccpars2(yeccgoto('<string_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(141, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<positive_int_const>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(143, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #string{length = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<string_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(145, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(146, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(147, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #sequence{type = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<sequence_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 150, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(150, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #sequence{type = __3, length = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('<sequence_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(151, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<complex_declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #case_dcl{type = __1, id = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<element_spec>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(156, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('OorM_<fixed_array_size>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #array{id = __1, size = reverse(__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<array_declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(159, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 160, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(160, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<fixed_array_size>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<fixed_array_size>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(163, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(164, __Cat, [163 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 ++ (__3 ++ (__5 ++ (__7 ++ [__4#case_dcl{label = reverse(__2)}]))),
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('<case>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ ([__3|__1] ++ __4),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('OorM_<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(169, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 173, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(170, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(171, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #pragma{type = __4, to = followed, apply = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('OE_pragma', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(172, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(173, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(174, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #pragma{type = __4, to = __5, apply = __6},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('OE_pragma', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #pragma{type = __4, to = __5, apply = icgen:float_to_version(__6)},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('OE_pragma', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<case>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #union{id = __2, type = __5, body = __8},
 __Nss = lists:nthtail(8, __Ss),
 yeccpars2(yeccgoto('<union_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  icgen:scoped_id_new_global(__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<scoped_name>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<type_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(183, __Cat, [181 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #typedef{type = __1, id = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<type_declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [183 | __Ss], [__T | __Stack]);
yeccpars2(183, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|reverse(__2)],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<declarators>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(185, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(187, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(191, __Cat, [187 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(188, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('OorM_<member>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(190, '}', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  reverse(__1),
 yeccpars2(yeccgoto('<member_list>', hd(__Ss)), '}', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(191, __Cat, [190 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(192, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(193, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(194, __Cat, [193 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 195, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(195, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(196, __Cat, [195 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 ++ (__4 ++ (__6 ++ [#member{type = __2, id = __3}])),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('<member>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<member>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #struct{id = __2, body = __4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<struct_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(200, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 203, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(202, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #module{id = __2, body = reverse(__4)},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<module>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, ';', __Ss, [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #forward{id = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<forward_dcl>', hd(__Nss)), ';', __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(207, __Cat, [204 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<inheritance_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<interface_header>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(208, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(209, __Cat, [208 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(209, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|reverse(__3)],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<inheritance_spec>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(211, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<scoped_name>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(212, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(214, __Cat, [213 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ZorM_<member>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(191, __Cat, [215 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ZorM_<member>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #except{id = __2, body = reverse(__4)},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<except_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(218, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 219, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(219, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(220, __Cat, [219 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(221, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(223, __Cat, [221 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(222, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #enumerator{id = __1},
 yeccpars2(yeccgoto('<enumerator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(224, __Cat, [223 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(224, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(225, __Cat, [224 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(226, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(228, __Cat, [226 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, __Cat, __Ss,  [__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #enum{id = __2, body = __4 ++ (__6 ++ (__8 ++ [__5|reverse(__7)]))},
 __Nss = lists:nthtail(8, __Ss),
 yeccpars2(yeccgoto('<enum_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(228, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(229, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ (__4 ++ [__5|__1]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('ZorM_<enumerator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(231, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(232, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 237, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(233, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(237, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(238, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(239, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #const{type = __2, id = __3, val = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<const_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(240, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(244, __Cat, [242 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 312, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, readonly, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 257, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, oneway, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('<interface_body>', hd(__Ss)), '}', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(244, attribute, __Ss, __Stack, __T, __Ts, __Tzr) ->
 __Val =  nil,
 yeccpars2(255, attribute, [244 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(244, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  nil,
 yeccpars2(254, __Cat, [244 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 311, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(246, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 310, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(247, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 309, [247 | __Ss], [__T | __Stack]);
yeccpars2(247, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(248, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  if list(__2) -> __2 ++ __1; true -> [__2|__1] end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ZorM_<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<op_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 308, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(251, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(252, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<export>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(253, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<export>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, void, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 270, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(255, attribute, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 258, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(256, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<op_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(257, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_readonly', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(259, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(260, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 263, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(261, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(262, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(265, __Cat, [264 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(265, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #attr{readonly = __1, type = __3, id = [__4|reverse(__5)]},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<attr_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(266, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 263, [266 | __Ss], [__T | __Stack]);
yeccpars2(266, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(267, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<simple_declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(268, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 271, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(269, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<op_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(270, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<op_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(271, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 272, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(272, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(290, __Cat, [272 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(273, raises, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 276, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(275, __Cat, [273 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(274, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<raises_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(275, context, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 283, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(282, __Cat, [275 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(276, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(277, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [277 | __Ss], [__T | __Stack]);
yeccpars2(277, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [277 | __Ss], [__T | __Stack]);
yeccpars2(277, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(278, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(279, __Cat, [278 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(279, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 280, [279 | __Ss], [__T | __Stack]);
yeccpars2(279, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [279 | __Ss], [__T | __Stack]);
yeccpars2(279, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(280, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|reverse(__4)],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<raises_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(281, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<context_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(282, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #op{oneway = __1, type = __2, id = __3, params = __4, raises = __5, ctx = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('<op_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(283, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(284, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 285, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(286, __Cat, [285 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(286, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 288, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(287, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|reverse(__4)],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<context_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(288, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(289, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<string_literal>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(290, inout, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 295, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, out, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, in, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(291, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<parameter_dcls>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(292, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(293, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(297, __Cat, [293 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(294, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(295, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(296, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(297, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ZorM_<param_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 299, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(300, __Cat, [298 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(299, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ [__3|reverse(__4)],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<parameter_dcls>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(300, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [300 | __Ss], [__T | __Stack]);
yeccpars2(300, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [300 | __Ss], [__T | __Stack]);
yeccpars2(300, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(301, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(302, __Cat, [301 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(302, inout, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 295, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, out, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, in, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(303, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(304, __Cat, [303 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(304, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ (__4 ++ (__6 ++ [__5|__1])),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('ZorM_<param_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(305, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 263, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(306, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #param{inout = __1, type = __2, id = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<param_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(307, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(308, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(309, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(311, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(312, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #interface{id = element(1,__1), inherit = element(2,__1), body = lists:reverse(__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<interface_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(313, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(314, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(315, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(316, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OE_preproc', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 318, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(318, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [318 | __Ss], [__T | __Stack]);
yeccpars2(318, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(320, __Cat, [318 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(319, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(322, __Cat, [319 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(320, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 321, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(321, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case __4 of [] -> case __2 of {_,_,"1"} -> #preproc{cat = line_nr, id = __3, aux = __4}; _ -> [] end; _ -> #preproc{cat = line_nr, id = __3, aux = __4} end,
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('OE_preproc', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(322, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ZorM_<integer_literal>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto('<op_type_spec>', 254) ->
 268;
yeccgoto('<enumerator>', 220) ->
 221;
yeccgoto('<enumerator>', 228) ->
 229;
yeccgoto('<switch_body>', 51) ->
 53;
yeccgoto('OorM_<case>', 51) ->
 54;
yeccgoto('<member_list>', 187) ->
 189;
yeccgoto('<struct_type>', 0) ->
 12;
yeccgoto('<struct_type>', 114) ->
 130;
yeccgoto('<struct_type>', 24) ->
 130;
yeccgoto('<struct_type>', 191) ->
 130;
yeccgoto('<struct_type>', 200) ->
 12;
yeccgoto('<struct_type>', 201) ->
 12;
yeccgoto('<struct_type>', 17) ->
 12;
yeccgoto('<struct_type>', 244) ->
 12;
yeccgoto('<unsigned_int>', 28) ->
 40;
yeccgoto('<unsigned_int>', 114) ->
 40;
yeccgoto('<unsigned_int>', 145) ->
 40;
yeccgoto('<unsigned_int>', 24) ->
 40;
yeccgoto('<unsigned_int>', 191) ->
 40;
yeccgoto('<unsigned_int>', 18) ->
 40;
yeccgoto('<unsigned_int>', 258) ->
 40;
yeccgoto('<unsigned_int>', 254) ->
 40;
yeccgoto('<unsigned_int>', 292) ->
 40;
yeccgoto('<constr_type_spec>', 114) ->
 120;
yeccgoto('<constr_type_spec>', 24) ->
 120;
yeccgoto('<constr_type_spec>', 191) ->
 120;
yeccgoto('<shift_expr>', 61) ->
 79;
yeccgoto('<shift_expr>', 87) ->
 79;
yeccgoto('<shift_expr>', 89) ->
 90;
yeccgoto('<shift_expr>', 108) ->
 79;
yeccgoto('<shift_expr>', 64) ->
 79;
yeccgoto('<shift_expr>', 141) ->
 79;
yeccgoto('<shift_expr>', 147) ->
 79;
yeccgoto('<shift_expr>', 158) ->
 79;
yeccgoto('<shift_expr>', 238) ->
 79;
yeccgoto('<or_expr>', 61) ->
 76;
yeccgoto('<or_expr>', 64) ->
 76;
yeccgoto('<or_expr>', 141) ->
 76;
yeccgoto('<or_expr>', 147) ->
 76;
yeccgoto('<or_expr>', 158) ->
 76;
yeccgoto('<or_expr>', 238) ->
 76;
yeccgoto('<inheritance_spec>', 204) ->
 206;
yeccgoto('ZorM_<param_dcl>', 293) ->
 298;
yeccgoto('Opt_<context_expr>', 275) ->
 282;
yeccgoto('<attr_dcl>', 244) ->
 245;
yeccgoto('<array_declarator>', 132) ->
 151;
yeccgoto('<array_declarator>', 180) ->
 151;
yeccgoto('<array_declarator>', 184) ->
 151;
yeccgoto('<array_declarator>', 192) ->
 151;
yeccgoto('<element_spec>', 114) ->
 121;
yeccgoto('<signed_int>', 28) ->
 36;
yeccgoto('<signed_int>', 114) ->
 36;
yeccgoto('<signed_int>', 145) ->
 36;
yeccgoto('<signed_int>', 24) ->
 36;
yeccgoto('<signed_int>', 191) ->
 36;
yeccgoto('<signed_int>', 18) ->
 36;
yeccgoto('<signed_int>', 258) ->
 36;
yeccgoto('<signed_int>', 254) ->
 36;
yeccgoto('<signed_int>', 292) ->
 36;
yeccgoto('<primary_expr>', 61) ->
 77;
yeccgoto('<primary_expr>', 87) ->
 77;
yeccgoto('<primary_expr>', 89) ->
 77;
yeccgoto('<primary_expr>', 92) ->
 77;
yeccgoto('<primary_expr>', 95) ->
 77;
yeccgoto('<primary_expr>', 99) ->
 77;
yeccgoto('<primary_expr>', 98) ->
 77;
yeccgoto('<primary_expr>', 97) ->
 77;
yeccgoto('<primary_expr>', 94) ->
 77;
yeccgoto('<primary_expr>', 91) ->
 77;
yeccgoto('<primary_expr>', 82) ->
 105;
yeccgoto('<primary_expr>', 108) ->
 77;
yeccgoto('<primary_expr>', 64) ->
 77;
yeccgoto('<primary_expr>', 141) ->
 77;
yeccgoto('<primary_expr>', 147) ->
 77;
yeccgoto('<primary_expr>', 158) ->
 77;
yeccgoto('<primary_expr>', 238) ->
 77;
yeccgoto('<interface_dcl>', 0) ->
 8;
yeccgoto('<interface_dcl>', 200) ->
 8;
yeccgoto('<interface_dcl>', 201) ->
 8;
yeccgoto('<interface_dcl>', 17) ->
 8;
yeccgoto('ZorM_<string_literal>', 285) ->
 286;
yeccgoto('Opt_<raises_expr>', 273) ->
 275;
yeccgoto('<integer_type>', 28) ->
 34;
yeccgoto('<integer_type>', 114) ->
 124;
yeccgoto('<integer_type>', 145) ->
 124;
yeccgoto('<integer_type>', 24) ->
 124;
yeccgoto('<integer_type>', 191) ->
 124;
yeccgoto('<integer_type>', 18) ->
 234;
yeccgoto('<integer_type>', 258) ->
 124;
yeccgoto('<integer_type>', 254) ->
 124;
yeccgoto('<integer_type>', 292) ->
 124;
yeccgoto('<signed_long_int>', 28) ->
 37;
yeccgoto('<signed_long_int>', 114) ->
 37;
yeccgoto('<signed_long_int>', 145) ->
 37;
yeccgoto('<signed_long_int>', 24) ->
 37;
yeccgoto('<signed_long_int>', 191) ->
 37;
yeccgoto('<signed_long_int>', 18) ->
 37;
yeccgoto('<signed_long_int>', 258) ->
 37;
yeccgoto('<signed_long_int>', 254) ->
 37;
yeccgoto('<signed_long_int>', 292) ->
 37;
yeccgoto('<literal>', 61) ->
 74;
yeccgoto('<literal>', 87) ->
 74;
yeccgoto('<literal>', 89) ->
 74;
yeccgoto('<literal>', 92) ->
 74;
yeccgoto('<literal>', 95) ->
 74;
yeccgoto('<literal>', 99) ->
 74;
yeccgoto('<literal>', 98) ->
 74;
yeccgoto('<literal>', 97) ->
 74;
yeccgoto('<literal>', 94) ->
 74;
yeccgoto('<literal>', 91) ->
 74;
yeccgoto('<literal>', 82) ->
 74;
yeccgoto('<literal>', 108) ->
 74;
yeccgoto('<literal>', 64) ->
 74;
yeccgoto('<literal>', 141) ->
 74;
yeccgoto('<literal>', 147) ->
 74;
yeccgoto('<literal>', 158) ->
 74;
yeccgoto('<literal>', 238) ->
 74;
yeccgoto('<export>', 244) ->
 248;
yeccgoto('<forward_dcl>', 0) ->
 6;
yeccgoto('<forward_dcl>', 200) ->
 6;
yeccgoto('<forward_dcl>', 201) ->
 6;
yeccgoto('<forward_dcl>', 17) ->
 6;
yeccgoto('OorM_<definition>', 0) ->
 17;
yeccgoto('OorM_<definition>', 200) ->
 201;
yeccgoto('<base_type_spec>', 114) ->
 116;
yeccgoto('<base_type_spec>', 145) ->
 116;
yeccgoto('<base_type_spec>', 24) ->
 116;
yeccgoto('<base_type_spec>', 191) ->
 116;
yeccgoto('<base_type_spec>', 258) ->
 259;
yeccgoto('<base_type_spec>', 254) ->
 259;
yeccgoto('<base_type_spec>', 292) ->
 259;
yeccgoto('<op_dcl>', 244) ->
 250;
yeccgoto('<const_exp>', 61) ->
 71;
yeccgoto('<const_exp>', 64) ->
 111;
yeccgoto('<const_exp>', 141) ->
 142;
yeccgoto('<const_exp>', 147) ->
 142;
yeccgoto('<const_exp>', 158) ->
 142;
yeccgoto('<const_exp>', 238) ->
 239;
yeccgoto('<case>', 51) ->
 52;
yeccgoto('<case>', 54) ->
 176;
yeccgoto('<any_type>', 114) ->
 115;
yeccgoto('<any_type>', 145) ->
 115;
yeccgoto('<any_type>', 24) ->
 115;
yeccgoto('<any_type>', 191) ->
 115;
yeccgoto('<any_type>', 258) ->
 115;
yeccgoto('<any_type>', 254) ->
 115;
yeccgoto('<any_type>', 292) ->
 115;
yeccgoto('<signed_short_int>', 28) ->
 38;
yeccgoto('<signed_short_int>', 114) ->
 38;
yeccgoto('<signed_short_int>', 145) ->
 38;
yeccgoto('<signed_short_int>', 24) ->
 38;
yeccgoto('<signed_short_int>', 191) ->
 38;
yeccgoto('<signed_short_int>', 18) ->
 38;
yeccgoto('<signed_short_int>', 258) ->
 38;
yeccgoto('<signed_short_int>', 254) ->
 38;
yeccgoto('<signed_short_int>', 292) ->
 38;
yeccgoto('<unary_expr>', 61) ->
 81;
yeccgoto('<unary_expr>', 87) ->
 81;
yeccgoto('<unary_expr>', 89) ->
 81;
yeccgoto('<unary_expr>', 92) ->
 81;
yeccgoto('<unary_expr>', 95) ->
 81;
yeccgoto('<unary_expr>', 99) ->
 100;
yeccgoto('<unary_expr>', 98) ->
 101;
yeccgoto('<unary_expr>', 97) ->
 102;
yeccgoto('<unary_expr>', 94) ->
 81;
yeccgoto('<unary_expr>', 91) ->
 81;
yeccgoto('<unary_expr>', 108) ->
 81;
yeccgoto('<unary_expr>', 64) ->
 81;
yeccgoto('<unary_expr>', 141) ->
 81;
yeccgoto('<unary_expr>', 147) ->
 81;
yeccgoto('<unary_expr>', 158) ->
 81;
yeccgoto('<unary_expr>', 238) ->
 81;
yeccgoto('<context_expr>', 275) ->
 281;
yeccgoto('ZorM_<scoped_name>', 208) ->
 209;
yeccgoto('ZorM_<scoped_name>', 278) ->
 279;
yeccgoto('<switch_type_spec>', 28) ->
 39;
yeccgoto('<complex_declarator>', 132) ->
 152;
yeccgoto('<complex_declarator>', 180) ->
 152;
yeccgoto('<complex_declarator>', 184) ->
 152;
yeccgoto('<complex_declarator>', 192) ->
 152;
yeccgoto('<declarators>', 180) ->
 182;
yeccgoto('<declarators>', 192) ->
 193;
yeccgoto('OorM_<member>', 187) ->
 190;
yeccgoto('<interface>', 0) ->
 7;
yeccgoto('<interface>', 200) ->
 7;
yeccgoto('<interface>', 201) ->
 7;
yeccgoto('<interface>', 17) ->
 7;
yeccgoto('<parameter_dcls>', 271) ->
 273;
yeccgoto('<op_attribute>', 244) ->
 249;
yeccgoto('<positive_int_const>', 141) ->
 143;
yeccgoto('<positive_int_const>', 147) ->
 149;
yeccgoto('<positive_int_const>', 158) ->
 159;
yeccgoto('OorM_<fixed_array_size>', 154) ->
 157;
yeccgoto('<sequence_type>', 114) ->
 127;
yeccgoto('<sequence_type>', 145) ->
 127;
yeccgoto('<sequence_type>', 24) ->
 127;
yeccgoto('<sequence_type>', 191) ->
 127;
yeccgoto('<case_label>', 59) ->
 60;
yeccgoto('<case_label>', 114) ->
 118;
yeccgoto('<octet_type>', 114) ->
 125;
yeccgoto('<octet_type>', 145) ->
 125;
yeccgoto('<octet_type>', 24) ->
 125;
yeccgoto('<octet_type>', 191) ->
 125;
yeccgoto('<octet_type>', 258) ->
 125;
yeccgoto('<octet_type>', 254) ->
 125;
yeccgoto('<octet_type>', 292) ->
 125;
yeccgoto('<type_dcl>', 0) ->
 13;
yeccgoto('<type_dcl>', 200) ->
 13;
yeccgoto('<type_dcl>', 201) ->
 13;
yeccgoto('<type_dcl>', 17) ->
 13;
yeccgoto('<type_dcl>', 244) ->
 251;
yeccgoto('<module>', 0) ->
 10;
yeccgoto('<module>', 200) ->
 10;
yeccgoto('<module>', 201) ->
 10;
yeccgoto('<module>', 17) ->
 10;
yeccgoto('<specification>', 0) ->
 11;
yeccgoto('<declarator>', 132) ->
 153;
yeccgoto('<declarator>', 180) ->
 181;
yeccgoto('<declarator>', 184) ->
 185;
yeccgoto('<declarator>', 192) ->
 181;
yeccgoto('<boolean_type>', 28) ->
 30;
yeccgoto('<boolean_type>', 114) ->
 117;
yeccgoto('<boolean_type>', 145) ->
 117;
yeccgoto('<boolean_type>', 24) ->
 117;
yeccgoto('<boolean_type>', 191) ->
 117;
yeccgoto('<boolean_type>', 18) ->
 230;
yeccgoto('<boolean_type>', 258) ->
 117;
yeccgoto('<boolean_type>', 254) ->
 117;
yeccgoto('<boolean_type>', 292) ->
 117;
yeccgoto('<union_type>', 0) ->
 14;
yeccgoto('<union_type>', 114) ->
 133;
yeccgoto('<union_type>', 24) ->
 133;
yeccgoto('<union_type>', 191) ->
 133;
yeccgoto('<union_type>', 200) ->
 14;
yeccgoto('<union_type>', 201) ->
 14;
yeccgoto('<union_type>', 17) ->
 14;
yeccgoto('<union_type>', 244) ->
 14;
yeccgoto('<add_expr>', 61) ->
 67;
yeccgoto('<add_expr>', 87) ->
 67;
yeccgoto('<add_expr>', 89) ->
 67;
yeccgoto('<add_expr>', 92) ->
 93;
yeccgoto('<add_expr>', 91) ->
 104;
yeccgoto('<add_expr>', 108) ->
 67;
yeccgoto('<add_expr>', 64) ->
 67;
yeccgoto('<add_expr>', 141) ->
 67;
yeccgoto('<add_expr>', 147) ->
 67;
yeccgoto('<add_expr>', 158) ->
 67;
yeccgoto('<add_expr>', 238) ->
 67;
yeccgoto('<interface_body>', 242) ->
 243;
yeccgoto('<except_dcl>', 0) ->
 5;
yeccgoto('<except_dcl>', 200) ->
 5;
yeccgoto('<except_dcl>', 201) ->
 5;
yeccgoto('<except_dcl>', 17) ->
 5;
yeccgoto('<except_dcl>', 244) ->
 247;
yeccgoto('<fixed_array_size>', 154) ->
 156;
yeccgoto('<fixed_array_size>', 157) ->
 161;
yeccgoto('<unsigned_short_int>', 28) ->
 42;
yeccgoto('<unsigned_short_int>', 114) ->
 42;
yeccgoto('<unsigned_short_int>', 145) ->
 42;
yeccgoto('<unsigned_short_int>', 24) ->
 42;
yeccgoto('<unsigned_short_int>', 191) ->
 42;
yeccgoto('<unsigned_short_int>', 18) ->
 42;
yeccgoto('<unsigned_short_int>', 258) ->
 42;
yeccgoto('<unsigned_short_int>', 254) ->
 42;
yeccgoto('<unsigned_short_int>', 292) ->
 42;
yeccgoto('<boolean_literal>', 61) ->
 69;
yeccgoto('<boolean_literal>', 87) ->
 69;
yeccgoto('<boolean_literal>', 89) ->
 69;
yeccgoto('<boolean_literal>', 92) ->
 69;
yeccgoto('<boolean_literal>', 95) ->
 69;
yeccgoto('<boolean_literal>', 99) ->
 69;
yeccgoto('<boolean_literal>', 98) ->
 69;
yeccgoto('<boolean_literal>', 97) ->
 69;
yeccgoto('<boolean_literal>', 94) ->
 69;
yeccgoto('<boolean_literal>', 91) ->
 69;
yeccgoto('<boolean_literal>', 82) ->
 69;
yeccgoto('<boolean_literal>', 108) ->
 69;
yeccgoto('<boolean_literal>', 64) ->
 69;
yeccgoto('<boolean_literal>', 141) ->
 69;
yeccgoto('<boolean_literal>', 147) ->
 69;
yeccgoto('<boolean_literal>', 158) ->
 69;
yeccgoto('<boolean_literal>', 238) ->
 69;
yeccgoto('<and_expr>', 61) ->
 68;
yeccgoto('<and_expr>', 87) ->
 88;
yeccgoto('<and_expr>', 108) ->
 68;
yeccgoto('<and_expr>', 64) ->
 68;
yeccgoto('<and_expr>', 141) ->
 68;
yeccgoto('<and_expr>', 147) ->
 68;
yeccgoto('<and_expr>', 158) ->
 68;
yeccgoto('<and_expr>', 238) ->
 68;
yeccgoto('Opt_<inheritance_spec>', 204) ->
 207;
yeccgoto('<scoped_name>', 28) ->
 35;
yeccgoto('<scoped_name>', 61) ->
 78;
yeccgoto('<scoped_name>', 87) ->
 78;
yeccgoto('<scoped_name>', 89) ->
 78;
yeccgoto('<scoped_name>', 92) ->
 78;
yeccgoto('<scoped_name>', 95) ->
 78;
yeccgoto('<scoped_name>', 99) ->
 78;
yeccgoto('<scoped_name>', 98) ->
 78;
yeccgoto('<scoped_name>', 97) ->
 78;
yeccgoto('<scoped_name>', 94) ->
 78;
yeccgoto('<scoped_name>', 91) ->
 78;
yeccgoto('<scoped_name>', 82) ->
 78;
yeccgoto('<scoped_name>', 108) ->
 78;
yeccgoto('<scoped_name>', 64) ->
 78;
yeccgoto('<scoped_name>', 114) ->
 126;
yeccgoto('<scoped_name>', 141) ->
 78;
yeccgoto('<scoped_name>', 145) ->
 126;
yeccgoto('<scoped_name>', 147) ->
 78;
yeccgoto('<scoped_name>', 158) ->
 78;
yeccgoto('<scoped_name>', 24) ->
 126;
yeccgoto('<scoped_name>', 191) ->
 126;
yeccgoto('<scoped_name>', 205) ->
 208;
yeccgoto('<scoped_name>', 210) ->
 211;
yeccgoto('<scoped_name>', 18) ->
 235;
yeccgoto('<scoped_name>', 238) ->
 78;
yeccgoto('<scoped_name>', 258) ->
 261;
yeccgoto('<scoped_name>', 254) ->
 261;
yeccgoto('<scoped_name>', 277) ->
 278;
yeccgoto('<scoped_name>', 292) ->
 261;
yeccgoto('<param_type_spec>', 258) ->
 260;
yeccgoto('<param_type_spec>', 254) ->
 269;
yeccgoto('<param_type_spec>', 292) ->
 305;
yeccgoto('ZorM_<member>', 213) ->
 215;
yeccgoto('<char_type>', 28) ->
 31;
yeccgoto('<char_type>', 114) ->
 119;
yeccgoto('<char_type>', 145) ->
 119;
yeccgoto('<char_type>', 24) ->
 119;
yeccgoto('<char_type>', 191) ->
 119;
yeccgoto('<char_type>', 18) ->
 231;
yeccgoto('<char_type>', 258) ->
 119;
yeccgoto('<char_type>', 254) ->
 119;
yeccgoto('<char_type>', 292) ->
 119;
yeccgoto('<const_dcl>', 0) ->
 2;
yeccgoto('<const_dcl>', 200) ->
 2;
yeccgoto('<const_dcl>', 201) ->
 2;
yeccgoto('<const_dcl>', 17) ->
 2;
yeccgoto('<const_dcl>', 244) ->
 246;
yeccgoto('<param_dcl>', 290) ->
 293;
yeccgoto('<param_dcl>', 302) ->
 303;
yeccgoto('ZorM_<simple_declarator>', 264) ->
 265;
yeccgoto('ZorM_<declarator>', 181) ->
 183;
yeccgoto('<const_type>', 18) ->
 232;
yeccgoto('<definition>', 0) ->
 3;
yeccgoto('<definition>', 200) ->
 3;
yeccgoto('<definition>', 201) ->
 202;
yeccgoto('<definition>', 17) ->
 202;
yeccgoto('<param_attribute>', 290) ->
 292;
yeccgoto('<param_attribute>', 302) ->
 292;
yeccgoto('<simple_declarator>', 132) ->
 155;
yeccgoto('<simple_declarator>', 180) ->
 155;
yeccgoto('<simple_declarator>', 184) ->
 155;
yeccgoto('<simple_declarator>', 192) ->
 155;
yeccgoto('<simple_declarator>', 260) ->
 264;
yeccgoto('<simple_declarator>', 266) ->
 267;
yeccgoto('<simple_declarator>', 305) ->
 306;
yeccgoto('Opt_readonly', 244) ->
 255;
yeccgoto('<simple_type_spec>', 114) ->
 128;
yeccgoto('<simple_type_spec>', 145) ->
 146;
yeccgoto('<simple_type_spec>', 24) ->
 128;
yeccgoto('<simple_type_spec>', 191) ->
 128;
yeccgoto('<enum_type>', 0) ->
 4;
yeccgoto('<enum_type>', 28) ->
 32;
yeccgoto('<enum_type>', 114) ->
 122;
yeccgoto('<enum_type>', 24) ->
 122;
yeccgoto('<enum_type>', 191) ->
 122;
yeccgoto('<enum_type>', 200) ->
 4;
yeccgoto('<enum_type>', 201) ->
 4;
yeccgoto('<enum_type>', 17) ->
 4;
yeccgoto('<enum_type>', 244) ->
 4;
yeccgoto('<type_spec>', 114) ->
 132;
yeccgoto('<type_spec>', 24) ->
 180;
yeccgoto('<type_spec>', 191) ->
 192;
yeccgoto('OorM_<case_label>', 55) ->
 58;
yeccgoto('<floating_pt_type>', 114) ->
 123;
yeccgoto('<floating_pt_type>', 145) ->
 123;
yeccgoto('<floating_pt_type>', 24) ->
 123;
yeccgoto('<floating_pt_type>', 191) ->
 123;
yeccgoto('<floating_pt_type>', 18) ->
 233;
yeccgoto('<floating_pt_type>', 258) ->
 123;
yeccgoto('<floating_pt_type>', 254) ->
 123;
yeccgoto('<floating_pt_type>', 292) ->
 123;
yeccgoto('<template_type_spec>', 114) ->
 131;
yeccgoto('<template_type_spec>', 145) ->
 131;
yeccgoto('<template_type_spec>', 24) ->
 131;
yeccgoto('<template_type_spec>', 191) ->
 131;
yeccgoto('<mult_expr>', 61) ->
 75;
yeccgoto('<mult_expr>', 87) ->
 75;
yeccgoto('<mult_expr>', 89) ->
 75;
yeccgoto('<mult_expr>', 92) ->
 75;
yeccgoto('<mult_expr>', 95) ->
 96;
yeccgoto('<mult_expr>', 94) ->
 103;
yeccgoto('<mult_expr>', 91) ->
 75;
yeccgoto('<mult_expr>', 108) ->
 75;
yeccgoto('<mult_expr>', 64) ->
 75;
yeccgoto('<mult_expr>', 141) ->
 75;
yeccgoto('<mult_expr>', 147) ->
 75;
yeccgoto('<mult_expr>', 158) ->
 75;
yeccgoto('<mult_expr>', 238) ->
 75;
yeccgoto('<xor_expr>', 61) ->
 83;
yeccgoto('<xor_expr>', 108) ->
 109;
yeccgoto('<xor_expr>', 64) ->
 83;
yeccgoto('<xor_expr>', 141) ->
 83;
yeccgoto('<xor_expr>', 147) ->
 83;
yeccgoto('<xor_expr>', 158) ->
 83;
yeccgoto('<xor_expr>', 238) ->
 83;
yeccgoto('<string_type>', 114) ->
 129;
yeccgoto('<string_type>', 145) ->
 129;
yeccgoto('<string_type>', 24) ->
 129;
yeccgoto('<string_type>', 191) ->
 129;
yeccgoto('<string_type>', 18) ->
 236;
yeccgoto('<string_type>', 258) ->
 262;
yeccgoto('<string_type>', 254) ->
 262;
yeccgoto('<string_type>', 292) ->
 262;
yeccgoto('<raises_expr>', 273) ->
 274;
yeccgoto('Opt_<op_attribute>', 244) ->
 254;
yeccgoto('ZorM_<enumerator>', 223) ->
 224;
yeccgoto('<member>', 187) ->
 188;
yeccgoto('<member>', 190) ->
 197;
yeccgoto('<member>', 215) ->
 216;
yeccgoto('<unsigned_long_int>', 28) ->
 41;
yeccgoto('<unsigned_long_int>', 114) ->
 41;
yeccgoto('<unsigned_long_int>', 145) ->
 41;
yeccgoto('<unsigned_long_int>', 24) ->
 41;
yeccgoto('<unsigned_long_int>', 191) ->
 41;
yeccgoto('<unsigned_long_int>', 18) ->
 41;
yeccgoto('<unsigned_long_int>', 258) ->
 41;
yeccgoto('<unsigned_long_int>', 254) ->
 41;
yeccgoto('<unsigned_long_int>', 292) ->
 41;
yeccgoto('<type_declarator>', 24) ->
 179;
yeccgoto('<unary_operator>', 61) ->
 82;
yeccgoto('<unary_operator>', 87) ->
 82;
yeccgoto('<unary_operator>', 89) ->
 82;
yeccgoto('<unary_operator>', 92) ->
 82;
yeccgoto('<unary_operator>', 95) ->
 82;
yeccgoto('<unary_operator>', 99) ->
 82;
yeccgoto('<unary_operator>', 98) ->
 82;
yeccgoto('<unary_operator>', 97) ->
 82;
yeccgoto('<unary_operator>', 94) ->
 82;
yeccgoto('<unary_operator>', 91) ->
 82;
yeccgoto('<unary_operator>', 108) ->
 82;
yeccgoto('<unary_operator>', 64) ->
 82;
yeccgoto('<unary_operator>', 141) ->
 82;
yeccgoto('<unary_operator>', 147) ->
 82;
yeccgoto('<unary_operator>', 158) ->
 82;
yeccgoto('<unary_operator>', 238) ->
 82;
yeccgoto('ZorM_<export>', 242) ->
 244;
yeccgoto('<interface_header>', 0) ->
 9;
yeccgoto('<interface_header>', 200) ->
 9;
yeccgoto('<interface_header>', 201) ->
 9;
yeccgoto('<interface_header>', 17) ->
 9;
yeccgoto('OE_preproc', 0) ->
 16;
yeccgoto('OE_preproc', 200) ->
 16;
yeccgoto('OE_preproc', 201) ->
 16;
yeccgoto('OE_preproc', 17) ->
 16;
yeccgoto('OE_preproc', 244) ->
 253;
yeccgoto('OE_pragma', 0) ->
 15;
yeccgoto('OE_pragma', 55) ->
 57;
yeccgoto('OE_pragma', 59) ->
 57;
yeccgoto('OE_pragma', 113) ->
 57;
yeccgoto('OE_pragma', 114) ->
 57;
yeccgoto('OE_pragma', 162) ->
 57;
yeccgoto('OE_pragma', 164) ->
 57;
yeccgoto('OE_pragma', 165) ->
 57;
yeccgoto('OE_pragma', 191) ->
 57;
yeccgoto('OE_pragma', 194) ->
 57;
yeccgoto('OE_pragma', 196) ->
 57;
yeccgoto('OE_pragma', 200) ->
 15;
yeccgoto('OE_pragma', 201) ->
 15;
yeccgoto('OE_pragma', 214) ->
 57;
yeccgoto('OE_pragma', 220) ->
 57;
yeccgoto('OE_pragma', 223) ->
 57;
yeccgoto('OE_pragma', 225) ->
 57;
yeccgoto('OE_pragma', 228) ->
 57;
yeccgoto('OE_pragma', 17) ->
 15;
yeccgoto('OE_pragma', 244) ->
 252;
yeccgoto('OE_pragma', 290) ->
 57;
yeccgoto('OE_pragma', 300) ->
 57;
yeccgoto('OE_pragma', 302) ->
 57;
yeccgoto('OE_pragma', 304) ->
 57;
yeccgoto('OE_pragma', 297) ->
 57;
yeccgoto('Uggly_pragmas', 51) ->
 55;
yeccgoto('Uggly_pragmas', 55) ->
 59;
yeccgoto('Uggly_pragmas', 60) ->
 113;
yeccgoto('Uggly_pragmas', 58) ->
 114;
yeccgoto('Uggly_pragmas', 121) ->
 162;
yeccgoto('Uggly_pragmas', 163) ->
 164;
yeccgoto('Uggly_pragmas', 118) ->
 165;
yeccgoto('Uggly_pragmas', 54) ->
 55;
yeccgoto('Uggly_pragmas', 187) ->
 191;
yeccgoto('Uggly_pragmas', 193) ->
 194;
yeccgoto('Uggly_pragmas', 195) ->
 196;
yeccgoto('Uggly_pragmas', 190) ->
 191;
yeccgoto('Uggly_pragmas', 213) ->
 214;
yeccgoto('Uggly_pragmas', 215) ->
 191;
yeccgoto('Uggly_pragmas', 219) ->
 220;
yeccgoto('Uggly_pragmas', 221) ->
 223;
yeccgoto('Uggly_pragmas', 224) ->
 225;
yeccgoto('Uggly_pragmas', 226) ->
 228;
yeccgoto('Uggly_pragmas', 272) ->
 290;
yeccgoto('Uggly_pragmas', 293) ->
 297;
yeccgoto('Uggly_pragmas', 298) ->
 300;
yeccgoto('Uggly_pragmas', 301) ->
 302;
yeccgoto('Uggly_pragmas', 303) ->
 304;
yeccgoto('ZorM_<integer_literal>', 318) ->
 320;
yeccgoto('ZorM_<integer_literal>', 319) ->
 322;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


